;;;
;;; Copyright (c) 2010, Jean-Claude Beaudoin.
;;; Copyright (c) 2005, Michael Goffioul (michael dot goffioul at swing dot be)
;;;
;;;   This program is free software; you can redistribute it and/or
;;;   modify it under the terms of the GNU Library General Public
;;;   License as published by the Free Software Foundation; either
;;;   version 2 of the License, or (at your option) any later version.
;;;
;;;   See file '../../Copyright' for full details.
;;;
;;; MKCL SPECIFIC OPERATIONS FOR ASDF
;;;

(in-package :asdf)
(require 'cmp)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (setq clos::*redefine-class-in-place* t) ;; Turn on strict ANSI class redefinition semantics.
  )

;;;
;;;

(defun system-depends-on (sys)
  (reverse (cdr (assoc 'load-op (component-depends-on 'load-op sys)))))

;;;
;;; BUNDLE-OP
;;;
;;; This operation takes all components from one or more systems and
;;; creates a single output file, which may be a FASL, a statically
;;; linked library, a shared library, etc The different targets are
;;; defined by specialization.
;;;

(defclass bundle-op (operation)
  ((fasl :initarg :fasl :initform t :reader bundle-op-fasl-p)
   (static-library :initarg :static-library :initform t :reader bundle-op-static-library-p)
   (build-args :initarg :args :initform nil :accessor bundle-op-build-args)))

(defvar *force-load-p* nil)

(defmethod operation-done-p :around ((operation load-op) c)
  (declare (ignore c))
  (if *force-load-p* nil (call-next-method)))

(defun gather-components (op-type system &key filter-system filter-type include-self)
  ;; This function creates a list of components, matched together with an
  ;; operation. This list may be restricted to sub-components of SYSTEM if
  ;; GATHER-ALL = NIL (default), and it may include the system itself.
  (let* ((operation (make-instance op-type))
         (*force-load-p* t)
         (tree (traverse (make-instance 'load-op) system)))
    (append
     (loop for (op . component) in tree
	when (and (typep op 'load-op)
		  (typep component filter-type)
		  (or (not filter-system) (eq (component-system component) filter-system)))
	collect (progn
		  (when (eq component system) (setf include-self nil))
		  (cons operation component)))
     (and include-self (list (cons operation system))))))

;;;
;;; BUNDLE-SUB-OPERATIONS
;;;
;;; Builds a list of pairs (operation . component) which contains all the
;;; dependencies of this bundle.
;;;

(defun bundle-sub-operations (sys)
  (gather-components 'compile-op sys
		     :filter-system sys
		     :filter-type '(not system))
  )

(defun files-to-bundle (sys &aux files)
  (dolist (op.compo-pair (bundle-sub-operations sys) (nreverse files))
    (let* ((sub-op (car op.compo-pair))
	   (sub-compo (cdr op.compo-pair))
	   (sub-files (output-files sub-op sub-compo))
	   )
      (when sub-files
	;;(push (car sub-files) files) ;; conses more.
	(setq files (rplacd sub-files files)) ;; conses less.
	))))

(defmethod component-depends-on ((o bundle-op) (c system))
  (cons `(compile-op ,(component-name c)) (call-next-method)))


(defmethod output-files ((o bundle-op) (c system))
  (let* ((name (component-name c))
	 (static-lib-name (merge-pathnames
			   (compile-file-pathname name :type :static-library)
			   (component-relative-pathname c)))
	 (fasl-name (merge-pathnames
		     (compile-file-pathname name :type :fasb)
		     (component-relative-pathname c)))
	 )
    (list static-lib-name fasl-name)
    )
  )

(defmethod perform ((o bundle-op) (c t))
  t)

(defmethod operation-done-p ((o bundle-op) (c source-file))
  t)

(defmethod perform ((o bundle-op) (c system))
  (let* ((object-files (files-to-bundle c))
	 (output (output-files o c)))
    (ensure-directories-exist (first output))
    (when (bundle-op-static-library-p o)
      (apply #'c::builder :static-library (first output) :lisp-files object-files
	   (bundle-op-build-args o)))
    (when (bundle-op-fasl-p o)
      (apply #'c::builder :fasl (second output) :lisp-files object-files
	     (bundle-op-build-args o)))))

(defun bundle-system (system &rest args &key force (verbose t) version 
			     &allow-other-keys)
  (declare (ignore force verbose version))
  (apply #'operate 'bundle-op system args))

(export '(bundle-system))


;;;
;;; BUNDLED FILES
;;;
;;; This component can be used to distribute ASDF libraries in bundled
;;; form.
;;;

(defclass bundle (component) ())

(defmethod source-file-type ((c bundle) (s system))
  "fasb")

(defmethod perform ((o load-op) (c bundle))
  (load (component-pathname c)))

(defmethod perform (o (c bundle))
  (declare (ignore o))
  nil)

;;;
;;; Final integration steps
;;;

;;(push '("fasb" . si::load-binary) si::*load-hooks*)

(defvar *require-asdf-operator* 'load-op) ;; Why is this a global? JCB

(defun module-provide-asdf (name)
  (handler-bind ((style-warning #'muffle-warning))
    (let* ((*verbose-out* (make-broadcast-stream))
           (system (asdf:find-system name nil)))
      (declare (special *verbose-out*))
      (when system
        ;;(asdf:operate *require-asdf-operator* name)
	(load-system name)
        t))))

(defun register-pre-built-system (name)
  (register-system name (make-instance 'system :name name)))

(setf si::*module-provider-functions*
      (loop for f in si::*module-provider-functions*
	    unless (eq f 'module-provide-asdf)
	    collect #'(lambda (name)
			(let ((l (multiple-value-list (funcall f name))))
			  (and (first l) (register-pre-built-system name))
			  (values-list l)))))

#+windows (push '("asd" . si::load-source) si::*load-hooks*)
(pushnew 'module-provide-asdf ext:*module-provider-functions*)

(locally
 (declare (special *central-registry*))
 (push '(si:getcwd) *central-registry*)
 ;;(pushnew (translate-logical-pathname "SYS:") *central-registry*)
 (pushnew (translate-logical-pathname "CONTRIB:") *central-registry*)
 )

(eval-when (:compile-toplevel :load-toplevel :execute)
  (setq clos::*redefine-class-in-place* nil)  ;; Turn off ANSI class redefinition semantics.
  )

;; Make sure our default "optimize" settings are still in effect!
;; In MKCL a "declaim" has pervasive effect across separate invocations
;; of the compiler.
;;(declaim (optimize (speed 2) (debug 3) (safety 3)))
(proclaim c::*default-optimize-settings*)

(provide 'asdf)
