/* -*- mode: c -*- */
/*
    alloc_2.c -- Memory allocation based on the Boehmn GC.
*/
/*
    Copyright (c) 2001, Juan Jose Garcia Ripoll.
    Copyright (c) 2011, Jean-Claude Beaudoin.

    MKCL is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file '../../Copyright' for full details.
*/


#include <mkcl/mkcl.h>
#include <mkcl/mkcl-gc.h>
#include <mkcl/mkcl-inl.h>
#include <mkcl/internal.h>

#include <stdlib.h> /* for access to native malloc */


static void finalize_queued(void);

/**********************************************************
 *		OBJECT ALLOCATION			  *
 **********************************************************/

void
_mkcl_set_max_heap_size(MKCL, mkcl_index new_size)
{
  mkcl_interrupt_status old_intr;

  mkcl_get_interrupt_status(env, old_intr);
  mkcl_disable_interrupts(env);
  GC_set_max_heap_size(mkcl_core.max_heap_size = new_size);
  if (new_size == 0) {
    mkcl_index size = mkcl_get_option(MKCL_OPT_HEAP_SAFETY_AREA);
    mkcl_core.safety_region = mkcl_alloc_atomic_unprotected(size);
  } else if (mkcl_core.safety_region) {
    GC_FREE(mkcl_core.safety_region);
    mkcl_core.safety_region = 0;
  }
  mkcl_set_interrupt_status(env, old_intr);
}

static volatile bool failure;

static void *
out_of_memory_check(size_t requested_bytes)
{
  failure = true;
  return NULL;
}

static void
no_warnings(char *msg, GC_word arg)
{
}

static pthread_mutex_t oom_handler_lock;

#define OUT_OF_MEMORY_LOCK() if (pthread_mutex_lock(&oom_handler_lock)) mkcl_lose("Failed in OUT_OF_MEMORY_LOCK()")
#define OUT_OF_MEMORY_UNLOCK() if (pthread_mutex_unlock(&oom_handler_lock)) mkcl_lose("Failed in OUT_OF_MEMORY_UNLOCK()")


static void *
out_of_memory(size_t requested_bytes)
{
  mkcl_env env = MKCL_ENV();
  mkcl_interrupt_status old_intr;
  int method = 0;

  mkcl_get_interrupt_status(env, old_intr);
  mkcl_disable_interrupts(env);

  /* The out of memory condition may happen in more than one thread */
  /* But then we have to ensure the error has not been solved */
  OUT_OF_MEMORY_LOCK();

  failure = false;
  GC_gcollect();
  GC_oom_fn = out_of_memory_check;
  {
    void *output = GC_MALLOC(requested_bytes);
    GC_oom_fn = out_of_memory;
    if (output != 0 && failure == false) {
      OUT_OF_MEMORY_UNLOCK();
      mkcl_set_interrupt_status(env, old_intr);
      return output;
    }
  }
  if (mkcl_core.max_heap_size == 0) {
    /* We did not set any limit in the amount of memory,
     * yet we failed, or we had some limits but we have
     * not reached them. */
    if (mkcl_core.safety_region) {
      /* We can free some memory and try handling the error */
      GC_FREE(mkcl_core.safety_region);
      env->string_pool = mk_cl_Cnil;
      mkcl_core.safety_region = 0;
      method = 0;
    } else {
      /* No possibility of continuing */
      method = 2;
    }
  } else {
    mkcl_core.max_heap_size += mkcl_get_option(MKCL_OPT_HEAP_SAFETY_AREA);
    GC_set_max_heap_size(mkcl_core.max_heap_size);
    method = 1;
  }
  OUT_OF_MEMORY_UNLOCK();
  mkcl_set_interrupt_status(env, old_intr);

  switch (method) {
  case 0:
    mk_cl_error(env, 1, @'ext::storage-exhausted');
    break;
  case 1:
    mk_cl_cerror(env, 2,
		 mkcl_make_simple_base_string(env, "Extend heap size"),
		 @'ext::storage-exhausted');
    break;
  default:
    mkcl_lose("Memory exhausted, quitting program.");
    break;
  }

  MKCL_NO_INTR(env, GC_set_max_heap_size(mkcl_core.max_heap_size
					 += mkcl_core.max_heap_size / 2));
  {
    void * output;
    /* Default allocation. Note that we do not allocate atomic. */
    MKCL_NO_INTR(env, output = GC_MALLOC(requested_bytes));
    return output;
  }
}


static size_t type_size[mkcl_t_end];

mkcl_object
mkcl_alloc_object(MKCL, mkcl_type t)
{
  /* GC_MALLOC already resets objects */
  switch (t) {
  case mkcl_t_fixnum:
    return MKCL_MAKE_FIXNUM(0); /* Immediate fixnum */
  case mkcl_t_character:
    return MKCL_CODE_CHAR(' '); /* Immediate character */
#ifdef MKCL_LONG_FLOAT
  case mkcl_t_longfloat:
#endif
  case mkcl_t_singlefloat:
  case mkcl_t_doublefloat: {
    mkcl_object obj;
    MKCL_NO_INTR(env, obj = (mkcl_object)GC_MALLOC_ATOMIC(type_size[t]));
    obj->d.t = t;
    return obj;
  }
  case mkcl_t_bignum:
  case mkcl_t_ratio:
  case mkcl_t_complex:
  case mkcl_t_symbol:
  case mkcl_t_package:
  case mkcl_t_hashtable:
  case mkcl_t_array:
  case mkcl_t_vector:
  case mkcl_t_base_string:
  case mkcl_t_string:
  case mkcl_t_bitvector:
  case mkcl_t_stream:
  case mkcl_t_random:
  case mkcl_t_readtable:
  case mkcl_t_pathname:
  case mkcl_t_bytecode:
  case mkcl_t_bclosure:
  case mkcl_t_cfun:
  case mkcl_t_cclosure:
  case mkcl_t_instance:
  case mkcl_t_structure:
  case mkcl_t_thread:
  case mkcl_t_lock:
  case mkcl_t_rwlock:
  case mkcl_t_semaphore:
  case mkcl_t_condition_variable:
  case mkcl_t_foreign:
  case mkcl_t_codeblock: {
    mkcl_object obj;
    MKCL_NO_INTR(env, obj = (mkcl_object)GC_MALLOC(type_size[t]));
    obj->d.t = t;
    return obj;
  }
  default:
    printf("\ttype = %d\n", t);
    mkcl_lose("alloc botch.");
  }
}

mkcl_object mkcl_alloc_cdisplay(MKCL, mkcl_index nb_levels)
{
  mkcl_object cd;

  MKCL_NO_INTR(env, cd = (mkcl_object)GC_MALLOC(sizeof(struct mkcl_cdisplay) + nb_levels * sizeof(mkcl_object)));
  cd->display.t = mkcl_t_cdisplay;
  cd->display.nb_levels = nb_levels;
  return cd;
}

mkcl_object mkcl_alloc_clevel_block(MKCL, mkcl_object producer, const union mkcl_lispunion * const outer, const mkcl_index nb_vars)
{
  mkcl_object cb;

  MKCL_NO_INTR(env, cb = (mkcl_object)GC_MALLOC(sizeof(struct mkcl_clevel_block) + nb_vars * sizeof(mkcl_object)));
  cb->lblock.t = mkcl_t_clevel_block;
  cb->lblock.producer = producer;
  cb->lblock.outer = (mkcl_object) outer;
  cb->lblock.nb_vars = nb_vars;
  return cb;
}

mkcl_object
mkcl_alloc_compact_object(MKCL, mkcl_type t, mkcl_index extra_space)
{
  mkcl_index size = type_size[t];
  mkcl_object x;
  MKCL_NO_INTR(env, x = (mkcl_object)GC_MALLOC_ATOMIC(size + extra_space));
  x->array.t = t;
  x->array.displaced = (void*)(((char*)x) + size);
  return x;
}

mkcl_object
mkcl_cons(MKCL, mkcl_object a, mkcl_object d)
{
  struct mkcl_cons *obj;
  MKCL_NO_INTR(env, obj = GC_MALLOC(sizeof(struct mkcl_cons)));
  obj->t = mkcl_t_cons;
  obj->car = a;
  obj->cdr = d;
  return (mkcl_object)obj;
}

mkcl_object
mkcl_list1(MKCL, mkcl_object a)
{
  struct mkcl_cons *obj;
  MKCL_NO_INTR(env, obj = GC_MALLOC(sizeof(struct mkcl_cons)));
  obj->t = mkcl_t_cons;
  obj->car = a;
  obj->cdr = mk_cl_Cnil;
  return (mkcl_object)obj;
}

mkcl_object
mkcl_alloc_instance(MKCL, mkcl_index nb_slots)
{
  mkcl_object i;
  i = mkcl_alloc_object(env, mkcl_t_instance);
  i->instance.slots = (mkcl_object *)mkcl_alloc(env, sizeof(mkcl_object) * nb_slots);
  i->instance.length = nb_slots;
  i->instance.f.entry = mkcl_FEnot_funcallable_vararg;
  i->instance.f._[0] = mkcl_FEnot_funcallable_fixed;
  i->instance.f._[1] = mkcl_FEnot_funcallable_fixed;
  i->instance.f._[2] = mkcl_FEnot_funcallable_fixed;
  i->instance.f._[3] = mkcl_FEnot_funcallable_fixed;
  i->instance.f._[4] = mkcl_FEnot_funcallable_fixed;
  return i;
}

void *
mkcl_alloc_uncollectable(MKCL, size_t size)
{
  void *output;
  MKCL_NO_INTR(env, output = GC_MALLOC_UNCOLLECTABLE(size));
  return output;
}

void
mkcl_free_uncollectable(MKCL, void *pointer)
{
  MKCL_NO_INTR(env, GC_FREE(pointer));
}

void *
mkcl_alloc_unprotected(mkcl_index n)
{
  return GC_MALLOC_IGNORE_OFF_PAGE(n);
}

void *
mkcl_alloc_atomic_unprotected(mkcl_index n)
{
  return GC_MALLOC_ATOMIC_IGNORE_OFF_PAGE(n);
}

void *
mkcl_alloc(MKCL, mkcl_index n)
{
  void *output;
  MKCL_NO_INTR(env, output = mkcl_alloc_unprotected(n));
  return output;
}

void *
mkcl_alloc_atomic(MKCL, mkcl_index n)
{
  void *output;
  MKCL_NO_INTR(env, output = mkcl_alloc_atomic_unprotected(n));
  return output;
}

void
mkcl_dealloc(MKCL, void *ptr)
{
  /* Intentionally left empty! Broken on any platform after all... */
}

void *
mkcl_foreign_malloc(MKCL, size_t size)
{
  void *output;
  MKCL_NO_INTR(env, output = malloc(size));
  return output;
}

void
mkcl_foreign_free(MKCL, void *pointer)
{
  MKCL_NO_INTR(env, free(pointer));
}

static int alloc_initialized = FALSE;


extern void (*GC_push_other_roots)(void); /* Internal and private to Boehm's GC. */
extern void (*GC_start_call_back)(void);  /* Internal and private to Boehm's GC. */


static void (*old_GC_push_other_roots)(void);
static void stacks_scanner(void);

void
mkcl_init_alloc(void)
{
  int i;
  if (alloc_initialized) return;
  alloc_initialized = TRUE;

#ifdef MKCL_WINDOWS
  oom_handler_lock = CreateMutex(NULL, FALSE, NULL);
  if ( oom_handler_lock == NULL )
    mkcl_lose("mkcl_init_alloc failed to create out-of-memory handler lock.");
#else
  {
    pthread_mutexattr_t attr;

    if (pthread_mutexattr_init(&attr))
      mkcl_C_lose("mkcl_init_alloc failed on pthread_mutexattr_init.");
    if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK))
      mkcl_C_lose("mkcl_init_alloc failed on pthread_mutexattr_settype.");
    if (pthread_mutex_init(&oom_handler_lock, &attr))
      mkcl_C_lose("mkcl_init_alloc failed on pthread_mutex_init of out-of-memory handler lock.");
    pthread_mutexattr_destroy(&attr);
  }
#endif


  /*
   * Garbage collector restrictions: we set up the garbage collector
   * library to work as follows
   *
   * 1) The garbage collector shall not scan shared libraries
   *    explicitely.
   * 2) We only detect objects that are referenced by a pointer to
   *    the begining or to the first byte.
   * 3) Out of the incremental garbage collector, we only use the
   *    generational component.
   */
  GC_set_no_dls(0);
  GC_set_all_interior_pointers(0);
  GC_set_time_limit(GC_TIME_UNLIMITED);
  GC_init();

  GC_clear_roots();
  GC_disable();
  GC_set_max_heap_size(mkcl_core.max_heap_size = mkcl_get_option(MKCL_OPT_HEAP_SIZE));
  /* Save some memory for the case we get tight. */
  if (mkcl_core.max_heap_size == 0) {
    mkcl_index size = mkcl_get_option(MKCL_OPT_HEAP_SAFETY_AREA);
    mkcl_core.safety_region = mkcl_alloc_atomic_unprotected(size);
  } else if (mkcl_core.safety_region) {
    mkcl_core.safety_region = 0;
  }

#define init_tm(x,y,z) type_size[x] = (z)
  for (i = 0; i < mkcl_t_end; i++) {
    type_size[i] = 0;
  }
  init_tm(mkcl_t_singlefloat, "SINGLE-FLOAT", /* 8 */ sizeof(struct mkcl_singlefloat));
  init_tm(mkcl_t_cons, "CONS", sizeof(struct mkcl_cons)); /* 12 */
  init_tm(mkcl_t_doublefloat, "DOUBLE-FLOAT", /* 16 */ sizeof(struct mkcl_doublefloat));
  init_tm(mkcl_t_bytecode, "BYTECODE", sizeof(struct mkcl_bytecode));
  init_tm(mkcl_t_bclosure, "BCLOSURE", sizeof(struct mkcl_bclosure));
  init_tm(mkcl_t_base_string, "BASE-STRING", sizeof(struct mkcl_base_string)); /* 20 */
  init_tm(mkcl_t_string, "STRING", sizeof(struct mkcl_string));
  init_tm(mkcl_t_array, "ARRAY", sizeof(struct mkcl_array)); /* 24 */
  init_tm(mkcl_t_pathname, "PATHNAME", sizeof(struct mkcl_pathname)); /* 28 */
  init_tm(mkcl_t_symbol, "SYMBOL", sizeof(struct mkcl_symbol)); /* 32 */
  init_tm(mkcl_t_package, "PACKAGE", sizeof(struct mkcl_package)); /* 36 */
  init_tm(mkcl_t_codeblock, "CODEBLOCK", sizeof(struct mkcl_codeblock));
  init_tm(mkcl_t_bignum, "BIGNUM", sizeof(struct mkcl_bignum));
  init_tm(mkcl_t_ratio, "RATIO", sizeof(struct mkcl_ratio));
  init_tm(mkcl_t_complex, "COMPLEX", sizeof(struct mkcl_complex));
  init_tm(mkcl_t_hashtable, "HASH-TABLE", sizeof(struct mkcl_hashtable));
  init_tm(mkcl_t_vector, "VECTOR", sizeof(struct mkcl_vector));
  init_tm(mkcl_t_bitvector, "BIT-VECTOR", sizeof(struct mkcl_vector));
  init_tm(mkcl_t_stream, "STREAM", sizeof(struct mkcl_stream));
  init_tm(mkcl_t_random, "RANDOM-STATE", sizeof(struct mkcl_random));
  init_tm(mkcl_t_readtable, "READTABLE", sizeof(struct mkcl_readtable));
  init_tm(mkcl_t_cfun, "CFUN", sizeof(struct mkcl_cfun));
  init_tm(mkcl_t_cclosure, "CCLOSURE", sizeof(struct mkcl_cclosure));
  init_tm(mkcl_t_structure, "STRUCTURE", sizeof(struct mkcl_structure));
  init_tm(mkcl_t_instance, "INSTANCE", sizeof(struct mkcl_instance));
  init_tm(mkcl_t_foreign, "FOREIGN", sizeof(struct mkcl_foreign));
  init_tm(mkcl_t_thread, "THREAD", sizeof(struct mkcl_thread));
  init_tm(mkcl_t_lock, "LOCK", sizeof(struct mkcl_lock));
  init_tm(mkcl_t_rwlock, "RWLOCK", sizeof(struct mkcl_rwlock));
  init_tm(mkcl_t_semaphore, "SEMAPHORE", sizeof(struct mkcl_semaphore));
  init_tm(mkcl_t_condition_variable, "CONDITION-VARIABLE", sizeof(struct mkcl_condition_variable));
#ifdef MKCL_LONG_FLOAT
  init_tm(mkcl_t_longfloat, "LONG-FLOAT", sizeof(struct mkcl_long_float));
#endif

  old_GC_push_other_roots = GC_push_other_roots;
  GC_push_other_roots = stacks_scanner;
  GC_start_call_back = (void (*)(void))finalize_queued;
  GC_java_finalization = 1;
  GC_oom_fn = out_of_memory;
  GC_set_warn_proc(no_warnings);
}

/**********************************************************
 *		FINALIZATION				  *
 **********************************************************/

static void
standard_finalizer(MKCL, mkcl_object o)
{
  switch (o->d.t)
    {
    case mkcl_t_codeblock:
      mkcl_library_close(env, o);
      break;
    case mkcl_t_stream:
      mk_cl_close(env, 1, o);
      break;
    case mkcl_t_lock:
      {
#if defined(MKCL_WINDOWS)
	MKCL_NO_INTR(env, CloseHandle(o->lock.mutex));  /* FIXME! return status? JCB */
#else
	MKCL_NO_INTR(env, pthread_mutex_destroy(&o->lock.mutex));  /* FIXME! return status? JCB */
#endif
      }
      break;
    case mkcl_t_rwlock:
      {
#if defined(MKCL_WINDOWS)
	MKCL_NO_INTR(env, CloseHandle(o->rwlock.rwlock));  /* FIXME! return status? JCB */
#else
	MKCL_NO_INTR(env, pthread_rwlock_destroy(&o->rwlock.rwlock));  /* FIXME! return status? JCB */
#endif
      }
      break;
    case mkcl_t_semaphore:
      {
#if defined(MKCL_WINDOWS)
	MKCL_NO_INTR(env, CloseHandle(o->semaphore.sem));  /* FIXME! return status? JCB */
#else
	MKCL_NO_INTR(env, sem_destroy(&o->semaphore.sem));  /* FIXME! return status? JCB */
#endif
      }
      break;
    case mkcl_t_condition_variable:
      {
#if defined(MKCL_WINDOWS)
	MKCL_NO_INTR(env, CloseHandle(o->condition_variable.cv));  /* FIXME! return status? JCB */
#else
	MKCL_NO_INTR(env, pthread_cond_destroy(&o->condition_variable.cv));  /* FIXME! return status? JCB */
#endif
      }
      break;
    default:;
    }
}

static void
group_finalizer(void * obj, void * client_data)
{
  mkcl_object l = (mkcl_object) obj;
  /* mkcl_object no_data = (mkcl_object) client_data; */

  MKCL_NEWENV_BEGIN {
    while (MKCL_CONSP(l)) {
      mkcl_object record = MKCL_CONS_CAR(l);
      mkcl_object o = MKCL_CONS_CAR(record);
      mkcl_object procedure = MKCL_CONS_CDR(record);
      l = MKCL_CONS_CDR(l);
      if (procedure != mk_cl_Ct) {
	mk_cl_funcall(env, 2, procedure, o);
      }
      standard_finalizer(env, o);
    }
  } MKCL_NEWENV_END;
}

static void
queueing_finalizer(void * obj, void * client_data)
{
  mkcl_object o = (mkcl_object) obj;
  mkcl_object finalizer = (mkcl_object) client_data;

  if (finalizer != mk_cl_Cnil && finalizer != NULL) {
    /* Only nonstandard finalizers are queued */
    MKCL_NEWENV_BEGIN {
      if (finalizer == mk_cl_Ct) {
	standard_finalizer(env, o);
      } else {
	/* Note the way we do this: finalizers might
	   get executed as a consequence of these calls. */
	volatile mkcl_object aux = MKCL_ACONS(env, o, finalizer, mk_cl_Cnil);
	mkcl_object l = mkcl_core.to_be_finalized;
	if (mkcl_Null(l)) {
	  GC_finalization_proc ofn;
	  void *odata;
	  mkcl_core.to_be_finalized = aux;
	  MKCL_NO_INTR(env, GC_register_finalizer_no_order(aux, group_finalizer, NULL, &ofn, &odata));
	} else {
	  MKCL_RPLACD(l, aux);
	}
      }
    } MKCL_NEWENV_END;
  }
}

mkcl_object
mk_si_get_finalizer(MKCL, mkcl_object o)
{
  mkcl_object output;
  GC_finalization_proc ofn;
  void *odata;
  mkcl_interrupt_status old_intr;

  mkcl_get_interrupt_status(env, old_intr);
  mkcl_disable_interrupts(env);
  GC_register_finalizer_no_order(o, NULL, 0, &ofn, &odata);
  if (ofn == 0) {
    output = mk_cl_Cnil;
  } else if (ofn == queueing_finalizer) {
    output = (mkcl_object)odata;
  } else {
    output = mk_cl_Cnil;
  }
  GC_register_finalizer_no_order(o, ofn, odata, &ofn, &odata);
  mkcl_set_interrupt_status(env, old_intr);
  @(return output);
}

mkcl_object
mk_si_set_finalizer(MKCL, mkcl_object o, mkcl_object finalizer)
{
  GC_finalization_proc ofn;
  void *odata;

  if (finalizer == mk_cl_Cnil) {
    MKCL_NO_INTR(env, GC_register_finalizer_no_order(o, NULL, 0, &ofn, &odata));
  } else {
    MKCL_NO_INTR(env, GC_register_finalizer_no_order(o, queueing_finalizer, finalizer, &ofn, &odata));
  }
  @(return);
}

mkcl_object
mk_si_gc_stats(MKCL, mkcl_object enable)
{
  mkcl_object old_status = mkcl_core.gc_stats? mk_cl_Ct : mk_cl_Cnil;

  mkcl_core.gc_stats = (enable != mk_cl_Cnil);
  if (mkcl_core.bytes_consed == mk_cl_Cnil) {
    mkcl_core.bytes_consed = mkcl_alloc_object(env, mkcl_t_bignum);
    mpz_init2(mkcl_core.bytes_consed->big.big_num, 128);
    mkcl_core.gc_counter = mkcl_alloc_object(env, mkcl_t_bignum);
    mpz_init2(mkcl_core.gc_counter->big.big_num, 128);
  }

  {
    /* This used to be in the finalizer but
       calls to GC routines from the finalizer do not work
       anymore since 7.2alpha4. JCB
    */
    /* This is not accurate and may wrap around. We try
       to detect this assuming that an overflow in an
       unsigned integer will produce an smaller
       integer.*/
    static size_t bytes = 0;
    size_t new_bytes = GC_get_total_bytes();

    if (bytes > new_bytes) {
      mkcl_index before_wrap = ~((mkcl_index)0) - bytes;
#if MKCL_LONG_BITS >= MKCL_WORD_BITS
      _mkcl_big_add_ui(mkcl_core.bytes_consed, mkcl_core.bytes_consed, before_wrap);
#else
      {
	mkcl_object x = _mkcl_big_register0();
	_mkcl_big_set_fixnum(x, before_wrap);
	_mkcl_big_add(mkcl_core.bytes_consed, mkcl_core.bytes_consed, x);
      }
#endif
    }
#if MKCL_LONG_BITS >= MKCL_WORD_BITS
    _mkcl_big_add_ui(mkcl_core.bytes_consed, mkcl_core.bytes_consed, new_bytes - bytes);
#else
    {
      mkcl_object x = _mkcl_big_register0();
      _mkcl_big_set_fixnum(x, new_bytes - bytes);
      _mkcl_big_add(mkcl_core.bytes_consed, mkcl_core.bytes_consed, x);
    }
#endif
    bytes = new_bytes;
  }
  
  @(return
    _mkcl_big_register_normalize(env, mkcl_core.bytes_consed)
    _mkcl_big_register_normalize(env, mkcl_core.gc_counter)
    old_status);
}

mkcl_object
mk_si_mem_stats(MKCL)
{
  size_t heap_size = GC_get_heap_size();
  size_t free_bytes = GC_get_free_bytes();

  @(return
    mkcl_make_unsigned_integer(env, heap_size)
    mkcl_make_unsigned_integer(env, free_bytes)
    (GC_get_parallel() ? mk_cl_Ct : mk_cl_Cnil)
    );
}

size_t mkcl_GC_get_total_bytes(void)
{
  return GC_get_total_bytes();
}

/*
 * This procedure is invoked after garbage collection. It invokes
 * finalizers for all objects that are to be reclaimed by the
 * colector. Note that we cannot cons because this procedure is
 * invoked with the garbage collection lock on.
 */
static void finalize_queued(void)
{
  mkcl_core.to_be_finalized = mk_cl_Cnil;
  if (mkcl_core.gc_stats) {
    mpz_add_ui(mkcl_core.gc_counter->big.big_num, mkcl_core.gc_counter->big.big_num, 1);
  }
}


/**********************************************************
 *		GARBAGE COLLECTOR			  *
 **********************************************************/
/* These three are (rogue!?) imports from gc/include/private/gc_priv.h */
void GC_push_conditional(char *, char *, int);
void GC_set_mark_bit(char *);
void GC_push_all(char *, char *);

static void
mkcl_mark_env(MKCL)
{
  if (env->stack) {
    GC_push_conditional((void *)env->stack, (void *)env->stack_top, 1);
    GC_set_mark_bit((void *)env->stack);
  }
  if (env->frs_top) {
    GC_push_conditional((void *)env->frs_org, (void *)(env->frs_top+1), 1);
    GC_set_mark_bit((void *)env->frs_org);
  }
  if (env->bds_top) {
    GC_push_conditional((void *)env->bds_org, (void *)(env->bds_top+1), 1);
    GC_set_mark_bit((void *)env->bds_org);
  }

  GC_push_conditional((void *)env, (void *)(env + 1), 1);
  GC_set_mark_bit((void *)env);
}

static void
stacks_scanner(void)
{
  mkcl_object l;
  l = mkcl_core.libraries;
  if (l) {
    for (; l != mk_cl_Cnil; l = MKCL_CONS_CDR(l)) {
      mkcl_object dll = MKCL_CONS_CAR(l);
      if (dll->cblock.locked) {
	GC_push_conditional((void *)dll, (void *)(&dll->cblock + 1), 1);
	GC_set_mark_bit((void *)dll);
      }
    }
  }
  GC_push_all((void *)(&mkcl_core), (void *)(&mkcl_core + 1));
  GC_push_all((void *)mkcl_root_symbols, (void *)(mkcl_root_symbols + mkcl_root_symbols_count));
  /* 	l = mkcl_core.threads; */
  /* 	if (l == MKCL_OBJNULL) { */
  /* 		mkcl_mark_env(&cl_env); */
  /* 	} else { */
  l = mkcl_core.threads;
  mkcl_loop_for_on_unsafe(l) {
    mkcl_object thread = MKCL_CONS_CAR(l);
    const mkcl_env env = thread->thread.env;
    if (env != NULL)
      mkcl_mark_env(env);
  } mkcl_end_loop_for_on;
  /* 	} */
  if (old_GC_push_other_roots)
    (*old_GC_push_other_roots)();
}

/**********************************************************
 *		GARBAGE COLLECTION			  *
 **********************************************************/

void mkcl_register_root(MKCL, mkcl_object *p) /* Does anyone call this? JCB */
{
  MKCL_NO_INTR(env, GC_add_roots((char*)p, (char*)(p+1)));
}

@(defun si::gc (&optional area)
@
  MKCL_NO_INTR(env, GC_gcollect());
  @(return);
@)

mkcl_object
mk_si_gc_dump(MKCL)
{
  MKCL_NO_INTR(env, GC_dump());
  @(return);
}

mkcl_object
mk_si_gc_off(MKCL)
{
  GC_disable();
  @(return);
}

mkcl_object
mk_si_gc_on(MKCL)
{
  GC_enable();
  @(return);
}


/******************/

mkcl_object mkcl_alloc_pin_bag(MKCL)
{
  mkcl_object bag;

  MKCL_NO_INTR(env, bag = (mkcl_object)GC_MALLOC(sizeof(struct mkcl_pin_bag)));

  bag->pin_bag.t = mkcl_t_pin_bag;
  bag->pin_bag.pins = mk_cl_Cnil;
  return bag;
}

mkcl_object mkcl_pin(MKCL, mkcl_object bag, mkcl_object obj)
{
  mkcl_object pin;

  MKCL_NO_INTR(env, pin = (mkcl_object)GC_MALLOC(sizeof(struct mkcl_pin)));

  pin->pin.t = mkcl_t_pin;
  pin->pin.this = obj;
  pin->pin.bag = bag;
  pin->pin.left = bag;
  pin->pin.right = bag->pin_bag.pins;

  bag->pin_bag.pins->pin.left = pin;
  bag->pin_bag.pins = pin;

  return pin;
}


mkcl_object mkcl_unpin(MKCL, mkcl_object pin)
{

  if (mkcl_type_of(pin->pin.left) == mkcl_t_pin_bag)
    {
      mkcl_object bag = pin->pin.left;
      mkcl_object right = pin->pin.right;

      bag->pin_bag.pins = right;
      right->pin.left = bag;
    }
  else
    {
      mkcl_object left = pin->pin.left;
      mkcl_object right = pin->pin.right;

      left->pin.right = right;
      right->pin.left = left;
    }

  return pin->pin.this;
}

