/* -*- mode: c -*- */
/*
    ffi.c -- User defined data types and foreign functions interface.
*/
/*
    Copyright (c) 2001, Juan Jose Garcia Ripoll.
    Copyright (c) 2011, Jean-Claude Beaudoin.

    MKCL is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file '../../Copyright' for full details.
*/

#include <mkcl/mkcl.h>
#include <mkcl/internal.h>
#include <string.h>

/* Note that mkcl_foreign_type_table[] and mkcl_foreign_type_size[]
   must have matching contents otherwise mayhem will result!
   Both must also match the order of declarations inside enum mkcl_ffi_tag!
   JCB
*/

static const mkcl_object mkcl_foreign_type_table[] = {
	@':char',
	@':unsigned-char',
	@':byte',
	@':unsigned-byte',
	@':short',
	@':unsigned-short',
	@':int',
	@':unsigned-int',
	@':long',
	@':unsigned-long',
        @':int16-t',
        @':uint16-t',
        @':int32-t',
        @':uint32-t',
        @':int64-t',
        @':uint64-t',
        @':long-long',
        @':unsigned-long-long',
	@':pointer-void',
	@':cstring',
	@':object',
	@':float',
	@':double',
	@':long-double',
#if 0 /* We'll be C99 compliant one day! JCB */
	@':float-complex',
	@':double-complex',
	@':long-double-complex',
	@':float-imaginary',
	@':double-imaginary',
	@':long-double-imaginary',
#endif
	@':void'
};

static unsigned int mkcl_foreign_type_size[] = {
	sizeof(char),
	sizeof(unsigned char),
	sizeof(mkcl_int8_t),
	sizeof(mkcl_uint8_t),
	sizeof(short),
	sizeof(unsigned short),
	sizeof(int),
	sizeof(unsigned int),
	sizeof(long),
	sizeof(unsigned long),
        sizeof(mkcl_int16_t),
        sizeof(mkcl_uint16_t),
        sizeof(mkcl_int32_t),
        sizeof(mkcl_uint32_t),
        sizeof(mkcl_int64_t),
        sizeof(mkcl_uint64_t),
        sizeof(long long),
        sizeof(unsigned long long),
	sizeof(void *),
	sizeof(char *),
	sizeof(mkcl_object),
	sizeof(float),
	sizeof(double),
	sizeof(long double),
#if 0 /* We'll be C99 compliant one day! JCB */
	sizeof(float _Complex),
	sizeof(double _Complex),
	sizeof(long double _Complex),
	sizeof(float _Imaginary),
	sizeof(double _Imaginary),
	sizeof(long double _Imaginary),
#endif
	0 /* sizeof(void) */
};

static const mkcl_object mkcl_foreign_cc_table[] = {
	@':cdecl',
	@':stdcall'
};

mkcl_object
mkcl_make_foreign_data(MKCL, mkcl_object tag, mkcl_index size, void *data)
{
  mkcl_object output = mkcl_alloc_object(env, mkcl_t_foreign);
  output->foreign.tag = tag == mk_cl_Cnil ? @':void' : tag;
  output->foreign.size = size;
  output->foreign.data = (char*)data;
  return output;
}

mkcl_object
mkcl_allocate_foreign_data(MKCL, mkcl_object tag, mkcl_index size)
{
  mkcl_object output = mkcl_alloc_object(env, mkcl_t_foreign);
  output->foreign.tag = tag;
  output->foreign.size = size;
  output->foreign.data = (char*)mkcl_alloc_atomic(env, size);
  return output;
}

void *
mkcl_foreign_data_pointer_safe(MKCL, mkcl_object f)
{
  if (mkcl_type_of(f) != mkcl_t_foreign)
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', f);
  return f->foreign.data;
}

char *
mkcl_base_string_pointer_safe(MKCL, mkcl_object f)
{
  unsigned char *s;
  /* FIXME! Is there a better function name? */
  f = mkcl_check_cl_type(env, @'si::make-foreign-data-from-array', f, mkcl_t_base_string);
  s = f->base_string.self;
  if (f->base_string.hasfillp && s[f->base_string.fillp] != 0) {
    mkcl_FEerror(env, "Cannot coerce a string with fill pointer to (char *)", 0);
  }
  return (char *)s;
}

mkcl_object
mkcl_null_terminated_base_string(MKCL, mkcl_object f)
{
  /* FIXME! Is there a better function name? */
  f = mkcl_check_cl_type(env, @'si::make-foreign-data-from-array', f, mkcl_t_base_string);
  if (f->base_string.hasfillp && f->base_string.self[f->base_string.fillp] != 0) {
    return mk_cl_copy_seq(env, f);
  } else {
    return f;
  }
}

mkcl_object
mk_si_allocate_foreign_data(MKCL, mkcl_object tag, mkcl_object size)
{
  mkcl_object output = mkcl_alloc_object(env, mkcl_t_foreign);
  mkcl_index bytes = mkcl_integer_to_index(env, size);

  output->foreign.tag = tag;
  output->foreign.size = bytes;
  /* FIXME! Should be atomic uncollectable or malloc, but we do not export
   * that garbage collector interface and malloc may be overwritten
   * by the GC library */
  output->foreign.data = bytes ? mkcl_foreign_malloc(env, bytes) : NULL;
  @(return output);
}

mkcl_object
mk_si_free_foreign_data(MKCL, mkcl_object f)
{
  if (mkcl_type_of(f) != mkcl_t_foreign) {
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', f);
  }
  if (f->foreign.size) {
    /* See mk_si_allocate_foreign_data() */
    mkcl_foreign_free(env, f->foreign.data);
  }
  f->foreign.size = 0;
  f->foreign.data = NULL;
  @(return );
}

mkcl_object
mk_si_make_foreign_data_from_array(MKCL, mkcl_object array)
{
  mkcl_object tag = mk_cl_Cnil;
  if (mkcl_type_of(array) != mkcl_t_array && mkcl_type_of(array) != mkcl_t_vector) {
    mkcl_FEwrong_type_argument(env, @'array', array);
  }
  switch (array->array.elttype) {
  case mkcl_aet_sf: tag = @':float'; break;
  case mkcl_aet_df: tag = @':double'; break;
#if MKCL_WORD_BITS > MKCL_LONG_BITS
  case mkcl_aet_word: tag = @':long-long'; break;
  case mkcl_aet_index: tag = @':unsigned-long-long'; break;
#else
  case mkcl_aet_word: tag = @':long'; break;
  case mkcl_aet_index: tag = @':unsigned-long'; break;
#endif
  default:
    mkcl_FEerror(env, "Cannot make foreign object from array with element type ~S.", 1, mkcl_elttype_to_symbol(env, array->array.elttype));
    break;
  }
  @(return mkcl_make_foreign_data(env, tag, 0, array->array.self.bc));
}

mkcl_object
mk_si_foreign_data_address(MKCL, mkcl_object f)
{
  if (mkcl_type_of(f) != mkcl_t_foreign) {
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', f);
  }
  @(return mkcl_make_unsigned_integer(env, (mkcl_index)f->foreign.data));
}

mkcl_object
mk_si_foreign_data_tag(MKCL, mkcl_object f)
{
  if (mkcl_type_of(f) != mkcl_t_foreign) {
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', f);
  }
  @(return f->foreign.tag);
}

mkcl_object
mk_si_foreign_data_pointer(MKCL, mkcl_object f, mkcl_object andx, mkcl_object asize,
			mkcl_object tag)
{
  mkcl_index ndx = mkcl_integer_to_index(env, andx);
  mkcl_index size = mkcl_integer_to_index(env, asize);
  mkcl_object output;

  if (mkcl_type_of(f) != mkcl_t_foreign) {
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', f);
  }
  if (ndx >= f->foreign.size || (f->foreign.size - ndx) < size) {
    mkcl_FEerror(env, "Out of bounds reference into foreign data type ~A.", 1, f);
  }
  output = mkcl_alloc_object(env, mkcl_t_foreign);
  output->foreign.tag = tag;
  output->foreign.size = size;
  output->foreign.data = f->foreign.data + ndx;
  @(return output);
}

mkcl_object
mk_si_foreign_data_ref(MKCL, mkcl_object f, mkcl_object andx, mkcl_object asize, mkcl_object tag)
{
  mkcl_index ndx = mkcl_integer_to_index(env, andx);
  mkcl_index size = mkcl_integer_to_index(env, asize);
  mkcl_object output;

  if (mkcl_type_of(f) != mkcl_t_foreign) {
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', f);
  }
  if (ndx >= f->foreign.size || (f->foreign.size - ndx) < size) {
    mkcl_FEerror(env, "Out of bounds reference into foreign data type ~A.", 1, f);
  }
  output = mkcl_allocate_foreign_data(env, tag, size);
  memcpy(output->foreign.data, f->foreign.data + ndx, size);
  @(return output);
}

mkcl_object
mk_si_foreign_data_set(MKCL, mkcl_object f, mkcl_object andx, mkcl_object value)
{
  mkcl_index ndx = mkcl_integer_to_index(env, andx);
  mkcl_index size, limit;

  if (mkcl_type_of(f) != mkcl_t_foreign) {
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', f);
  }
  if (mkcl_type_of(value) != mkcl_t_foreign) {
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', value);
  }
  size = value->foreign.size;
  limit = f->foreign.size;
  if (ndx >= limit || (limit - ndx) < size) {
    mkcl_FEerror(env, "Out of bounds reference into foreign data type ~A.", 1, f);
  }
  memcpy(f->foreign.data + ndx, value->foreign.data, size);
  @(return value);
}

enum mkcl_ffi_tag
mkcl_foreign_type_code(MKCL, mkcl_object type)
{
  int i;
  for (i = 0; i <= MKCL_FFI_VOID; i++) {
    if (type == mkcl_foreign_type_table[i])
      return (enum mkcl_ffi_tag)i;
  }
  mkcl_FEerror(env, "~A does not denote an elementary foreign type.", 1, type);
  return MKCL_FFI_VOID;
}

enum mkcl_ffi_calling_convention
mkcl_foreign_cc_code(MKCL, mkcl_object cc)
{
  int i;
  for (i = 0; i <= MKCL_FFI_CC_STDCALL; i++) {
    if (cc == mkcl_foreign_cc_table[i])
      return (enum mkcl_ffi_calling_convention)i;
  }
  mkcl_FEerror(env, "~A does no denote a valid calling convention.", 1, cc);
  return MKCL_FFI_CC_CDECL;
}

mkcl_object
mkcl_foreign_data_ref_elt(MKCL, void *p, enum mkcl_ffi_tag tag)
{
  switch (tag) {
  case MKCL_FFI_CHAR:
    return MKCL_CODE_CHAR(*(char *)p);
  case MKCL_FFI_UNSIGNED_CHAR:
    return MKCL_CODE_CHAR(*(unsigned char *)p);
  case MKCL_FFI_BYTE:
    return MKCL_MAKE_FIXNUM(*(int8_t *)p);
  case MKCL_FFI_UNSIGNED_BYTE:
    return MKCL_MAKE_FIXNUM(*(uint8_t *)p);
  case MKCL_FFI_SHORT:
    return MKCL_MAKE_FIXNUM(*(short *)p);
  case MKCL_FFI_UNSIGNED_SHORT:
    return MKCL_MAKE_FIXNUM(*(unsigned short *)p);
  case MKCL_FFI_INT:
    return mkcl_make_integer(env, *(int *)p);
  case MKCL_FFI_UNSIGNED_INT:
    return mkcl_make_unsigned_integer(env, *(unsigned int *)p);
  case MKCL_FFI_LONG:
    return mkcl_make_integer(env, *(long *)p);
  case MKCL_FFI_INT16_T:
    return mkcl_make_int16_t(env, *(mkcl_int16_t *)p);
  case MKCL_FFI_UINT16_T:
    return mkcl_make_uint16_t(env, *(mkcl_uint16_t *)p);
  case MKCL_FFI_INT32_T:
    return mkcl_make_int32_t(env, *(mkcl_int32_t *)p);
  case MKCL_FFI_UINT32_T:
    return mkcl_make_uint32_t(env, *(mkcl_uint32_t *)p);
  case MKCL_FFI_INT64_T:
    return mkcl_make_int64_t(env, *(mkcl_int64_t *)p);
  case MKCL_FFI_UINT64_T:
    return mkcl_make_uint64_t(env, *(mkcl_uint64_t *)p);
  case MKCL_FFI_LONG_LONG:
    return mkcl_make_long_long(env, *(mkcl_long_long_t *)p);
  case MKCL_FFI_UNSIGNED_LONG_LONG:
    return mkcl_make_ulong_long(env, *(mkcl_ulong_long_t *)p);
  case MKCL_FFI_UNSIGNED_LONG:
    return mkcl_make_unsigned_integer(env, *(unsigned long *)p);
  case MKCL_FFI_POINTER_VOID:
    return mkcl_make_foreign_data(env, @':pointer-void', 0, *(void **)p);
  case MKCL_FFI_CSTRING:
    return *(char **)p ? mkcl_make_simple_base_string(env, *(char **)p) : mk_cl_Cnil;
  case MKCL_FFI_OBJECT:
    return *(mkcl_object *)p;
  case MKCL_FFI_FLOAT:
    return mkcl_make_singlefloat(env, *(float *)p);
  case MKCL_FFI_DOUBLE:
    return mkcl_make_doublefloat(env, *(double *)p);
  case MKCL_FFI_LONG_DOUBLE:
#ifdef MKCL_LONG_FLOAT
    return mkcl_make_longfloat(env, *(long double *)p);
#else
    return mkcl_make_doublefloat(env, *(long double *)p);
#endif
  case MKCL_FFI_VOID:
    return mk_cl_Cnil;
  default:
    mkcl_lose("Unknown foreign data type tag received in mkcl_foreign_data_ref_elt");
  }
}

void
mkcl_foreign_data_set_elt(MKCL, void *p, enum mkcl_ffi_tag tag, mkcl_object value)
{
  switch (tag) {
  case MKCL_FFI_CHAR:
    *(char *)p = (char)mkcl_base_char_code(env, value);
    break;
  case MKCL_FFI_UNSIGNED_CHAR:
    *(unsigned char*)p = (unsigned char)mkcl_base_char_code(env, value);
    break;
  case MKCL_FFI_BYTE:
    *(int8_t *)p = mkcl_integer_to_word(env, value);
    break;
  case MKCL_FFI_UNSIGNED_BYTE:
    *(uint8_t *)p = mkcl_integer_to_index(env, value);
    break;
  case MKCL_FFI_SHORT:
    *(short *)p = mkcl_integer_to_word(env, value);
    break;
  case MKCL_FFI_UNSIGNED_SHORT:
    *(unsigned short *)p = mkcl_integer_to_index(env, value);
    break;
  case MKCL_FFI_INT:
    *(int *)p = mkcl_integer_to_word(env, value);
    break;
  case MKCL_FFI_UNSIGNED_INT:
    *(unsigned int *)p = mkcl_integer_to_index(env, value);
    break;
  case MKCL_FFI_LONG:
    *(long *)p = mkcl_integer_to_word(env, value);
    break;
  case MKCL_FFI_UNSIGNED_LONG:
    *(unsigned long *)p = mkcl_integer_to_index(env, value);
    break;
  case MKCL_FFI_INT16_T:
    *(mkcl_int16_t *)p = mkcl_to_int16_t(env, value);
    break;
  case MKCL_FFI_UINT16_T:
    *(mkcl_uint16_t *)p = mkcl_to_uint16_t(env, value);
    break;
  case MKCL_FFI_INT32_T:
    *(mkcl_int32_t *)p = mkcl_to_int32_t(env, value);
    break;
  case MKCL_FFI_UINT32_T:
    *(mkcl_uint32_t *)p = mkcl_to_uint32_t(env, value);
    break;
  case MKCL_FFI_INT64_T:
    *(mkcl_int64_t *)p = mkcl_to_int64_t(env, value);
    break;
  case MKCL_FFI_UINT64_T:
    *(mkcl_uint64_t *)p = mkcl_to_uint64_t(env, value);
    break;
  case MKCL_FFI_LONG_LONG:
    *(mkcl_long_long_t *)p = mkcl_to_long_long(env, value);
    break;
  case MKCL_FFI_UNSIGNED_LONG_LONG:
    *(mkcl_ulong_long_t *)p = mkcl_to_ulong_long(env, value);
    break;
  case MKCL_FFI_POINTER_VOID:
    *(void **)p = mkcl_foreign_data_pointer_safe(env, value);
    break;
  case MKCL_FFI_CSTRING:
    *(char **)p = value == mk_cl_Cnil ? NULL : (char*)value->base_string.self;
    break;
  case MKCL_FFI_OBJECT:
    *(mkcl_object *)p = value;
    break;
  case MKCL_FFI_FLOAT:
    *(float *)p = mkcl_to_float(env, value);
    break;
  case MKCL_FFI_DOUBLE:
    *(double *)p = mkcl_to_double(env, value);
    break;
  case MKCL_FFI_LONG_DOUBLE:
#ifdef MKCL_LONG_FLOAT
    *(long double *)p = mkcl_to_long_double(env, value);
#else
    *(long double *)p = mkcl_to_double(env, value);
#endif
    break;
  case MKCL_FFI_VOID:
    break;
  default:
    mkcl_lose("Unknown foreign data type tag received in mkcl_foreign_data_set_elt");
  }
}

mkcl_object
mk_si_foreign_data_ref_elt(MKCL, mkcl_object f, mkcl_object andx, mkcl_object type)
{
  mkcl_index ndx = mkcl_integer_to_index(env, andx);
  mkcl_index limit = f->foreign.size;
  enum mkcl_ffi_tag tag = mkcl_foreign_type_code(env, type);
  if (ndx >= limit || (ndx + mkcl_foreign_type_size[tag] > limit)) {
    mkcl_FEerror(env, "Out of bounds reference into foreign data type ~A.", 1, f);
  }
  if (mkcl_type_of(f) != mkcl_t_foreign) {
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', f);
  }
  @(return mkcl_foreign_data_ref_elt(env, (void*)(f->foreign.data + ndx), tag));
}

mkcl_object
mk_si_foreign_data_set_elt(MKCL, mkcl_object f, mkcl_object andx, mkcl_object type, mkcl_object value)
{
  mkcl_index ndx = mkcl_integer_to_index(env, andx);
  mkcl_index limit = f->foreign.size;
  enum mkcl_ffi_tag tag = mkcl_foreign_type_code(env, type);
  if (ndx >= limit || ndx + mkcl_foreign_type_size[tag] > limit) {
    mkcl_FEerror(env, "Out of bounds reference into foreign data type ~A.", 1, f);
  }
  if (mkcl_type_of(f) != mkcl_t_foreign) {
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', f);
  }
  mkcl_foreign_data_set_elt(env, (void*)(f->foreign.data + ndx), tag, value);
  @(return value);
}

mkcl_object
mk_si_size_of_foreign_elt_type(MKCL, mkcl_object type)
{
  enum mkcl_ffi_tag tag = mkcl_foreign_type_code(env, type);
  @(return MKCL_MAKE_FIXNUM(mkcl_foreign_type_size[tag]));
}

mkcl_object
mk_si_null_pointer_p(MKCL, mkcl_object f)
{
  if (mkcl_type_of(f) != mkcl_t_foreign)
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', f);
  @(return ((f->foreign.data == NULL)? mk_cl_Ct : mk_cl_Cnil));
}

mkcl_object
mk_si_foreign_data_recast(MKCL, mkcl_object f, mkcl_object size, mkcl_object tag)
{
  if (mkcl_type_of(f) != mkcl_t_foreign)
    mkcl_FEwrong_type_argument(env, @'si::foreign-data', f);
  f->foreign.size = mkcl_integer_to_index(env, size);
  f->foreign.tag = tag;
  @(return f);
}

mkcl_object
mk_si_load_foreign_module(MKCL, mkcl_object filename)
{
  mkcl_object output = mk_cl_Cnil;
  mkcl_object l_c_lock = mkcl_symbol_value(env, @'mt::+load-compile-lock+');
  volatile mkcl_object locked = mk_cl_Cnil;

  MKCL_UNWIND_PROTECT_BEGIN(env) {
    mkcl_interrupt_status old_intr;

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    locked = mk_mt_get_lock(env, 1, l_c_lock);
    mkcl_set_interrupt_status(env, old_intr);

    output = mkcl_library_open(env, filename, 0);
    if (output->cblock.handle == NULL) {
      mkcl_object msg = mkcl_library_error(env, output);
      mkcl_library_close(env, output);
      output = msg;
    }
    ; /* MSVC complains about missing ';' before '}' */
  } MKCL_UNWIND_PROTECT_EXIT {
    if (!mkcl_Null(locked)) mk_mt_giveup_lock(env, l_c_lock);
  } MKCL_UNWIND_PROTECT_END;
  if (mkcl_type_of(output) != mkcl_t_codeblock) {
    mkcl_FEerror(env, "LOAD-FOREIGN-MODULE: Could not load foreign module ~S (Error: ~S)", 2, filename, output);
  }
  output->cblock.locked |= 1;
  @(return output);
}

mkcl_object
mk_si_find_foreign_symbol(MKCL, mkcl_object var, mkcl_object module, mkcl_object type, mkcl_object size)
{
  mkcl_object block;
  mkcl_object output = mk_cl_Cnil;
  void *sym;

  block = (module == @':default' ? module : mk_si_load_foreign_module(env, module));
  var = mkcl_null_terminated_base_string(env, var);
  sym = mkcl_library_symbol(env, block, (char*)var->base_string.self, 1);
  if (sym == NULL) {
      output = mkcl_library_error(env, block);
    goto OUTPUT;
  }
  output = mkcl_make_foreign_data(env, type, mkcl_integer_to_index(env, size), sym);
 OUTPUT:
  if (mkcl_type_of(output) != mkcl_t_foreign)
    mkcl_FEerror(env, "FIND-FOREIGN-SYMBOL: Could not load foreign symbol ~S from module ~S (Error: ~S)", 3, var, module, output);
  @(return output);
}

static void
mkcl_fficall_overflow(MKCL)
{
  mkcl_FEerror(env, "Stack overflow on SI:CALL-CFUN", 0);
}

void
mkcl_fficall_prepare(MKCL, mkcl_object return_type, mkcl_object arg_type, mkcl_object cc_type)
{
  struct mkcl_fficall *fficall = env->fficall;
  fficall->buffer_sp = fficall->buffer;
  fficall->buffer_size = 0;
  fficall->cstring = mk_cl_Cnil;
  fficall->cc = mkcl_foreign_cc_code(env, cc_type);
  fficall->registers = mkcl_fficall_prepare_extra(env, fficall->registers);
}

void
mkcl_fficall_push_bytes(MKCL, void *data, size_t bytes)
{
  struct mkcl_fficall *fficall = env->fficall;
  fficall->buffer_size += bytes;
  if (fficall->buffer_size >= MKCL_FFICALL_LIMIT)
    mkcl_fficall_overflow(env);
  memcpy(fficall->buffer_sp, (char*)data, bytes);
  fficall->buffer_sp += bytes;
}

void
mkcl_fficall_push_int(MKCL, int data)
{
  mkcl_fficall_push_bytes(env, &data, sizeof(int));
}

void
mkcl_fficall_align(MKCL, int data)
{ /* Note that the value of 'data' must be a power of 2 for this code to work. */
  struct mkcl_fficall *fficall = env->fficall;
  if (data == 1)
    return;
  else {
    size_t sp = fficall->buffer_sp - fficall->buffer;
    size_t mask = data - 1;
    size_t new_sp = (sp + mask) & ~mask;

    if (new_sp >= MKCL_FFICALL_LIMIT)
      mkcl_fficall_overflow(env);
    fficall->buffer_sp = fficall->buffer + new_sp;
    fficall->buffer_size = new_sp;
  }
}

@(defun si::call-cfun (fun return_type arg_types args &optional (cc_type @':cdecl'))
  struct mkcl_fficall *fficall = env->fficall;
  void *cfun = mkcl_foreign_data_pointer_safe(env, fun);
  mkcl_object object;
  enum mkcl_ffi_tag return_type_tag = mkcl_foreign_type_code(env, return_type);
@

  mkcl_fficall_prepare(env, return_type, arg_types, cc_type);
  while (MKCL_CONSP(arg_types)) {
    enum mkcl_ffi_tag type;
    if (!MKCL_CONSP(args)) {
      mkcl_FEerror(env, "In SI:CALL-CFUN, mismatch between argument types and argument list: ~A vs ~A", 0);
    }
    type = mkcl_foreign_type_code(env, MKCL_CAR(arg_types));
    if (type == MKCL_FFI_CSTRING) {
      object = mkcl_null_terminated_base_string(env, MKCL_CAR(args));
      if (MKCL_CAR(args) != object)
	fficall->cstring =
	  MKCL_CONS(env, object, fficall->cstring);
    } else {
      object = MKCL_CAR(args);
    }
    mkcl_foreign_data_set_elt(env, &fficall->output, type, object);
    mkcl_fficall_push_arg(env, &fficall->output, type);
    arg_types = MKCL_CDR(arg_types);
    args = MKCL_CDR(args);
  }
  mkcl_fficall_execute(env, cfun, fficall, return_type_tag);
  object = mkcl_foreign_data_ref_elt(env, &fficall->output, return_type_tag);
  
  fficall->buffer_size = 0;
  fficall->buffer_sp = fficall->buffer;
  fficall->cstring = mk_cl_Cnil;
  
  @(return object);
@)

@(defun si::make-dynamic-callback (fun sym rtype argtypes &optional (cctype @':cdecl'))
	mkcl_object data;
	mkcl_object cbk;
@
  data = mk_cl_list(env, 3, fun, rtype, argtypes);
  cbk  = mkcl_make_foreign_data(env, @':pointer-void', 0,
				mkcl_dynamic_callback_make(env, data, mkcl_foreign_cc_code(env, cctype)));

  mk_si_put_sysprop(env, sym, @':callback', MKCL_CONS(env, cbk, data));
  @(return cbk);
@)

