/* -*- mode: c -*- */
/*
    load.d -- Binary loader.
*/
/*
    Copyright (c) 1990, Giuseppe Attardi and William F. Schelter.
    Copyright (c) 2001, Juan Jose Garcia Ripoll.
    Copyright (c) 2011, Jean-Claude Beaudoin.

    MKCL is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file '../../Copyright' for full details.
*/

#include <mkcl/mkcl.h>
#include <string.h>
#include <stdio.h>

#ifdef HAVE_DLFCN_H
# include <dlfcn.h>
# define INIT_PREFIX "mkcl_init_fas_"
#endif

#ifdef HAVE_LINK_H
# include <link.h>
#endif

#if defined(MKCL_WINDOWS)
# include <windows.h>
# include <windef.h>
# include <winbase.h>
# include <tlhelp32.h>
# define INIT_PREFIX "mkcl_init_fas_"
#else
# include <unistd.h>
#endif

#include <mkcl/mkcl-inl.h>
#include <mkcl/internal.h>


static mkcl_object
copy_object_file(MKCL, mkcl_object original)
{
  mkcl_interrupt_status old_intr;
  int err;
  mkcl_object copy = mkcl_make_simple_base_string(env, "TMP:MKCL");
  copy = mk_si_coerce_to_filename(env, mk_si_mkstemp(env, copy));

  mkcl_get_interrupt_status(env, old_intr);
  mkcl_disable_interrupts(env);
  /*
   * We either have to make a full copy to convince the loader to load this object
   * file again, or we want to retain the possibility of overwriting the object
   * file we load later on (case of Windows, which locks files that are loaded).
   * The symlinks do not seem to work in latest versions of Linux.
   */
#if defined(MKCL_WINDOWS)
  err = !CopyFile((char *) original->base_string.self, (char *) copy->base_string.self, 0);
#else
  err = mkcl_Null(mk_si_copy_file(env, original, copy));
#endif
  mkcl_set_interrupt_status(env, old_intr);

  if (err) {
    mk_cl_delete_file(env, copy);
    mkcl_FEerror(env, "Unable to copy file ~A to ~A", 2, original, copy);
  }
  return copy;
}

static mkcl_object
mkcl_library_find_by_name(MKCL, mkcl_object filename)
{
  mkcl_object l;
  for (l = mkcl_core.libraries; l != mk_cl_Cnil; l = MKCL_CONS_CDR(l)) {
    mkcl_object other = MKCL_CONS_CAR(l);
    mkcl_object name = other->cblock.name;
    if (!mkcl_Null(name) && mkcl_string_E(env, name, filename)) {
      return other;
    }
  }
  return mk_cl_Cnil;
}

static mkcl_object
mkcl_library_find_by_handle(void *handle)
{
  mkcl_object l;
  for (l = mkcl_core.libraries; l != mk_cl_Cnil; l = MKCL_CONS_CDR(l)) {
    mkcl_object other = MKCL_CONS_CAR(l);
    if (handle == other->cblock.handle) {
      return other;
    }
  }
  return mk_cl_Cnil;
}

mkcl_object
mkcl_library_open(MKCL, mkcl_object filename, bool force_reload) 
{
  mkcl_object block;
  bool self_destruct = 0;
  char *filename_string;

  /* Coerces to a file name but does not merge with cwd */
  filename = mk_si_coerce_to_filename(env, filename);

  if (!force_reload) {
    /* When loading a foreign library, such as a dll or a
     * so, it cannot contain any executable top level
     * code. In that case force_reload=0 and there is no
     * need to reload it if it has already been loaded. */
    block = mkcl_library_find_by_name(env, filename);
    if (!mkcl_Null(block)) {
      return block;
    }
  } else {
    /* We are using shared libraries as modules and
     * force_reload=1.  Here we have to face the problem
     * that many operating systems do not allow to load a
     * shared library twice, even if it has changed. Hence
     * we have to make a unique copy to be able to load
     * the same FASL twice. In Windows this copy is
     * _always_ made because otherwise it cannot be
     * overwritten. In Unix we need only do that when the
     * file has been previously loaded. */
#if defined(MKCL_WINDOWS)
    filename = copy_object_file(env, filename);
#else
    block = mkcl_library_find_by_name(env, filename);
    if (!mkcl_Null(block)) {
      filename = copy_object_file(env, filename);
      self_destruct = 1;
    }
#endif
  }
  block = mkcl_alloc_object(env, mkcl_t_codeblock);
  block->cblock.self_destruct = self_destruct;
  block->cblock.locked = 0;
  block->cblock.handle = NULL;
  block->cblock.entry = NULL;
  block->cblock.data = NULL;
  block->cblock.data_size = 0;
  block->cblock.temp_data = NULL;
  block->cblock.temp_data_size = 0;
  block->cblock.data_text = NULL;
  block->cblock.data_text_size = 0;
  block->cblock.name = filename;
  block->cblock.next = mk_cl_Cnil;
  block->cblock.links = mk_cl_Cnil;
  block->cblock.cfuns_size = 0;
  block->cblock.cfuns = NULL;
  block->cblock.source = mk_cl_Cnil;
  block->cblock.fun_ref_syms = NULL;
  block->cblock.fun_refs = NULL;
  block->cblock.cfun_objs = NULL;
  filename_string = (char*)filename->base_string.self;

#ifdef HAVE_DLFCN_H
  MKCL_NO_INTR(env, block->cblock.handle = dlopen(filename_string, RTLD_NOW|RTLD_GLOBAL));
  if (block->cblock.handle == NULL)
    {
      return block;
    }
#elif defined(MKCL_WINDOWS)
  {
    mkcl_interrupt_status old_intr;

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    {
      UINT error = SetErrorMode(SetErrorMode(0) || SEM_NOOPENFILEERRORBOX);
      block->cblock.handle = LoadLibrary(filename_string);
      SetErrorMode(error);
    }
    mkcl_set_interrupt_status(env, old_intr);

    if (block->cblock.handle == NULL)
      {
	return block;
      }
  }
#else
# error "mkcl_library_open is imcomplete on this platform."
#endif

  /*
   * A second pass to ensure that the dlopen routine has not
   * returned a library that we had already loaded. If this is
   * the case, we close the new copy to ensure we do refcounting
   * right.
   *
   * INV: We can modify "libraries" in a multithread environment
   * because we have already taken the +load-compile-lock+
   */
  {
    mkcl_object other = mkcl_library_find_by_handle(block->cblock.handle);
    if (other != mk_cl_Cnil) {
      mkcl_library_close(env, block);
      block = other;
    } else {
      mk_si_set_finalizer(env, block, mk_cl_Ct);
      mkcl_core.libraries = MKCL_CONS(env, block, mkcl_core.libraries);
    }
  }
  return block;
}

void *
mkcl_library_symbol(MKCL, mkcl_object block, const char *symbol, bool lock) 
{
  void *p;

  if (block == @':default') {
    mkcl_object l;
    for (l = mkcl_core.libraries; l != mk_cl_Cnil; l = MKCL_CONS_CDR(l)) {
      mkcl_object block = MKCL_CONS_CAR(l);
      p = mkcl_library_symbol(env, block, symbol, lock);
      if (p) return p;
    }
#if defined(MKCL_WINDOWS)
    {
      mkcl_interrupt_status old_intr;

      HANDLE hndSnap = NULL;
      HANDLE hnd = NULL;
      
      mkcl_get_interrupt_status(env, old_intr);
      mkcl_disable_interrupts(env);
      hndSnap = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, GetCurrentProcessId());
      if (hndSnap != INVALID_HANDLE_VALUE)
	{
	  MODULEENTRY32 me32;
	  me32.dwSize = sizeof(MODULEENTRY32);
	  if (Module32First(hndSnap, &me32))
	    {
	      do
		hnd = GetProcAddress(me32.hModule, symbol);
	      while (hnd == NULL && Module32Next(hndSnap, &me32));
	    }
	  /* GetLastError() == ERROR_NO_MORE_FILES ?? JCB */
	  CloseHandle(hndSnap);
	}
      p = (void*)hnd;
      mkcl_set_interrupt_status(env, old_intr);
    }
#elif defined(HAVE_DLFCN_H)
    /* Is this a valid call? 0 looks suspicious. JCB */
    MKCL_NO_INTR(env, (dlerror(), p = dlsym(0, symbol)));
#else /* !defined(MKCL_WINDOWS) && !defined(HAVE_DLFCN_H) */
# error "mkcl_library_symbol is imcomplete on this platform."
    p = 0;
#endif
  } else {
#ifdef HAVE_DLFCN_H
    MKCL_NO_INTR(env, (dlerror(), p = dlsym(block->cblock.handle, symbol)));
#elif defined(MKCL_WINDOWS)
    {
      HMODULE h = (HMODULE)(block->cblock.handle);
      MKCL_NO_INTR(env, p = GetProcAddress(h, symbol));
    }
#else
# error "mkcl_library_symbol is incomplete on this platform."
#endif
    /* Libraries whose symbols are being referenced by the FFI should not
     * get garbage collected. Until we find a better solution we simply lock
     * them for the rest of the runtime */
    if (p) {
      block->cblock.locked |= lock;
    }
  }
  return p;
}

mkcl_object
mkcl_library_error(MKCL, mkcl_object block) 
{
  mkcl_object output;
#ifdef HAVE_DLFCN_H
  MKCL_NO_INTR(env, output = mkcl_make_base_string_copy(env, dlerror()));
#endif
#if defined(MKCL_WINDOWS)
  {
    mkcl_interrupt_status old_intr;
    void * message;
    DWORD error_code;
    DWORD msg_size;

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    error_code = GetLastError();
    if ((msg_size = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
				  FORMAT_MESSAGE_ALLOCATE_BUFFER,
				  0, error_code, 0, (LPTSTR) &message, 0, NULL)))
      {
	char * msg_ctl = "(%d) %s";
	output = mkcl_alloc_simple_base_string(env, strlen(message) + sizeof(msg_ctl) + 15);
	sprintf((char *) output->base_string.self, msg_ctl, error_code, message);
	output->base_string.fillp = strlen((char *) output->base_string.self);
	LocalFree(message);
      }
    else
      output = mkcl_make_simple_base_string(env, "");
    mkcl_set_interrupt_status(env, old_intr);
  }
#endif
  return output;
}

void
mkcl_library_close(MKCL, mkcl_object block) 
{
  const char *filename;
  bool verbose = mkcl_symbol_value(env, @'si::*gc-verbose*') != mk_cl_Cnil;

  if (mkcl_Null(block->cblock.name))
    filename = "<anonymous>";
  else
    filename = (char*)block->cblock.name->base_string.self;
  if (block->cblock.handle != NULL) {
    if (verbose) {
      fprintf(stderr, ";;; Freeing library %s\n", filename);
    }
#ifdef HAVE_DLFCN_H
    MKCL_NO_INTR(env, dlclose(block->cblock.handle));  /* Return value not handled! FIXME! JCB */
#elif defined(MKCL_WINDOWS)
    MKCL_NO_INTR(env, FreeLibrary(block->cblock.handle));
#else
# error "mkcl_library_close is incomplete on this platform."
#endif
  }
  if (block->cblock.self_destruct) {
    if (verbose) {
      fprintf(stderr, ";;; Removing file %s\n", filename);
    }
    unlink(filename);
  }
  mkcl_core.libraries = mkcl_remove_eq(env, block, mkcl_core.libraries);
}

void
mkcl_library_close_all(MKCL)
{
  while (mkcl_core.libraries != mk_cl_Cnil) {
    mkcl_library_close(env, MKCL_CONS_CAR(mkcl_core.libraries));
  }
}

mkcl_object
mk_si_load_binary(MKCL, mkcl_object filename, mkcl_object verbose, mkcl_object print)
{
  mkcl_object block;
  mkcl_object basename;
  mkcl_object prefix;
  volatile mkcl_object output = mk_cl_Cnil;
  mkcl_object l_c_lock;
  volatile mkcl_object locked = mk_cl_Cnil;

  /* We need the full pathname */
  filename = mk_cl_truename(env, filename);

  /* Loading binary code is not thread safe. When another thread tries
     to load the same file, we may end up initializing twice the same
     module. */
  l_c_lock = mkcl_symbol_value(env, @'mt::+load-compile-lock+');
  MKCL_UNWIND_PROTECT_BEGIN(env) {
    mkcl_interrupt_status old_intr;

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    locked = mk_mt_get_lock(env, 1, l_c_lock);
    mkcl_set_interrupt_status(env, old_intr);

    /* Try to load shared object file */
    block = mkcl_library_open(env, filename, 1);
    if (block->cblock.handle == NULL) {
      output = mkcl_library_error(env, block);
      goto OUTPUT;
    }

    /* Fist try to call "mkcl_init_fas_CODE()" */
    block->cblock.entry = mkcl_library_symbol(env, block, INIT_PREFIX "CODE", 0);
    if (block->cblock.entry != NULL)
      goto GO_ON;

    /* Next try to call "init_FILE()" where FILE is the file name */
    prefix = mkcl_symbol_value(env, @'si::*init-function-prefix*');
    if (mkcl_Null(prefix))
      prefix = mkcl_make_simple_base_string(env, INIT_PREFIX);
    else
      prefix = @si::concatenate-base-strings(env, 3,
					     mkcl_make_simple_base_string(env, INIT_PREFIX),
					     prefix,
					     mkcl_make_simple_base_string(env, "_"));
    basename = mk_cl_pathname_name(env, 1,filename);
    basename = @si::concatenate-base-strings(env, 2, 
					     prefix,
					     @string-upcase(env, 1,
							    mk_cl_funcall(env, 4,
									  @'nsubstitute',
									  MKCL_CODE_CHAR('_'),
									  MKCL_CODE_CHAR('-'),
									  basename)));
    block->cblock.entry = mkcl_library_symbol(env, block, (char*)basename->base_string.self, 0);

    if (block->cblock.entry == NULL) {
      mkcl_object s = mkcl_make_simple_base_string(env, ("This file is probably not "
							 "a proper MKCL binary file: "));
      output = mkcl_library_error(env, block);
      output = mk_si_concatenate_base_strings(env, 2, s, output);
      mkcl_library_close(env, block);
      goto OUTPUT;
    }

    /* Finally, perform initialization */
  GO_ON:	
    mkcl_read_VV(env, block, block->cblock.entry, filename);
    output = mk_cl_Cnil;
  OUTPUT: ;
  } MKCL_UNWIND_PROTECT_EXIT {
    if (!mkcl_Null(locked)) mk_mt_giveup_lock(env, l_c_lock);
  } MKCL_UNWIND_PROTECT_END;
  @(return output);
}

mkcl_object
mk_si_load_source(MKCL, mkcl_object source, mkcl_object verbose, mkcl_object print)
{
  volatile mkcl_object locked = mk_cl_Cnil;
  mkcl_object l_c_lock;
  mkcl_object x, strm;
  mkcl_object form_end_position = MKCL_MAKE_FIXNUM(0);
  mkcl_object location = MKCL_CONS(env, source, form_end_position);

  /* Source may be either a stream or a filename */
  if (mkcl_type_of(source) != mkcl_t_pathname && mkcl_type_of(source) != mkcl_t_base_string) {
    /* INV: if "source" is not a valid stream, file.d will complain */
    strm = source;
  } else {
    strm = mkcl_open_stream(env, source, mkcl_smm_input, mk_cl_Cnil, mk_cl_Cnil, 8,
			    MKCL_STREAM_DEFAULT_FORMAT | MKCL_STREAM_STDIO_STREAM, mk_cl_Cnil);
    if (mkcl_Null(strm))
      @(return mk_cl_Cnil);
  }
  l_c_lock = mkcl_symbol_value(env, @'mt::+load-compile-lock+');
  MKCL_UNWIND_PROTECT_BEGIN(env) {
    mkcl_interrupt_status old_intr;

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    locked = mk_mt_get_lock(env, 1, l_c_lock);
    mkcl_set_interrupt_status(env, old_intr);

    mkcl_bds_bind(env, @'ext::*source-location*', location);
    for (;;) {
      x = mk_si_read_object_or_ignore(env, strm, MKCL_OBJNULL);
      form_end_position = mkcl_file_position(env, strm);
      MKCL_RPLACD(location, form_end_position);
      if (x == MKCL_OBJNULL)
	break;
      if (env->nvalues) {
	mk_si_eval_with_env(env, 1, x);
	if (print != mk_cl_Cnil) {
	  @write(env, 1, x);
	  @terpri(env, 0);
	}
      }
    }
    mkcl_bds_unwind1(env);
  } MKCL_UNWIND_PROTECT_EXIT {
    if (!mkcl_Null(locked)) mk_mt_giveup_lock(env, l_c_lock);
    /* We do not want to come back here if close_stream fails,
       therefore, first we frs_pop() current jump point, then
       try to close the stream, and then jump to next catch
       point */
    if (strm != source)
      mk_cl_close(env, 3, strm, @':abort', @'t');
  } MKCL_UNWIND_PROTECT_END;
  @(return mk_cl_Cnil);
}

@(defun load (source
	      &key (verbose mkcl_symbol_value(env, @'*load-verbose*'))
	      (print mkcl_symbol_value(env, @'*load-print*'))
	      (if_does_not_exist @':error')
	      (search_list mkcl_symbol_value(env, @'si::*load-search-list*'))
	      &aux pathname pntype hooks filename function ok)
  bool not_a_filename = 0;
@
  ok = mk_cl_Cnil;
  /* If source is a stream, read conventional lisp code from it */
  if (mkcl_type_of(source) != mkcl_t_pathname && !mkcl_stringp(env, source)) {
    /* INV: if "source" is not a valid stream, file.d will complain */
    filename = source;
    function = mk_cl_Cnil;
    not_a_filename = 1;
    goto NOT_A_FILENAME;
  }
  /* INV: mkcl_coerce_to_file_pathname() creates a fresh new pathname object */
  source   = mk_cl_merge_pathnames(env, 1, source);
  pathname = mkcl_coerce_to_file_pathname(env, source);
  pntype   = pathname->pathname.type;
  
  filename = mk_cl_Cnil;
  hooks = mkcl_symbol_value(env, @'si::*load-hooks*');
  if (mkcl_Null(pathname->pathname.directory) &&
      mkcl_Null(pathname->pathname.host) &&
      mkcl_Null(pathname->pathname.device) &&
      !mkcl_Null(search_list))
    {
      mkcl_loop_for_in(env, search_list) {
	mkcl_object d = MKCL_CAR(search_list);
	mkcl_object f = mk_cl_merge_pathnames(env, 2, pathname, d);
	mkcl_object ok = mk_cl_load(env, 9,
			       f, @':verbose', verbose,
			       @':print', print,
			       @':if-does-not-exist', mk_cl_Cnil,
			       @':search-list', mk_cl_Cnil);
	if (!mkcl_Null(ok)) {
	  @(return ok);
	}
      } mkcl_end_loop_for_in;
    }
  if (!mkcl_Null(pntype) && (pntype != @':wild')) {
    /* If filename already has an extension, make sure
       that the file exists */
    mkcl_object kind;
    filename = mk_si_coerce_to_filename(env, pathname);
    kind = mk_si_file_kind(env, filename, mk_cl_Ct);
    if (kind != @':file' && kind != @':special') {
      filename = mk_cl_Cnil;
    } else {
      function = mk_cl_cdr(env, mkcl_assoc(env, pathname->pathname.type, hooks));
    }
  } else mkcl_loop_for_in(env, hooks) {
      /* Otherwise try with known extensions until a matching
	 file is found */
      mkcl_object kind;
      filename = pathname;
      filename->pathname.type = MKCL_CAAR(hooks);
      function = MKCL_CDAR(hooks);
      kind = mk_si_file_kind(env, filename, mk_cl_Ct);
      if (kind == @':file' || kind == @':special')
	break;
      else
	filename = mk_cl_Cnil;
    } mkcl_end_loop_for_in;
  if (mkcl_Null(filename)) {
    if (mkcl_Null(if_does_not_exist))
      { @(return mk_cl_Cnil); }
    else
      mkcl_FEcannot_open(env, source);
  }
 NOT_A_FILENAME:
  if (verbose != mk_cl_Cnil) {
    mk_cl_format(env, 3, mk_cl_Ct, mkcl_make_simple_base_string(env, "~&;;; Loading ~s~%"),
	      filename);
  }
  mkcl_bds_bind(env, @'*package*', mkcl_symbol_value(env, @'*package*'));
  mkcl_bds_bind(env, @'*readtable*', mkcl_symbol_value(env, @'*readtable*'));
  mkcl_bds_bind(env, @'*load-pathname*', not_a_filename ? mk_cl_Cnil : source);
  mkcl_bds_bind(env, @'*load-truename*',mk_cl_Cnil);
  MKCL_SETQ(env, @'*load-truename*', (not_a_filename 
				      ? mk_cl_Cnil : (filename = mk_cl_truename(env, filename))));
  if (!mkcl_Null(function)) {
    mkcl_object l_c_lock = mkcl_symbol_value(env, @'mt::+load-compile-lock+');
    volatile mkcl_object locked = mk_cl_Cnil;
    MKCL_UNWIND_PROTECT_BEGIN(env) {
      mkcl_interrupt_status old_intr;

      mkcl_get_interrupt_status(env, old_intr);
      mkcl_disable_interrupts(env);
      locked = mk_mt_get_lock(env, 1, l_c_lock);
      mkcl_set_interrupt_status(env, old_intr);

      ok = mk_cl_funcall(env, 4, function, filename, verbose, print);
    } MKCL_UNWIND_PROTECT_EXIT {
      if (!mkcl_Null(locked)) mk_mt_giveup_lock(env, l_c_lock);
    } MKCL_UNWIND_PROTECT_END;
  } else {
    if (not_a_filename) {
      ok = mk_cl_Ct;
    } else {
      ok = mk_si_load_binary(env, filename, verbose, print);
    }
    if (!mkcl_Null(ok))
      ok = mk_si_load_source(env, filename, verbose, print);
  }
  mkcl_bds_unwind_n(env, 4);

  if (!mkcl_Null(ok))
    mkcl_FEerror(env, "LOAD: Could not load file ~S (Error: ~S)", 2, filename, ok);
  if (print != mk_cl_Cnil) {
    mk_cl_format(env, 3,
		 mk_cl_Ct,
		 mkcl_make_simple_base_string(env, "~&;;; Loaded ~s~%"),
		 filename);
  }
  @(return filename);
@)

mkcl_object mk_si_list_libraries(MKCL)
{
  @(return mk_cl_copy_list(env, mkcl_core.libraries));
}

