/* -*- mode: c -*- */
/*
    main.c --
*/
/*
    Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
    Copyright (c) 1990, Giuseppe Attardi.
    Copyright (c) 2001, Juan Jose Garcia Ripoll.
    Copyright (c) 2010-2011, Jean-Claude Beaudoin.

    MKCL is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file '../../Copyright' for full details.
*/

/******************************** GLOBALS *****************************/

#include <mkcl/mkcl.h>
#include <limits.h>

#if defined(MKCL_WINDOWS)
# include <windows.h>
# include <shellapi.h>
# define MAXPATHLEN 512
# include <winbase.h>
#endif

#ifndef MAXPATHLEN
# ifdef PATH_MAX
#   define MAXPATHLEN PATH_MAX
# else
#   define NO_PATH_MAX
#   include <unistd.h>
# endif
#endif
#   include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <mkcl/internal.h>


/******************************* EXPORTS ******************************/

struct mkcl_core_struct mkcl_core;
static const char *mkcl_self;


/************************ GLOBAL INITIALIZATION ***********************/

static int ARGC;
static char **ARGV;

/* When setting the value of MKCL_OPT_C_STACK_SIZE make sure you do NOT
   exceed the system allocated stack size (minus twice the value of
   MKCL_OPT_STACK_SAFETY_AREA). If you do otherwise you will very likely
   get some "very interesting" results, most probably in the form
   of a SIGSEGV or an internal error of some sort... JCB

   This is particularly important on Microsoft Windows where this
   system has no mean to query the effective size of the initial stack.
   By default the stack size is 2MB on MINGW32 but we grew it to 8MB.
*/
static mkcl_word option_values[MKCL_OPT_LIMIT] = {
	0,		/* MKCL_OPT_INCREMENTAL_GC */
	1,		/* MKCL_OPT_CHAIN_SIGSEGV */
	1,		/* MKCL_OPT_CHAIN_SIGFPE */
	1,		/* MKCL_OPT_CHAIN_SIGINT */
	1,		/* MKCL_OPT_CHAIN_SIGILL */
	1,		/* MKCL_OPT_CHAIN_SIGBUS */
	0,		/* MKCL_OPT_BOOTED */
	4* 8192,		/* MKCL_OPT_BIND_STACK_SIZE */
	128,		/* MKCL_OPT_BIND_STACK_SAFETY_AREA */
	4* 2048,		/* MKCL_OPT_FRAME_STACK_SIZE */
	128,		/* MKCL_OPT_FRAME_STACK_SAFETY_AREA */
	4* 32768,		/* MKCL_OPT_LISP_STACK_SIZE */
	128,		/* MKCL_OPT_LISP_STACK_SAFETY_AREA */
	3*512*1024,	/* MKCL_OPT_C_STACK_SIZE */ /* This is a lower bound. JCB */
	16*1024,	/* MKCL_OPT_C_STACK_SAFETY_AREA */
	1,		/* MKCL_OPT_SIGALTSTACK_SIZE */
	0,              /* MKCL_OPT_HEAP_SIZE */
	1024*1024, 	/* MKCL_OPT_HEAP_SAFETY_AREA */
        0,		/* MKCL_OPT_THREAD_INTERRUPT_SIGNAL */
        0,		/* MKCL_OPT_THREAD_RESUME_SIGNAL */
        1,		/* MKCL_OPT_SET_GMP_MEMORY_FUNCTIONS */
};

#define NB_FEATURES (sizeof(feature_names)/sizeof(feature_names[0]))

static const char *feature_names[] = {
        "MKCL", "COMMON", "COMMON-LISP", "ANSI-CL",
#if defined(__i386) || defined(__pentium)
	"X86",
#elif defined(__x86_64)
	"X86-64",
#endif
#ifdef MKCL_IEEE_FP
        "IEEE-FLOATING-POINT",
#endif
#ifdef unix
	"UNIX",
#endif
#ifdef linux
        "LINUX",
#endif
#ifdef __MINGW32__
	"WINDOWS",
# ifdef __MINGW64__
	"MINGW64",
# else
	"MINGW32",
# endif
#endif
#ifdef _MSC_VER
	"WINDOWS",
	"MSVC",
#endif
#ifdef MKCL_UNICODE
	"UNICODE",
#endif
#ifdef MKCL_RELATIVE_PACKAGE_NAMES
	"RELATIVE-PACKAGE-NAMES",
#endif
};

mkcl_word
mkcl_get_option(mkcl_option option)
{
  return option_values[option];
}

int
mkcl_set_option(mkcl_option option, mkcl_word value)
{
  if (option > MKCL_OPT_LIMIT || option < 0)
    return MKCL_BAD_OPTION;
  else if (option < MKCL_OPT_BOOTED && option_values[MKCL_OPT_BOOTED])
    return MKCL_IMMUTABLE_OPTION;
  else
    option_values[option] = value;
  return 0;
}

void
mkcl_init_env(MKCL, mkcl_env new_env)
{
  new_env->c_env = NULL;

  new_env->string_pool = mk_cl_Cnil;

  new_env->stack = NULL;
  new_env->stack_top = NULL;
  new_env->stack_limit = NULL;
  new_env->stack_size = 0;
  mkcl_stack_set_size(env, mkcl_get_option(MKCL_OPT_LISP_STACK_SIZE));

  new_env->fficall = mkcl_alloc(env, sizeof(struct mkcl_fficall));
  ((struct mkcl_fficall*)new_env->fficall)->registers = 0;

  new_env->method_hash = mk_cl_Cnil;
  new_env->method_spec_vector = mk_cl_Cnil;
  new_env->method_generation = 0;
  _mkcl_set_method_hash_size(new_env, 4096);
  new_env->method_hash_clear_list = mk_cl_Cnil;

  mkcl_init_stacks(env, new_env);

  {
    int i;
    for (i = 0; i < 3; i++)
      {
	mkcl_object x = mkcl_alloc_object(env, mkcl_t_bignum);
	_mkcl_big_init2(x, MKCL_BIG_REGISTER_SIZE);
	new_env->big_register[i] = x;
      }
  }

  new_env->fpe_control_bits = 0;
}

#include "iso_latin_names.h"

void
_mkcl_dealloc_env(MKCL)
{
  int i;
  for (i = 0; i < 3; i++)
    {
      _mkcl_big_clear(env->big_register[i]);
    }
  mkcl_dealloc(env, env);
}

const mkcl_env _mkcl_alloc_env(MKCL)
{
  const mkcl_env new_env 
    = ((env == NULL) 
       ? mkcl_alloc_unprotected(sizeof(struct mkcl_env_struct))
       : mkcl_alloc(env, sizeof(struct mkcl_env_struct)));

  /*
   * An uninitialized environment _always_ disables interrupts. They
   * are activated later on by the thread entry point or mkcl_init_unixint().
   */
  mkcl_disable_interrupts(new_env);

  new_env->nvalues = 0;
  {
    int i;

    for (i = 0; i < MKCL_MULTIPLE_VALUES_LIMIT; i++)
      new_env->values[i] = mk_cl_Cnil;
  }

  new_env->function = mk_cl_Cnil;

  new_env->stack_size = 0;
  new_env->stack = NULL;
  new_env->stack_top = NULL;
  new_env->stack_limit = NULL;

  new_env->bds_size = 0;
  new_env->bds_org = NULL;
  new_env->bds_top = NULL;
  new_env->bds_limit = NULL;

  new_env->specials_size = MKCL_STARTUP_SPECIALS_SIZE;
  new_env->specials =
    ((env == NULL) 
     ? mkcl_alloc_unprotected(MKCL_STARTUP_SPECIALS_SIZE * sizeof(mkcl_object))
     : mkcl_alloc(env, MKCL_STARTUP_SPECIALS_SIZE * sizeof(mkcl_object)));

  {
    mkcl_index i;
    
    for (i = 0; i < MKCL_STARTUP_SPECIALS_SIZE; i++)
      new_env->specials[i] = MKCL_END_OF_BDS_CHAIN;
  }

  new_env->ihs_top = NULL;

  new_env->frs_size = 0;
  new_env->frs_org = NULL;
  new_env->frs_top = NULL;
  new_env->frs_limit = NULL;
  new_env->nlj_fr = NULL;
  new_env->frame_id = 0;

  new_env->cs_org = NULL;
  new_env->cs_limit = NULL;
  new_env->cs_barrier = NULL;
  new_env->cs_size = 0;

  new_env->string_pool = mk_cl_Cnil;

  new_env->c_env = NULL;

  new_env->fmt_aux_stream = mk_cl_Cnil;

  new_env->print_pretty = FALSE;
  new_env->queue = NULL;
  new_env->indent_stack = NULL;
  new_env->qh = new_env->qt = new_env->qc = new_env->isp = new_env->iisp = 0;

  new_env->big_register[0] = mk_cl_Cnil;
  new_env->big_register[1] = mk_cl_Cnil;
  new_env->big_register[2] = mk_cl_Cnil;

  new_env->own_thread = mk_cl_Cnil;

  new_env->method_hash_clear_list = mk_cl_Cnil;
  new_env->method_hash = mk_cl_Cnil;
  new_env->method_spec_vector = mk_cl_Cnil;
  new_env->method_generation = 0;

  new_env->fficall = NULL;

  new_env->altstack = NULL;
  new_env->altstack_size = 0;

  new_env->fpe_control_bits = 0;

  return new_env;
}

void
mkcl_shutdown(MKCL)
{
  if (env != NULL && mkcl_get_option(MKCL_OPT_BOOTED) > 0) {
    mkcl_object l = mkcl_symbol_value(env, @'si::*exit-hooks*');
    mkcl_object form = mk_cl_list(env, 2, @'funcall', mk_cl_Cnil);
    while (MKCL_CONSP(l)) {
      mkcl_elt_set(env, form, 1, MKCL_CONS_CAR(l));
      MKCL_CATCH_ALL_BEGIN(env) {
	mk_si_safe_eval(env, 3, form, mk_cl_Cnil, MKCL_OBJNULL);
      } MKCL_CATCH_ALL_END;
      l = MKCL_CDR(l);
      MKCL_SET(@'si::*exit-hooks*', l);
    }
    mk_si_gc_off(env);
    mkcl_library_close_all(env);
    mkcl_set_option(MKCL_OPT_BOOTED, MKCL_UNBOUND_OPTION);
  }
}

static void mkcl_shutdown_at_exit(void)
{
  mkcl_shutdown(MKCL_ENV());
}

#ifdef MKCL_UNICODE
static void
read_char_database(MKCL)
{
  mkcl_object s = mk_si_concatenate_base_strings(env, 2,
						 mk_si_get_library_pathname(env),
						 mkcl_make_simple_base_string(env, "ucd.dat"));
  mkcl_object output = mk_cl_Cnil;
  FILE *f = fopen(s->base_string.self, "rb");
  if (f) {
    mkcl_index size, read;
    if (!fseek(f, 0, SEEK_END)) {
      size = ftell(f);
      fseek(f, 0, SEEK_SET);
      output = mk_si_make_vector(env, @'ext::byte8', MKCL_MAKE_FIXNUM(size),
				 mk_cl_Cnil, mk_cl_Cnil, mk_cl_Cnil, mk_cl_Cnil);
	    
      read = 0;
      while (read < size) {
	mkcl_index res;
	res = fread(output->vector.self.b8 + read, 1, size - read, f);
	if (res > 0) {
	  read += res;
	} else {
	  output = mk_cl_Cnil;
	  break;
	}
      }
    }
    mkcl_safe_fclose(env, f, mk_cl_Cnil);
  }
  if (output == mk_cl_Cnil) {
    printf("Unable to read Unicode database: %s\n", s->base_string.self);
    abort();
  } else {
    uint8_t *p = output->vector.self.b8;
    mkcl_core.unicode_database = output;
    mkcl_core.ucd_misc = p + 2;
    mkcl_core.ucd_pages = mkcl_core.ucd_misc + (p[0] + (p[1]<<8));
    mkcl_core.ucd_data = mkcl_core.ucd_pages + (0x110000 / 256);
  }
  MKCL_SET(@'si::+unicode-database+', output);
}
#else
#define read_char_database(env)
#endif

static mkcl_object mkcl_true_self(MKCL)
{
  mkcl_object self_truename;

#ifdef linux
  {
    /* The value of 3 here is a rough upper bound of the number of decimal
       digits needed to print the value of 8 binary bits. A more precise
       value would have been 2.409 but we needed an integer. */
    static char buf[sizeof("/proc//exe") + (3 * sizeof(pid_t))];

    sprintf(buf, "/proc/%d/exe", getpid());

    self_truename = mkcl_make_base_string_copy(env, buf);
  }
#elif defined(MKCL_WINDOWS)
  {
    char buf[2 * MAX_PATH];
    self_truename = mkcl_core.self;
    DWORD rc;

    MKCL_NO_INTR(env, rc = GetModuleFileName(NULL, buf, sizeof(buf)));
    switch (rc)
      {
      case 0:
	mkcl_FEwin32_error(env, "Could not get MKCL true name!", 0);
	break;
      case sizeof(buf):
	mkcl_lose("mkcl_true_self(): Buffer overflow for GetModuleFileName.");
	break;
      default:
	break;
      }

    self_truename = mkcl_make_base_string_copy(env, buf);
  }
#else
# error "This is almost sure to be wrong and may lead to a core dump! Fix it!"
  self_truename = mkcl_core.self;
#endif

  self_truename = mk_cl_probe_file(env, self_truename);
  return self_truename;
}

mkcl_object mk_si_self_truename(MKCL)
{
  @(return mkcl_core.self_truename);
}

static mkcl_object initial_thread_bindings[] =
  {
    @'*break-on-signals*',
    @'*compile-print*',
    @'*compile-verbose*',
    @'*debugger-hook*',
    @'*default-pathname-defaults*',
    @'*gensym-counter*',
    @'*load-print*',
    @'*load-verbose*',
    @'*macroexpand-hook*',
    @'*package*',
    @'*print-array*',
    @'*print-base*',
    @'*print-case*',
    @'*print-circle*',
    @'*print-escape*',
    @'*print-gensym*',
    @'*print-length*',
    @'*print-level*',
    @'*print-lines*',
    @'*print-miser-width*',
    @'*print-pprint-dispatch*',
    @'*print-pretty*',
    @'*print-radix*',
    @'*print-readably*',
    @'*print-right-margin*',
    @'*random-state*',
    @'*read-base*',
    @'*read-default-float-format*',
    @'*read-eval*',
    @'*read-suppress*',
    @'*readtable*',
    @'mt::*thread*',
    @'si::*dynamic-cons-stack*',
    @'si::*default-floating-point-exception-set*'
  };

static void setup_initial_thread_bindings(MKCL)
{
  int i;
  int nb_bindings = sizeof(initial_thread_bindings)/sizeof(initial_thread_bindings[0]);

  for (i = 0; i < nb_bindings; i++)
    mkcl_bds_push(env, initial_thread_bindings[i]);
}

static mkcl_object mkcl_T; /* debugging convenience. */

mkcl_env
mkcl_boot(int argc, char **argv)
{
  mkcl_env env;
  mkcl_object aux;
  mkcl_object features;
  int i;

  i = option_values[MKCL_OPT_BOOTED];
  if (i) {
    if (i < 0) {
      /* We have called mkcl_shutdown and want to use MKCL again. */
      option_values[MKCL_OPT_BOOTED] = 1;
    }
    return MKCL_ENV();
  }

#ifdef HASHTABLE_STATS /* JCB */
  mkcl_core.hashtables[mkcl_htt_eq] = mk_cl_Cnil;
  mkcl_core.hashtables[mkcl_htt_eql] = mk_cl_Cnil;
  mkcl_core.hashtables[mkcl_htt_equal] = mk_cl_Cnil;
  mkcl_core.hashtables[mkcl_htt_equalp] = mk_cl_Cnil;
  mkcl_core.hashtables[mkcl_htt_package] = mk_cl_Cnil;
#endif

  ARGC = argc;
  ARGV = argv;
  mkcl_self = argv[0];

  mkcl_init_alloc();
  env = _mkcl_alloc_env(NULL);
  mkcl_init_early_unixint(env);
  mkcl_init_early_threads(env);

  /*
   * Initialize the per-thread data.
   * This cannot come later, because some routines need the
   * frame stack immediately (for instance SI:PATHNAME-TRANSLATIONS).
   */
  mkcl_init_big(env);
  mkcl_init_env(env, env);
  mkcl_init_c_stack_control(env, &env);

  /*
   * 1) Initialize symbols and packages
   */

  mk_cl_Cnil_symbol->symbol.t = mkcl_t_symbol;
  mk_cl_Cnil_symbol->symbol.dynamic = 0;
  mk_cl_Cnil_symbol->symbol.special_index = MKCL_NOT_A_SPECIAL_INDEX;
  mk_cl_Cnil_symbol->symbol.value = mk_cl_Cnil;
  mk_cl_Cnil_symbol->symbol.name = mkcl_make_simple_base_string(env, "NIL");
  mk_cl_Cnil_symbol->symbol.gfdef = mk_cl_Cnil;
  mk_cl_Cnil_symbol->symbol.plist = mk_cl_Cnil;
  mk_cl_Cnil_symbol->symbol.sys_plist = mk_cl_Cnil;
  mk_cl_Cnil_symbol->symbol.hpack = mk_cl_Cnil;
  mk_cl_Cnil_symbol->symbol.properly_named_class = mk_cl_Cnil;
  mk_cl_Cnil_symbol->symbol.stype = mkcl_stp_constant;
  mkcl_root_symbols_count=1;

  mk_cl_Ct->symbol.t = mkcl_t_symbol;
  mk_cl_Ct->symbol.dynamic = 0;
  mk_cl_Ct->symbol.special_index = MKCL_NOT_A_SPECIAL_INDEX;
  mk_cl_Ct->symbol.value = mk_cl_Ct;
  mk_cl_Ct->symbol.name = mkcl_make_simple_base_string(env, "T");
  mk_cl_Ct->symbol.gfdef = mk_cl_Cnil;
  mk_cl_Ct->symbol.plist = mk_cl_Cnil;
  mk_cl_Ct->symbol.sys_plist = mk_cl_Cnil;
  mk_cl_Ct->symbol.hpack = mk_cl_Cnil;
  mk_cl_Ct->symbol.properly_named_class = mk_cl_Cnil;
  mk_cl_Ct->symbol.stype = mkcl_stp_constant;
  mkcl_root_symbols_count=2;

  mkcl_T = mk_cl_Ct;

#ifdef NO_PATH_MAX
  mkcl_core.path_max = sysconf(_PC_PATH_MAX);
#else
  mkcl_core.path_max = MAXPATHLEN;
#endif
#ifdef unix
  mkcl_core.arg_max = sysconf(_SC_ARG_MAX);
#endif

  mkcl_core.packages = mk_cl_Cnil;
  mkcl_core.packages_to_be_created = mk_cl_Cnil;

  mkcl_core.lisp_package =
    mkcl_make_package(env, mkcl_make_simple_base_string(env, "COMMON-LISP"),
		     mk_cl_list(env, 2, mkcl_make_simple_base_string(env, "CL"),
			     mkcl_make_simple_base_string(env, "LISP")),
		     mk_cl_Cnil);
  mkcl_core.user_package =
    mkcl_make_package(env, mkcl_make_simple_base_string(env, "COMMON-LISP-USER"),
		     mk_cl_list(env, 2, mkcl_make_simple_base_string(env, "CL-USER"),
			     mkcl_make_simple_base_string(env, "USER")),
		     mkcl_list1(env, mkcl_core.lisp_package));
  mkcl_core.keyword_package =
    mkcl_make_package(env, mkcl_make_simple_base_string(env, "KEYWORD"),
		     mk_cl_Cnil, mk_cl_Cnil);
  mkcl_core.system_package =
    mkcl_make_package(env, mkcl_make_simple_base_string(env, "SI"),
		      mk_cl_list(env, 3,
				 mkcl_make_simple_base_string(env, "SYSTEM"),
				 mkcl_make_simple_base_string(env, "SYS"),
				 mkcl_make_simple_base_string(env, "EXT")),
		      mkcl_list1(env, mkcl_core.lisp_package));
  mkcl_core.clos_package =
    mkcl_make_package(env, mkcl_make_simple_base_string(env, "CLOS"),
		      mk_cl_Cnil, mkcl_list1(env, mkcl_core.lisp_package));
  mkcl_core.mt_package =
    mkcl_make_package(env, mkcl_make_simple_base_string(env, "MT"),
		      mk_cl_list(env, 3,
				 mkcl_make_simple_base_string(env, "MULTI-THREADING"),
				 mkcl_make_simple_base_string(env, "MP"),
				 mkcl_make_simple_base_string(env, "MULTIPROCESSING")),
		      mkcl_list1(env, mkcl_core.lisp_package));
  mkcl_core.gray_package = mkcl_make_package(env, mkcl_make_simple_base_string(env, "GRAY"),
					     mk_cl_Cnil,
					     MKCL_CONS(env, mkcl_core.lisp_package, mk_cl_Cnil));

  mk_cl_Cnil_symbol->symbol.hpack = mkcl_core.lisp_package;
  mkcl_import2(env, mk_cl_Cnil, mkcl_core.lisp_package);
  mkcl_export2(env, mk_cl_Cnil, mkcl_core.lisp_package);

  mk_cl_Ct->symbol.hpack = mkcl_core.lisp_package;
  mkcl_import2(env, mk_cl_Ct, mkcl_core.lisp_package);
  mkcl_export2(env, mk_cl_Ct, mkcl_core.lisp_package);

  /* At exit, clean up */
  atexit(mkcl_shutdown_at_exit);

  /* These must come _after_ the packages and NIL/T have been created */
  mkcl_init_all_symbols(env);
  mkcl_init_gentemp(env);

  mk_si_gc_on(env);

  /*
   * Initialize constants (strings, numbers and time).
   */

  /* LIBRARIES is an adjustable vector of objects. It behaves as
     a vector of weak pointers thanks to the magic in
     gbc.d/alloc_2.d */
  mkcl_core.libraries = mk_cl_Cnil;
  mkcl_core.to_be_finalized = mk_cl_Cnil;
  mkcl_core.bytes_consed = mk_cl_Cnil;
  mkcl_core.gc_counter = mk_cl_Cnil;
  mkcl_core.gc_stats = FALSE;

  mkcl_core.null_string = mkcl_make_simple_base_string(env, "");
  mkcl_core.dot_string = mkcl_make_simple_base_string(env, ".");
  mkcl_core.dot_dot_string = mkcl_make_simple_base_string(env, "..");
  mkcl_core.localhost_string = mkcl_make_simple_base_string(env, "localhost");

  mkcl_core.null_stream = mk_cl_Cnil; /* Filled in file.d */

  mkcl_init_system_properties(env);

  mkcl_core.gensym_prefix = mkcl_make_simple_base_string(env, "G");
  mkcl_core.gentemp_prefix = mkcl_make_simple_base_string(env, "T");
  mkcl_core.gentemp_counter = MKCL_MAKE_FIXNUM(0);


  mkcl_init_error(env);

  /*
   * Initialize default pathnames
   */
  mkcl_core.pathname_translations = mk_cl_Cnil;
  mkcl_core.library_pathname = mk_cl_Cnil;
  mkcl_core.empty_default_pathname_defaults = mkcl_make_pathname(env,
								 mk_cl_Cnil, mk_cl_Cnil, mk_cl_Cnil,
								 mk_cl_Cnil, mk_cl_Cnil, mk_cl_Cnil);

  MKCL_SET(@'*default-pathname-defaults*', mkcl_core.empty_default_pathname_defaults);

  mkcl_core.self = mkcl_make_base_string_copy(env, mkcl_self);
  mkcl_core.self_truename = mkcl_true_self(env);

  /*
   * Initialize Unicode character database.
   */
  read_char_database(env);

  /*
   * Load character names. The following hash table is a map
   * from names to character codes and viceversa. Note that we
   * need EQUALP because it has to be case insensitive.
   */
  mkcl_core.char_names = aux
    = mk_cl__make_hash_table(env, @'equalp', MKCL_MAKE_FIXNUM(128), /* size */
			     mkcl_make_singlefloat(env, 1.5f), /* rehash-size */
			     mkcl_make_singlefloat(env, 0.5f)); /* rehash-threshold */
  
  for (i = 0; char_names[i]; i++) {
    mkcl_object name = mkcl_make_simple_base_string(env, char_names[i]);
    mkcl_object code = MKCL_MAKE_FIXNUM(i);
    mkcl_sethash(env, name, aux, code);
    mkcl_sethash(env, code, aux, name);
  }
  {
    /* Linefeed is redundant with one of the names given
     * in iso_latin_names.h, but it can not be associated
     * to the code 10, because the default name must be
     * Newline. */
    mkcl_object name = mkcl_make_simple_base_string(env, "Linefeed");
    mkcl_sethash(env, name, aux, MKCL_MAKE_FIXNUM(10));
  }

  /*
   * Initialize logical pathname translations. This must come after
   * the character database has been filled.
   */
  @si::pathname-translations(env, 2, mkcl_make_simple_base_string(env, "SYS"),
			     mk_cl_list(env, 1,
					mk_cl_list(env, 2,
						   mkcl_make_simple_base_string(env, "*.*"),
						   mkcl_make_simple_base_string(env, "./*.*"))));

  /*
   * Ininitialize numbers
   */
  MKCL_SET(@'si::c-int-max', mkcl_make_integer(env, INT_MAX));
  MKCL_SET(@'si::c-int-min', mkcl_make_integer(env, INT_MIN));
  MKCL_SET(@'si::c-long-max', mkcl_make_integer(env, LONG_MAX));
  MKCL_SET(@'si::c-long-min', mkcl_make_integer(env, LONG_MIN));
  MKCL_SET(@'si::c-uint-max', mkcl_make_unsigned_integer(env, UINT_MAX));
  MKCL_SET(@'si::c-ulong-max', mkcl_make_unsigned_integer(env, ULONG_MAX));

  mkcl_init_number(env);

  MKCL_SET(@'si::c-long-long-min', mkcl_make_int64_t(env, INT64_MIN));
  MKCL_SET(@'si::c-long-long-max', mkcl_make_int64_t(env, INT64_MAX));
  MKCL_SET(@'si::c-ulong-long-max', mkcl_make_uint64_t(env, UINT64_MAX));

  mkcl_init_unixtime(env);

  MKCL_SET(@'mt::*thread*', env->own_thread);

  /*
   * Initialize I/O subsystem.
   */
  mkcl_init_file(env);
  mkcl_init_read(env);

  MKCL_SET(@'*print-case*', @':upcase');

  /*
   * Set up hooks for LOAD, errors and macros.
   */
  MKCL_SET(@'mt::+forward-reference-lock+',
	   mk_mt_make_lock(env, 2, @':name', @'mt::+forward-reference-lock+'));
  MKCL_SET(@'mt::+load-compile-lock+',
	   mk_mt_make_lock(env, 2, @':name', @'mt::+load-compile-lock+'));
  aux = mk_cl_list(env, 11,
		   MKCL_CONS(env, mkcl_make_simple_base_string(env, "fasb"), @'si::load-binary'),
		   MKCL_CONS(env, mkcl_make_simple_base_string(env, "FASB"), @'si::load-binary'),
		   MKCL_CONS(env, mkcl_make_simple_base_string(env, "fas"), @'si::load-binary'),
		   MKCL_CONS(env, mkcl_make_simple_base_string(env, "FAS"), @'si::load-binary'),
		   MKCL_CONS(env, mkcl_make_simple_base_string(env, "fasl"), @'si::load-binary'),
		   MKCL_CONS(env, mkcl_make_simple_base_string(env, "FASL"), @'si::load-binary'),
		   MKCL_CONS(env, mkcl_make_simple_base_string(env, "lsp"), @'si::load-source'),
		   MKCL_CONS(env, mkcl_make_simple_base_string(env, "LSP"), @'si::load-source'),
		   MKCL_CONS(env, mkcl_make_simple_base_string(env, "lisp"), @'si::load-source'),
		   MKCL_CONS(env, mkcl_make_simple_base_string(env, "LISP"), @'si::load-source'),
		   MKCL_CONS(env, mk_cl_Cnil, @'si::load-source'));
  MKCL_SET(@'si::*load-hooks*', aux);
  MKCL_SET(@'*load-verbose*', mk_cl_Cnil);

  mkcl_init_macros(env);

  /*
   * Set up infrastructure for CLOS.
   */
  MKCL_SET(@'si::*class-name-hash-table*',
	   mk_cl__make_hash_table(env, @'eq', MKCL_MAKE_FIXNUM(1024), /* size */
				  mkcl_make_singlefloat(env, 1.5f), /* rehash-size */
				  mkcl_make_singlefloat(env, 0.75f))); /* rehash-threshold */
  /*
   * Features.
   */

  MKCL_SET(@'LAMBDA-LIST-KEYWORDS',
	   mk_cl_list(env, 8,
		      @'&optional', @'&rest', @'&key', @'&allow-other-keys',
		      @'&aux', @'&whole', @'&environment', @'&body'));

  for (i = 0, features = mk_cl_Cnil; i < NB_FEATURES; i++) {
    features = MKCL_CONS(env, mkcl_make_keyword(env, feature_names[i]),features);
  }

  MKCL_SET(@'*features*', features);

  MKCL_SET(@'*package*', mkcl_core.lisp_package);

  /* This has to come before init_LSP/CLOS, because we need
   * mkcl_clear_compiler_properties() to work in mkcl_init_CLOS(). */
  mkcl_set_option(MKCL_OPT_BOOTED, TRUE);


  {
    mkcl_object fpe_set = mk_si_initial_floating_point_exception_set(env);
    
    MKCL_SET(@'si::*default-floating-point-exception-set*', fpe_set); /* global */
  }

  setup_initial_thread_bindings(env);

 /* This initializes the dynamic cons stack and must be done right
    after the initial thread binding of si:*dynamic-cons-stack*. */
  mk_si_trim_dynamic_cons_stack(env);

  mkcl_read_VV(env, MKCL_OBJNULL, mkcl_init_lib_LSP, mk_cl_Cnil);

  MKCL_SET(@'*package*', mkcl_core.user_package); /* global */
  MKCL_SETQ(env, @'*package*', mkcl_core.user_package); /* thread-local */

  /* Jump to top level */
  mkcl_init_late_unixint(env);
  mkcl_init_late_threads(env);
  return env;
}

/************************* ENVIRONMENT ROUTINES ***********************/

mkcl_object mk_si_raw_exit(MKCL, mkcl_object code)
{
  if (!MKCL_FIXNUMP(code))
    mkcl_FEerror(env, "Illegal exit code: ~S.", 1, code);
  exit((MKCL_FIXNUMP(code) ? mkcl_fixnum_to_word(code) : EXIT_FAILURE));
}

mkcl_object mk_si_halt(MKCL)
{
  _Exit(EXIT_FAILURE);
}

mkcl_object
mk_si_argc(MKCL)
{
  @(return MKCL_MAKE_FIXNUM(ARGC));
}

mkcl_object
mk_si_argv(MKCL, mkcl_object index)
{
  if (MKCL_FIXNUMP(index)) {
    mkcl_word i = mkcl_fixnum_to_word(index);
    if (i >= 0 && i < ARGC)
      @(return mkcl_make_base_string_copy(env, ARGV[i]));
  }
  mkcl_FEerror(env, "Illegal argument index: ~S.", 1, index);
}

mkcl_object
mk_si_getenv(MKCL, mkcl_object var)
{
  const char *value;

  var = mkcl_check_cl_type(env, @'ext::getenv', var, mkcl_t_base_string);
  value = getenv((char*)var->base_string.self);
  @(return ((value == NULL)? mk_cl_Cnil : mkcl_make_base_string_copy(env, value)));
}

mkcl_object
mk_si_setenv(MKCL, mkcl_object var, mkcl_object value)
{
#if defined(HAVE_SETENV) || defined(HAVE_PUTENV)
  mkcl_word ret_val = 0;

  /* Strings have to be null terminated base strings */
  var = mk_si_copy_to_simple_base_string(env, var);
  if (value == mk_cl_Cnil) {
#ifdef HAVE_SETENV
    /* Remove the variable when setting to nil, so that
     * (si:setenv "foo" nil), then (si:getenv "foo") returns
     * the right thing. */
    unsetenv((char*)var->base_string.self);
#else
#ifdef MKCL_WINDOWS
    mk_si_setenv(env, var, mkcl_make_simple_base_string(env, ""));
#else
    putenv((char*)var->base_string.self);
#endif
#endif
    ret_val = 0;
  } else {
#ifdef HAVE_SETENV
    value = mk_si_copy_to_simple_base_string(env, value);
    ret_val = setenv((char*)var->base_string.self, (char*)value->base_string.self, 1);
#else
    value = mk_cl_format(env, 4, mk_cl_Cnil, mkcl_make_simple_base_string(env, "~A=~A"), var, value);
    value = mk_si_copy_to_simple_base_string(env, value);
    ret_val = putenv((char*)value->base_string.self);
#endif
  }
  if (ret_val == -1)
    mkcl_CEerror(env, mk_cl_Ct, "SI:SETENV failed: insufficient space in environment.", 0);
  @(return value);
#else
  @(return mk_cl_Cnil);
#endif
}

mkcl_object
mk_si_pointer(MKCL, mkcl_object x)
{
  @(return mkcl_make_unsigned_integer(env, (mkcl_index)x));
}

mkcl_object
mk_si_gdb(MKCL)
{
  /* A simple do-nothing function that allows for an easy escape
     from the REPL down into gdb. */
  @(return mk_cl_Cnil);
}

#ifdef MKCL_WINDOWS
void
mkcl_get_commandline_args(int* argc, char*** argv)
{
  LPWSTR *wArgs;
  int i;

  if (argc == NULL || argv == NULL)
    return;

  wArgs = CommandLineToArgvW(GetCommandLineW(), argc);
  *argv = (char**)malloc(sizeof(char*)*(*argc));
  for (i=0; i<*argc; i++) {
    int len = wcslen(wArgs[i]);
    (*argv)[i] = (char*)malloc(2*(len+1));
    wcstombs((*argv)[i], wArgs[i], len+1);
  }
  LocalFree(wArgs);
}
#endif
