/* -*- mode: c -*- */
/*
    pathname.d -- Pathnames.
*/
/*
    Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
    Copyright (c) 1990, Giuseppe Attardi.
    Copyright (c) 2001, Juan Jose Garcia Ripoll.
    Copyright (c) 2010-2011, Jean-Claude Beaudoin

    MKCL is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file '../../Copyright' for full details.
*/

/*
	This file contains those functions that interpret namestrings.
*/

#include <mkcl/mkcl.h>
#include <mkcl/mkcl-inl.h>
#include <mkcl/internal.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>

typedef int (*delim_fn)(int);

static void
push_substring(MKCL, mkcl_object buffer, mkcl_object string, mkcl_index start, mkcl_index end)
{
  string = mk_cl_string(env, string);
  while (start < end) {
    mkcl_string_push_extend(env, buffer, mkcl_char(env, string, start));
    start++;
  }
}

static void
push_string(MKCL, mkcl_object buffer, mkcl_object string)
{
  push_substring(env, buffer, string, 0, mkcl_length(env, string));
}

static mkcl_object
destructively_check_directory(MKCL, mkcl_object directory, bool logical)
{
  /* This function performs two tasks
   * 1) It ensures that the list is a valid directory list
   * 2) It ensures that all strings in the list are valid C strings without fill pointer
   *    All strings are copied, thus avoiding problems with the user modifying the
   *    list that was passed to MAKE-PATHNAME.
   */
  /* INV: directory is always a list */
  mkcl_object ptr;
  int i;

  if (!MKCL_LISTP(directory))
    return @':error'; /* return mk_cl_Cnil; JCB */
  if (mkcl_Null(directory))
    return directory; /* return mk_cl_Ct; JCB */
  if (MKCL_CONS_CAR(directory) != @':absolute' &&
      MKCL_CONS_CAR(directory) != @':relative')
    return @':error'; /* return mk_cl_Cnil; JCB */
 BEGIN:
  for (i=0, ptr=directory; MKCL_CONSP(ptr); ptr = MKCL_CONS_CDR(ptr), i++) {
    mkcl_object item = MKCL_CONS_CAR(ptr);
    if (item == @':back') {
      if (i == 0)
	return @':error';
      item = mkcl_nth(env, i-1, directory);
      if (item == @':absolute' || item == @':wild-inferiors')
	return @':error';
      if (i >= 2)
	MKCL_RPLACD(mkcl_nthcdr(env, i-2, directory),
		   MKCL_CONS_CDR(ptr));
    } else if (item == @':up') {
      if (i == 0)
	return @':error';
      item = mkcl_nth(env, i-1, directory);
      if (item == @':absolute' || item == @':wild-inferiors')
	return @':error';
    } else if (item == @':relative' || item == @':absolute') {
      if (i > 0)
	return @':error';
    } else if (mkcl_stringp(env, item)) {
      mkcl_index l = mkcl_length(env, item);
#ifdef MKCL_UNICODE
      if (mkcl_fits_in_base_string(env, item)) {
	item = mk_si_copy_to_simple_base_string(env, item);
      } else
#endif
	item = mk_cl_copy_seq(env, item);
      MKCL_RPLACA(ptr, item);
      if (logical)
	continue;
      if (l && mkcl_char(env, item,0) == '.') {
	if (l == 1) {
	  /* Single dot */
	  if (i == 0)
	    return @':error';
	  MKCL_RPLACD(mkcl_nthcdr(env, i-1, directory), MKCL_CONS_CDR(ptr));
	  goto BEGIN;
	} else if (l == 2 && mkcl_char(env, item,1) == '.') {
	  MKCL_RPLACA(ptr, @':back');
	  goto BEGIN;
	}
      }
    } else if (item != @':wild' && item != @':wild-inferiors') {
      return @':error';
    }
  }
  return directory;
}

mkcl_object
mkcl_make_pathname(MKCL, mkcl_object host, mkcl_object device, mkcl_object directory,
		  mkcl_object name, mkcl_object type, mkcl_object version)
{
  mkcl_object x = mk_cl_Cnil, p, component;

  p = mkcl_alloc_object(env, mkcl_t_pathname);
  if (mkcl_stringp(env, host))
    p->pathname.logical = mkcl_logical_hostname_p(env, host);
  else if (host == mk_cl_Cnil)
    p->pathname.logical = FALSE;
  else {
    x = directory;
    component = @':host';
    goto _MKCL_ERROR;
  }
  if (device != mk_cl_Cnil && device != @':unspecific' &&
      !(!p->pathname.logical && mkcl_stringp(env, device))
      && !(p->pathname.logical && (device == @':unspecific'))) {
    x = device;
    component = @':device';
    goto _MKCL_ERROR;
  }
  if (name != mk_cl_Cnil && name != @':wild' && !mkcl_stringp(env, name)) {
    x = name;
    component = @':name';
    goto _MKCL_ERROR;
  }
  if (type != mk_cl_Cnil && type != @':unspecific' && type != @':wild' && !mkcl_stringp(env, type)) {
    x = type;
    component = @':type';
    goto _MKCL_ERROR;
  }
  if (version != @':unspecific' && version != @':newest' &&
      version != @':wild' && version != mk_cl_Cnil && !MKCL_FIXNUMP(version))
    {
      x = version;
      component = @':version';
    _MKCL_ERROR:
      mkcl_FEerror(env, "~s is not a valid pathname-~a component", 2, x, component);
    }
  switch (mkcl_type_of(directory)) {
#ifdef MKCL_UNICODE
  case mkcl_t_string:
#endif
  case mkcl_t_base_string:
    directory = mk_cl_list(env, 2, @':absolute', directory);
    break;
  case mkcl_t_symbol:
    if ( mkcl_Null(directory) )
      break;
    if (directory == @':wild') {
      directory = mk_cl_list(env, 2, @':absolute', @':wild-inferiors');
      break;
    }
    component = @':directory';
    goto _MKCL_ERROR;
  case mkcl_t_cons:
    directory = mk_cl_copy_list(env, directory);
    break;
  case mkcl_t_null:
    break;
  default:
    x = directory;
    component = @':directory';
    goto _MKCL_ERROR;
  }
  p->pathname.host      = host;
  p->pathname.device    = device;
  p->pathname.directory = directory;
  p->pathname.name      = name;
  p->pathname.type      = type;
  p->pathname.version   = version;
  if (destructively_check_directory(env, directory, p->pathname.logical) == @':error') {
    mk_cl_error(env, 3, @'file-error', @':pathname', p);
  }
  return(p);
}

static mkcl_object
tilde_expand(MKCL, mkcl_object pathname)
{
  /*
   * If the pathname is a physical one, without hostname, without device
   * and the first element is either a tilde '~' or '~' followed by
   * a user name, we merge the user homedir pathname with this one.
   */
  mkcl_object directory, head;
  if (pathname->pathname.logical || pathname->pathname.host != mk_cl_Cnil
      || pathname->pathname.device != mk_cl_Cnil) {
    return pathname;
  }
  directory = pathname->pathname.directory;
  if (!MKCL_CONSP(directory) || MKCL_CONS_CAR(directory) != @':relative'
      || MKCL_CONS_CDR(directory) == mk_cl_Cnil) {
    return pathname;
  }
  head = MKCL_CADR(directory);
  if (mkcl_stringp(env, head) && mkcl_length(env, head) > 0 && 
      mkcl_char(env, head,0) == '~') {
    /* Remove the tilde component */
    MKCL_RPLACD(directory, MKCL_CDDR(directory));
    pathname = mk_cl_merge_pathnames(env, 2, pathname,
				  mkcl_homedir_pathname(env, head));
  }
  return pathname;
}

#define WORD_INCLUDE_DELIM 1
#define WORD_ALLOW_ASTERISK  2
#define WORD_EMPTY_IS_NIL 4
#define WORD_LOGICAL 8
#define WORD_SEARCH_LAST_DOT 16
#define WORD_ALLOW_LEADING_DOT 32
#define WORD_DISALLOW_SLASH 64
#define WORD_DISALLOW_SEMICOLON 128

static mkcl_object
make_one(MKCL, mkcl_object s, mkcl_index start, mkcl_index end)
{
  return mk_cl_subseq(env, 3, s, MKCL_MAKE_FIXNUM(start), MKCL_MAKE_FIXNUM(end));
}

static int is_colon(int c) { return c == ':'; }
static int is_slash(int c) { return MKCL_IS_DIR_SEPARATOR(c); }
static int is_semicolon(int c) { return c == ';'; }
static int is_dot(int c) { return c == '.'; }
static int is_null(int c) { return c == '\0'; }

/*
 * Translates a string into the host's preferred case.
 * See CLHS 19.2.2.1.2.2 Common Case in Pathname Components.
 */

static mkcl_object
translate_common_case(MKCL, mkcl_object str)
{
  int string_case;
  if (!mkcl_stringp(env, str)) {
    /* Pathnames may contain some other objects, such as symbols,
     * numbers, etc, which need not be translated */
    return str;
  }
  string_case = mkcl_string_case(str);
  if (string_case > 0) { /* ALL_UPPER */
    /* We use UN*X conventions, so lower case is default.
     * However, this really should be conditionalised to the OS type,
     * and it should translate to the _local_ case.
     */
    return mk_cl_string_downcase(env, 1, str);
  } else if (string_case < 0) { /* ALL_LOWER */
    /* We use UN*X conventions, so lower case is default.
     * However, this really should be conditionalised to the OS type,
     * and it should translate to _opposite_ of the local case.
     */
    return mk_cl_string_upcase(env, 1, str);
  } else {
    /* Mixed case goes unchanged */
    return str;
  }
}

static mkcl_object
translate_pathname_case(MKCL, mkcl_object str, mkcl_object scase)
{
	if (scase == @':common') {
	  return translate_common_case(env, str);
	} else if (scase == @':local') {
		return str;
	} else {
	  mkcl_FEerror(env, "~S is not a valid pathname case specificer.~S"
			"Only :COMMON or :LOCAL are accepted.", 1, scase);
	}
}

static mkcl_object
translate_directory_case(MKCL, mkcl_object list, mkcl_object scase)
{
  /* If the argument is really a list, translate all strings in it and
   * return this new list, else assume it is a string and translate it.
   */
  if (!MKCL_CONSP(list)) {
    return translate_pathname_case(env, list, scase);
  } else {
    mkcl_object l;
    list = mk_cl_copy_list(env, list);
    for (l = list; !mkcl_endp(env, l); l = MKCL_CDR(l)) {
      /* It is safe to pass anything to translate_pathname_case,
       * because it will only transform strings, leaving other
       * object (such as symbols) unchanged.*/
      mkcl_object name = MKCL_CONS_CAR(l);
      name = translate_pathname_case(env, name, scase);
      MKCL_RPLACA(l, name);
    }
    return list;
  }
}


/*
 * Parses a word from string `S' until either:
 *	1) character `DELIM' is found
 *	2) end of string is reached
 *	3) a non valid character is found
 * Output is either
 *	1) :error in case (3) above
 *	2) :wild, :wild-inferiors, :up
 *	3) "" or mk_cl_Cnil when word has no elements
 *	5) A non empty string
 */
static mkcl_object
parse_word(MKCL, mkcl_object s, delim_fn delim, int flags, mkcl_index start,
	   mkcl_index end, mkcl_index *end_of_word)
{
  mkcl_index i, j, last_delim = end;
  bool wild_inferiors = FALSE;

  i = j = start;
  for (; i < end; i++) {
    bool valid_char;
    mkcl_index c = mkcl_char(env, s, i);
    if (delim(c)) {
      if ((i == start) && (flags & WORD_ALLOW_LEADING_DOT)) {
	/* Leading dot is included */
	continue;
      }
      last_delim = i;
      if (!(flags & WORD_SEARCH_LAST_DOT)) {
	break;
      }
    }
    if (c == '*') {
      if (!(flags & WORD_ALLOW_ASTERISK))
	valid_char = FALSE; /* Asterisks not allowed in this word */
      else {
	wild_inferiors = (i > start && mkcl_char(env, s, i-1) == '*');
	valid_char = TRUE; /* single "*" */
      }
    } else if (c == ';' && (flags & WORD_DISALLOW_SEMICOLON)) {
      valid_char = FALSE;
    } else if (c == '/' && (flags & WORD_DISALLOW_SLASH)) {
      valid_char = FALSE;
    } else {
      valid_char = c != 0;
    }
    if (!valid_char) {
      *end_of_word = start;
      return @':error';
    }
  }
  if (i > last_delim) {
    /* Go back to the position of the last delimiter */
    i = last_delim;
  }
  if (i < end) {
    *end_of_word = i+1;
  } else {
    *end_of_word = end;
    /* We have reached the end of the string without finding
       the proper delimiter */
    if (flags & WORD_INCLUDE_DELIM) {
      *end_of_word = start;
      return mk_cl_Cnil;
    }
  }
  switch(i-j) {
  case 0:
    if (flags & WORD_EMPTY_IS_NIL)
      return mk_cl_Cnil;
    return mkcl_core.null_string;
  case 1:
    if (mkcl_char(env, s,j) == '*')
      return @':wild';
    break;
  case 2: {
    mkcl_index c0 = mkcl_char(env, s,j);
    mkcl_index c1 = mkcl_char(env, s,j+1);
    if (c0 == '*' && c1 == '*')
      return @':wild-inferiors';
    if (!(flags & (WORD_LOGICAL | WORD_ALLOW_LEADING_DOT)) && c0 == '.' && c1 == '.')
      return @':up';
    break;
  }
  default:
    if (wild_inferiors)	/* '**' surrounded by other characters */
      return @':error';
  }
  return make_one(env, s, j, i);
}

/*
 * Parses a logical or physical directory tree. Output is always a
 * list of valid directory components, which may be just NIL.
 *
 * INV: When parsing of directory components has failed, a valid list
 * is also returned, and it will be later in the parsing of
 * pathname-name or pathname-type when the same error is detected.
 */

static mkcl_object
parse_directories(MKCL, mkcl_object s, int flags, mkcl_index start, mkcl_index end,
		  mkcl_index *end_of_dir)
{
  mkcl_index i, j;
  mkcl_object path = mk_cl_Cnil;
  delim_fn delim = (flags & WORD_LOGICAL) ? is_semicolon : is_slash;

  flags |= WORD_INCLUDE_DELIM | WORD_ALLOW_ASTERISK;
  *end_of_dir = start;
  for (i = j = start; i < end; j = i) {
    mkcl_object part = parse_word(env, s, delim, flags, j, end, &i);
    if (part == @':error' || part == mk_cl_Cnil)
      break;
    if (part == mkcl_core.null_string) {  /* "/", ";" */
      if (j != start) {
	if (flags & WORD_LOGICAL)
	  return @':error';
	*end_of_dir = i;
	continue;
      }
      part = (flags & WORD_LOGICAL) ? @':relative' : @':absolute';
    }
    *end_of_dir = i;
    path = mkcl_cons(env, part, path);
  }
  return mk_cl_nreverse(env, path);
}

bool
mkcl_logical_hostname_p(MKCL, mkcl_object host)
{
  if (!mkcl_stringp(env, host))
    return FALSE;
  return !mkcl_Null(@assoc(env, 4,
		      host,
		      mkcl_core.pathname_translations,
		      @':test',
		      @'string-equal'));
}

/*
 * Parses a lisp namestring until the whole substring is parsed or an
 * error is found. It returns a valid pathname or NIL, plus the place
 * where parsing ended in *END_OF_PARSING.
 *
 * The rules are as follows:
 *
 * 1) If a hostname is supplied it determines whether the namestring
 *    will be parsed as logical or as physical.
 *
 * 2) If no hostname is supplied, first it tries parsing using logical
 *    pathname rules and, if no logical hostname is found, then it
 *    tries the physical pathname format.
 *
 * 3) Logical pathname syntax:
 *	[logical-hostname:][;][logical-directory-component;][pathname-name][.pathname-type]
 *
 * 4) Physical pathname syntax:
 *	[device:][[//hostname]/][directory-component/]*[pathname-name][.pathname-type]
 *
 *	logical-hostname, device, hostname = word
 *	logical-directory-component = word | wildcard-word
 *	directory-component = word | wildcard-word | '..' | '.'
 *	pathname-name, pathname-type = word | wildcard-word | ""
 *
 */
mkcl_object
mkcl_parse_namestring(MKCL, mkcl_object s, mkcl_index start, mkcl_index end, mkcl_index *ep,
		     mkcl_object default_host)
{
  mkcl_object host, device, path, name, type, aux, version;
  bool logical;

  if (start == end) {
    host = device = path = name = type = aux = version = @'nil';
    logical = 0;
    goto make_it;
  }
  /* We first try parsing as logical-pathname. In case of
   * failure, physical-pathname parsing is performed only when
   * there is no supplied *logical* host name. All other failures
   * result in mk_cl_Cnil as output.
   */
  host = parse_word(env, s, is_colon, WORD_LOGICAL | WORD_INCLUDE_DELIM |
		    WORD_DISALLOW_SEMICOLON, start, end, ep);
  if (default_host != mk_cl_Cnil) {
    if (host == mk_cl_Cnil || host == @':error')
      host = default_host;
  }
  if (mkcl_logical_hostname_p(env, host))
    {
      /*
       * Logical pathname format:
       *	[logical-hostname:][;][logical-directory-component;][pathname-name][.pathname-type]
       */
      logical = TRUE;
      device = @':unspecific';
      path = parse_directories(env, s, WORD_LOGICAL, *ep, end, ep);
      if (MKCL_CONSP(path)) {
	if (MKCL_CONS_CAR(path) != @':relative' &&
	    MKCL_CONS_CAR(path) != @':absolute')
	  path = MKCL_CONS(env, @':absolute', path);
	path = destructively_check_directory(env, path, TRUE);
      } else {
	path = MKCL_CONS(env, @':absolute', path);
      }
      if (path == @':error')
	return mk_cl_Cnil;
      name = parse_word(env, s, is_dot, WORD_LOGICAL | WORD_ALLOW_ASTERISK |
			WORD_EMPTY_IS_NIL, *ep, end, ep);
      if (name == @':error')
	return mk_cl_Cnil;
      type = mk_cl_Cnil;
      version = mk_cl_Cnil;
      if (*ep == start || mkcl_char(env, s, *ep-1) != '.')
	goto make_it;
      type = parse_word(env, s, is_dot, WORD_LOGICAL | WORD_ALLOW_ASTERISK |
			WORD_EMPTY_IS_NIL, *ep, end, ep);
      if (type == @':error')
	return mk_cl_Cnil;
      if (*ep == start || mkcl_char(env, s, *ep-1) != '.')
	goto make_it;
      aux = parse_word(env, s, is_null, WORD_LOGICAL | WORD_ALLOW_ASTERISK |
		       WORD_EMPTY_IS_NIL, *ep, end, ep);
      if (aux == @':error') {
	return mk_cl_Cnil;
      } else if (MKCL_SYMBOLP(aux)) {
	version = aux;
      } else {
	version = mk_cl_parse_integer(env, 3, aux, @':junk-allowed', mk_cl_Ct);
	if (mk_cl_integerp(env, version) != mk_cl_Cnil && mkcl_plusp(env, version) &&
	    mkcl_fixnum_to_word(MKCL_VALUES(1)) == mkcl_length(env, aux))
	  ;
	else if (mk_cl_string_equal(env, 2, aux, @':newest') != mk_cl_Cnil)
	  version = @':newest';
	else
	  return mk_cl_Cnil;
      }
    }
  else
    {
    /* physical: */
      /*
       * Physical pathname format:
       *	[[device:[//hostname]]/][directory-component/]*[pathname-name][.pathname-type]
       *
       * We would be a lot happier with this one on MS-Windows. JCB
       *  [device:[/]|//hostname/|/]{directory/}[.name|{name}+[.type]]
       *
       * And this one on Unix.
       *  [/]{directory/}[.name|{name}+[.type]]
       *
       * And give up on the idea of a single format!
       * If the CL ANSI standard renounced that idea
       * long ago as most probably impossible why should
       * we even try? JCB
       */
      logical = FALSE;
      /* We only parse a hostname when the device was present. This
       * requisite is a bit stupid and only applies to the Unix port,
       * where "//home/" is equivalent to "/home". However, in Windows
       * we need "//FOO/" to be separately handled, for it is a shared
       * resource.
       */
#if defined(MKCL_WINDOWS)
      if ((start+1 <= end) && is_slash(mkcl_char(env, s, start))) {
	device = mk_cl_Cnil;
	goto maybe_parse_host;
      }
#endif
      device = parse_word(env, s, is_colon,
			  WORD_INCLUDE_DELIM
			  | WORD_EMPTY_IS_NIL
			  | WORD_DISALLOW_SLASH,
			  start, end, ep);
      if (device == @':error')
	{
	  device = mk_cl_Cnil;
	  host = mk_cl_Cnil;
	  goto done_device_and_host;
	}
      else if (device == mk_cl_Cnil)
	{
	  start = *ep;
	  host = mk_cl_Cnil;
	  goto done_device_and_host;
	}
      if (!mkcl_stringp(env, device)) {
	return mk_cl_Cnil;
      }
    maybe_parse_host:
      /* Files have no effective device. */
      if (@string-equal(env, 2, device, @':file') == mk_cl_Ct) /* What is this for? JCB */
	device = mk_cl_Cnil;
      if (mkcl_length(env, device) == 0)
	device = mk_cl_Cnil;
#ifdef MKCL_WINDOWS
      if (!mkcl_Null(device))
	{ /* On MS-Windows host and device are mutually exclusive. */
	  host = mkcl_core.localhost_string;
	}
      else
#endif
      if ((start+2) <= end && is_slash(mkcl_char(env, s, start)) &&
	  is_slash(mkcl_char(env, s, start+1)))
	{
	  host = parse_word(env, s, is_slash, WORD_EMPTY_IS_NIL,
			    start+2, end, ep);
	  if (host == @':error') {
	    host = mk_cl_Cnil;
	  } else if (host != mk_cl_Cnil) {
	    if (!mkcl_stringp(env, host))
	      return mk_cl_Cnil;
#ifdef MKCL_WINDOWS
	    /* parse share name */
	    device = parse_word(env, s, is_slash, WORD_EMPTY_IS_NIL,
				*ep, end, ep);
#endif
	    start = *ep - 1;
	    if (is_slash(mkcl_char(env, s, start)))
	      *ep = start;
	  }
	}
      else
	host = mk_cl_Cnil;

      
    done_device_and_host:
      path = parse_directories(env, s, 0, *ep, end, ep);
      if (MKCL_CONSP(path)) {
	if (MKCL_CONS_CAR(path) != @':relative' &&
	    MKCL_CONS_CAR(path) != @':absolute')
	  path = MKCL_CONS(env, @':relative', path);
	path = destructively_check_directory(env, path, FALSE);
      }
      if (path == @':error')
	return mk_cl_Cnil;
      start = *ep;
      name = parse_word(env, s, is_dot,
			WORD_ALLOW_LEADING_DOT | WORD_SEARCH_LAST_DOT |
			WORD_ALLOW_ASTERISK | WORD_EMPTY_IS_NIL,
			start, end, ep);
      if (name == @':error')
	return mk_cl_Cnil;
      if ((*ep - start) <= 1 || mkcl_char(env, s, *ep-1) != '.') {
	type = mk_cl_Cnil;
	if (mkcl_Null(path) && !mkcl_Null(mk_cl_stringE(env, 2, name, mkcl_core.dot_string)))
	  {
	    path = mkcl_list1(env, @':relative');
	    name = mk_cl_Cnil;
	  }
      } else {
	type = parse_word(env, s, is_null, WORD_ALLOW_ASTERISK, *ep, end, ep);
	if (type == @':error')
	  return mk_cl_Cnil;
	if (!mkcl_Null(mk_cl_stringE(env, 2, type, mkcl_core.null_string))
	    && !mkcl_Null(mk_cl_stringE(env, 2, name, mkcl_core.dot_string)))
	  {
	    if (mkcl_Null(path))
	      path = MKCL_CONS(env, @':relative', mkcl_list1(env, @':up'));
	    else
	      path = mkcl_nconc(env, path, mkcl_list1(env, @':up'));
	    type = name = mk_cl_Cnil;
	  }
      }
      version = (name != mk_cl_Cnil || type != mk_cl_Cnil) ? @':newest' : mk_cl_Cnil;
    }

 make_it:
  if (*ep >= end) *ep = end;
  path = mkcl_make_pathname(env, host, device, path, name, type, version);
  path->pathname.logical = logical;
  return tilde_expand(env, path);
}

mkcl_object
mk_si_default_pathname_defaults(MKCL)
{
  /* This routine outputs the value of *default-pathname-defaults*
   * coerced to type PATHNAME. Special care is taken so that we do
   * not enter an infinite loop when using PARSE-NAMESTRING, because
   * this routine might itself try to use the value of this variable. */
  mkcl_object path = mkcl_symbol_value(env, @'*default-pathname-defaults*');
  while (mkcl_type_of(path) != mkcl_t_pathname) {
#if 0
      mkcl_bds_bind(env, @'*default-pathname-defaults*', mk_si_getcwd(env, 0));
#else
      mkcl_bds_bind(env, @'*default-pathname-defaults*', mkcl_core.empty_default_pathname_defaults);
#endif
      path = mkcl_type_error(env, @'pathname', "*default-pathname-defaults*", path, @'pathname');
      mkcl_bds_unwind1(env);
  }
  @(return path);
}

mkcl_object
mk_cl_pathname(MKCL, mkcl_object x)
{
 L:
  switch (mkcl_type_of(x)) {
#ifdef MKCL_UNICODE
  case mkcl_t_string:
#endif
  case mkcl_t_base_string:
    x = mk_cl_parse_namestring(env, 1, x);
  case mkcl_t_pathname:
    break;
  case mkcl_t_stream:
    switch ((enum mkcl_smmode)x->stream.mode) {
    case mkcl_smm_input:
    case mkcl_smm_output:
    case mkcl_smm_probe:
    case mkcl_smm_io:
    case mkcl_smm_input_file:
    case mkcl_smm_output_file:
    case mkcl_smm_io_file:
      x = MKCL_IO_STREAM_FILENAME(x);
      goto L;
    case mkcl_smm_synonym:
      x = MKCL_SYNONYM_STREAM_STREAM(env, x);
      goto L;
    default:
      ;/* Fall through to error message */
    }
  default:
    mkcl_FEwrong_type_argument(env,
			       mk_cl_list(env, 4, @'or', @'file-stream', @'string', @'pathname'),
			       x);
  }
  @(return x);
}

mkcl_object
mk_cl_logical_pathname(MKCL, mkcl_object x)
{
  x = mk_cl_pathname(env, x);
  if (!x->pathname.logical) {
    mk_cl_error(env, 9, @'simple-type-error', @':format-control',
		mkcl_make_simple_base_string(env, "~S cannot be coerced to a logical pathname."),
		@':format-arguments', mk_cl_list(env, 1, x),
		@':expected-type', @'logical-pathname',
		@':datum', x);
  }
  @(return x);
}

@(defun wild-pathname-p (pathname &optional component)
  bool checked = 0;
@
  pathname = mk_cl_pathname(env, pathname);
  if (component == mk_cl_Cnil || component == @':host') {
    if (pathname->pathname.host == @':wild')
      @(return mk_cl_Ct);
    checked = 1;
  }
  if (component == mk_cl_Cnil || component == @':device') {
    if (pathname->pathname.device == @':wild')
      @(return mk_cl_Ct);
    checked = 1;
  }
  if (component == mk_cl_Cnil || component == @':version') {
    if (pathname->pathname.version == @':wild')
      @(return mk_cl_Ct);
    checked = 1;
  }
  if (component == mk_cl_Cnil || component == @':name') {
    mkcl_object name = pathname->pathname.name;
    if (name != mk_cl_Cnil &&
	(name == @':wild' || 
	 (!MKCL_SYMBOLP(name) && mkcl_member_char(env, '*', name))))
      @(return mk_cl_Ct);
    checked = 1;
  }
  if (component == mk_cl_Cnil || component == @':type') {
    mkcl_object name = pathname->pathname.type;
    if (name != mk_cl_Cnil &&
	(name == @':wild' || 
	 (!MKCL_SYMBOLP(name) && mkcl_member_char(env, '*', name))))
      @(return mk_cl_Ct);
    checked = 1;
  }
  if (component == mk_cl_Cnil || component == @':directory') {
    mkcl_object list = pathname->pathname.directory;
    checked = 1;
    mkcl_loop_for_on_unsafe(list) {
      mkcl_object name = MKCL_CONS_CAR(list);
      if (name != mk_cl_Cnil &&
	  (name == @':wild' || name == @':wild-inferiors' ||
	   (!MKCL_SYMBOLP(name) && mkcl_member_char(env, '*', name))))
	{
	  @(return mk_cl_Ct);
	}
    } mkcl_end_loop_for_on;
  }
  if (checked == 0) {
    mkcl_FEerror(env, "~A is not a valid pathname component", 1, component);
  }
  @(return mk_cl_Cnil);
@)

/*
 * mkcl_coerce_to_file_pathname(P) converts P to a physical pathname,
 * for a file which is accesible in our filesystem.
 * INV: Wildcards are allowed.
 * INV: A fresh new copy of the pathname is created.
 * INV: The pathname is absolute. (And why should that be so? JCB)
 */
mkcl_object
mkcl_coerce_to_file_pathname(MKCL, mkcl_object pathname)
{
  pathname = mkcl_coerce_to_physical_pathname(env, pathname);
  pathname = mk_cl_merge_pathnames(env, 1, pathname);
  return pathname;
}

/*
 * mkcl_coerce_to_physical_pathname(P) converts P to a physical pathname,
 * performing the appropiate transformation if P was a logical pathname.
 */
mkcl_object
mkcl_coerce_to_physical_pathname(MKCL, mkcl_object x)
{
  x = mk_cl_pathname(env, x);
  if (x->pathname.logical)
    return mk_cl_translate_logical_pathname(env, 1, x);
  return x;
}

/*
 * mk_si_coerce_to_filename(P) converts P to a physical pathname and then to
 * a namestring. The output must always be a new simple-string which can
 * be used by the C library.
 * INV: No wildcards are allowed.
 */
mkcl_object
mk_si_coerce_to_filename(MKCL, mkcl_object pathname_orig)
{
  mkcl_object namestring, pathname;

  /* We always go through the pathname representation and thus
   * mk_cl_namestring() always outputs a fresh new string */
  pathname = mkcl_coerce_to_file_pathname(env, pathname_orig);
  if (mk_cl_wild_pathname_p(env, 1,pathname) != mk_cl_Cnil)
    mk_cl_error(env, 3, @'file-error', @':pathname', pathname_orig);
  namestring = mk_cl_namestring(env, pathname);
  if (namestring == mk_cl_Cnil) {
    mkcl_FEerror(env, "Pathname ~A does not have a physical namestring", 1, pathname_orig);
  }
  if (mkcl_core.path_max != -1 &&
      mkcl_length(env, namestring) >= mkcl_core.path_max - 16)
    mkcl_FEerror(env, "Too long filename: ~S.", 1, namestring);
#ifdef MKCL_UNICODE
  if (mkcl_type_of(namestring) == mkcl_t_string) {
    if (!mkcl_fits_in_base_string(env, namestring))
      mkcl_FEerror(env, "The filesystem does not accept filenames with extended characters: ~S",
		   1, namestring);
    namestring = mk_si_copy_to_simple_base_string(env, namestring);
  }
#endif
  return namestring;
}

#define default_device(host) mk_cl_Cnil

mkcl_object
mkcl_merge_pathnames(MKCL, mkcl_object path, mkcl_object defaults, mkcl_object default_version)
{
  mkcl_object host, device, directory, name, type, version;

  defaults = mk_cl_pathname(env, defaults);
  path = mk_cl_parse_namestring(env, 1, path, mk_cl_Cnil, defaults);
  if (mkcl_Null(host = path->pathname.host))
    host = defaults->pathname.host;
  if (mkcl_Null(path->pathname.device))
    if (mkcl_Null(path->pathname.host))
      device = defaults->pathname.device;
    else if (path->pathname.host == defaults->pathname.host)
      device = defaults->pathname.device;
    else
      device = default_device(path->pathname.host);
  else
    device = path->pathname.device;
  if (mkcl_Null(path->pathname.directory))
    directory = defaults->pathname.directory;
  else if (MKCL_CONS_CAR(path->pathname.directory) == @':absolute')
    directory = path->pathname.directory;
  else if (!mkcl_Null(defaults->pathname.directory))
    directory = mkcl_append(env, defaults->pathname.directory,
			   MKCL_CDR(path->pathname.directory));
  else
    directory = path->pathname.directory;
  if (mkcl_Null(name = path->pathname.name))
    name = defaults->pathname.name;
  if (mkcl_Null(type = path->pathname.type))
    type = defaults->pathname.type;
  version = path->pathname.version;
  if (mkcl_Null(path->pathname.name)) {
    if (mkcl_Null(version))
      version = defaults->pathname.version;
  }
  if (mkcl_Null(version))
    version = default_version;
  /*
    In this implementation, version is not considered
  */
  defaults = mkcl_make_pathname(env, host, device, directory, name, type, version);
  return defaults;
}

/*
	mkcl_namestring(x, flag) converts a pathname to a namestring.
	if flag is true, then the pathname may be coerced to the requirements
	of the filesystem, removing fields that have no meaning (such as
	version, or type, etc); otherwise, when it is not possible to
	produce a readable representation of the pathname, NIL is returned.
*/
mkcl_object
mkcl_namestring(MKCL, mkcl_object x, int truncate_if_unreadable)
{
  bool logical;
  mkcl_object l, y;
  mkcl_object buffer, host;

  x = mk_cl_pathname(env, x);

  /* INV: Pathnames can only be created by merging or parsing namestrings
   * or using mkcl_make_pathname(). In all of these cases MKCL will complain
   * at creation time if the pathname has wrong components.
   */
  buffer = mkcl_make_string_output_stream(env, 128, 1);
  logical = x->pathname.logical;
  host = x->pathname.host;
  if (logical) {
    if ((y = x->pathname.device) != @':unspecific' &&
	truncate_if_unreadable)
      return mk_cl_Cnil; /* FIXME: This is in fact a breach of invariant.
			    Logical pathnames device cannot have any other
			    value than :unspecific and thus by definition.
			    This should be flagged as an internal error.
			    JCB
			 */
    if (host != mk_cl_Cnil) {
      mk_si_do_write_sequence(env, host, buffer, MKCL_MAKE_FIXNUM(0), mk_cl_Cnil);
      mkcl_write_cstr(env, ":", buffer);
    }
  } else {
#ifdef unix
    /* Physical "device" is a nonsense on Unix. JCB */
    if ((y = x->pathname.device) != mk_cl_Cnil) {
      mk_si_do_write_sequence(env, y, buffer, MKCL_MAKE_FIXNUM(0), mk_cl_Cnil);
      mkcl_write_cstr(env, ":", buffer);
    }
    if (host != mk_cl_Cnil) {
#ifndef MKCL_WINDOWS
      if (y == mk_cl_Cnil) {
	mkcl_write_cstr(env, "file:", buffer);
      }
#endif
      mkcl_write_cstr(env, "//", buffer);
      mk_si_do_write_sequence(env, host, buffer, MKCL_MAKE_FIXNUM(0), mk_cl_Cnil);
    }
#elif defined(MKCL_WINDOWS)
    if ((host != mk_cl_Cnil)
	&& !mkcl_string_E(env, host, mkcl_core.localhost_string))
      {
	mkcl_write_cstr(env, "//", buffer);
	mk_si_do_write_sequence(env, host, buffer, MKCL_MAKE_FIXNUM(0), mk_cl_Cnil);
	mkcl_write_cstr(env, "/", buffer);
	y = x->pathname.device;
	mk_si_do_write_sequence(env, y, buffer, MKCL_MAKE_FIXNUM(0), mk_cl_Cnil);
      }
    else if ((y = x->pathname.device) != mk_cl_Cnil)
      {
	mk_si_do_write_sequence(env, y, buffer, MKCL_MAKE_FIXNUM(0), mk_cl_Cnil);
	mkcl_write_cstr(env, ":", buffer);
      }
#endif
  }
  l = x->pathname.directory;
  if (mkcl_endp(env, l))
    goto NO_DIRECTORY;
  y = MKCL_CONS_CAR(l);
  if (y == @':relative') {
    if (logical)
      mkcl_write_char(env, ';', buffer);
    else
      {
	mkcl_write_char(env, '.', buffer);
	mkcl_write_char(env, MKCL_DIR_SEPARATOR, buffer);
      }
  } else {
    if (!logical)
      mkcl_write_char(env, MKCL_DIR_SEPARATOR, buffer);
  }
  l = MKCL_CONS_CDR(l);
  mkcl_loop_for_in(env, l) {
    y = MKCL_CONS_CAR(l);
    if (y == @':up') {
      mkcl_write_cstr(env, "..", buffer);
    } else if (y == @':wild') {
      mkcl_write_cstr(env, "*", buffer);
    } else if (y == @':wild-inferiors') {
      mkcl_write_cstr(env, "**", buffer);
    } else if (y != @':back') {
      mk_si_do_write_sequence(env, y, buffer, MKCL_MAKE_FIXNUM(0), mk_cl_Cnil);
    } else {
      /* Directory :back has no namestring representation */
      return mk_cl_Cnil; /* This is not true, :back does have a representation
			    at least according to CLTL2 and ANSI-CL!
			    This is the syntactic vs semantic distinguo.
			    Fix me! JCB
			 */
    }
    mkcl_write_char(env, (logical ? ';' : MKCL_DIR_SEPARATOR), buffer);
  } mkcl_end_loop_for_in;
 NO_DIRECTORY:
  if (mkcl_file_position(env, buffer) == MKCL_MAKE_FIXNUM(0)) {
    if ((mkcl_stringp(env, x->pathname.name) &&
	 mkcl_member_char(env, ':', x->pathname.name)) ||
	(mkcl_stringp(env, x->pathname.type) &&
	 mkcl_member_char(env, ':', x->pathname.type)))
      mkcl_write_cstr(env, ":", buffer);
  }
  y = x->pathname.name;
  if (y != mk_cl_Cnil) {
    if (y == @':wild') {
      mkcl_write_cstr(env, "*", buffer);
    } else {
      mk_si_do_write_sequence(env, y, buffer, MKCL_MAKE_FIXNUM(0), mk_cl_Cnil);
    }
  }
  y = x->pathname.type;
  if (y != mk_cl_Cnil && y != @':unspecific') {
    if (y == @':wild') {
      mkcl_write_cstr(env, ".*", buffer);
    } else {
      mkcl_write_cstr(env, ".", buffer);
      mk_si_do_write_sequence(env, y, buffer, MKCL_MAKE_FIXNUM(0), mk_cl_Cnil);
    }
  }
  y = x->pathname.version;
  if (logical) {
    if (y != mk_cl_Cnil) {
      mkcl_write_cstr(env, ".", buffer);
      if (y == @':wild') {
	mkcl_write_cstr(env, "*", buffer);
      } else if (y == @':newest') {
	mk_si_do_write_sequence(env, mkcl_symbol_name(env, y), buffer,
			     MKCL_MAKE_FIXNUM(0), mk_cl_Cnil);
      } else {
	/* Since the printer is not reentrant,
	 * we cannot use mk_cl_write and friends.
	 */
	int n = mkcl_fixnum_to_word(y), i;
	char b[MKCL_WORD_BITS/2];
	for (i = 0; n; i++) {
	  b[i] = n%10 + '0';
	  n = n/10;
	}
	if (i == 0)
	  b[i++] = '0';
	while (i--) {
	  mkcl_write_char(env, b[i], buffer);
	}
      }
    }
  } else if (!truncate_if_unreadable) {
    /* Namestrings of physical pathnames have restrictions... */
#if !(defined(unix) || defined(MKCL_WINDOWS))
    /*
      Either version matters and we do this or
      it does not matter (as stated in merge-pathnames!)
      and then simply drop all of this business.
      I go with merge-pathnames for sake of coherence. JCB
     */
    if (mkcl_Null(x->pathname.name) && mkcl_Null(x->pathname.type)) {
      /* Directories cannot have a version number */
      if (y != mk_cl_Cnil)
	return mk_cl_Cnil;
    } else if (y != @':newest') {
      /* Filenames have an implicit version :newest */
      return mk_cl_Cnil;
    }
#endif
  }
  /* we copy to get a simple-string. */
  return mkcl_copy_string(env, mk_cl_get_output_stream_string(env, buffer));
}

mkcl_object
mkcl_namestring_base(MKCL, mkcl_object x, int truncate_if_unreadable)
{
  mkcl_object buffer = mkcl_namestring(env, x, truncate_if_unreadable);
#ifdef MKCL_UNICODE
  if (mkcl_type_of(buffer) == mkcl_t_string) {
    if (!mkcl_fits_in_base_string(env, buffer))
      mkcl_FEerror(env, "The filesystem does not accept filenames with extended characters: ~S",
		   1, buffer);
    buffer = mk_si_copy_to_simple_base_string(env, buffer);
  }
#endif
  return buffer;
}

mkcl_object
mk_cl_namestring(MKCL, mkcl_object x)
{
  @(return mkcl_namestring(env, x, TRUE));
}

@(defun parse_namestring (thing
			  &o host (defaults mk_si_default_pathname_defaults(env))
			  &k (start MKCL_MAKE_FIXNUM(0)) end junk_allowed
			  &a output)
  mkcl_index s, e, ee;
@
  if (host != mk_cl_Cnil) {
    host = mk_cl_string(env, host);
  }
  if (!mkcl_stringp(env, thing)) {
    output = mk_cl_pathname(env, thing);
  } else {
    mkcl_object default_host = host;
    if (default_host == mk_cl_Cnil && defaults != mk_cl_Cnil) {
      defaults = mk_cl_pathname(env, defaults);
      default_host = defaults->pathname.host;
    }
#ifdef MKCL_UNICODE
    thing = mk_si_coerce_to_base_string(env, thing);
#endif
    mkcl_get_string_start_end(env, thing, start, end, &s, &e);
    output = mkcl_parse_namestring(env, thing, s, e, &ee, default_host);
    start = MKCL_MAKE_FIXNUM(ee);
    if (output == mk_cl_Cnil || ee != e) {
      if (mkcl_Null(junk_allowed)) {
	mkcl_FEparse_error(env, "Cannot parse the namestring ~S~%"
		      "from ~S to ~S.", mk_cl_Cnil,
		      3, thing, start, end);
      }
      goto OUTPUT;
    }
  }
  if (host != mk_cl_Cnil && !mkcl_equal(env, output->pathname.host, host)) {
    mkcl_FEerror(env, "The pathname ~S does not contain the required host ~S.",
	    2, thing, host);
  }
 OUTPUT:
  @(return output start);
@)

@(defun merge_pathnames (path
	&o (defaults mk_si_default_pathname_defaults(env))
 	   (default_version @':newest'))
@
  path = mk_cl_pathname(env, path);
  defaults = mk_cl_pathname(env, defaults);
  @(return mkcl_merge_pathnames(env, path, defaults, default_version));
@)

@(defun make_pathname (&key (host MKCL_OBJNULL) (device MKCL_OBJNULL) (directory MKCL_OBJNULL)
			    (name MKCL_OBJNULL) (type MKCL_OBJNULL) (version MKCL_OBJNULL)
		            ((:case scase) @':local')
		            defaults
		       &aux x)
@
  if (mkcl_Null(defaults)) {
    defaults = mk_si_default_pathname_defaults(env);
    defaults = mkcl_make_pathname(env, defaults->pathname.host,
				 mk_cl_Cnil, mk_cl_Cnil, mk_cl_Cnil, mk_cl_Cnil, mk_cl_Cnil);
  } else {
    defaults = mk_cl_pathname(env, defaults);
  }
  x = mkcl_make_pathname(env,
			 host != MKCL_OBJNULL ? translate_pathname_case(env, host,scase)
			 : defaults->pathname.host,
			 device != MKCL_OBJNULL ? translate_pathname_case(env, device,scase)
			 : defaults->pathname.device,
			 directory != MKCL_OBJNULL ? translate_directory_case(env, directory,scase)
			 : defaults->pathname.directory,
			 name != MKCL_OBJNULL ? translate_pathname_case(env, name,scase)
			 : defaults->pathname.name,
			 type != MKCL_OBJNULL ? translate_pathname_case(env, type,scase)
			 : defaults->pathname.type,
			 version != MKCL_OBJNULL ? version : defaults->pathname.version);
  @(return x);
@)

mkcl_object
mk_cl_pathnamep(MKCL, mkcl_object pname)
{
  @(return ((mkcl_type_of(pname) == mkcl_t_pathname)? mk_cl_Ct : mk_cl_Cnil));
}

mkcl_object
mk_si_logical_pathname_p(MKCL, mkcl_object pname)
{
  @(return ((mkcl_type_of(pname) == mkcl_t_pathname && pname->pathname.logical) ? mk_cl_Ct : mk_cl_Cnil));
}

@(defun pathname_host (pname &key ((:case scase) @':local'))
@
  pname = mk_cl_pathname(env, pname);
  @(return translate_pathname_case(env, pname->pathname.host,scase));
@)

@(defun pathname_device (pname &key ((:case scase) @':local'))
@
  pname = mk_cl_pathname(env, pname);
  @(return translate_pathname_case(env, pname->pathname.device,scase));
@)

@(defun pathname_directory (pname &key ((:case scase) @':local'))
@
  pname = mk_cl_pathname(env, pname);
  @(return translate_directory_case(env, pname->pathname.directory,scase));
@)

@(defun pathname_name(pname &key ((:case scase) @':local'))
@
  pname = mk_cl_pathname(env, pname);
  @(return translate_pathname_case(env, pname->pathname.name,scase));
@)

@(defun pathname_type(pname &key ((:case scase) @':local'))
@
  pname = mk_cl_pathname(env, pname);
  @(return translate_pathname_case(env, pname->pathname.type,scase));
@)

mkcl_object
mk_cl_pathname_version(MKCL, mkcl_object pname)
{
  pname = mk_cl_pathname(env, pname);
  @(return  pname->pathname.version);
}

mkcl_object
mk_cl_file_namestring(MKCL, mkcl_object pname)
{
  pname = mk_cl_pathname(env, pname);
  @(return mkcl_namestring(env, mkcl_make_pathname(env, mk_cl_Cnil, mk_cl_Cnil, mk_cl_Cnil,
						   pname->pathname.name,
						   pname->pathname.type,
						   pname->pathname.version),
			   TRUE));
}

mkcl_object
mk_cl_directory_namestring(MKCL, mkcl_object pname)
{
  pname = mk_cl_pathname(env, pname);
  @(return mkcl_namestring(env, mkcl_make_pathname(env, mk_cl_Cnil, mk_cl_Cnil,
						   pname->pathname.directory,
						   mk_cl_Cnil, mk_cl_Cnil, mk_cl_Cnil),
			   TRUE));
}

mkcl_object
mk_cl_host_namestring(MKCL, mkcl_object pname)
{
  pname = mk_cl_pathname(env, pname);
  pname = pname->pathname.host;
  if (mkcl_Null(pname) || pname == @':wild')
    pname = mkcl_core.null_string;
  @(return pname);
}

#define EN_MATCH(e,p1,p2,el) (mkcl_equalp(e,p1->pathname.el, p2->pathname.el) ? mk_cl_Cnil : p1->pathname.el)

@(defun enough_namestring (path
			   &o (defaults mk_si_default_pathname_defaults(env)))
  mkcl_object newpath, pathdir, defaultdir, fname;
@
  defaults = mk_cl_pathname(env, defaults);
  path = mk_cl_pathname(env, path);
  pathdir = path->pathname.directory;
  defaultdir = defaults->pathname.directory;
  if (mkcl_Null(pathdir)) {
    pathdir = mkcl_list1(env, @':relative');
  } else if (mkcl_Null(defaultdir)) {
    /* The defaults pathname does not have a directory. */
  } else if (MKCL_CONS_CAR(pathdir) == @':relative') {
    /* The pathname is relative to the default one, so we just output the original one. */
  } else {
    /* The new pathname is an absolute one. We compare it with the defaults
       and if they have some common elements, we just output the remaining ones. */
    mkcl_object dir_begin = mk_cl_funcall(env, 5, @'mismatch', pathdir, defaultdir, @':test', @'equal');
    if (dir_begin == mk_cl_Cnil) {
      pathdir = mk_cl_Cnil;
    } else if (dir_begin == mk_cl_length(env, defaultdir)) {
      pathdir = mk_cl_funcall(env, 3, @'subseq', pathdir, dir_begin);
      pathdir = MKCL_CONS(env, @':relative', pathdir);
    }
  }
  fname = EN_MATCH(env, path, defaults, name);
  if (fname == mk_cl_Cnil) fname = path->pathname.name;
  /* Create a path with all elements that do not match the default */
  newpath = mkcl_make_pathname(env, 
			      EN_MATCH(env, path, defaults, host),
			      EN_MATCH(env, path, defaults, device),
			      pathdir, fname,
			      EN_MATCH(env, path, defaults, type),
			      EN_MATCH(env, path, defaults, version));
  newpath->pathname.logical = path->pathname.logical;
  @(return mkcl_namestring(env, newpath, TRUE));
@)

#undef EN_MATCH

/* --------------- PATHNAME MATCHING ------------------ */

static bool path_item_match(MKCL, mkcl_object a, mkcl_object mask);

static bool
do_path_item_match(MKCL, mkcl_object s, mkcl_index j, mkcl_object p, mkcl_index i)
{
  mkcl_index ls = mkcl_length(env, s), lp = mkcl_length(env, p);
  while (i < lp) {
    mkcl_index cp = mkcl_char(env, p, i);
    if (cp == '*') {
      /* An asterisk in the pattern matches any number
       * of characters. We try the shortest sequence
       * that matches. */
      mkcl_index cn = 0, next;
      for (next = i+1;
	   next < lp && ((cn = mkcl_char(env, p, next)) == '*');
	   next++)
	;
      if (next == lp) {
	return TRUE;
      }
      while (j < ls) {
	if (do_path_item_match(env, s, j, p, next)) {
	  return TRUE;
	}
	j++;
      }
      return FALSE;
    }
    if ((j >= ls) || (cp != mkcl_char(env, s, j))) {
      /* Either there are no characters left in "s"
       * or the next character does not match. */
      return FALSE;
    }
    i++; j++;
  }
  return (j >= ls);
}

static bool
path_item_match(MKCL, mkcl_object a, mkcl_object mask)
{
  if (mask == @':wild')
    return TRUE;
  /* If a component in the tested path is a wildcard field, this
     can only be matched by the same wildcard field in the mask */
  if (!mkcl_stringp(env, a) || mask == mk_cl_Cnil)
    return (a == mask);
  if (!mkcl_stringp(env, mask))
    mkcl_FEerror(env, "~S is not supported as mask for pathname-match-p", 1, mask);
  return do_path_item_match(env, a, 0, mask, 0);
}

static bool
path_list_match(MKCL, mkcl_object a, mkcl_object mask)
{
  mkcl_object item_mask;
  while (!mkcl_endp(env, mask)) {
    item_mask = MKCL_CAR(mask);
    mask = MKCL_CDR(mask);
    if (item_mask == @':wild-inferiors') {
      if (mkcl_endp(env, mask))
	return TRUE;
      while (!mkcl_endp(env, a)) {
	if (path_list_match(env, a, mask))
	  return TRUE;
	a = MKCL_CDR(a);
      }
      return FALSE;
    } else if (mkcl_endp(env, a)) {
      /* A NIL directory should match against :absolute
	 or :relative, in order to perform suitable translations. */
      if (item_mask != @':absolute' && item_mask != @':relative')
	return FALSE;
    } else if (!path_item_match(env, MKCL_CAR(a), item_mask)) {
      return FALSE;
    } else {
      a = MKCL_CDR(a);
    }
  }
  if (!mkcl_endp(env, a))
    return FALSE;
  return TRUE;
}

mkcl_object
mk_cl_pathname_match_p(MKCL, mkcl_object path, mkcl_object mask)
{
  mkcl_object output = mk_cl_Cnil;
  path = mk_cl_pathname(env, path);
  mask = mk_cl_pathname(env, mask);
  if (path->pathname.logical != mask->pathname.logical)
    goto OUTPUT;

  /* Missing components default to :WILD */
  if (!mkcl_Null(mask->pathname.directory) &&
      !path_list_match(env, path->pathname.directory, mask->pathname.directory))
    goto OUTPUT;
  if (!path_item_match(env, path->pathname.name, mask->pathname.name))
    goto OUTPUT;
  if (!path_item_match(env, path->pathname.type, mask->pathname.type))
    goto OUTPUT;
  if (mkcl_Null(mask->pathname.version) ||
      path_item_match(env, path->pathname.version, mask->pathname.version))
    output = mk_cl_Ct;
 OUTPUT:
  @(return output);
}

/* --------------- PATHNAME TRANSLATIONS ------------------ */

static mkcl_object
coerce_to_from_pathname(MKCL, mkcl_object x, mkcl_object host)
{
  switch (mkcl_type_of(x)) {
#ifdef MKCL_UNICODE
  case mkcl_t_string:
#endif
  case mkcl_t_base_string:
    x = mk_cl_parse_namestring(env, 2, x, host);
  case mkcl_t_pathname:
    if (x->pathname.logical)
      return x;
  default:
    mkcl_FEerror(env, "~S is not a valid from-pathname translation", 1, x);
  }
}

@(defun si::pathname_translations (host &optional (set MKCL_OBJNULL))
  mkcl_index parsed_len, len;
  mkcl_object pair, l;
@
  /* Check that host is a valid host name */
  host = mkcl_check_type_string(env, @'si::pathname-translations', host);
  len = mkcl_length(env, host);
  parse_word(env, host, is_null, WORD_LOGICAL, 0, len, &parsed_len);
  if (parsed_len < len) {
    mkcl_FEerror(env, "Wrong host syntax ~S", 1, host);
  }
  /* Find its translation list */
  pair = @assoc(env, 4, host, mkcl_core.pathname_translations, @':test', @'string-equal');
  if (set == MKCL_OBJNULL) {
    @(return ((pair == mk_cl_Cnil)? mk_cl_Cnil : MKCL_CADR(pair)));
  }
  /* Set the new translation list */
  mkcl_assert_type_list(env, set);
  if (pair == mk_cl_Cnil) {
    pair = MKCL_CONS(env, host, MKCL_CONS(env, mk_cl_Cnil, mk_cl_Cnil));
    mkcl_core.pathname_translations = MKCL_CONS(env, pair, mkcl_core.pathname_translations);
  }
  for (l = set, set = mk_cl_Cnil; !mkcl_endp(env, l); l = MKCL_CDR(l)) {
    mkcl_object item = MKCL_CAR(l);
    mkcl_object from = coerce_to_from_pathname(env, mk_cl_car(env, item), host);
    mkcl_object to = mk_cl_pathname(env, mk_cl_cadr(env, item));
    set = MKCL_CONS(env, MKCL_CONS(env, from, MKCL_CONS(env, to, mk_cl_Cnil)), set);
  }
  set = mk_cl_nreverse(env, set);
  MKCL_RPLACA(MKCL_CONS_CDR(pair), set);
  @(return set);
@)

mkcl_object mk_si_all_logical_pathname_translations(MKCL)
{
  mkcl_object x = mk_cl_copy_alist(env, mkcl_core.pathname_translations);
  @(return x);
}

static mkcl_object
find_wilds(MKCL, mkcl_object l, mkcl_object source, mkcl_object match)
{
  mkcl_index i, j, k, ls, lm;

  if (match == @':wild')
    return mkcl_list1(env, source);
  if (!mkcl_stringp(env, match) || !mkcl_stringp(env, source)) {
    if (match != source)
      return @':error';
    return l;
  }
  ls = mkcl_length(env, source);
  lm = mkcl_length(env, match);
  for(i = j = 0; i < ls && j < lm; ) {
    mkcl_index pattern_char = mkcl_char(env, match,j);
    if (pattern_char == '*') {
      for (j++, k = i;
	   k < ls && mkcl_char(env, source,k) != pattern_char;
	   k++)
	;
      l = MKCL_CONS(env, make_one(env, source, i, k), l);
      i = k;
      continue;
    }
    if (mkcl_char(env, source,i) != pattern_char)
      return @':error';
    i++, j++;
  }
  if (i < ls || j < lm)
    return @':error';
  return l;
}

static mkcl_object
find_list_wilds(MKCL, mkcl_object a, mkcl_object mask)
{
  mkcl_object l = mk_cl_Cnil, l2;

  while (!mkcl_endp(env, mask)) {
    mkcl_object item_mask = MKCL_CAR(mask);
    mask = MKCL_CDR(mask);
    if (item_mask == @':wild-inferiors') {
      l2 = mk_cl_Cnil;
      while (!path_list_match(env, a, mask)) {
	if (mkcl_endp(env, a))
	  return @':error';
	l2 = MKCL_CONS(env, MKCL_CAR(a),l2);
	a = MKCL_CDR(a);
      }
      l = MKCL_CONS(env, l2, l);
    } else if (mkcl_endp(env, a)) {
      /* A NIL directory should match against :absolute
	 or :relative, in order to perform suitable translations. */
      if (item_mask != @':absolute' && item_mask != @':relative')
	return @':error';
    } else {
      l2 = find_wilds(env, l, MKCL_CAR(a), item_mask);
      if (l == @':error')
	return @':error';
      if (!mkcl_Null(l2))
	l = MKCL_CONS(env, l2, l);
      a = MKCL_CDR(a);
    }
  }
  return @nreverse(env, l);
}

static mkcl_object
copy_wildcards(MKCL, mkcl_object *wilds_list, mkcl_object pattern)
{
  mkcl_index i, l, j;
  bool new_string;
  mkcl_object wilds = *wilds_list, token;

  if (pattern == @':wild') {
    if (mkcl_endp(env, wilds))
      return @':error';
    pattern = MKCL_CAR(wilds);
    *wilds_list = MKCL_CDR(wilds);
    return pattern;
  }
  if (pattern == @':wild-inferiors')
    return @':error';
  if (!mkcl_stringp(env, pattern))
    return pattern;

  new_string = FALSE;
  l = mkcl_length(env, pattern);
  token = mk_si_get_buffer_string(env);
  for (j = i = 0; i < l; ) {
    mkcl_index c = mkcl_char(env, pattern, i);
    if (c != '*') {
      i++;
      continue;
    }
    if (i != j) {
      push_substring(env, token, pattern, j, i);
    }
    new_string = TRUE;
    if (mkcl_endp(env, wilds)) {
      return @':error';
    }
    push_string(env, token, MKCL_CAR(wilds));
    wilds = MKCL_CDR(wilds);
    j = i++;
  }
  /* Only create a new string when needed */
  if (new_string) {
    if (mkcl_fits_in_base_string(env, token)) {
      pattern = mk_si_copy_to_simple_base_string(env, token);
    } else {
      pattern = mk_cl_copy_seq(env, token);
    }
  }
  mk_si_put_buffer_string(env, token);
  *wilds_list = wilds;
  return pattern;
}

static mkcl_object
copy_list_wildcards(MKCL, mkcl_object *wilds, mkcl_object to)
{
  mkcl_object l = mk_cl_Cnil;

  while (!mkcl_endp(env, to)) {
    mkcl_object d, mask = MKCL_CAR(to);
    if (mask == @':wild-inferiors') {
      mkcl_object list = *wilds;
      if (mkcl_endp(env, list))
	return @':error';
      else {
	mkcl_object dirlist = MKCL_CAR(list);
	if (MKCL_CONSP(dirlist))
	  l = mkcl_append(env, MKCL_CAR(list), l);
	else if (!mkcl_Null(MKCL_CAR(list)))
	  return @':error';
      }
      *wilds = MKCL_CDR(list);
    } else {
      d = copy_wildcards(env, wilds, MKCL_CAR(to));
      if (d == @':error')
	return d;
      l = MKCL_CONS(env, d, l);
    }
    to = MKCL_CDR(to);
  }
  if (MKCL_CONSP(l))
    l = @nreverse(env, l);
  return l;
}

@(defun translate-pathname (source from to &key)
	mkcl_object wilds, out, d;
@
  /* The pathname from which we get the data */
  source = mk_cl_pathname(env, source);
  /* The mask applied to the source pathname */
  from = mk_cl_pathname(env, from);
  /* The pattern which says what the output should look like */
  to = mk_cl_pathname(env, to);
  
  if (source->pathname.logical != from->pathname.logical)
    goto error;
  out = mkcl_alloc_object(env, mkcl_t_pathname);
  out->pathname.logical = to->pathname.logical;
  
  /* Match host names */
  if (mk_cl_string_equal(env, 2, source->pathname.host, from->pathname.host) == mk_cl_Cnil)
    goto error;
  out->pathname.host = to->pathname.host;
  
  /* Logical pathnames do not have devices. We just overwrite it. */
  out->pathname.device = to->pathname.device;
  
  /* Match directories */
  wilds = find_list_wilds(env, source->pathname.directory,
			  from->pathname.directory);
  if (wilds == @':error')	goto error;
  d = copy_list_wildcards(env, &wilds, to->pathname.directory);
  if (d == @':error') goto error;
  if (wilds != mk_cl_Cnil) goto error2;
  out->pathname.directory = d;
  
  /* Match name */
  wilds = find_wilds(env, mk_cl_Cnil, source->pathname.name, from->pathname.name);
  if (wilds == @':error') goto error2;
  d = copy_wildcards(env, &wilds, to->pathname.name);
  if (d == @':error') goto error;
  if (wilds != mk_cl_Cnil) goto error2;
  out->pathname.name = d;
  
  /* Match type */
  wilds = find_wilds(env, mk_cl_Cnil, source->pathname.type, from->pathname.type);
  if (wilds == @':error') goto error2;
  d = copy_wildcards(env, &wilds, to->pathname.type);
  if (d == @':error') goto error;
  if (wilds != mk_cl_Cnil) goto error2;
  out->pathname.type = d;
  
  /* Match version */
  out->pathname.version = to->pathname.version;
  if (from->pathname.version == @':wild') {
    if (to->pathname.version == @':wild') {
      out->pathname.version = source->pathname.version;
    }
  }
  return out;
  
 error:
  mkcl_FEerror(env, "~S is not a specialization of path ~S", 2, source, from);
 error2:
  mkcl_FEerror(env, "Number of wildcards in ~S do not match  ~S", 2, from, to);
@)

@(defun translate-logical-pathname (source &key)
	mkcl_object l, pair;
	mkcl_object pathname;
@
  pathname = mk_cl_pathname(env, source);
 begin:
  if (!pathname->pathname.logical) {
    @(return pathname);
  }
  l = @si::pathname-translations(env, 1, pathname->pathname.host);
  for(; !mkcl_endp(env, l); l = MKCL_CDR(l)) {
    pair = MKCL_CAR(l);
    if (!mkcl_Null(mk_cl_pathname_match_p(env, pathname, MKCL_CAR(pair)))) {
      pathname = mk_cl_translate_pathname(env, 3, pathname, MKCL_CAR(pair), MKCL_CADR(pair));
      goto begin;
    }
  }
  mkcl_FEerror(env, "~S admits no logical pathname translations", 1, pathname);
@)
