/* -*- mode: c -*- */
/*
    stacks.c -- Binding/History/Frame stacks.
*/
/*
    Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
    Copyright (c) 1990, Giuseppe Attardi.
    Copyright (c) 2001, Juan Jose Garcia Ripoll.
    Copyright (c) 2010, Jean-Claude Beaudoin.

    MKCL is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file '../../Copyright' for full details.
*/

#include <mkcl/mkcl.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#ifdef HAVE_SYS_RESOURCE_H
# include <sys/time.h>
# include <sys/resource.h>
#endif
#include <mkcl/internal.h>


mkcl_object mk_si_disable_interrupts(MKCL)
{ /* Returns T if interrupts were enabled, NIL otherwise. */
  @(return ((env->disable_interrupts) ? mk_cl_Cnil : ((env->disable_interrupts = 1), mk_cl_Ct)));
}

mkcl_object mk_si_enable_interrupts(MKCL)
{
  @(return ((env->disable_interrupts = 0), mk_cl_Cnil));
}

/************************ C STACK ***************************/

static void
cs_set_size(MKCL, mkcl_index new_size)
{
  if (new_size)
    {
      volatile char foo = 0;
      mkcl_index safety_area = mkcl_get_option(MKCL_OPT_C_STACK_SAFETY_AREA);

      new_size += 2*safety_area;

#ifdef MKCL_DOWN_STACK
      /* FIXME: This pointer magic can hit "bottom of memory"
	 without warning and thus wrap around and obviously
	 give a very wrong result! JCB */
      if (&foo > env->cs_org - new_size + 16) {
	env->cs_limit = (char *) (env->cs_org - new_size + 2*safety_area);
	if (env->cs_limit < env->cs_barrier)
	  env->cs_barrier = env->cs_limit;
      }
#else /* def MKCL_DOWN_STACK */
      if (&foo < env->cs_org + new_size - 16) {
	env->cs_limit = (char *) (env->cs_org + new_size - 2*safety_area);
	if (env->cs_limit > env->cs_barrier)
	  env->cs_barrier = env->cs_limit;
      }
#endif /* def MKCL_DOWN_STACK */
      else
	mkcl_lose("can't reset env->cs_limit.");
    }
  else
    { /* This is the "unlimited" case. */
#ifdef MKCL_DOWN_STACK
      env->cs_barrier = env->cs_limit = NULL;
#else /* def MKCL_DOWN_STACK */
      env->cs_limit = NULL;
      env->cs_limit--; /* Wrap around */
      env->cs_barrier = env->cs_limit;
#endif /* def MKCL_DOWN_STACK */
    }
  env->cs_size = new_size;
}

void
mkcl_cs_overflow(MKCL)
{
  errno = 0;
  mkcl_lose("C Stack Overflow!");
}


/********************* BINDING STACK ************************/

mkcl_index mkcl_alloc_new_special_index(MKCL, mkcl_object sym)
{
  mkcl_index index;
  mkcl_interrupt_status old_intr;

  mkcl_get_interrupt_status(env, old_intr);
  mkcl_disable_interrupts(env);
  if (pthread_mutex_lock(&mkcl_core.special_index_lock))
    mkcl_lose("mkcl_alloc_new_special_index failed on pthread_mutex_lock().");

  index = mkcl_core.top_special_index;

  if (index == MKCL_NOT_A_SPECIAL_INDEX)
    mkcl_lose("Out of special variables!");

  mkcl_core.top_special_index = index + 1;

  if (pthread_mutex_unlock(&mkcl_core.special_index_lock))
    mkcl_lose("mkcl_alloc_new_special_index failed on pthread_mutex_unlock().");
  mkcl_set_interrupt_status(env, old_intr);

  return index;
}

void mkcl_grow_specials(MKCL, mkcl_index new_size)
{
  mkcl_object * new = (mkcl_object *) mkcl_alloc(env, new_size * sizeof(mkcl_object));

  memcpy(new, env->specials, env->specials_size * sizeof(mkcl_object));

  {
    mkcl_index i;

    for (i = env->specials_size; i < new_size; i++)
      new[i] = MKCL_END_OF_BDS_CHAIN;
  }

  env->specials = new;
  env->specials_size = new_size;
}


void
mkcl_bds_push(MKCL, mkcl_object s)
{
  mkcl_bds_bind(env, s, _mkcl_sym_val(env, s));
}


mkcl_object
mkcl_set_symbol_value(MKCL, mkcl_object s, mkcl_object value)
{
  mkcl_type type_of_s = mkcl_type_of(s);

  if (type_of_s != mkcl_t_symbol)
    mkcl_FEillegal_variable_name(env, s);
  else if (mkcl_Null(s) || s->symbol.stype & mkcl_stp_constant)
    mkcl_FEprogram_error(env, "Tried to bind a value to the constant ~S.", 1, s);
  else
    {
      mkcl_index index = s->symbol.special_index;

      if (index < env->specials_size)
	{
	  mkcl_object v = env->specials[index];
	  if (v != MKCL_END_OF_BDS_CHAIN)
	    return env->specials[index] = value;
	}
      return s->symbol.value = value;
    }
}

static void
mkcl_bds_set_size(MKCL, mkcl_index size)
{
  mkcl_bds_ptr old_org = env->bds_org;
  mkcl_index limit = env->bds_top - old_org;
  if (size <= limit) {
    mkcl_FEerror(env, "Cannot shrink the binding stack below ~D.", 1,
	    mkcl_make_unsigned_integer(env, limit));
  } else {
    mkcl_interrupt_status old_intr;
    mkcl_index margin = mkcl_get_option(MKCL_OPT_BIND_STACK_SAFETY_AREA);
    mkcl_bds_ptr org;
    org = mkcl_alloc(env, size * sizeof(*org));

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    memcpy(org, old_org, (limit + 1) * sizeof(*org));
    env->bds_top = org + limit;
    env->bds_org = org;
    env->bds_limit = org + (size - 2*margin);
    env->bds_size = size;
    mkcl_set_interrupt_status(env, old_intr);

    mkcl_dealloc(env, old_org);
  }
}

void
mkcl_bds_overflow(MKCL)
{
  mkcl_index margin = mkcl_get_option(MKCL_OPT_BIND_STACK_SAFETY_AREA);
  mkcl_index size = env->bds_size;
  mkcl_bds_ptr org = env->bds_org;
  mkcl_bds_ptr last = org + size;
  if (env->bds_limit >= last) {
    mkcl_lose("Bind stack overflow, cannot grow larger.");
  }
  env->bds_limit += margin;
  mk_cl_cerror(env, 6, mkcl_make_simple_base_string(env, "Extend stack size"),
	       @'ext::stack-overflow', @':size', MKCL_MAKE_FIXNUM(size),
	       @':type', @'ext::binding-stack');
  mkcl_bds_set_size(env, size + (size / 2));
}

void
mkcl_bds_unwind(MKCL, mkcl_index new_bds_top_index)
{
  mkcl_bds_ptr new_bds_top = new_bds_top_index + env->bds_org;
  mkcl_bds_ptr bds = env->bds_top;
  for (;  bds > new_bds_top;  bds--)
    mkcl_bds_unwind1(env);
  env->bds_top = new_bds_top;
}

static mkcl_bds_ptr
get_bds_ptr(MKCL, mkcl_object x)
{
  if (MKCL_FIXNUMP(x)) {
    mkcl_bds_ptr p = env->bds_org + mkcl_fixnum_to_word(x);
    if (env->bds_org <= p && p <= env->bds_top)
      return(p);
  }
  mkcl_FEerror(env, "~S is an illegal bds index. BDS_TOP = ~S.", 2, x, mk_si_bds_top(env));
}

mkcl_object
mk_si_bds_top(MKCL)
{
  @(return MKCL_MAKE_FIXNUM(env->bds_top - env->bds_org));
}

mkcl_object
mk_si_bds_var(MKCL, mkcl_object arg)
{
  @(return get_bds_ptr(env, arg)->symbol);
}

mkcl_object
mk_si_bds_val(MKCL, mkcl_object arg)
{
  mkcl_object v = get_bds_ptr(env, arg)->value;
  @(return ((v == MKCL_OBJNULL || v == MKCL_END_OF_BDS_CHAIN) ? MKCL_UNBOUND : v));
}

/******************** INVOCATION STACK **********************/

static mkcl_object
ihs_function_name(mkcl_object x)
{
  mkcl_object y;

  switch (mkcl_type_of(x)) {
  case mkcl_t_symbol:
    return(x);

  case mkcl_t_bclosure:
    x = x->bclosure.code;
  case mkcl_t_bytecode:
    y = x->bytecode.name;
    if (mkcl_Null(y))
      return(@'lambda');
    else
      return y;

  case mkcl_t_cfun:
    return(x->cfun.name);
  case mkcl_t_cclosure:
    return (x->cclosure.name);
  default:
    return(mk_cl_Cnil);
  }
}

static mkcl_ihs_ptr
get_ihs_ptr(MKCL, mkcl_index n)
{
  mkcl_ihs_ptr p = env->ihs_top;
  if (n > p->index)
    mkcl_FEerror(env, "~D is an illegal IHS index. IHS_TOP = ~D.",
		 2, MKCL_MAKE_FIXNUM(n), MKCL_MAKE_FIXNUM(p->index));
  while (n < p->index)
    p = p->next;
  return p;
}

mkcl_object
mk_si_ihs_top_function_name(MKCL)
{
  return ihs_function_name(env->ihs_top->function);
}

mkcl_object
mk_si_ihs_top(MKCL)
{
  @(return MKCL_MAKE_FIXNUM(env->ihs_top->index));
}

mkcl_object
mk_si_ihs_prev(MKCL, mkcl_object x)
{
  @(return mk_cl_1M(env, x));
}

mkcl_object
mk_si_ihs_next(MKCL, mkcl_object x)
{
  @(return mk_cl_1P(env, x));
}

mkcl_object
mk_si_ihs_fun(MKCL, mkcl_object arg)
{
  @(return get_ihs_ptr(env, mkcl_integer_to_index(env, arg))->function);
}

mkcl_object
mk_si_ihs_env(MKCL, mkcl_object arg)
{
  @(return get_ihs_ptr(env, mkcl_integer_to_index(env, arg))->lex_env);
}

mkcl_object
mk_si_ihs_bds_marker(MKCL, mkcl_object arg)
{
  @(return MKCL_MAKE_FIXNUM(get_ihs_ptr(env, mkcl_integer_to_index(env, arg))->bds_marker));
}

/********************** FRAME STACK *************************/

static void
frs_set_size(MKCL, mkcl_index size)
{
  mkcl_frame_ptr old_org = env->frs_org;
  mkcl_index limit = env->frs_top - old_org;

  if (size <= limit) {
    mkcl_FEerror(env, "Cannot shrink frame stack below ~D.", 1,
	    mkcl_make_unsigned_integer(env, limit));
  } else {
    mkcl_interrupt_status old_intr;
    mkcl_index margin = mkcl_get_option(MKCL_OPT_FRAME_STACK_SAFETY_AREA);
    mkcl_frame_ptr org;
    size += 2*margin;

    org = mkcl_alloc(env, size * sizeof(*org));

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    memcpy(org, old_org, (limit + 1) * sizeof(*org));
    env->frs_top = org + limit;
    env->frs_org = org;
    env->frs_limit = org + (size - 2*margin);
    env->frs_size = size;
    mkcl_set_interrupt_status(env, old_intr);

    mkcl_dealloc(env, old_org);
  }
}

static void
frs_overflow(MKCL)
{
  mkcl_index margin = mkcl_get_option(MKCL_OPT_FRAME_STACK_SAFETY_AREA);
  mkcl_index size = env->frs_size;
  mkcl_frame_ptr org = env->frs_org;
  mkcl_frame_ptr last = org + size;
  if (env->frs_limit >= last) {
    mkcl_lose("Frame stack overflow, cannot grow larger.");
  }
  env->frs_limit += margin;
  mk_cl_cerror(env, 6, mkcl_make_simple_base_string(env, "Extend stack size"),
	       @'ext::stack-overflow', @':size', MKCL_MAKE_FIXNUM(size),
	       @':type', @'ext::frame-stack');
  frs_set_size(env, size + size / 2);
}

mkcl_frame_ptr
_mkcl_frs_push(MKCL, mkcl_object val)
{
  mkcl_frame_ptr output = ++env->frs_top;
  if (output >= env->frs_limit) {
    frs_overflow(env);
    output = env->frs_top;
  }
  output->frs_bds_top_index = env->bds_top - env->bds_org;
  output->frs_val = val;
  output->frs_ihs = env->ihs_top;
  output->frs_sp = MKCL_STACK_INDEX(env);

  mkcl_get_interrupt_status(env, output->frs_intr);

  return output;
}

void
mkcl_unwind(MKCL, mkcl_frame_ptr fr)
{
  if ( env->frs_top < env->frs_org )
    mkcl_lose("Tried to unwind an empty (FRS) frame stack!");

  mkcl_disable_interrupts(env); /* interrupts are re-enabled by the longjmp target. */
  env->nlj_fr = fr;
  while (env->frs_top != fr && env->frs_top->frs_val != MKCL_PROTECT_TAG)
    {
      --env->frs_top;
      if (env->frs_top < env->frs_org)
	mkcl_lose("Frame stack (FRS) undeflow!");
    }
  env->ihs_top = env->frs_top->frs_ihs;
  mkcl_bds_unwind(env, env->frs_top->frs_bds_top_index);
  MKCL_STACK_SET_INDEX(env, env->frs_top->frs_sp);
  mkcl_longjmp(env->frs_top->frs_jmpbuf, 1);
  /* never reached */
}

mkcl_frame_ptr
mkcl_frs_sch (MKCL, mkcl_object frame_id)
{
  mkcl_frame_ptr top;
  for (top = env->frs_top;  top >= env->frs_org;  top--)
    if (top->frs_val == frame_id)
      return(top);
  return(NULL);
}

static mkcl_frame_ptr
get_frame_ptr(MKCL, mkcl_object x)
{
  if (MKCL_FIXNUMP(x)) {
    mkcl_frame_ptr p = env->frs_org + mkcl_fixnum_to_word(x);
    if (env->frs_org <= p && p <= env->frs_top)
      return p;
  }
  mkcl_FEerror(env, "~S is an illegal frs index. FRS_TOP = ~S.", 2, x, mk_si_frs_top(env));
}

mkcl_object
mk_si_frs_top(MKCL)
{
  @(return MKCL_MAKE_FIXNUM(env->frs_top - env->frs_org));
}

mkcl_object
mk_si_frs_bds(MKCL, mkcl_object arg)
{
  @(return MKCL_MAKE_FIXNUM(get_frame_ptr(env, arg)->frs_bds_top_index));
}

mkcl_object
mk_si_frs_tag(MKCL, mkcl_object arg)
{
  @(return get_frame_ptr(env, arg)->frs_val);
}

mkcl_object
mk_si_frs_ihs(MKCL, mkcl_object arg)
{
  @(return MKCL_MAKE_FIXNUM(get_frame_ptr(env, arg)->frs_ihs->index));
}

mkcl_object
mk_si_sch_frs_base(MKCL, mkcl_object fr, mkcl_object ihs)
{
  mkcl_frame_ptr x;
  mkcl_index y = mkcl_integer_to_index(env, ihs);
  for (x = get_frame_ptr(env, fr); 
       x <= env->frs_top && x->frs_ihs->index < y;
       x++);
  @(return ((x > env->frs_top) ? mk_cl_Cnil : MKCL_MAKE_FIXNUM(x - env->frs_org)));
}

/************************************************************/

mkcl_object *
_mkcl_va_sp(MKCL, mkcl_narg narg)
{
  return env->stack_top - narg;
}


/********************* INITIALIZATION ***********************/

mkcl_object
mk_si_set_limit(MKCL, mkcl_object type, mkcl_object size)
{
  mkcl_index the_size = mkcl_integer_to_index(env, size);
  if (type == @'ext::frame-stack') {
    frs_set_size(env, the_size);
  } else if (type == @'ext::binding-stack') {
    mkcl_bds_set_size(env, the_size);
  } else if (type == @'ext::c-stack') {
    cs_set_size(env, the_size);
  } else if (type == @'ext::lisp-stack') {
    mkcl_stack_set_size(env, the_size);
  } else {
    _mkcl_set_max_heap_size(env, the_size);
  }
  return mk_si_get_limit(env, type);
}

mkcl_object
mk_si_get_limit(MKCL, mkcl_object type)
{
  mkcl_index output;
  if (type == @'ext::frame-stack') {
    output = env->frs_size;
  } else if (type == @'ext::binding-stack') {
    output = env->bds_size;
  } else if (type == @'ext::c-stack') {
    output = env->cs_size;
  } else if (type == @'ext::lisp-stack') {
    output = env->stack_size;
  } else {
    output = mkcl_core.max_heap_size;
  }
  @(return mkcl_make_unsigned_integer(env, output));
}

void mkcl_init_c_stack_control(MKCL, const void * new_cs_org)
{
  env->cs_org = (char *) new_cs_org;
  env->cs_barrier = (char *) new_cs_org;
#if defined(HAVE_SYS_RESOURCE_H) && defined(RLIMIT_STACK)
  {
    struct rlimit rl;
    mkcl_index size;
    if (getrlimit(RLIMIT_STACK, &rl))
      mkcl_C_lose("mkcl_init_c_stack_control failed on getrlimit().");
    if (rl.rlim_cur != RLIM_INFINITY) {
      size = rl.rlim_cur - (rl.rlim_cur / 16); 
      if (size > (mkcl_index)mkcl_get_option(MKCL_OPT_C_STACK_SIZE))
	mkcl_set_option(MKCL_OPT_C_STACK_SIZE, size);
#ifdef MKCL_DOWN_STACK
      env->cs_barrier = (char *) (env->cs_org - rl.rlim_cur - 1024);
#else
      env->cs_barrier = (char *) (env->cs_org + rl.rlim_cur + 1024);
#endif
    }
    else
      {
	size = 0; /* the OS said this was unlimited and we believe it! JCB */
	mkcl_set_option(MKCL_OPT_C_STACK_SIZE, size);
      }
  }
#else
#ifndef MKCL_WINDOWS
  /* On MS-Windows we fly blind! */
#error "Don't know how to query stack size!"
#endif
#endif
  cs_set_size(env, mkcl_get_option(MKCL_OPT_C_STACK_SIZE));
}

void
mkcl_init_stacks(MKCL, mkcl_env new_env)
{
  static struct mkcl_ihs_frame ihs_org = { NULL, NULL, NULL, 0};
  mkcl_index size, margin;

  margin = mkcl_get_option(MKCL_OPT_FRAME_STACK_SAFETY_AREA);
  size = mkcl_get_option(MKCL_OPT_FRAME_STACK_SIZE) + 2 * margin;
  new_env->frs_size = size;

  new_env->frs_org = (mkcl_frame_ptr)mkcl_alloc(env, size * sizeof(*new_env->frs_org));

  new_env->frs_top = new_env->frs_org-1; /* This value has undefined behavior in C99. JCB */
  new_env->frs_limit = &new_env->frs_org[size - 2*margin];

  margin = mkcl_get_option(MKCL_OPT_BIND_STACK_SAFETY_AREA);
  size = mkcl_get_option(MKCL_OPT_BIND_STACK_SIZE) + 2 * margin;
  new_env->bds_size = size;

  new_env->bds_org = (mkcl_bds_ptr)mkcl_alloc(env, size * sizeof(*new_env->bds_org));

  new_env->bds_top = new_env->bds_org-1; /* This value has undefined behavior in C99. JCB */
  new_env->bds_limit = &new_env->bds_org[size - 2*margin];

  new_env->ihs_top = &ihs_org;
  ihs_org.function = mk_cl_Cnil;
  ihs_org.lex_env = mk_cl_Cnil;
  ihs_org.index = 0;

#if defined(HAVE_SIGPROCMASK) && defined(SA_SIGINFO)
  if (mkcl_get_option(MKCL_OPT_SIGALTSTACK_SIZE)) {
    stack_t new_stack;
    mkcl_index size = mkcl_get_option(MKCL_OPT_SIGALTSTACK_SIZE);
    if (size < SIGSTKSZ) {
      size = SIGSTKSZ + (sizeof(double)*16) +
	(sizeof(mkcl_object)*4);
    }
    new_env->altstack_size = size;

    new_env->altstack = mkcl_alloc(env, size);

    new_stack.ss_size = new_env->altstack_size;
    new_stack.ss_sp = new_env->altstack;
    new_stack.ss_flags = 0;
    sigaltstack(&new_stack, NULL);
    /* DO NOT hope to ever be able to run any lisp code
       on this alternate stack! It has none of the C stack
       control set up for it and will send you in a tail spin
       on the first stack overflow check...
       This is quite clearly a C code only stack. JCB */
  }
#endif
}

