/* -*- mode: c -*- */
/*
    string.d -- String routines.
*/
/*
    Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
    Copyright (c) 1990, Giuseppe Attardi.
    Copyright (c) 2001, Juan Jose Garcia Ripoll.
    Copyright (c) 2011, Jean-Claude Beaudoin.

    MKCL is free software; you can redistribute it and/or
    modify it under thep terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file '../../Copyright' for full details.
*/


#include <mkcl/mkcl.h>
#include <string.h>
#include <mkcl/mkcl-inl.h>

#ifndef MKCL_WINDOWS
# ifdef MKCL_UNICODE
#  include <wchar.h>
# endif

#define mkcl_wmemcpy(d,s,l) ((mkcl_character *) wmemcpy((wchar_t *) d, (wchar_t *) s, l))
#define mkcl_wmemmove(d,s,l) ((mkcl_character *) wmemmove((wchar_t *) d, (wchar_t *) s, l))
#define mkcl_wmemchr(p,v,l) ((mkcl_character *) wmemchr((wchar_t *) p, (wchar_t) v, l))
#define mkcl_wmemset(p,v,l) ((mkcl_character *) wmemset((wchar_t *) p, (wchar_t) v, l))
#define mkcl_wmemcmp(p1,p2,l) wmemcmp((wchar_t *) p1, (wchar_t *) p2, l)
#else
static inline mkcl_character * mkcl_wmemcpy(mkcl_character * dest, const mkcl_character * src, size_t len)
{
  size_t i = 0;

  for (i = 0; i < len; i++)
    dest[i] = src[i];

  return dest;
}

static inline mkcl_character * mkcl_wmemmove(mkcl_character * dest, const mkcl_character * src, size_t len)
{
  size_t i;

  if ( dest < src )
    for (i = 0; i < len; i++)
      dest[i] = src[i];
  else if ( dest > src )
    for (i = len; i; i--)
      dest[i-1] = src[i-1];

  return dest;
}

static inline mkcl_character * mkcl_wmemchr(const mkcl_character * ptr, mkcl_character val, size_t len)
{
  size_t i;

  for (i = 0; i < len; i++)
    if (ptr[i] == val)
      return (mkcl_character *) (ptr + i);

  return NULL;
}

static inline mkcl_character * mkcl_wmemset(mkcl_character * ptr, mkcl_character val, size_t len)
{
  size_t i;

  for (i = 0; i < len; i++)
    ptr[i] = val;
  return ptr;
}

static inline int mkcl_wmemcmp(const mkcl_character * ptr1, const mkcl_character * ptr2, size_t len)
{
  size_t i;

  for (i = 0; i < len; i++)
    if (ptr1[i] != ptr2[i])
      return ptr1[i] - ptr2[i];
  
  return 0;
}
#endif /* MKCL_WINDOWS */

typedef mkcl_character (*mkcl_casefun)(mkcl_character, bool *);

static mkcl_object
do_make_base_string(MKCL, mkcl_index s, mkcl_base_char code)
{
  mkcl_object x = mkcl_alloc_simple_base_string(env, s);
  mkcl_index i;
  for (i = 0;  i < s;  i++)
    x->base_string.self[i] = code;
  return x;
}

#ifdef MKCL_UNICODE
static mkcl_object
do_make_string(MKCL, mkcl_index s, mkcl_character code)
{
  mkcl_object x = mkcl_alloc_simple_extended_string(env, s);
  mkcl_index i;
  for (i = 0;  i < s;  i++)
    x->string.self[i] = code;
  return x;
}
#else
#define do_make_string do_make_base_string
#endif

@(defun make_string (size &key (initial_element MKCL_CODE_CHAR(' '))
		     (element_type @'character'))
  mkcl_index s;
  mkcl_object x;
@
  s = mkcl_to_array_index(env, size);
  /* INV: mkcl_[base_]char_code() checks the type of initial_element() */
  if (element_type == @'base-char' || element_type == @'standard-char') {
    int code = mkcl_base_char_code(env, initial_element);
    x = do_make_base_string(env, s, code);
  } else if (element_type == @'character') {
    mkcl_index code = mkcl_char_code(env, initial_element);
    x = do_make_string(env, s, code);
  } else if (mk_cl_funcall(env, 3, @'subtypep', element_type, @'base-char') == mk_cl_Ct) {
    int code = mkcl_base_char_code(env, initial_element);
    x = do_make_base_string(env, s, code);
  } else if (mk_cl_funcall(env, 3, @'subtypep', element_type, @'character') == mk_cl_Ct) {
    mkcl_index code = mkcl_char_code(env, initial_element);
    x = do_make_string(env, s, code);
  } else {
    mkcl_FEerror(env, "The type ~S is not a valid string char type.", 1, element_type);
  }
  @(return x);
@)

mkcl_object
mkcl_alloc_simple_base_string(MKCL, mkcl_index length)
{
  mkcl_object x;

  x = mkcl_alloc_object(env, mkcl_t_base_string);
  x->base_string.hasfillp     = FALSE;
  x->base_string.adjustable   = FALSE;
  x->base_string.displaced    = mk_cl_Cnil;
  x->base_string.dim          = (x->base_string.fillp = length);
  x->base_string.elem         = mkcl_base_char_index;
  x->base_string.set          = mkcl_base_char_set_index;
  x->base_string.self         = (mkcl_base_char *)mkcl_alloc_atomic(env, length+1);
  x->base_string.self[length] = x->base_string.self[0] = 0;
  return x;
}

#ifdef MKCL_UNICODE
mkcl_object
mkcl_alloc_simple_extended_string(MKCL, mkcl_index length)
{
  mkcl_object x;

  /* should this call mk_si_make_vector? */
  x = mkcl_alloc_object(env, mkcl_t_string);
  x->string.hasfillp   = FALSE;
  x->string.adjustable = FALSE;
  x->string.displaced  = mk_cl_Cnil;
  x->string.dim        = x->string.fillp = length;
  x->string.elem       = mkcl_character_index;
  x->string.set        = mkcl_character_set_index;
  x->string.self       
    = (mkcl_character *) mkcl_alloc_align(env,
					  sizeof(mkcl_character)*(length + 1), /* FIXME! +1 to mask a bug! JCB */
					  sizeof(mkcl_character));
  return(x);
}
#endif

/*
  Make a string of a certain size, with some leading zeros to
  keep C happy. The string must be adjustable, to allow further
  growth. (See unixfsys.c for its use).
*/
mkcl_object
mkcl_alloc_adjustable_base_string(MKCL, mkcl_index l)
{
  mkcl_object output = mkcl_alloc_simple_base_string(env, l);
  output->base_string.fillp = 0;
  output->base_string.hasfillp = TRUE;
  output->base_string.adjustable = TRUE;
  return output;
}

#ifdef MKCL_UNICODE
mkcl_object
mkcl_alloc_adjustable_extended_string(MKCL, mkcl_index l)
{
  mkcl_object output = mkcl_alloc_simple_extended_string(env, l);
  output->base_string.fillp = 0;
  output->base_string.hasfillp = TRUE;
  output->base_string.adjustable = TRUE;
  return output;
}
#endif

/*
  mkcl_make_simple_base_string(s) makes a simple-base string from C string s.
*/
mkcl_object
mkcl_make_simple_base_string(MKCL, char *s)
{
  mkcl_object x;
  mkcl_index l = (s ? strlen(s) : 0);

  x = mkcl_alloc_object(env, mkcl_t_base_string);
  x->base_string.hasfillp = FALSE;
  x->base_string.adjustable = FALSE;
  x->base_string.displaced = mk_cl_Cnil;
  x->base_string.dim = (x->base_string.fillp = l);
  x->base_string.elem = mkcl_base_char_index;
  x->base_string.set = mkcl_base_char_set_index;
  x->base_string.self = (mkcl_base_char *)s;
	
  return x;
}

mkcl_object
mkcl_make_base_string_copy(MKCL, const char *s)
{
  mkcl_object x;
  mkcl_index l = (s ? strlen(s) : 0);
  
  x = mkcl_alloc_simple_base_string(env, l);
  if (s != NULL)
    memcpy(x->base_string.self, s, l);
  return x;
}

mkcl_object mkcl_copy_base_string(MKCL, mkcl_object s)
{
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(s))) mkcl_FEtype_error_base_string(env, s);

  mkcl_index len = s->base_string.fillp;
  mkcl_object new = mkcl_alloc_simple_base_string(env, len);
  if (len)
    memcpy(new->base_string.self, s->base_string.self, len);
  return new;
}

#ifdef MKCL_UNICODE
mkcl_object mkcl_copy_string(MKCL, mkcl_object s)
{
  mkcl_index len;
  mkcl_object new;

  if (mkcl_unlikely(!MKCL_STRINGP(s))) mkcl_FEtype_error_string(env, s);

  if (s->string.t == mkcl_t_string)
    {
      len = s->string.fillp;
      new = mkcl_alloc_simple_extended_string(env, len);
      if (len)
	mkcl_wmemcpy(new->string.self, s->string.self, len);
    }
  else
    {
      len = s->base_string.fillp;
      new = mkcl_alloc_simple_base_string(env, len);
      if (len)
	memcpy(new->base_string.self, s->base_string.self, len);
    }
  return new;
}
#endif

mkcl_object
mkcl_cstring_to_base_string_or_nil(MKCL, const char *s)
{
  if (s == NULL)
    return mk_cl_Cnil;
  else
    return mkcl_make_base_string_copy(env, s);
}

bool
mkcl_fits_in_base_string(MKCL, mkcl_object s)
{
 AGAIN:
  switch (mkcl_type_of(s)) {
#ifdef MKCL_UNICODE
  case mkcl_t_string: {
    mkcl_index i;
    for (i = 0; i < s->string.fillp; i++) {
      if (!MKCL_BASE_CHAR_CODE_P(s->string.self[i]))
	return false;
    }
    return true;
  }
#endif
  case mkcl_t_base_string:
    return true;
  default:
    s = mkcl_type_error(env, @'si::copy-to-simple-base-string',"",s,@'string');
    goto AGAIN;
  }
}

mkcl_object
mkcl_coerce_list_to_base_string(MKCL, mkcl_object l)
{
  mkcl_object s = mkcl_alloc_adjustable_base_string(env, 10);
  mkcl_loop_for_in(env, l) {
    mkcl_object maybe_char = MKCL_CAR(l);
    if (MKCL_BASE_CHAR_P(maybe_char))
      mkcl_base_string_push_extend(env, s, MKCL_CHAR_CODE(maybe_char));
    else
      mkcl_FEtype_error_base_char(env, maybe_char);
  } mkcl_end_loop_for_in;
  return s;
}

#ifdef MKCL_UNICODE
mkcl_object
mkcl_coerce_list_to_string(MKCL, mkcl_object l)
{
  mkcl_object s = mkcl_alloc_adjustable_extended_string(env, 10);
  mkcl_loop_for_in(env, l) {
    mkcl_object maybe_char = MKCL_CAR(l);
    if (MKCL_CHARACTERP(maybe_char))
      mkcl_string_push_extend(env, s, MKCL_CHAR_CODE(maybe_char));
    else
      mkcl_FEtype_error_character(env, maybe_char);
  } mkcl_end_loop_for_in;
  return s;
}
#endif

mkcl_object
mk_si_copy_to_simple_base_string(MKCL, mkcl_object x)
{
  mkcl_object y;
 AGAIN:
  switch(mkcl_type_of(x)) {
  case mkcl_t_symbol:
    x = x->symbol.name;
    goto AGAIN;
  case mkcl_t_character:
    x = mk_cl_string(env, x);
    goto AGAIN;
#ifdef MKCL_UNICODE
  case mkcl_t_string: {
    mkcl_index index, length = x->string.fillp;
    y = mkcl_alloc_simple_base_string(env, length);
    for (index=0; index < length; index++) {
      mkcl_character c = x->string.self[index];
      if (!MKCL_BASE_CHAR_CODE_P(c))
	mkcl_FEerror(env, "Cannot coerce string ~A to a base-string", 1, x);
      y->base_string.self[index] = c;
    }
    break;
  }
#endif
  case mkcl_t_base_string: {
    mkcl_index length = x->base_string.fillp;
    y = mkcl_alloc_simple_base_string(env, length);
    memcpy(y->base_string.self, x->base_string.self, length);
    break;
  }
  case mkcl_t_cons:
    y = mkcl_coerce_list_to_base_string(env, x);
    break;
  case mkcl_t_null:
    x = mk_cl_Cnil_symbol->symbol.name;
    goto AGAIN;
  default:
    x = mkcl_type_error(env, @'si::copy-to-simple-base-string',"",x,@'string');
    goto AGAIN;
  }
  @(return y);
}

mkcl_object
mk_cl_string(MKCL, mkcl_object x)
{
 AGAIN:
  switch (mkcl_type_of(x)) {
  case mkcl_t_symbol:
    x = x->symbol.name;
    break;
  case mkcl_t_character: {
    mkcl_object y;
    mkcl_character c = MKCL_CHAR_CODE(x);
#ifdef MKCL_UNICODE
    if (MKCL_BASE_CHAR_CODE_P(c)) {
      y = mkcl_alloc_simple_base_string(env, 1);
      y->base_string.self[0] = c;
      x = y;
    } else {
      y = mkcl_alloc_simple_extended_string(env, 1);
      y->string.self[0] = c;
      x = y;
    }
#else
    y = mkcl_alloc_simple_base_string(env, 1);
    y->base_string.self[0] = c;
    x = y;
    break;
#endif
  }
#ifdef MKCL_UNICODE
  case mkcl_t_string:
#endif
  case mkcl_t_base_string:
    break;
  case mkcl_t_null:
    x = mk_cl_Cnil_symbol->symbol.name;
    break;
  default:
    x = mkcl_type_error(env, @'string',"",x,@'string');
    goto AGAIN;
  }
  @(return x);
}

mkcl_object
mk_si_coerce_to_base_string(MKCL, mkcl_object x)
{
  if (mkcl_type_of(x) != mkcl_t_base_string) {
    x = mk_si_copy_to_simple_base_string(env, x);
  }
  @(return x);
}


#ifdef MKCL_UNICODE
mkcl_object
mk_si_coerce_to_extended_string(MKCL, mkcl_object x)
{
  mkcl_object y;
 AGAIN:
  switch (mkcl_type_of(x))
    {
    case mkcl_t_base_string:
      {
	mkcl_index index; const mkcl_index len = x->base_string.fillp;
	y = mkcl_alloc_simple_extended_string(env, len);
	for(index=0; index < len; index++) {
	  y->string.self[index] = x->base_string.self[index];
	}
	y->string.fillp = len;
      }
      break;
    case mkcl_t_string:
      y = x;
      break;
    case mkcl_t_character:
      y = mkcl_alloc_simple_extended_string(env, 1);
      y->string.self[0] = MKCL_CHAR_CODE(x);
      break;
    case mkcl_t_symbol:
      x = x->symbol.name;
      goto AGAIN;
    case mkcl_t_null:
      x = mk_cl_Cnil_symbol->symbol.name;
      goto AGAIN;
    case mkcl_t_cons:
      y = mkcl_coerce_list_to_string(env, x);
      break;
    default:
      x = mkcl_type_error(env, @'si::coerce-to-extended-string',"",x,@'string');
      goto AGAIN;
    }
  @(return y);
}
#endif

mkcl_object
mk_cl_char(MKCL, mkcl_object object, mkcl_object index)
{
  mkcl_index i;

  for (;;)
    switch(mkcl_type_of(object)) {
#ifdef MKCL_UNICODE
    case mkcl_t_string:
      if (MKCL_FIXNUMP(index) && ((i = mkcl_fixnum_to_word(index)) < object->string.dim))
	{ @(return MKCL_CODE_CHAR(object->string.self[i])); }
      else
	{
	  i = mkcl_fixnum_to_word(mkcl_ensure_valid_array_index_type(env, object, index));
	  @(return MKCL_CODE_CHAR(object->string.self[i]));
	}
#endif
    case mkcl_t_base_string:
      if (MKCL_FIXNUMP(index) && ((i = mkcl_fixnum_to_word(index)) < object->base_string.dim))
	{ @(return MKCL_CODE_CHAR(object->base_string.self[i])); }
      else
	{
	  i = mkcl_fixnum_to_word(mkcl_ensure_valid_array_index_type(env, object, index));
	  @(return MKCL_CODE_CHAR(object->base_string.self[i]));
	}
    default:
      object = mkcl_type_error(env, @'char',"",object,@'string');
      break;
    }
}

mkcl_character
mkcl_char(MKCL, mkcl_object object, mkcl_index index)
{
  /* CHAR bypasses fill pointers when accessing strings */

  for (;;)
    switch(mkcl_type_of(object)) {
#ifdef MKCL_UNICODE
    case mkcl_t_string:
      if (index >= object->string.dim)
	mkcl_FEillegal_index(env, object, MKCL_MAKE_FIXNUM(index));
      return object->string.self[index];
#endif
    case mkcl_t_base_string:
      if (index >= object->base_string.dim)
	mkcl_FEillegal_index(env, object, MKCL_MAKE_FIXNUM(index));
      return object->base_string.self[index];
    default:
      object = mkcl_type_error(env, @'char',"",object,@'string');
      break;
    }
}

mkcl_object
mk_si_char_set(MKCL, mkcl_object object, mkcl_object index, mkcl_object value)
{
  mkcl_index i;

  for (;;)
    switch(mkcl_type_of(object)) {
#ifdef MKCL_UNICODE
    case mkcl_t_string:
      {
	mkcl_character c;

	if (MKCL_CHARACTERP(value))
	  c = MKCL_CHAR_CODE(value);
	else
	  mkcl_FEtype_error_character(env, value);
      
	if (MKCL_FIXNUMP(index) && ((i = mkcl_fixnum_to_word(index)) < object->string.dim))
	  {
	    object->string.self[i] = c;
	    @(return value);
	  }
	else
	  {
	    i = mkcl_fixnum_to_word(mkcl_ensure_valid_array_index_type(env, object, index));
	    object->string.self[i] = c;
	    @(return value);
	  }
      }
#endif
    case mkcl_t_base_string:
      {
	mkcl_base_char c;

	if (MKCL_BASE_CHAR_P(value))
	  c = MKCL_CHAR_CODE(value);
	else
	  mkcl_FEwrong_type_argument(env, @'base-char', value);
	
	if (MKCL_FIXNUMP(index) && ((i = mkcl_fixnum_to_word(index)) < object->base_string.dim))
	  {
	    object->base_string.self[i] = c;
	    @(return value);
	  }
	else
	  {
	    i = mkcl_fixnum_to_word(mkcl_ensure_valid_array_index_type(env, object, index));
	    object->base_string.self[i] = c;
	    @(return value);
	  }
      }
    default:
      object = mkcl_type_error(env, @'char',"",object,@'string');
      break;
    }
}

mkcl_character
mkcl_char_set(MKCL, mkcl_object object, mkcl_index index, mkcl_character value)
{
 AGAIN:
  /* CHAR bypasses fill pointers when accessing strings */
  switch(mkcl_type_of(object)) {
#ifdef MKCL_UNICODE
  case mkcl_t_string:
    if (index >= object->string.dim)
      mkcl_FEillegal_index(env, object, MKCL_MAKE_FIXNUM(index));
    return object->string.self[index] = value;
#endif
  case mkcl_t_base_string:
    if (index >= object->base_string.dim)
      mkcl_FEillegal_index(env, object, MKCL_MAKE_FIXNUM(index));
    return object->base_string.self[index] = value;
  default:
    object = mkcl_type_error(env, @'si::char-set', "", object, @'string');
    goto AGAIN;
  }
}

void
mkcl_get_string_start_end(MKCL, mkcl_object string, mkcl_object start, mkcl_object end,
			  mkcl_index *ps, mkcl_index *pe)
{
  /* INV: works on both mkcl_t_base_string and mkcl_t_string */
  /* INV: Works with either string or symbol */
  if (!MKCL_FIXNUMP(start) || MKCL_FIXNUM_MINUSP(start))
    goto E;
  else
    *ps = mkcl_fixnum_to_word(start);
  if (mkcl_Null(end)) {
    *pe = string->vector.fillp;
    if (*pe < *ps)
      goto E;
  } else if (!MKCL_FIXNUMP(end) || MKCL_FIXNUM_MINUSP(end))
    goto E;
  else {
    *pe = mkcl_fixnum_to_word(end);
    if (*pe < *ps || *pe > string->vector.fillp)
      goto E;
  }
  return;

 E:
  mkcl_FEerror(env,
	       "~S and ~S are illegal as :START and :END~%"
	       "for the string designator ~S.",
	       3, start, end, string);
}

#ifdef MKCL_UNICODE
static int
compare_strings(MKCL, mkcl_object string1, mkcl_index s1, mkcl_index e1,
		mkcl_object string2, mkcl_index s2, mkcl_index e2,
		int case_sensitive, mkcl_index *m)
{
  mkcl_index c1, c2;
  for (; s1 < e1; s1++, s2++) {
    if (s2 >= e2) { /* s1 is longer than s2, therefore s2 < s1 */
      *m = s1;
      return +1;
    }
    c1 = mkcl_char(env, string1, s1);
    c2 = mkcl_char(env, string2, s2);
    if (!case_sensitive) {
      c1 = mkcl_char_upcase(c1);
      c2 = mkcl_char_upcase(c2);
    }
    if (c1 < c2) {
      *m = s1;
      return -1;
    } else if (c1 > c2) {
      *m = s1;
      return +1;
    }
  }
  *m = s1;
  if (s2 >= e2) {
    return 0;
  } else { /* s1 is shorter than s2, hence s1 < s2 */
    return -1;
  }
}
#endif

static int
compare_base(unsigned char *s1, mkcl_index l1, unsigned char *s2, mkcl_index l2,
	     int case_sensitive, mkcl_index *m)
{
  mkcl_index l, c1, c2;
  for (l = 0; l < l1; l++, s1++, s2++) {
    if (l == l2) { /* s1 is longer than s2, therefore s2 < s1 */
      *m = l;
      return +1;
    }
    c1 = *s1;
    c2 = *s2;
    if (!case_sensitive) {
      c1 = mkcl_char_upcase(c1);
      c2 = mkcl_char_upcase(c2);
    }
    if (c1 < c2) {
      *m = l;
      return -1;
    } else if (c1 > c2) {
      *m = l;
      return +1;
    }
  }
  *m = l;
  if (l1 == l2) 
    return 0;
  else { /* s1 is shorter than s2, hence s1 < s2 */
    return -1;
  }
}

@(defun string= (string1 string2 &key (start1 MKCL_MAKE_FIXNUM(0)) end1
		                      (start2 MKCL_MAKE_FIXNUM(0)) end2)
	mkcl_index s1, e1, s2, e2;
@
  /* AGAIN: */
  string1 = mk_cl_string(env, string1);
  string2 = mk_cl_string(env, string2);
  mkcl_get_string_start_end(env, string1, start1, end1, &s1, &e1);
  mkcl_get_string_start_end(env, string2, start2, end2, &s2, &e2);
  if (e1 - s1 != e2 - s2)
    { @(return mk_cl_Cnil); }
#ifdef MKCL_UNICODE
  switch(mkcl_type_of(string1)) {
  case mkcl_t_string:
    switch(mkcl_type_of(string2)) {
    case mkcl_t_string:
      while (s1 < e1)
	if (string1->string.self[s1++] != string2->string.self[s2++])
	  { @(return mk_cl_Cnil); }
      @(return mk_cl_Ct);
    case mkcl_t_base_string:
      while (s1 < e1)
	if (string1->string.self[s1++] != string2->base_string.self[s2++])
	  { @(return mk_cl_Cnil); }
      @(return mk_cl_Ct);
    }
    break;
  case mkcl_t_base_string:
    switch(mkcl_type_of(string2)) {
    case mkcl_t_string:
      while (s1 < e1)
	if (string1->base_string.self[s1++] != string2->string.self[s2++])
	  { @(return mk_cl_Cnil); }
      @(return mk_cl_Ct);
    case mkcl_t_base_string:
      while (s1 < e1)
	if (string1->base_string.self[s1++] != string2->base_string.self[s2++])
	  { @(return mk_cl_Cnil); }
      @(return mk_cl_Ct);
    }
    break;
  }
#else
  while (s1 < e1)
    if (string1->base_string.self[s1++] != string2->base_string.self[s2++])
      { @(return mk_cl_Cnil); }
#endif
  @(return mk_cl_Ct);
@)

/*
	This correponds to string= (just the string equality).
*/
bool
mkcl_string_E(MKCL, mkcl_object x, mkcl_object y)
{
  mkcl_index i, j;
 AGAIN:
#ifdef MKCL_UNICODE
  switch(mkcl_type_of(x)) {
  case mkcl_t_string:
    switch(mkcl_type_of(y)) {
    case mkcl_t_string: {
      i = x->string.fillp;
      j = y->string.fillp;
      if (mkcl_unlikely(i != j)) return FALSE;
      return mkcl_wmemcmp(x->string.self, y->string.self, i) == 0;
    }
    case mkcl_t_base_string: {
      mkcl_index index;
      i = x->string.fillp;
      j = y->base_string.fillp;
      if (mkcl_unlikely(i != j)) return FALSE;
      for(index=0; index<i; index++)
	if (x->string.self[index] != y->base_string.self[index])
	  return FALSE;
      return TRUE;
    }
    default:
      y = mkcl_type_error(env, @'string=', "", y, @'string');
      goto AGAIN;
    }
    break;
  case mkcl_t_base_string:
    switch(mkcl_type_of(y)) {
    case mkcl_t_string: {
      mkcl_index index;
      i = x->base_string.fillp;
      j = y->string.fillp;
      if (mkcl_unlikely(i != j)) return FALSE;
      for(index=0; index<i; index++)
	if (x->base_string.self[index] != y->string.self[index])
	  return FALSE;
      return TRUE;
    }
    case mkcl_t_base_string: {
      i = x->base_string.fillp;
      j = y->base_string.fillp;
      if (mkcl_unlikely(i != j)) return FALSE;
      return memcmp(x->base_string.self, y->base_string.self, i) == 0;
    }
    default:
      y = mkcl_type_error(env, @'string=', "", y, @'string');
      goto AGAIN;
    }
    break;
  default:
    x = mkcl_type_error(env, @'string=', "", x, @'string');
    goto AGAIN;
  }
#else
  while (mkcl_unlikely(!MKCL_STRINGP(x))) x = mkcl_type_error(env, @'string=', "", x, @'string');
  while (mkcl_unlikely(!MKCL_STRINGP(y))) y = mkcl_type_error(env, @'string=', "", y, @'string');

  i = x->base_string.fillp;
  j = y->base_string.fillp;

  if (mkcl_unlikely(i != j))
    return FALSE;
  else
    return memcmp(x->base_string.self, y->base_string.self, i) == 0;
#endif
}


@(defun string_equal (string1 string2 &key (start1 MKCL_MAKE_FIXNUM(0)) end1
		                           (start2 MKCL_MAKE_FIXNUM(0)) end2)
	mkcl_index s1, e1, s2, e2;
	int output;
@
/* AGAIN: */
  string1 = mk_cl_string(env, string1);
  string2 = mk_cl_string(env, string2);
  mkcl_get_string_start_end(env, string1, start1, end1, &s1, &e1);
  mkcl_get_string_start_end(env, string2, start2, end2, &s2, &e2);
  if (e1 - s1 != e2 - s2)
    { @(return mk_cl_Cnil); }
#ifdef MKCL_UNICODE
  if (mkcl_type_of(string1) != mkcl_t_base_string || mkcl_type_of(string2) != mkcl_t_base_string) {
    output = compare_strings(env, string1, s1, e1, string2, s2, e2, 0, &e1);
  } else
#endif
    output = compare_base(string1->base_string.self + s1, e1 - s1,
			  string2->base_string.self + s2, e2 - s2,
			  0, &e1);
  @(return ((output == 0) ? mk_cl_Ct : mk_cl_Cnil));
@)

static mkcl_object
string_compare(MKCL, mkcl_narg narg, int sign1, int sign2, int case_sensitive, mkcl_va_list ARGS)
{
  mkcl_object string1 = mkcl_va_arg(ARGS);
  mkcl_object string2 = mkcl_va_arg(ARGS);
  mkcl_index s1, e1, s2, e2;
  int output;
  mkcl_object result;
  mkcl_object KEYS[4];
#define start1 KEY_VARS[0]
#define end1 KEY_VARS[1]
#define start2 KEY_VARS[2]
#define end2 KEY_VARS[3]
#define start1p KEY_VARS[4]
#define start2p KEY_VARS[6]
  mkcl_object KEY_VARS[8];

  if (narg < 2) mkcl_FEwrong_num_arguments_anonym(env);
  KEYS[0]=@':start1';
  KEYS[1]=@':end1';
  KEYS[2]=@':start2';
  KEYS[3]=@':end2';
  mk_cl_parse_key(env, ARGS, 4, KEYS, KEY_VARS, NULL, FALSE, FALSE);

  string1 = mk_cl_string(env, string1);
  string2 = mk_cl_string(env, string2);
  if (start1p == mk_cl_Cnil) start1 = MKCL_MAKE_FIXNUM(0);
  if (start2p == mk_cl_Cnil) start2 = MKCL_MAKE_FIXNUM(0);
  mkcl_get_string_start_end(env, string1, start1, end1, &s1, &e1);
  mkcl_get_string_start_end(env, string2, start2, end2, &s2, &e2);
#ifdef MKCL_UNICODE
  if (mkcl_type_of(string1) != mkcl_t_base_string || mkcl_type_of(string2) != mkcl_t_base_string) {
    output = compare_strings(env, string1, s1, e1, string2, s2, e2, case_sensitive, &e1);
  } else
#endif
    {
      output = compare_base(string1->base_string.self + s1, e1 - s1,
			    string2->base_string.self + s2, e2 - s2,
			    case_sensitive, &e1);
      e1 += s1;
    }
  if (output == sign1 || output == sign2) {
    result = MKCL_MAKE_FIXNUM(e1);
  } else {
    result = mk_cl_Cnil;
  }
  @(return result);
#undef start1p
#undef start2p
#undef start1
#undef end1
#undef start2
#undef end2
}

@(defun string< (&rest args)
@
  return string_compare(env, narg, -1, -1, 1, args);
@)

@(defun string> (&rest args)
@
  return string_compare(env, narg, +1, +1, 1, args);
@)

@(defun string<= (&rest args)
@
  return string_compare(env, narg, -1, 0, 1, args);
@)

@(defun string>= (&rest args)
@
  return string_compare(env, narg, 0, +1, 1, args);
@)

@(defun string/= (&rest args)
@
  return string_compare(env, narg, -1, +1, 1, args);
@)

@(defun string-lessp (&rest args)
@
  return string_compare(env, narg, -1, -1, 0, args);
@)

@(defun string-greaterp (&rest args)
@
  return string_compare(env, narg, +1, +1, 0, args);
@)

@(defun string-not-greaterp (&rest args)
@
  return string_compare(env, narg, -1, 0, 0, args);
@)

@(defun string-not-lessp (&rest args)
@
  return string_compare(env, narg, 0, +1, 0, args);
@)

@(defun string-not-equal (&rest args)
@
  return string_compare(env, narg, -1, +1, 0, args);
@)


mkcl_object mkcl_base_string_NE(MKCL, mkcl_object str1, mkcl_object str2)
{ /* returns the "mismatch-index" on success or ((mkcl_index)-1) on failure. */

}

mkcl_object mkcl_base_string_L(MKCL, mkcl_object str1, mkcl_object str2)
{ /* returns the "mismatch-index" on success or ((mkcl_index)-1) on failure. */
  while (mkcl_unlikely(!MKCL_STRINGP(str1))) str1 = mkcl_type_error(env, @'string=', "", str1, @'string');
  while (mkcl_unlikely(!MKCL_STRINGP(str2))) str2 = mkcl_type_error(env, @'string=', "", str2, @'string');

  mkcl_index i = str1->base_string.fillp;
  mkcl_index j = str2->base_string.fillp;

  if (mkcl_unlikely(i != j))
    return mk_cl_Cnil;
  else
    return (memcmp(str1->base_string.self, str2->base_string.self, i) == 0) ? mk_cl_Ct : mk_cl_Cnil;
}

mkcl_object mkcl_base_string_G(MKCL, mkcl_object str1, mkcl_object str2)
{ /* returns the "mismatch-index" on success or ((mkcl_index)-1) on failure. */

}

mkcl_object mkcl_base_string_LE(MKCL, mkcl_object str1, mkcl_object str2)
{ /* returns the "mismatch-index" on success or ((mkcl_index)-1) on failure. */

}

mkcl_object mkcl_base_string_GE(MKCL, mkcl_object str1, mkcl_object str2)
{ /* returns the "mismatch-index" on success or ((mkcl_index)-1) on failure. */

}



/* ***************************************** */


bool
mkcl_member_char(MKCL, mkcl_character c, mkcl_object char_bag)
{
  mkcl_index i, f;

  if ( mkcl_Null(char_bag) ) return(FALSE);

 AGAIN:
  switch (mkcl_type_of(char_bag)) {
  case mkcl_t_cons:
    mkcl_loop_for_in(env, char_bag) {
      mkcl_object other = MKCL_CAR(char_bag);
      if (MKCL_CHARACTERP(other) && c == MKCL_CHAR_CODE(other))
	return(TRUE);
    } mkcl_end_loop_for_in;
    return(FALSE);
  case mkcl_t_vector:
    for (i = 0, f = char_bag->vector.fillp;  i < f;  i++) {
      mkcl_object other = char_bag->vector.self.t[i];
      if (MKCL_CHARACTERP(other) && c == MKCL_CHAR_CODE(other))
	return(TRUE);
    }
    return(FALSE);
#ifdef MKCL_UNICODE
  case mkcl_t_string:
    for (i = 0, f = char_bag->string.fillp;  i < f;  i++) {
      if (c == char_bag->string.self[i])
	return(TRUE);
    }
    return(FALSE);
#endif
  case mkcl_t_base_string:
    for (i = 0, f = char_bag->base_string.fillp;  i < f;  i++) {
      if (c == char_bag->base_string.self[i])
	return(TRUE);
    }
    return(FALSE);
  case mkcl_t_bitvector:
    return(FALSE);
  default:
    char_bag = mkcl_type_error(env, @'member',"",char_bag,@'sequence');
    goto AGAIN;
  }
}

static mkcl_object
string_trim0(MKCL, bool left_trim, bool right_trim, mkcl_object char_bag, mkcl_object strng)
{
  mkcl_index i, j;

  strng = mk_cl_string(env, strng);
  i = 0;
  j = mkcl_length(env, strng);
  if (left_trim) {
    for (;  i < j;  i++) {
      mkcl_index c = mkcl_char(env, strng, i);
      if (!mkcl_member_char(env, c, char_bag))
	break;
    }
  }
  if (right_trim) {
    for (; j > i; j--) {
      mkcl_index c = mkcl_char(env, strng, j-1);
      if (!mkcl_member_char(env, c, char_bag)) {
	break;
      }
    }
  }
  return mk_cl_subseq(env, 3, strng, MKCL_MAKE_FIXNUM(i), MKCL_MAKE_FIXNUM(j));
}

mkcl_object
mk_cl_string_trim(MKCL, mkcl_object char_bag, mkcl_object strng)
{
  return string_trim0(env, TRUE, TRUE, char_bag, strng);
}

mkcl_object
mk_cl_string_left_trim(MKCL, mkcl_object char_bag, mkcl_object strng)
{
  return string_trim0(env, TRUE, FALSE, char_bag, strng);
}

mkcl_object
mk_cl_string_right_trim(MKCL, mkcl_object char_bag, mkcl_object strng)
{
  return string_trim0(env, FALSE, TRUE, char_bag, strng);
}

static mkcl_object
string_case(MKCL, mkcl_narg narg, mkcl_casefun casefun, mkcl_va_list ARGS)
{
  mkcl_object strng = mkcl_va_arg(ARGS);
  mkcl_index s, e, i;
  bool b;
  mkcl_object KEYS[2];
#define start KEY_VARS[0]
#define end KEY_VARS[1]
#define startp KEY_VARS[2]
  mkcl_object conv;
  mkcl_object KEY_VARS[4];

  if (narg < 1) mkcl_FEwrong_num_arguments_anonym(env);
  KEYS[0]=@':start';
  KEYS[1]=@':end';
  mk_cl_parse_key(env, ARGS, 2, KEYS, KEY_VARS, NULL, FALSE, FALSE);

  strng = mk_cl_string(env, strng);
  conv  = mk_cl_copy_seq(env, strng);
  if (startp == mk_cl_Cnil)
    start = MKCL_MAKE_FIXNUM(0);
  mkcl_get_string_start_end(env, conv, start, end, &s, &e);
  b = TRUE;
#ifdef MKCL_UNICODE
  switch(mkcl_type_of(conv)) {
  case mkcl_t_string:
    for (i = s;  i < e;  i++)
      conv->string.self[i] = (*casefun)(conv->string.self[i], &b);
    break;
  case mkcl_t_base_string:
    for (i = s;  i < e;  i++)
      conv->base_string.self[i] = (*casefun)(conv->base_string.self[i], &b);
    break;
  }
#else
  for (i = s;  i < e;  i++)
    conv->base_string.self[i] = (*casefun)(conv->base_string.self[i], &b);
#endif
  @(return conv);
#undef startp
#undef start
#undef end
}

static mkcl_character
char_upcase(mkcl_character c, bool *bp)
{
  return mkcl_char_upcase(c);
}

@(defun string-upcase (&rest args)
@
  return string_case(env, narg, char_upcase, args);
@)

static mkcl_character
char_downcase(mkcl_character c, bool *bp)
{
  return mkcl_char_downcase(c);
}

@(defun string-downcase (&rest args)
@
  return string_case(env, narg, char_downcase, args);
@)

static mkcl_character
char_capitalize(mkcl_character c, bool *bp)
{
  if (mkcl_lower_case_p(c)) {
    if (*bp)
      c = mkcl_char_upcase(c);
    *bp = FALSE;
  } else if (mkcl_upper_case_p(c)) {
    if (!*bp)
      c = mkcl_char_downcase(c);
    *bp = FALSE;
  } else {
    *bp = !mkcl_alphanumericp(c);
  }
  return c;
}

@(defun string-capitalize (&rest args)
@
  return string_case(env, narg, char_capitalize, args);
@)


static mkcl_object
nstring_case(MKCL, mkcl_narg narg, mkcl_object fun, mkcl_casefun casefun, mkcl_va_list ARGS)
{
  mkcl_object strng = mkcl_va_arg(ARGS);
  mkcl_index s, e, i;
  bool b;
  mkcl_object KEYS[2];
#define start KEY_VARS[0]
#define end KEY_VARS[1]
#define startp KEY_VARS[2]
  mkcl_object KEY_VARS[4];

  if (narg < 1) mkcl_FEwrong_num_arguments_anonym(env);
  KEYS[0]=@':start';
  KEYS[1]=@':end';
  mk_cl_parse_key(env, ARGS, 2, KEYS, KEY_VARS, NULL, FALSE, FALSE);

  strng = mkcl_check_type_string(env, fun,strng);
  if (startp == mk_cl_Cnil) start = MKCL_MAKE_FIXNUM(0);
  mkcl_get_string_start_end(env, strng, start, end, &s, &e);
  b = TRUE;
#ifdef MKCL_UNICODE
  if (mkcl_type_of(strng) == mkcl_t_string) {
    for (i = s;  i < e;  i++)
      strng->string.self[i] = (*casefun)(strng->string.self[i], &b);
  } else {
    for (i = s;  i < e;  i++)
      strng->base_string.self[i] = (*casefun)(strng->base_string.self[i], &b);
  }
#else
  for (i = s;  i < e;  i++)
    strng->base_string.self[i] = (*casefun)(strng->base_string.self[i], &b);
#endif
  @(return strng);
#undef startp
#undef start
#undef end
}

@(defun nstring-upcase (&rest args)
@
  return nstring_case(env, narg, @'nstring-upcase', char_upcase, args);
@)

@(defun nstring-downcase (&rest args)
@
  return nstring_case(env, narg, @'nstring-downcase', char_downcase, args);
@)

@(defun nstring-capitalize (&rest args)
@
  return nstring_case(env, narg, @'nstring-capitalize', char_capitalize, args);
@)

@(defun si::concatenate_base_strings (&rest args)
  mkcl_index l;
  int i;
  mkcl_object output;
@
  if (narg == 1) {
    mkcl_object s = mkcl_va_arg(args);
    
    if (MKCL_BASE_STRING_P(s))
      output = mkcl_copy_base_string(env, s);
    else
      output = mk_si_coerce_to_base_string(env, s);
  } else {
    /* Compute final size and store NONEMPTY coerced strings. */
    for (i = 0, l = 0; i < narg; i++) {
      mkcl_object s = mk_si_coerce_to_base_string(env, mkcl_va_arg(args));
      if (s->base_string.fillp) {
	MKCL_STACK_PUSH(env, s);
	l += s->base_string.fillp;
      }
    }
    /* Do actual copying by recovering those strings */
    output = mkcl_alloc_simple_base_string(env, l);
    while (l) {
      mkcl_object s = MKCL_STACK_POP_UNSAFE(env);
      size_t bytes = s->base_string.fillp;
      l -= bytes;
      memcpy(output->base_string.self + l, s->base_string.self, bytes);
    }
  }
  @(return output);
@)

mkcl_object mkcl_concatenate_2_base_strings(MKCL, mkcl_object str1, mkcl_object str2)
{
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str1))) mkcl_FEtype_error_base_string(env, str1);
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str2))) mkcl_FEtype_error_base_string(env, str2);

  mkcl_index len1 = str1->base_string.fillp;
  mkcl_index len2 = str2->base_string.fillp;

  mkcl_object new = mkcl_alloc_simple_base_string(env, len1 + len2);

  memcpy(new->base_string.self, str1->base_string.self, len1);
  memcpy(&(new->base_string.self[len1]), str2->base_string.self, len2);
  return new;
}

mkcl_object mkcl_concatenate_3_base_strings(MKCL, mkcl_object str1, mkcl_object str2, mkcl_object str3)
{
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str1))) mkcl_FEtype_error_base_string(env, str1);
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str2))) mkcl_FEtype_error_base_string(env, str2);
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str3))) mkcl_FEtype_error_base_string(env, str3);

  mkcl_index len1 = str1->base_string.fillp;
  mkcl_index len2 = str2->base_string.fillp;
  mkcl_index len3 = str3->base_string.fillp;

  mkcl_object new = mkcl_alloc_simple_base_string(env, len1 + len2 + len3);

  memcpy(new->base_string.self, str1->base_string.self, len1);
  memcpy(&(new->base_string.self[len1]), str2->base_string.self, len2);
  memcpy(&(new->base_string.self[len1 + len2]), str3->base_string.self, len3);
  return new;
}

#ifdef MKCL_UNICODE
@(defun si::concatenate_strings (&rest args)
  mkcl_index l;
  int i;
  mkcl_object output;
@
  if (narg == 1) {
    mkcl_object s = mkcl_va_arg(args);
    
    if (MKCL_EXTENDED_STRING_P(s))
      output = mkcl_copy_string(env, s);
    else
      output = mk_si_coerce_to_extended_string(env, s);
  } else {
    /* Compute final size and store NONEMPTY coerced strings. */
    for (i = 0, l = 0; i < narg; i++) {
      mkcl_object s = mk_si_coerce_to_extended_string(env, mkcl_va_arg(args));
      if (s->string.fillp) {
	MKCL_STACK_PUSH(env, s);
	l += s->string.fillp;
      }
    }
    /* Do actual copying by recovering those strings */
    output = mkcl_alloc_simple_extended_string(env, l);
    while (l) {
      mkcl_object s = MKCL_STACK_POP_UNSAFE(env);
      size_t bytes = s->string.fillp;
      l -= bytes;
      mkcl_wmemcpy((output->string.self + l), s->string.self, bytes);
    }
  }
  @(return output);
@)

mkcl_object mkcl_concatenate_2_strings(MKCL, mkcl_object str1, mkcl_object str2)
{
  if (mkcl_unlikely(!MKCL_STRINGP(str1))) mkcl_FEtype_error_string(env, str1);
  if (mkcl_unlikely(!MKCL_STRINGP(str2))) mkcl_FEtype_error_string(env, str2);

  if (str1->d.t == mkcl_t_base_string) str1 = mk_si_coerce_to_extended_string(env, str1);
  if (str2->d.t == mkcl_t_base_string) str2 = mk_si_coerce_to_extended_string(env, str2);

  mkcl_index len1 = str1->string.fillp;
  mkcl_index len2 = str2->string.fillp;

  mkcl_object new = mkcl_alloc_simple_extended_string(env, len1 + len2);

  mkcl_wmemcpy(new->string.self, str1->string.self, len1);
  mkcl_wmemcpy(&(new->string.self[len1]), str2->string.self, len2);
  return new;
}

mkcl_object mkcl_concatenate_3_strings(MKCL, mkcl_object str1, mkcl_object str2, mkcl_object str3)
{
  if (mkcl_unlikely(!MKCL_STRINGP(str1))) mkcl_FEtype_error_string(env, str1);
  if (mkcl_unlikely(!MKCL_STRINGP(str2))) mkcl_FEtype_error_string(env, str2);
  if (mkcl_unlikely(!MKCL_STRINGP(str3))) mkcl_FEtype_error_string(env, str3);

  if (str1->d.t == mkcl_t_base_string) str1 = mk_si_coerce_to_extended_string(env, str1);
  if (str2->d.t == mkcl_t_base_string) str2 = mk_si_coerce_to_extended_string(env, str2);
  if (str3->d.t == mkcl_t_base_string) str3 = mk_si_coerce_to_extended_string(env, str3);

  mkcl_index len1 = str1->string.fillp;
  mkcl_index len2 = str2->string.fillp;
  mkcl_index len3 = str3->string.fillp;

  mkcl_object new = mkcl_alloc_simple_extended_string(env, len1 + len2 + len3);

  mkcl_wmemcpy(new->string.self, str1->string.self, len1);
  mkcl_wmemcpy(&(new->string.self[len1]), str2->string.self, len2);
  mkcl_wmemcpy(&(new->string.self[len1 + len2]), str3->string.self, len3);
  return new;
}

#endif

mkcl_base_char * mkcl_extend_base_string(MKCL, mkcl_object s)
{
  mkcl_object other;
  mkcl_index new_length;
  const mkcl_index fillp = s->base_string.fillp;

  if (!s->base_string.adjustable)
    mkcl_FEerror(env, "string-push-extend: the string ~S is not adjustable.", 1, s);

  /* The following code expects that (2 * MKCL_ADIMLIM) fits inside a mkcl_index without overflow! JCB */
  if (s->base_string.dim >= MKCL_ADIMLIM)
    mkcl_FEerror(env, "Can't extend the string.", 0);
  new_length = 1 + s->base_string.dim + (s->base_string.dim / 2);
  if (new_length > MKCL_ADIMLIM)
    new_length = MKCL_ADIMLIM;

  other = mkcl_alloc_adjustable_base_string(env, new_length);
  if (fillp) memcpy(other->base_string.self, s->base_string.self, fillp);
  other->base_string.fillp = fillp;

  s = mk_si_replace_array(env, s, other);
  return(s->base_string.self);
}

#ifdef MKCL_UNICODE
mkcl_character * mkcl_extend_string(MKCL, mkcl_object s)
{
  mkcl_object other;
  mkcl_index new_length;
  const mkcl_index fillp = s->string.fillp;

  if (!s->string.adjustable)
    mkcl_FEerror(env, "string-push-extend: the string ~S is not adjustable.", 1, s);

  /* The following code expects that (2 * MKCL_ADIMLIM) fits inside a mkcl_index without overflow! JCB */
  if (s->string.dim >= MKCL_ADIMLIM)
    mkcl_FEerror(env, "Can't extend the string.", 0);
  new_length = 1 + s->string.dim + (s->string.dim / 2);
  if (new_length > MKCL_ADIMLIM)
    new_length = MKCL_ADIMLIM;

  other = mkcl_alloc_adjustable_extended_string(env, new_length);
  if (fillp) mkcl_wmemcpy(other->string.self, s->string.self, fillp);
  other->string.fillp = fillp;
  s = mk_si_replace_array(env, s, other);

  return(s->string.self);  
}
#endif

union mkcl_array_data mkcl_extend_vector(MKCL, mkcl_object s)
{
  mkcl_object other;
  mkcl_index new_length;
  const mkcl_index fillp = s->vector.fillp;

  if (!s->vector.adjustable)
    mkcl_FEerror(env, "string-push-extend: the string ~S is not adjustable.", 1, s);

  /* The following code expects that (2 * MKCL_ADIMLIM) fits inside a mkcl_index without overflow! JCB */
  if (s->vector.dim >= MKCL_ADIMLIM)
    mkcl_FEerror(env, "Can't extend the string.", 0);
  new_length = 1 + s->vector.dim + (s->vector.dim / 2);
  if (new_length > MKCL_ADIMLIM)
    new_length = MKCL_ADIMLIM;

  other = mk_si_make_vector(env, mk_cl_array_element_type(env, s),
			    MKCL_MAKE_FIXNUM(new_length), mk_cl_Ct,
			    MKCL_MAKE_FIXNUM(s->vector.fillp),
			    mk_cl_Cnil, MKCL_MAKE_FIXNUM(0));
  mkcl_copy_subarray(env, other, 0, s, 0, s->vector.fillp);

  s = mk_si_replace_array(env, s, other);
  return(s->vector.self);
}


/********************************/

/* Sequence specializers for string and base-string. */

mkcl_object mkcl_base_substring(MKCL, mkcl_object str, mkcl_index start, mkcl_index end)
{
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str))) mkcl_FEtype_error_base_string(env, str);

  if (end == -1) end = str->base_string.fillp;

  if (start > str->base_string.fillp || start > end || end > str->base_string.fillp)
    mkcl_FEerror(env, "~S and ~S are invalid as :START and :END~%for the string sequence ~S.",
		 3, mkcl_make_unsigned_integer(env, start), mkcl_make_unsigned_integer(env, end), str);

  mkcl_index len = end - start;
  mkcl_object new = mkcl_alloc_simple_base_string(env, len);
  if (len)
    memcpy(new->base_string.self, &(str->base_string.self[start]), len);
  return new;
}

#ifdef MKCL_UNICODE
mkcl_object mkcl_substring(MKCL, mkcl_object str, mkcl_index start, mkcl_index end)
{
  mkcl_index len;
  mkcl_object new;

  if (mkcl_unlikely(!MKCL_STRINGP(str))) mkcl_FEtype_error_string(env, str);

  if (end == -1) end = str->base_string.fillp;

  if (start > str->base_string.fillp || start > end || end > str->base_string.fillp)
    mkcl_FEerror(env, "~S and ~S are invalid as :START and :END~%for the string sequence ~S.",
		 3, mkcl_make_unsigned_integer(env, start), mkcl_make_unsigned_integer(env, end), str);

  len = end - start;
  if (str->string.t == mkcl_t_string)
    {
      new = mkcl_alloc_simple_extended_string(env, len);
      if (len)
	memcpy(new->string.self, &(str->string.self[start]), len * sizeof(new->string.self[0]));
    }
  else
    {
      new = mkcl_alloc_simple_base_string(env, len);
      if (len)
	memcpy(new->base_string.self, &(str->base_string.self[start]), len * sizeof(new->base_string.self[0]));
    }

  return new;
}
#endif


mkcl_object
mkcl_fill_base_string(MKCL, mkcl_object str, mkcl_base_char ch)
{
  mkcl_index start = 0;
  mkcl_index end;

  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str))) mkcl_FEtype_error_base_string(env, str);

  end = str->base_string.fillp;

  if (end)
    memset(&(str->base_string.self[start]), ch, end);
  return str;
}

mkcl_object
mkcl_fill_base_string_k(MKCL, mkcl_object str, mkcl_base_char ch, mkcl_index start, mkcl_index end)
{
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str))) mkcl_FEtype_error_base_string(env, str);

  if (end == -1) end = str->base_string.fillp;

  if (start > str->base_string.fillp || start > end || end > str->base_string.fillp)
    mkcl_FEerror(env, "~S and ~S are invalid as :START and :END~%for the string sequence ~S.",
		 3, mkcl_make_unsigned_integer(env, start), mkcl_make_unsigned_integer(env, end), str);

  if (end)
    memset(&(str->base_string.self[start]), ch, end);
  return str;
}

#ifdef MKCL_UNICODE
mkcl_object
mkcl_fill_string(MKCL, mkcl_object str, mkcl_character ch)
{
  mkcl_index start = 0;
  mkcl_index end;

  if (mkcl_unlikely(!MKCL_STRINGP(str))) mkcl_FEtype_error_string(env, str);

  if (str->string.t == mkcl_t_string)
    {
      end = str->string.fillp;
      if (end)
	mkcl_wmemset(&(str->string.self[start]), ch, end);
    }
  else
    {
      end = str->base_string.fillp;
      if (!MKCL_BASE_CHAR_CODE_P(ch))
	mkcl_FEerror(env, "Cannot insert character ~A into a base-string", 1, MKCL_CODE_CHAR(ch));
      if (end)
	memset(&(str->base_string.self[start]), ch, end);
    }

  return str;
}

mkcl_object
mkcl_fill_string_k(MKCL, mkcl_object str, mkcl_character ch, mkcl_index start, mkcl_index end)
{
  if (mkcl_unlikely(!MKCL_STRINGP(str))) mkcl_FEtype_error_string(env, str);

  if (end == -1) end = str->string.fillp;

  if (start > str->string.fillp || start > end || end > str->string.fillp)
    mkcl_FEerror(env, "~S and ~S are invalid as :START and :END~%for the string sequence ~S.",
		 3, mkcl_make_unsigned_integer(env, start), mkcl_make_unsigned_integer(env, end), str);

  if (str->string.t == mkcl_t_string)
    {
      if (end)
	mkcl_wmemset(&(str->string.self[start]), ch, end);
    }
  else
    {
      end = str->base_string.fillp;
      if (!MKCL_BASE_CHAR_CODE_P(ch))
	mkcl_FEerror(env, "Cannot insert character ~A into a base-string", 1, MKCL_CODE_CHAR(ch));
      if (end)
	memset(&(str->base_string.self[start]), ch, end);
    }

  return str;
}
#endif

mkcl_object
mkcl_search_in_base_string(MKCL, mkcl_object str1, mkcl_object str2)
{
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str1))) mkcl_FEtype_error_base_string(env, str1);
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str2))) mkcl_FEtype_error_base_string(env, str2);
  
  const mkcl_index start1 = 0;
  const mkcl_index end1 = str1->base_string.fillp;
  mkcl_index start2 = 0;
  const mkcl_index end2 = str2->base_string.fillp;
  const mkcl_base_char * const self1 = str1->base_string.self;
  const mkcl_base_char * const self2 = str2->base_string.self;

  const mkcl_index len1 = end1 - start1;
  const mkcl_base_char ch0 = self1[start1];

  if (len1 == 0) return MKCL_MAKE_FIXNUM(start2);

  while (start2 + len1 <= end2)
    {
      mkcl_base_char * base = memchr(&(self2[start2]), ch0, end2 - start2);

      if (base && ((start2 = base - self2) + len1 <= end2))
	{
	  bool match = TRUE;
	  mkcl_index j = 0;
	  const mkcl_base_char * sub_self2 = self2 + ++start2;

	  if (0 == memcmp(self1 + 1, sub_self2, len1 - 1))
	    return MKCL_MAKE_FIXNUM(start2 - 1);
	}
      else
	return mk_cl_Cnil;
    }
  return mk_cl_Cnil;
}


#ifdef MKCL_UNICODE
mkcl_object
mkcl_search_in_string(MKCL, mkcl_object str1, mkcl_object str2)
{
  if (mkcl_unlikely(!MKCL_STRINGP(str1))) mkcl_FEtype_error_string(env, str1);
  if (mkcl_unlikely(!MKCL_STRINGP(str2))) mkcl_FEtype_error_string(env, str2);

  if (str2->string.t == mkcl_t_base_string)
    {
      if (str1->string.t == mkcl_t_string)
	str1 = mk_si_coerce_to_base_string(env, str1);
      return mkcl_search_in_base_string(env, str1, str2);
    }

  if (str1->string.t == mkcl_t_base_string)
    str1 = mk_si_coerce_to_extended_string(env, str1);
  
  const mkcl_index start1 = 0;
  const mkcl_index end1 = str1->string.fillp;
  mkcl_index start2 = 0;
  const mkcl_index end2 = str2->string.fillp;
  const mkcl_character * const self1 = str1->string.self;
  const mkcl_character * const self2 = str2->string.self;

  const mkcl_index len1 = end1 - start1;
  const mkcl_character ch0 = self1[start1];

  if (len1 == 0) return MKCL_MAKE_FIXNUM(start2);

  while (start2 + len1 <= end2)
    {
      mkcl_character * base = mkcl_wmemchr(&(self2[start2]), ch0, end2 - start2);

      if (base && ((start2 = base - self2) + len1 <= end2))
	{
	  bool match = TRUE;
	  mkcl_index j = 0;
	  const mkcl_character * sub_self2 = self2 + ++start2;

	  if (0 == mkcl_wmemcmp((self1 + 1), sub_self2, len1 - 1))
	    return MKCL_MAKE_FIXNUM(start2 - 1);
	}
      else
	return mk_cl_Cnil;
    }
  return mk_cl_Cnil;
}
#endif


mkcl_object mkcl_replace_in_base_string(MKCL, mkcl_object str1, mkcl_object str2)
{
  mkcl_index start1 = 0;
  mkcl_index start2 = 0;

  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str1))) mkcl_FEtype_error_base_string(env, str1);
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str2))) mkcl_FEtype_error_base_string(env, str2);

  mkcl_index end1 = str1->base_string.fillp;
  mkcl_index end2 = str2->base_string.fillp;
  mkcl_index len1 = end1 - start1;
  mkcl_index len2 = end2 - start2;

  if (len1 != 0 && len2 != 0)
    if (len1 < len2)
      memmove(&(str1->base_string.self[start1]), &(str2->base_string.self[start2]), len1);
    else
      memmove(&(str1->base_string.self[start1]), &(str2->base_string.self[start2]), len2);

  return str1;
}

mkcl_object mkcl_replace_in_base_string_k(MKCL, mkcl_object str1, mkcl_object str2,
					  mkcl_index start1, mkcl_index end1,
					  mkcl_index start2, mkcl_index end2)
{
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str1))) mkcl_FEtype_error_base_string(env, str1);
  if (mkcl_unlikely(!MKCL_BASE_STRING_P(str2))) mkcl_FEtype_error_base_string(env, str2);

  if (end1 == -1) end1 = str1->base_string.fillp;

  if (start1 > str1->base_string.fillp || start1 > end1 || end1 > str1->base_string.fillp)
    mkcl_FEerror(env, "~S and ~S are invalid as :START and :END~%for the string sequence ~S.",
		 3, mkcl_make_unsigned_integer(env, start1), mkcl_make_unsigned_integer(env, end1), str1);

  if (end2 == -1) end2 = str2->base_string.fillp;

  if (start2 > str2->base_string.fillp || start2 > end2 || end2 > str2->base_string.fillp)
    mkcl_FEerror(env, "~S and ~S are invalid as :START and :END~%for the string sequence ~S.",
		 3, mkcl_make_unsigned_integer(env, start2), mkcl_make_unsigned_integer(env, end2), str2);

  mkcl_index len1 = end1 - start1;
  mkcl_index len2 = end2 - start2;

  if (len1 != 0 && len2 != 0)
    if (len1 < len2)
      memmove(&(str1->base_string.self[start1]), &(str2->base_string.self[start2]), len1);
    else
      memmove(&(str1->base_string.self[start1]), &(str2->base_string.self[start2]), len2);

  return str1;
}

#ifdef MKCL_UNICODE
mkcl_object mkcl_replace_in_string(MKCL, mkcl_object str1, mkcl_object str2)
{
  mkcl_index start1 = 0;
  mkcl_index start2 = 0;

  if (mkcl_unlikely(!MKCL_STRINGP(str1))) mkcl_FEtype_error_string(env, str1);
  if (mkcl_unlikely(!MKCL_STRINGP(str2))) mkcl_FEtype_error_string(env, str2);

  if (str1->string.t != str2->string.t)
    {
      if (str1->string.t == mkcl_t_string)
	str2 = mk_si_coerce_to_extended_string(env, str2);
      else
	str2 = mk_si_coerce_to_base_string(env, str2);
    }

  mkcl_index end1 = str1->string.fillp;
  mkcl_index end2 = str2->string.fillp;
  mkcl_index len1 = end1 - start1;
  mkcl_index len2 = end2 - start2;

  if (len1 != 0 && len2 != 0)
    if (len1 < len2)
      if (str1->string.t == mkcl_t_string)
	mkcl_wmemmove(&(str1->string.self[start1]), &(str2->string.self[start2]), len1);
      else
	memmove(&(str1->base_string.self[start1]), &(str2->base_string.self[start2]), len1);
    else
      if (str1->string.t == mkcl_t_string)
	mkcl_wmemmove(&(str1->string.self[start1]), &(str2->string.self[start2]), len2);
      else
	memmove(&(str1->base_string.self[start1]), &(str2->base_string.self[start2]), len2);

  return str1;
}

mkcl_object mkcl_replace_in_string_k(MKCL, mkcl_object str1, mkcl_object str2,
				     mkcl_index start1, mkcl_index end1,
				     mkcl_index start2, mkcl_index end2)
{
  if (mkcl_unlikely(!MKCL_STRINGP(str1))) mkcl_FEtype_error_string(env, str1);
  if (mkcl_unlikely(!MKCL_STRINGP(str2))) mkcl_FEtype_error_string(env, str2);

  if (str1->string.t != str2->string.t)
    {
      if (str1->string.t == mkcl_t_string)
	str2 = mk_si_coerce_to_extended_string(env, str2);
      else
	str2 = mk_si_coerce_to_base_string(env, str2);
    }

  if (end1 == -1) end1 = str1->string.fillp;

  if (start1 > str1->string.fillp || start1 > end1 || end1 > str1->string.fillp)
    mkcl_FEerror(env, "~S and ~S are invalid as :START and :END~%for the string sequence ~S.",
		 3, mkcl_make_unsigned_integer(env, start1), mkcl_make_unsigned_integer(env, end1), str1);

  if (end2 == -1) end2 = str2->string.fillp;

  if (start2 > str2->string.fillp || start2 > end2 || end2 > str2->string.fillp)
    mkcl_FEerror(env, "~S and ~S are invalid as :START and :END~%for the string sequence ~S.",
		 3, mkcl_make_unsigned_integer(env, start2), mkcl_make_unsigned_integer(env, end2), str2);

  mkcl_index len1 = end1 - start1;
  mkcl_index len2 = end2 - start2;

  if (len1 != 0 && len2 != 0)
    if (len1 < len2)
      if (str1->string.t == mkcl_t_string)
	mkcl_wmemmove(&(str1->string.self[start1]), &(str2->string.self[start2]), len1);
      else
	memmove(&(str1->base_string.self[start1]), &(str2->base_string.self[start2]), len1);
    else
      if (str1->string.t == mkcl_t_string)
	mkcl_wmemmove(&(str1->string.self[start1]), &(str2->string.self[start2]), len2);
      else
	memmove(&(str1->base_string.self[start1]), &(str2->base_string.self[start2]), len2);
  
  return str1;
}
#endif /* def MKCL_UNICODE */

