/* -*- mode: c -*- */
/*
    threads.d -- Posix threads.
*/
/*
    Copyright (c) 2003, Juan Jose Garcia Ripoll.
    Copyright (c) 2010-2011 Jean-Claude Beaudoin.

    MKCL is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file '../../Copyright' for full details.
*/

#include <mkcl/mkcl.h>
#include <mkcl/mkcl-gc.h>
#include <mkcl/mkcl-math.h>
#include <mkcl/mkcl-inl.h>
#include <mkcl/internal.h>

#include <limits.h>

#include <errno.h>
#include <time.h>
#include <stdlib.h>
#include <signal.h>

#ifdef HAVE_GETTIMEOFDAY
# include <sys/time.h>
#endif
#ifdef HAVE_SCHED_YIELD
# include <sched.h>
#endif

#ifdef MKCL_WINDOWS
static DWORD cl_env_key;
#else
static pthread_key_t cl_env_key;
#endif


#ifdef linux
/* This _true_pthread_join gives you direct access to the real pthread_join,
   not the Boehm's GC wrapped one.
*/
static int _true_pthread_join(pthread_t thread, void ** retval);
#endif /* linux */

static void setup_thread_bindings(MKCL, mkcl_object initial_bindings);


const mkcl_env mkcl_thread_env(void)
{
#ifdef MKCL_WINDOWS
  return TlsGetValue(cl_env_key);
#else
  return pthread_getspecific(cl_env_key);
#endif
}

static void
mkcl_set_thread_env(MKCL)
{
#ifdef WITH___THREAD
  mk_cl_env_p = env;
#elif defined(MKCL_WINDOWS)
  if (!TlsSetValue(cl_env_key, env))
    mkcl_FEwin32_error(env, "mkcl_set_thread_env failed on TlsSetValue", 0);
#else
  if (pthread_setspecific(cl_env_key, env))
    mkcl_FElibc_error(env, "mkcl_set_thread_env failed on pthread_setspecific().", 0);
#endif
}

inline mkcl_object
mk_mt_current_thread(MKCL)
{
  @(return env->own_thread);
}

/*----------------------------------------------------------------------
 * THREAD OBJECT
 */

inline static void
mkcl_assert_type_thread(MKCL, mkcl_object o)
{
  if (mkcl_type_of(o) != mkcl_t_thread)
    mkcl_FEwrong_type_argument(env, @'mt::thread', o);
}

static void
thread_cleanup(void *aux)
{
  /* This routine performs some cleanup before a thread is completely
   * killed. For instance, it has to remove the associated thread
   * object from the list, an it has to dealloc some memory.
   *
   * Well, it used to do that. JCB
   *
   * NOTE: thread_cleanup() does not provide enough "protection". In
   * order to ensure that all UNWIND-PROTECT forms are properly
   * executed, never use pthread_cancel() to kill a thread, but
   * rather use the lisp functions mk_mt_interrupt_thread() and
   * mk_mt_thread_kill().
   */
  mkcl_object thread = (mkcl_object)aux;
  const mkcl_env env = thread->thread.env;

  thread->thread.active = 0;

  if (thread->thread.result_value == @':detached'
      || thread->thread.result_value == @':imported'
      )
    {
      volatile bool locked = false;

      MKCL_UNWIND_PROTECT_BEGIN(env) {
	mkcl_interrupt_status old_intr;

	mkcl_get_interrupt_status(env, old_intr);
	mkcl_disable_interrupts(env);
	MKCL_THREAD_LIST_LOCK(); locked = true;
	mkcl_set_interrupt_status(env, old_intr);
	
	mkcl_core.threads = mkcl_remove_eq(env, thread, mkcl_core.threads);
      } MKCL_UNWIND_PROTECT_EXIT {
	if (locked) MKCL_THREAD_LIST_UNLOCK();
      } MKCL_UNWIND_PROTECT_END;

#ifdef MKCL_WINDOWS      
      CloseHandle(thread->thread.thread);
#endif
    }

  /* We clear all references from the thread structure
     in an effort to prevent the GC from walking into
     an unmapped memory area and thus crash.
  */
  mkcl_set_thread_env(NULL);
  thread->thread.interrupt = mk_cl_Cnil;
  thread->thread.sigmask_frs_marker = NULL;
  {
    int i;

    for (i = 0; i < MKCL_MAX_INTERRUPTS; i++)
      {
	thread->thread.interrupted_threads[i].tid = 0;
	thread->thread.interrupted_threads[i].cs_org = NULL;
	thread->thread.interrupted_threads[i].cs_limit = NULL;
	thread->thread.interrupted_threads[i].cs_barrier = NULL;
	thread->thread.interrupted_threads[i].cs_size = 0;
      }
  }
}

#ifdef MKCL_WINDOWS
# if 0
/* For use with direct call to CreateThread */
typedef DWORD thread_value_t;
#  define CALL_CONV WINAPI
# else
/* For calls to _beginthreadex */
typedef unsigned thread_value_t;
#  define CALL_CONV __stdcall
# endif
#elif defined(linux)
typedef void * thread_value_t;
# define CALL_CONV
#endif

static thread_value_t CALL_CONV thread_entry_point(void *arg)
{
  volatile bool locked = false;
  mkcl_object thread = (mkcl_object)arg;
  const mkcl_env env = thread->thread.env;
  thread_value_t status = (thread_value_t) MKCL_THREAD_NORMAL_EXIT;

  /* 1) Setup the environment for the execution of the thread */
  mkcl_set_thread_env(env);
  thread->thread.tid = mkcl_gettid();
  mkcl_init_c_stack_control(env, &env);

#ifndef MKCL_WINDOWS
  pthread_cleanup_push(thread_cleanup, (void *)thread);
#endif

  MKCL_UNWIND_PROTECT_BEGIN(env) {
    mkcl_interrupt_status old_intr;

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    MKCL_THREAD_LIST_LOCK(); locked = true;
    mkcl_set_interrupt_status(env, old_intr);

    mkcl_core.threads = MKCL_CONS(env, thread, mkcl_core.threads);
  } MKCL_UNWIND_PROTECT_EXIT {
    if (locked) MKCL_THREAD_LIST_UNLOCK();
  } MKCL_UNWIND_PROTECT_END;

  if (!mkcl_Null(thread->thread.initial_bindings))
    setup_thread_bindings(env, thread->thread.initial_bindings);

  mk_si_clear_all_fpe(env);
  mk_si_enable_fpe(env, @':default');

  /* 2) Execute the code. The CATCH_ALL point
   *     provides us with an elegant way to exit the thread: we just
   *     do an unwind up to frs_top.
   */
  /* (((env)->frs_top < (env)->frs_org) 
      ? fprintf(stderr, "\n*** Start of thread_entry_point, frs_top UNDERFLOW! (%ld)\n",
       (env)->frs_org - (env)->frs_top), fflush(NULL)
      : 0); */
  MKCL_CATCH_ALL_BEGIN(env) {
    static mkcl_object top_apply_fun = mk_cl_Cnil;
    mkcl_object value;

    mkcl_bds_bind(env, @'mt::*thread*', thread);
    mkcl_bds_bind(env, @'si::*dynamic-cons-stack*', mk_cl_Cnil);
    mk_si_trim_dynamic_cons_stack(env);
    thread->thread.active = 1;
    mkcl_enable_interrupts(env);
    if (mkcl_Null(top_apply_fun))
      {
	top_apply_fun = MKCL_SYM_FUN(@'si::top-apply');
	if (mkcl_Null(top_apply_fun))
	  value = mk_cl_apply(env, 2, thread->thread.function, thread->thread.args);
	else
	  value = mk_cl_funcall(env, 3,
			     top_apply_fun,
			     thread->thread.function, thread->thread.args);
      }
    else
      value = mk_cl_funcall(env, 3,
			 top_apply_fun,
			 thread->thread.function, thread->thread.args);
    /* mkcl_bds_unwind1(env); */
    /* mkcl_bds_unwind1(env); */
    if (thread->thread.result_value == MKCL_OBJNULL)
      thread->thread.result_value = value;
    status = (thread_value_t) MKCL_THREAD_NORMAL_EXIT;
  } MKCL_CATCH_ALL_IF_CAUGHT {
    /* mkcl_bds_unwind1(env); */
    /* mkcl_bds_unwind1(env); */
    status = (thread_value_t) MKCL_THREAD_KILLED;
  } MKCL_CATCH_ALL_END;
  thread->thread.active = 0;

  /* 3) If everything went right, we should be exiting the thread
   *    through this point. thread_cleanup is automatically invoked.
   */
#ifdef MKCL_WINDOWS
  thread_cleanup(thread);
#else
  pthread_cleanup_pop(1); /* This executes the top-most clean-up handler.
			     Not necessarily the one we intended if there
			     was somewhere an unbalanced push. JCB
			     We could be a bit more defensive here...
			  */
#endif
  return status;
}

#ifdef linux

static void * signal_servicing_thread_entry_point(void *arg)
{
  volatile bool locked = false;
  mkcl_object thread = MKCL_CAR((mkcl_object) arg);
  const mkcl_env env = thread->thread.env;
  int sig = mkcl_fixnum_to_word(MKCL_CDR((mkcl_object) arg));

  /* 1) Setup the environment for the execution of the thread */
  mkcl_set_thread_env(env);
  thread->thread.tid = mkcl_gettid();
  mkcl_init_c_stack_control(env, &env);

  pthread_cleanup_push(thread_cleanup, (void *)thread);

  MKCL_UNWIND_PROTECT_BEGIN(env) {
    mkcl_interrupt_status old_intr;

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    MKCL_THREAD_LIST_LOCK(); locked = true;
    mkcl_set_interrupt_status(env, old_intr);

    mkcl_core.threads = MKCL_CONS(env, thread, mkcl_core.threads);
  } MKCL_UNWIND_PROTECT_EXIT {
    if (locked) MKCL_THREAD_LIST_UNLOCK();
  } MKCL_UNWIND_PROTECT_END;
  
  if (!mkcl_Null(thread->thread.initial_bindings))
    setup_thread_bindings(env, thread->thread.initial_bindings);

  mk_si_clear_all_fpe(env);
  mk_si_enable_fpe(env, @':default');

  /* 2) Execute the code. The CATCH_ALL point
   *     provides us with an elegant way to exit the thread: we just
   *     do an unwind up to frs_top.
   */
  MKCL_CATCH_ALL_BEGIN(env) {
    mkcl_bds_bind(env, @'mt::*thread*', thread);
    mkcl_bds_bind(env, @'si::*dynamic-cons-stack*', mk_cl_Cnil);
    mk_si_trim_dynamic_cons_stack(env);
    thread->thread.active = 1;
    mkcl_enable_interrupts(env);

    for (;;)
      {
	int rc = 0;
	int received_signo;

	do {
	  rc = sem_wait(&mkcl_signals[sig].sem); /* a cancel point */
	} while ( rc && errno == EINTR );
	if ( rc ) mkcl_C_lose("signal_servicing_thread_entry_point failed on sem_wait.");

	received_signo = ((sig == 0) ? mkcl_terminal_signal_number : sig);
	mk_cl_funcall(env, 2, thread->thread.function, MKCL_MAKE_FIXNUM(received_signo));
      }
    /* mkcl_bds_unwind1(env); */
    /* mkcl_bds_unwind1(env); */
  } MKCL_CATCH_ALL_IF_CAUGHT {
    /* mkcl_bds_unwind1(env); */
    /* mkcl_bds_unwind1(env); */
    /* We could add some special action here to help stack unwinding
       but we don't need to do anything yet. */
  } MKCL_CATCH_ALL_END;
  thread->thread.active = 0;  

  pthread_cleanup_pop(1);
  return NULL;
}

#endif /* linux */

#ifdef linux
static pthread_mutex_t mkcl_interrupt_thread_lock;
static sigset_t mkcl_standard_sigmask;
#elif defined(MKCL_WINDOWS)
static HANDLE mkcl_interrupt_thread_lock;
#endif


#ifdef linux

static sem_t mkcl_run_interrupt_function;

static void *
interrupt_thread_entry_point(void *arg)
{
  int rc = 0;
  char c_stack_mark = 0;
  mkcl_object thread = (mkcl_object)arg;
  const mkcl_env env = thread->thread.env;

  pthread_sigmask(SIG_SETMASK, &mkcl_standard_sigmask, NULL);
  pthread_detach(pthread_self()); /* no join for interrupt threads. */

  do {
    rc = sem_wait(&mkcl_run_interrupt_function);
  } while ( rc && errno == EINTR );
  if ( rc ) mkcl_C_lose("interrupt_thread_entry_point failed on sem_wait.");

  /* 1) Setup the environment for the execution of the thread */
  mkcl_set_thread_env(env);
	
  /* We need to adjust the C stack here. JCB */
  mkcl_init_c_stack_control(env, &c_stack_mark);

  mk_si_clear_all_fpe(env);
  mkcl_reactivate_fpe_set(env);

  /* 2) Execute the code. The CATCH_ALL point is the destination
   *     provides us with an elegant way to exit the thread: we just
   *     do an unwind up to frs_top.
   */
  /* We should have a special unwind-protect here. */
  MKCL_CATCH_ALL_BEGIN(env) {
    mkcl_enable_interrupts(env);
    mk_cl_funcall(env, 1, thread->thread.interrupt);
    mkcl_disable_interrupts(env);
    env->nlj_fr = NULL; /* make sure to signal a normal exit */
#if 0
  } MKCL_CATCH_ALL_IF_CAUGHT {
    /* We could add some special action here to help stack unwinding
       but we don't need to do anything yet. */
#endif
  } MKCL_CATCH_ALL_END;

  if (pthread_mutex_lock(&mkcl_interrupt_thread_lock))
    mkcl_lose("interrupt_thread_entry_point failed on pthread_mutex_lock.");

  {
    int rc = 0;
    int sig = mkcl_get_option(MKCL_OPT_THREAD_RESUME_SIGNAL);
    int i = thread->thread.interrupt_count - 1;
    struct interrupted_thread_ctrl * p 
      = &(thread->thread.interrupted_threads[i]);
	  
    thread->thread.thread = p->tid;
    env->cs_org = p->cs_org;
    env->cs_limit = p->cs_limit;
    env->cs_barrier = p->cs_barrier;
    env->cs_size = p->cs_size;

    if ( env->nlj_fr == NULL )
      { /* This is a normal (not unwinding) return. */
	env->disable_interrupts = p->disable_interrupts;
#if MKCL_DEBUG_INTERRUPT_MASK	
	env->interrupt_disabler_file = p->interrupt_disabler_file;
	env->interrupt_disabler_lineno = p->interrupt_disabler_lineno;
#endif
      }

    thread->thread.interrupt_count = i; /* interrupted_threads stack poped */

    mkcl_interrupted_thread_env = env;
    env->own_thread->thread.resume_handler_ran = FALSE;

    /* resume the interrupted thread */
    if ((rc = pthread_kill(thread->thread.thread, sig)))
      switch (rc /* errno */)
	{
	case ESRCH: /* the interrupted thread died on us unexpectedly! */
	  /* Should we flag this untimely death or simply wimper away? */
	  /* Let's be quiet for now. */
	  break;
	case EINVAL:
	  mkcl_lose("interrupt_thread_entry_point: invalid signal used with pthread_kill.");
	  break;
	default:
	  mkcl_lose("interrupt_thread_entry_point failed on pthread_kill.");
	  break;
	}
    else
      {
	do {
	  rc = sem_wait(&mkcl_interrupted_thread_resumed);
	} while ( rc && errno == EINTR );
	if ( rc ) mkcl_C_lose("interrupt_thread_entry_point failed on sem_wait.");
      }
    mkcl_interrupted_thread_env = NULL;
  }

  if (pthread_mutex_unlock(&mkcl_interrupt_thread_lock))
    mkcl_lose("interrupt_thread_entry_point failed on pthread_mutex_unlock.");

  return NULL;
}

#elif defined(MKCL_WINDOWS)

static void resume_thread_unwinding(void)
{ /* This function is never called through normal means.
     Its call is the result a direct instruction pointer
     manipulation through the "CONTEXT" structure.
  */
  /* In this function we cannot assume that we have a valid
     stack frame to work with, so no automatic variable allowed!
  */

  mkcl_unwind(MKCL_ENV(), MKCL_ENV()->nlj_fr);
  /* We should never return from the line just above! */
  mkcl_lose("resume_thread_unwinding failed!");
}

static thread_value_t CALL_CONV interrupt_thread_entry_point(void * arg)
{
  char c_stack_mark = 0;
  mkcl_object thread = (mkcl_object)arg;
  const mkcl_env env = thread->thread.env;

  /* 1) Setup the environment for the execution of the thread */
  mkcl_set_thread_env(env);
	
  /* We need to adjust the C stack here. JCB */
  mkcl_init_c_stack_control(env, &c_stack_mark);

  mk_si_clear_all_fpe(env);
  mkcl_reactivate_fpe_set(env);

  /* 2) Execute the code. The CATCH_ALL point is the destination
   *     provides us with an elegant way to exit the thread: we just
   *     do an unwind up to frs_top.
   */
  /* We should have a special unwind-protect here. */
  MKCL_CATCH_ALL_BEGIN(env) {
    mkcl_enable_interrupts(env);
    mk_cl_funcall(env, 1, thread->thread.interrupt);
    env->nlj_fr = NULL; /* make sure to signal a normal exit */
#if 0
  } MKCL_CATCH_ALL_IF_CAUGHT {
    /* We could add some special action here to help stack unwinding
       but we don't need to do anything yet. */
#endif
  } MKCL_CATCH_ALL_END;

  /* There is a hole here where we could get interrupted with
     our pants down! JCB
  */

  switch (WaitForSingleObject(mkcl_interrupt_thread_lock, INFINITE))
    {
    case WAIT_OBJECT_0: break;
    case WAIT_TIMEOUT:
    case WAIT_ABANDONED:
    case WAIT_FAILED:
    default:
      mkcl_FEwin32_error(env, "interrupt-thread failed to acquire lock.", 0);
    }

  {
    HANDLE os_thread;
    int i = thread->thread.interrupt_count - 1;
    struct interrupted_thread_ctrl * p 
      = &(thread->thread.interrupted_threads[i]);
	  
    os_thread = thread->thread.thread = p->tid;
    env->cs_org = p->cs_org;
    env->cs_limit = p->cs_limit;
    env->cs_barrier = p->cs_barrier;
    env->cs_size = p->cs_size;

    if ( env->nlj_fr == NULL )
      { /* This is a normal (not unwinding) return. */
	env->disable_interrupts = p->disable_interrupts;
#if MKCL_DEBUG_INTERRUPT_MASK	
	env->interrupt_disabler_file = p->interrupt_disabler_file;
	env->interrupt_disabler_lineno = p->interrupt_disabler_lineno;
#endif
      }

    thread->thread.interrupt_count = i; /* interrupted_threads stack poped */

    env->own_thread->thread.resume_handler_ran = FALSE; /* needed? No. */

    if ( env->nlj_fr != NULL )
      {
	CONTEXT context;
	context.ContextFlags = CONTEXT_CONTROL | CONTEXT_INTEGER;
	if (!GetThreadContext(os_thread, &context))
	  mkcl_FEwin32_error(env, "Cannot get context for thread ~A", 1, thread);
#if defined(_X86_)
	context.Eip = (DWORD) resume_thread_unwinding;
#elif defined(__x86_64)
	context.Rip = (DWORD64) resume_thread_unwinding;
#else
#error Unknown processor architecture
#endif
	if (!SetThreadContext(os_thread, &context))
	  mkcl_FEwin32_error(env, "Cannot set context for thread ~A", 1, thread);
      }

    {
      DWORD count;

      if ((count = ResumeThread(os_thread)) == (DWORD)-1)
	mkcl_FEwin32_error(env, "Cannot resume interrupted thread ~A", 1, thread);
    }
  }

  if (!ReleaseMutex(mkcl_interrupt_thread_lock))
    mkcl_FEwin32_error(env, "interrupt thread failed to release interrupt lock.", 0);

  return 0;
}

#endif /* linux */

static void setup_thread_bindings(MKCL, mkcl_object initial_bindings)
{
  mkcl_object l = initial_bindings;
  mkcl_object table = mk_cl__make_hash_table(env, @'eq',
					     MKCL_MAKE_FIXNUM(1024),
					     MKCL_MAKE_FIXNUM(1024),
					     MKCL_MAKE_FIXNUM(1));

  mkcl_loop_for_in(env, l)
    {
      mkcl_object cell = MKCL_CONS_CAR(l);

      mkcl_object sym = MKCL_CONS_CAR(cell);
      mkcl_object value = MKCL_CONS_CDR(cell);

      mkcl_check_symbol(env, sym);
      if (NULL == mkcl_search_hash(env, sym, table))
	{
	  mkcl_sethash(env, sym, table, value);
	  mkcl_bds_bind(env, sym, value);
	}
    }
  mkcl_end_loop_for_in;
}


static mkcl_object
_mkcl_specials_snapshot(MKCL)
{
  struct mkcl_bds_bd * bds_top = env->bds_top;
  struct mkcl_bds_bd * bds = env->bds_org;
  mkcl_object init_bind = mk_cl_Cnil;

  for (; bds <= bds_top; bds++)
    init_bind = mk_cl_acons(env, bds->symbol, MKCL_SYM_VAL(env, bds->symbol), init_bind);
  return(init_bind);
}

static mkcl_object
alloc_thread(MKCL, mkcl_object name, mkcl_object initial_bindings)
{
  const mkcl_env new_env = _mkcl_alloc_env(env);
  mkcl_object thread = mkcl_alloc_object(env, mkcl_t_thread);
  thread->thread.active = 0;
  thread->thread.name = name;
  thread->thread.function = mk_cl_Cnil;
  thread->thread.args = mk_cl_Cnil;
  thread->thread.result_value = MKCL_OBJNULL;
  thread->thread.thread = 0;
  thread->thread.base_thread = 0;
  thread->thread.tid = 0;
  thread->thread.interrupt = mk_cl_Cnil;
  thread->thread.sleeping = false;
  thread->thread.env = new_env;

#ifdef MKCL_WINDOWS
  thread->thread.saved_sigmask = 0;
#else
  sigemptyset(&thread->thread.saved_sigmask);
#endif
  thread->thread.resume_handler_ran = FALSE;
  thread->thread.sigmask_frs_marker = NULL;
  thread->thread.interrupt_count = 0;
  {
    int i;

    for (i = 0; i < MKCL_MAX_INTERRUPTS; i++)
      {
	thread->thread.interrupted_threads[i].tid = 0;
	thread->thread.interrupted_threads[i].cs_org = NULL;
	thread->thread.interrupted_threads[i].cs_limit = NULL;
	thread->thread.interrupted_threads[i].cs_barrier = NULL;
	thread->thread.interrupted_threads[i].cs_size = 0;
	thread->thread.interrupted_threads[i].disable_interrupts = FALSE;
	thread->thread.interrupted_threads[i].interrupt_disabler_file = NULL;
	thread->thread.interrupted_threads[i].interrupt_disabler_lineno = 0;
      }   
  }

  thread->thread.initial_bindings = NULL;

  if (initial_bindings == mk_cl_Ct)
    { /* New thread inherits its initial bindings from its creator thread. */
      thread->thread.initial_bindings = _mkcl_specials_snapshot(env);
    }
  else
    thread->thread.initial_bindings = mk_cl_copy_alist(env, initial_bindings);

  new_env->own_thread = thread;
  mkcl_init_env(env, new_env);

  return thread;
}

static pthread_mutex_t mkcl_imported_thread_pool_lock;
static sem_t mkcl_imported_thread_pool_empty;
static sem_t mkcl_imported_thread_pool_full;


static void fill_imported_thread_pool(MKCL)
{
  volatile bool locked = false;
  mkcl_object head;
  int i;

  MKCL_UNWIND_PROTECT_BEGIN(env) {
    mkcl_interrupt_status old_intr;

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    if (pthread_mutex_lock(&mkcl_imported_thread_pool_lock))
      mkcl_lose("fill_imported_thread_pool failed on pthread_mutex_lock.");
    locked = true;
    mkcl_set_interrupt_status(env, old_intr);

    head = mkcl_core.imported_thread_pool;
    for (i = 0; i < 5; i++)
      head = MKCL_CONS(env, alloc_thread(env, mk_cl_Cnil, mk_cl_Cnil), head);
    mkcl_core.imported_thread_pool = head;

  } MKCL_UNWIND_PROTECT_EXIT {
    if (locked)
      if (pthread_mutex_unlock(&mkcl_imported_thread_pool_lock))
	mkcl_lose("fill_imported_thread_pool failed on pthread_mutex_unlock.");  
  } MKCL_UNWIND_PROTECT_END;
}

static thread_value_t CALL_CONV imported_thread_pool_filler(void * arg)
{
  volatile bool locked = false;
  mkcl_object thread = (mkcl_object) arg;
  const mkcl_env env = thread->thread.env;
  thread_value_t status = (thread_value_t) MKCL_THREAD_NORMAL_EXIT;

  mkcl_set_thread_env(env);
  thread->thread.tid = mkcl_gettid();

  mkcl_init_c_stack_control(env, &env);

#ifndef MKCL_WINDOWS
  pthread_cleanup_push(thread_cleanup, (void *)thread);
#endif

  MKCL_UNWIND_PROTECT_BEGIN(env) {
    mkcl_interrupt_status old_intr;

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    MKCL_THREAD_LIST_LOCK(); locked = true;
    mkcl_set_interrupt_status(env, old_intr);

    mkcl_core.threads = MKCL_CONS(env, thread, mkcl_core.threads);
  } MKCL_UNWIND_PROTECT_EXIT {
    if (locked) MKCL_THREAD_LIST_UNLOCK();
  } MKCL_UNWIND_PROTECT_END;

  MKCL_CATCH_ALL_BEGIN(env) {
    mkcl_bds_bind(env, @'mt::*thread*', thread);
    mkcl_bds_bind(env, @'si::*dynamic-cons-stack*', mk_cl_Cnil);
    mk_si_trim_dynamic_cons_stack(env);
    thread->thread.active = 1;
    mk_mt_thread_detach(env, thread);
    mkcl_enable_interrupts(env);

    for (;;) /* forever until we're killed. */
      {
#ifdef MKCL_WINDOWS
	switch (WaitForSingleObject(mkcl_imported_thread_pool_empty, INFINITE))
	  {
	  case WAIT_OBJECT_0: break;
	  case WAIT_TIMEOUT:
	  case WAIT_ABANDONED:
	  case WAIT_FAILED:
	  default:
	    mkcl_FEwin32_error(env, ("imported_thread_pool_filler failed "
				     "on wait for empty semaphore."), 0);
	  }
#else
	int rc;

	do {
	  rc = sem_wait(&mkcl_imported_thread_pool_empty);
	} while ( rc && errno == EINTR );

	if (rc)
	  mkcl_C_lose("imported_thread_pool_filler failed on sem_wait.");
#endif

	fill_imported_thread_pool(env);

#ifdef MKCL_WINDOWS
	if (!ReleaseSemaphore(mkcl_imported_thread_pool_full, 1, NULL))
	  mkcl_FEwin32_error(env, "imported_thread_pool_filler failed on ReleaseSemaphore.", 0);
#else
	if (sem_post(&mkcl_imported_thread_pool_full))
	  mkcl_C_lose("imported_thread_pool_filler failed on sem_post.");
#endif
      }
    /* mkcl_bds_unwind1(env); */
    /* mkcl_bds_unwind1(env); */
  } MKCL_CATCH_ALL_IF_CAUGHT {
    status = (thread_value_t) MKCL_THREAD_KILLED;
  } MKCL_CATCH_ALL_END;
  thread->thread.active = 0;

#ifdef MKCL_WINDOWS
  thread_cleanup(thread);
#else
  pthread_cleanup_pop(1);
#endif
  return status;
}

static void mkcl_create_imported_thread_pool_filler_thread(MKCL)
{
  char * name = "Imported thread pool filler daemon";
  mkcl_object thread_name = mkcl_make_base_string_copy(env, name);
  mkcl_object thread = alloc_thread(env, thread_name, mk_cl_Cnil);

  thread->thread.function = mk_cl_Cnil;
  thread->thread.args = mk_cl_Cnil;

#ifdef MKCL_WINDOWS
  {
    unsigned threadId;
    HANDLE code = (HANDLE) _beginthreadex(NULL, 0,
					  imported_thread_pool_filler,
					  thread, 0, &threadId);

    if (code == 0)
      mkcl_FEwin32_error(env,
		    "mt::thread-enable failed on "
		    "_beginthreadex for thread ~S",
		    1, thread);
    else
      thread->thread.base_thread = thread->thread.thread = code;
  }

  /* Trigger the initial fill-up. */
  if (!ReleaseSemaphore(mkcl_imported_thread_pool_empty, 1, NULL))
    mkcl_FEwin32_error(env, "mkcl_create_imported_thread_pool_filler failed on ReleaseSemaphore.", 0);
#else
  /* There is a race condition on thread.thread here. JCB */
  int result = pthread_create(&thread->thread.thread, NULL,
			      imported_thread_pool_filler,
			      thread);
  thread->thread.base_thread = thread->thread.thread;
  if ( result != 0 )
    mkcl_lose("Cannot create signal servicing thread.");

  /* Trigger the initial fill-up. */
  if (sem_post(&mkcl_imported_thread_pool_empty))
    mkcl_C_lose("mkcl_create_imported_thread_pool_filler_thread failed on sem_post.");
#endif
}

static void push_in_imported_thread_pool(MKCL, mkcl_object thread)
{
  volatile bool locked = false;
  MKCL_UNWIND_PROTECT_BEGIN(env) {
    mkcl_interrupt_status old_intr;

    mkcl_get_interrupt_status(env, old_intr);
    mkcl_disable_interrupts(env);
    if (pthread_mutex_lock(&mkcl_imported_thread_pool_lock))
      mkcl_lose("push_in_imported_thread_pool failed on pthread_mutex_lock.");
    locked = true;
    mkcl_set_interrupt_status(env, old_intr);

    mkcl_core.imported_thread_pool = MKCL_CONS(env, thread, mkcl_core.imported_thread_pool);
  } MKCL_UNWIND_PROTECT_EXIT {
    if (locked)
      if (pthread_mutex_unlock(&mkcl_imported_thread_pool_lock))
	mkcl_lose("push_in_imported_thread_pool failed on pthread_mutex_unlock.");
  } MKCL_UNWIND_PROTECT_END;
}

static mkcl_object pop_from_imported_thread_pool(void)
{
  static unsigned int level = 0;
  mkcl_object thread = NULL;

  if (pthread_mutex_lock(&mkcl_imported_thread_pool_lock))
    mkcl_lose("pop_from_imported_thread_pool failed on pthread_mutex_lock.");

  thread = mkcl_core.imported_thread_pool;

  if (mkcl_Null(thread))
    { /* the pool is empty. */
      if (level > 0)
	mkcl_lose("Imported thread pool failure!");

      level++;

#ifdef MKCL_WINDOWS
      if (!ReleaseSemaphore(mkcl_imported_thread_pool_empty, 1, NULL))
	mkcl_lose("pop_from_imported_thread_pool failed on ReleaseSemaphore.");

      switch (WaitForSingleObject(mkcl_imported_thread_pool_full, INFINITE))
	{
	case WAIT_OBJECT_0: break;
	case WAIT_TIMEOUT:
	case WAIT_ABANDONED:
	case WAIT_FAILED:
	default:
	  mkcl_lose("pop_from_imported_thread_pool failed"
		    " on wait for full semaphore.");
	}
#else
      if (sem_post(&mkcl_imported_thread_pool_empty))
	mkcl_C_lose("create_interrupt_thread failed on sem_post.");

      {
	int rc;

	do
	  rc = sem_wait(&mkcl_imported_thread_pool_full);
	while (rc && errno == EINTR);
      }
#endif

      thread = pop_from_imported_thread_pool();
      level--;
    }
  else
    {
      mkcl_core.imported_thread_pool = MKCL_CONS_CDR(thread);
      thread = MKCL_CONS_CAR(thread);
    }

  /* This function is called before we have a proper thread environment set up
     so we cannot wrap the lock/unlock in an unwind-protect.
     Hopefully the body of this function cannot raise any exception. JCB
   */
  if (pthread_mutex_unlock(&mkcl_imported_thread_pool_lock))
    mkcl_lose("pop_from_imported_thread_pool failed on pthread_mutex_unlock.");
  
  return(thread);
}

static mkcl_object assign_imported_thread(mkcl_object name, mkcl_object bindings)
{
  mkcl_object thread = pop_from_imported_thread_pool();

  thread->thread.active = false;
  thread->thread.name = name;
  thread->thread.function = mk_cl_Cnil;
  thread->thread.args = mk_cl_Cnil;
  thread->thread.result_value = MKCL_OBJNULL;
  thread->thread.initial_bindings = (bindings == mk_cl_Ct ? mk_cl_Cnil : bindings);
  thread->thread.interrupt_count = 0;
  thread->thread.sleeping = false;

  return thread;
}

const mkcl_env
mkcl_import_current_thread(mkcl_object name, mkcl_object bindings)
{
  mkcl_object l;
  pthread_t current;

#ifdef MKCL_WINDOWS
# define pthread_equal(a,b) ((a)==(b))
# define _close_handle(h) CloseHandle(h)
  DuplicateHandle(GetCurrentProcess(),
 		  GetCurrentThread(),
 		  GetCurrentProcess(),
 		  &current,
 		  0,
 		  FALSE,
 		  DUPLICATE_SAME_ACCESS);
#else
# define _close_handle(h)
  current = pthread_self();
#endif

  for (l = mkcl_core.threads; l != mk_cl_Cnil; l = MKCL_CONS_CDR(l)) {
    mkcl_object p = MKCL_CONS_CAR(l);
    int i;
    if (pthread_equal(p->thread.thread, current)) {
      { _close_handle(current); return(NULL); }
    }
    /* could be an interrupted thread. JCB */
    for (i = 0; i < p->thread.interrupt_count; i++)
      if (pthread_equal(current, p->thread.interrupted_threads[i].tid))
	{ _close_handle(current); return(NULL); }
  }
#ifdef MKCL_WINDOWS
  _close_handle(current);
# undef pthread_equal
# undef _close_handle
#endif

  GC_register_my_thread((void *) &name);

  {
    mkcl_object thread = assign_imported_thread(name, bindings);
    const mkcl_env env = thread->thread.env;
    volatile bool locked = false;

    mkcl_set_thread_env(env);
    mkcl_init_c_stack_control(env, &env);

    thread->thread.thread = current;
    thread->thread.base_thread = current;
    thread->thread.tid = mkcl_gettid();

    /* imported threads are presumed not to be under lisp full (ultimate) control. */
    thread->thread.result_value = @':imported';

    MKCL_UNWIND_PROTECT_BEGIN(env) {
      mkcl_interrupt_status old_intr;

      mkcl_get_interrupt_status(env, old_intr);
      mkcl_disable_interrupts(env);
      MKCL_THREAD_LIST_LOCK(); locked = true;
      mkcl_set_interrupt_status(env, old_intr);

      mkcl_core.threads = MKCL_CONS(env, thread, mkcl_core.threads);
    } MKCL_UNWIND_PROTECT_EXIT {
      if (locked) MKCL_THREAD_LIST_UNLOCK();
    } MKCL_UNWIND_PROTECT_END;

    if (!mkcl_Null(thread->thread.initial_bindings))
      setup_thread_bindings(env, thread->thread.initial_bindings);

    mk_si_clear_all_fpe(env);
    mk_si_enable_fpe(env, @':default');

    return(env);
  }
}

void
mkcl_release_current_thread(MKCL)
{
  mkcl_object thread = env->own_thread;

  thread_cleanup(thread);
  push_in_imported_thread_pool(env, thread);
  GC_unregister_my_thread();
}

@(defun mt::make-thread (&key name ((:initial-bindings initial_bindings) mk_cl_Ct))
  mkcl_object thread;
@
  thread = alloc_thread(env, name, initial_bindings);
  @(return thread);
@)

#ifdef linux
static void print_sig_mask(sigset_t * set)
{
  int i;

  fprintf(stderr, "Blocked: ");
  for (i = 1; i < NSIG; i++) {
    if (sigismember(set, i)) { fprintf(stderr, "%d ", i); }
  }
  fprintf(stderr, "\n");
  fflush(stderr);
}

static void print_thread_sig_mask(void)
{
  sigset_t blocked;

  pthread_sigmask(SIG_BLOCK, NULL, &blocked);
  print_sig_mask(&blocked);
}
#endif /* linux */

mkcl_object
mk_mt_show_sigmask(MKCL)
{
#ifdef linux
  print_thread_sig_mask();
#endif
  @(return mk_cl_Cnil);
}

mkcl_object
mk_mt_reset_sigmask(MKCL)
{
#ifdef linux
  sigset_t all_signals;

  sigfillset(&all_signals);

  if (pthread_sigmask(SIG_UNBLOCK, &all_signals, NULL))
    mkcl_lose("mk_mt_reset_sigmask failed on pthread_sigmask.");
#endif
  @(return mk_cl_Cnil);
}


mkcl_object
mk_mt_block_signals(MKCL)
{
#ifdef linux
  mkcl_object this_thread = mk_mt_current_thread(env);

  if ( this_thread->thread.sigmask_frs_marker == NULL )
    {
      sigset_t all_signals;

      sigfillset(&all_signals); /* block everything! */

      if (pthread_sigmask(SIG_SETMASK,
			  &all_signals,
			  &(this_thread->thread.saved_sigmask)))
	mkcl_lose("mk_mt_block_signals failed on pthread_sigmask.");
      
      this_thread->thread.sigmask_frs_marker
	= this_thread->thread.env->frs_top;
    }
#endif
  @(return mk_cl_Cnil);
}

mkcl_object
mk_mt_unblock_signals(MKCL)
{
#ifdef linux
  mkcl_object this_thread = mk_mt_current_thread(env);

  if ( this_thread->thread.sigmask_frs_marker 
       >= this_thread->thread.env->frs_top )
    {
      if (pthread_sigmask(SIG_SETMASK, &(this_thread->thread.saved_sigmask), NULL))
	mkcl_lose("mk_mt_unblock_signals failed on pthread_sigmask.");

      this_thread->thread.sigmask_frs_marker = NULL;
    }
#endif
  @(return mk_cl_Cnil);
}

mkcl_object
mk_mt_thread_preset(MKCL, mkcl_narg narg, mkcl_object thread, mkcl_object function, ...)
{
  mkcl_va_list args;
  mkcl_va_start(env, args, function, narg, 2);
  if (narg < 2)
    mkcl_FEwrong_num_arguments(env, @'mt::thread-preset');
  mkcl_assert_type_thread(env, thread);
  thread->thread.function = function;
  thread->thread.args = mk_cl_grab_rest_args(env, args, FALSE);
  @(return thread);
}

#ifdef linux
void mkcl_create_signal_servicing_thread(MKCL,
					 char * thread_cname, 
					 int sig,
					 mkcl_object func_designator)
{
  mkcl_object thread_name = mkcl_make_base_string_copy(env, thread_cname);
  mkcl_object thread = alloc_thread(env, thread_name, mk_cl_Cnil);
  int interrupt_sig = mkcl_get_option(MKCL_OPT_THREAD_INTERRUPT_SIGNAL);
  sigset_t servicing_thread_sigmask, current_sigmask;
  int result;

  sigfillset(&servicing_thread_sigmask); /* Block everything. */
  if (sigdelset(&servicing_thread_sigmask, interrupt_sig)) /* almost everything */
    mkcl_lose("mkcl_create_signal_servicing_thread failed on sigdelset.");
  pthread_sigmask(SIG_SETMASK, &servicing_thread_sigmask, &current_sigmask);

  thread->thread.function = func_designator;
  thread->thread.args = mk_cl_Cnil;

  /* Is there a race condition on thread.thread here? JCB */
  result = pthread_create(&thread->thread.thread, NULL,
			  signal_servicing_thread_entry_point,
			  MKCL_CONS(env, thread, MKCL_MAKE_FIXNUM(sig)));
  thread->thread.base_thread = thread->thread.thread;
  pthread_sigmask(SIG_SETMASK, &current_sigmask, NULL);
  if ( result != 0 )
    mkcl_lose("Cannot create signal servicing thread.");

}
#endif /* linux */

#ifdef linux
mkcl_env_ptr mkcl_interrupted_thread_env = NULL;
bool mkcl_interrupt_refused;
bool mkcl_interrupt_forcefully;
sem_t mkcl_interrupted_thread_suspended;
sem_t mkcl_interrupted_thread_resumed;

static sem_t mkcl_run_interrupt_function;
static pthread_mutex_t mkcl_interrupt_thread_lock;
static sigset_t mkcl_standard_sigmask;
#endif /* linux */

#define MAX_INTERRUPT_RETRIES 50

#ifdef linux
static void create_interrupt_thread(mkcl_object thread)
{
  int result;

  result = pthread_create(&thread->thread.thread, NULL,
			  interrupt_thread_entry_point, thread);

  if ( result != 0 )
    mkcl_lose("create_interrupt_thread failed on pthread_create.");

  if (sem_post(&mkcl_run_interrupt_function))
    mkcl_C_lose("create_interrupt_thread failed on sem_post.");
}

#elif defined(MKCL_WINDOWS)

static void create_interrupt_thread(mkcl_object thread)
{
  HANDLE code;
  unsigned threadId;

  code = (HANDLE) _beginthreadex(NULL, 0, interrupt_thread_entry_point, thread, CREATE_SUSPENDED, &threadId);
  if (code == 0)
    mkcl_FEwin32_error(MKCL_ENV(), "create_interrupt_thread failed on _beginthreadex for thread ~A", 1, thread);
  else
    thread->thread.thread = code;

  
  switch (ResumeThread(thread->thread.thread))
    {
    case 0: /* This one should in fact be an error. It means it was not suspended! */
    case 1: /* This is the normal case. */
      break;
    default:
      mkcl_FEwin32_error(MKCL_ENV(), "create_interrupt_thread failed on ResumeThead for thread ~A", 1, thread);
      break;
    }

  SwitchToThread();
}
#endif

#ifdef MKCL_WINDOWS

static HANDLE mkcl_interrupt_thread_lock;

/* mk_mt_interrupt_thread_internal() is to be run with interrupts disabled. */
static mkcl_object
mk_mt_interrupt_thread_internal(MKCL, mkcl_object thread,
			      mkcl_object function, mkcl_object force)
{
  if (mk_mt_thread_active_p(env, thread) == mk_cl_Cnil)
    mkcl_FEerror(env, "Cannot interrupt an inactive thread ~A", 1, thread);

  if ( thread == mk_mt_current_thread(env) )
    {
      mk_cl_funcall(env, 1, function);
      return mk_cl_Ct;
    }

  {
    HANDLE os_thread = thread->thread.thread;
    bool success = FALSE, failure = FALSE;
    mkcl_object reason = mk_cl_Cnil;
    int retry_count = 0;

    do {
      if (mk_mt_thread_active_p(env, thread) == mk_cl_Cnil)
	mkcl_FEerror(env, "Cannot interrupt an inactive thread ~A", 1, thread);

      switch (WaitForSingleObject(mkcl_interrupt_thread_lock, 1000))
	{
	case WAIT_OBJECT_0: break;
	case WAIT_TIMEOUT:
	  mkcl_FEwin32_error(env, "interrupt-thread: interrupt lock timeout.", 0);
	case WAIT_ABANDONED:
	case WAIT_FAILED:
	default:
	  mkcl_FEwin32_error(env, "interrupt-thread failed to acquire lock.", 0);
	}

      if ( thread->thread.sleeping )
	{
	  reason = @'mt::thread-sleeping';
	  failure = TRUE;
	}
      else if ( thread->thread.interrupt_count < MKCL_MAX_INTERRUPTS )
	{
	  DWORD suspend_count;

	  if ((suspend_count = SuspendThread(os_thread)) == (DWORD)-1)
	    mkcl_FEwin32_error(env, "Cannot suspend thread ~A", 1, thread);

	  mkcl_env_ptr target_env = thread->thread.env;

	  if ( suspend_count != 0 )
	    {
	      /* The thread is already suspended!
		 What should we do? This should never happen!!
	       */
	      DWORD count;
	      failure = TRUE;
	      reason = @':suspended';
	      /* "resume" to set suspend_count back to its previous value. */
	      if ((count = ResumeThread(os_thread)) == (DWORD)-1)
		mkcl_FEwin32_error(env, "Cannot resume thread ~A", 1, thread);
	    }
	  else if (target_env == NULL /* The thread died unexpectedly! */
		   || (target_env->own_thread->thread.active == 0) )
	    { /* The thread is dying on us! */
	      reason = @':dead';
	      failure = TRUE;
	    }
	  else if ((target_env->disable_interrupts > 1) /* being interrupted already! */
		   || (target_env->disable_interrupts && (force == mk_cl_Cnil)) )
	    {	/* The interruption is refused. */
	      DWORD previous_suspend_count;
	      retry_count++;
	      
	      if ((previous_suspend_count = ResumeThread(os_thread)) == (DWORD)-1)
		mkcl_FEwin32_error(env, "Cannot resume thread ~A", 1, thread);
	    }
	  else
	    {	/* The interruption is accepted. */
		/* Push interrupted thread on stack */
	      int i = thread->thread.interrupt_count;
	      struct interrupted_thread_ctrl * p 
		= &(thread->thread.interrupted_threads[i]);

	      p->tid = thread->thread.thread;
	      p->cs_org = target_env->cs_org;
	      p->cs_limit = target_env->cs_limit;
	      p->cs_barrier = target_env->cs_barrier;
	      p->cs_size = target_env->cs_size;
	      p->disable_interrupts = target_env->disable_interrupts;
#if MKCL_DEBUG_INTERRUPT_MASK
	      p->interrupt_disabler_file = target_env->interrupt_disabler_file;
	      p->interrupt_disabler_lineno = target_env->interrupt_disabler_lineno;
#endif

	      thread->thread.interrupt_count = i + 1; /* Pushed on stack.*/

	      thread->thread.interrupt = function;

	      target_env->disable_interrupts = 2; /* This flags an interrupt in progress. */
#if MKCL_DEBUG_INTERRUPT_MASK
	      target_env->interrupt_disabler_lineno = __LINE__;
	      target_env->interrupt_disabler_file = __FILE__;
#endif

	      /* Create interruption sub-thread here */
	      create_interrupt_thread(thread);

	      success = TRUE;
	    }
	}
      else
	{
	  retry_count++;
	}

      if (!ReleaseMutex(mkcl_interrupt_thread_lock))
	mkcl_FEwin32_error(env, "interrupt-thread failed to release lock.", 0);

      if ( !(success || failure) )
	SwitchToThread();
    } while ( !(retry_count >= MAX_INTERRUPT_RETRIES || success || failure) );

    if ( retry_count >= MAX_INTERRUPT_RETRIES)
      {
	mk_cl_error(env, 3, @'mt::interrupt-refused', @':thread', thread);
	return mk_cl_Cnil;
      }
    else if ( failure )
      {
	if ( reason == @'mt::thread-sleeping')
	  mk_cl_error(env, 3, @'mt::thread-sleeping', @':thread', thread);
	else
	  mk_cl_error(env, 5, @'mt::invalid-thread',
		   @':thread', thread, @':reason', reason);
	return mk_cl_Cnil;
      }
    else if ( success )
      { 
	return mk_cl_Ct;
      }
    else
      mkcl_lose("interrupt-thread is really confused.");
  }
  return mk_cl_Ct;
}

#elif defined(linux)

static mkcl_object
mk_mt_interrupt_thread_internal(MKCL, mkcl_object thread,
			      mkcl_object function, mkcl_object force)
{
  if (mk_mt_thread_active_p(env, thread) == mk_cl_Cnil)
    mkcl_FEerror(env, "Cannot interrupt an inactive thread ~A", 1, thread);

  if ( thread == mk_mt_current_thread(env) )
    {
      mk_cl_funcall(env, 1, function);
      return mk_cl_Ct;
    }

  {
    int sig = mkcl_get_option(MKCL_OPT_THREAD_INTERRUPT_SIGNAL);
    sigset_t no_interrupt_sigmask, current_sigmask;
    bool success = FALSE, failure = FALSE;
    int retry_count = 0;

    mkcl_interrupt_forcefully = ((force == mk_cl_Cnil) ? FALSE : TRUE);

    sigemptyset(&no_interrupt_sigmask);
    if (sigaddset(&no_interrupt_sigmask, sig))
      mkcl_lose("mk_mt_interrupt_thread failed on sigaddset.");

    do {
      if (mk_mt_thread_active_p(env, thread) == mk_cl_Cnil)
	mkcl_FEerror(env, "Cannot interrupt an inactive thread ~A", 1, thread);

      /* adjust sigmask to prevent interrupts during the critical section */
      pthread_sigmask(SIG_BLOCK, &no_interrupt_sigmask, &current_sigmask);

      if (pthread_mutex_lock(&mkcl_interrupt_thread_lock))
	mkcl_lose("mk_mt_interrupt_thread failed on pthread_mutex_lock.");

      if ( thread->thread.interrupt_count < MKCL_MAX_INTERRUPTS )
	{
	  int rc = 0;
	  struct mkcl_env_struct *env = thread->thread.env;
	  mkcl_interrupted_thread_env = env;
	  mkcl_interrupt_refused = FALSE;

	  /* send signal to interrupt (suspend) thread. */
	  if ((rc = pthread_kill(thread->thread.thread, sig)))
	    switch (rc /* errno */)
	      {
	      case ESRCH:
		/* The targetted thread does not exist. */
		failure = TRUE;
		mkcl_interrupted_thread_env = NULL; /* done with it. clear it. */
		break;
	      case EINVAL:
		mkcl_lose("mk_mt_interrupt_thread: invalid signal used with pthread_kill.");
		break;
	      default:
		mkcl_lose("mk_mt_interrupt_thread failed on pthread_kill.");
		break;
	      }
	  else
	    {
	      do { /* should this be timed? What if the interruptee dies on us? JCB */
		rc = sem_wait(&mkcl_interrupted_thread_suspended);
	      } while ( rc && errno == EINTR );
	      if ( rc ) mkcl_C_lose("mk_mt_interrupt_thread failed on sem_wait.");
	      mkcl_interrupted_thread_env = NULL; /* done with it. clear it. */
	  
	      if ( mkcl_interrupt_refused )
		retry_count++;
	      else
		{ /* Push interrupted thread on stack */
		  int i = thread->thread.interrupt_count;
		  struct interrupted_thread_ctrl * p 
		    = &(thread->thread.interrupted_threads[i]);

		  p->tid = thread->thread.thread;
		  p->cs_org = env->cs_org;
		  p->cs_limit = env->cs_limit;
		  p->cs_barrier = env->cs_barrier;
		  p->cs_size = env->cs_size;
		  p->disable_interrupts = env->disable_interrupts;
#if MKCL_DEBUG_INTERRUPT_MASK
		  p->interrupt_disabler_file = env->interrupt_disabler_file;
		  p->interrupt_disabler_lineno = env->interrupt_disabler_lineno;
#endif

		  thread->thread.interrupt_count = i + 1; /* Pushed on stack.*/

		  thread->thread.interrupt = function;

		  env->disable_interrupts = 2; /* This flags an interrupt in progress. */
#if MKCL_DEBUG_INTERRUPT_MASK
		  env->interrupt_disabler_lineno = __LINE__;
		  env->interrupt_disabler_file = __FILE__;
#endif

		  /* Create interruption sub-thread here */
		  create_interrupt_thread(thread);

		  success = TRUE;
		}
	    }
	}
      else
	{
	  retry_count++;
	}

      if (pthread_mutex_unlock(&mkcl_interrupt_thread_lock))
	mkcl_lose("mk_mt_interrupt_thread failed on pthread_mutex_unlock.");

      pthread_sigmask(SIG_SETMASK, &current_sigmask, NULL);

      if ( !(success || failure) )
	sched_yield();
    } while ( !(retry_count >= MAX_INTERRUPT_RETRIES || success || failure) );

    if ( retry_count >= MAX_INTERRUPT_RETRIES)
      {
#if MKCL_DEBUG_INTERRUPT_MASK
	mk_cl_error(env, 7, @'mt::interrupt-refused',
		    @':thread', thread,
		    @':file',
		    mkcl_make_simple_base_string
		    (env, (char *) thread->thread.env->interrupt_disabler_file),
		    @':lineno',
		    MKCL_MAKE_FIXNUM(thread->thread.env->interrupt_disabler_lineno));
#else
	mk_cl_error(env, 3, @'mt::interrupt-refused', @':thread', thread);
#endif
	return mk_cl_Cnil;
      }
    else if ( failure )
      {
	mk_cl_error(env, 5, @'mt::invalid-thread', @':thread', thread);
	return mk_cl_Cnil;
      }
  }
  return mk_cl_Ct;
}

#endif /* defined(linux) */

@(defun mt::interrupt-thread (thread function &key (force mk_cl_Cnil))
@
  mkcl_object val;

  MKCL_NO_INTR(env, val = mk_mt_interrupt_thread_internal(env, thread, function, force));
  @(return val);
@)

mkcl_object
mk_mt_thread_kill(MKCL, mkcl_object thread)
{
  mkcl_object val;

  MKCL_NO_INTR(env, val = mk_mt_interrupt_thread_internal(env, thread, @'mt::terminate-thread', mk_cl_Ct));
  @(return val);
}

mkcl_object mk_mt_thread_detach(MKCL, mkcl_object thread)
{
  if (mkcl_Null(thread))
    thread = mk_mt_current_thread(env);
  else if (mkcl_type_of(thread) != mkcl_t_thread)
    thread = mkcl_type_error(env, @'mt::thread-detach', "argument", thread, @'mt::thread');

  if ( thread->thread.result_value != MKCL_OBJNULL )
    mkcl_CEerror(env, mk_cl_Ct, "Cannot detach an already detached or terminated thread.", 0);
  else
    {
#ifdef linux
      volatile bool locked = false;
      pthread_t base_thread;
      int rc;

      MKCL_UNWIND_PROTECT_BEGIN(env) {
	mkcl_interrupt_status old_intr;

	mkcl_get_interrupt_status(env, old_intr);
	mkcl_disable_interrupts(env);
	if ((rc = pthread_mutex_lock(&mkcl_interrupt_thread_lock)))
	  mkcl_lose("mk_mt_thread_detach failed on pthread_mutex_lock.");
	else
	  locked = true;
	
	if ( thread->thread.interrupt_count )
	  base_thread = thread->thread.interrupted_threads[0].tid;
	else
	  base_thread = thread->thread.thread;
	
	rc = pthread_detach(base_thread);
	mkcl_set_interrupt_status(env, old_intr);
	if (rc)
	  mkcl_FElibc_error(env, "mk_mt_thread_detach failed on pthread_detach().", 0);

      } MKCL_UNWIND_PROTECT_EXIT {
	if (locked)
	  if ((rc = pthread_mutex_unlock(&mkcl_interrupt_thread_lock)))
	    mkcl_lose("mk_mt_thread_detach failed on pthread_mutex_unlock.");
      } MKCL_UNWIND_PROTECT_END;
#endif
      thread->thread.result_value = @':detached';
    }
    
  @(return thread->thread.result_value);
}

#ifdef linux

mkcl_object
mk_mt_thread_join(MKCL, mkcl_object thread)
{
  pthread_t base_thread;
  mkcl_object result_value = mk_cl_Cnil;
  void * status;
  int rc = 0;

  if (mk_mt_thread_active_p(env, thread) != mk_cl_Cnil )
    {
      mkcl_interrupt_status old_intr;

      mkcl_get_interrupt_status(env, old_intr);
      mkcl_disable_interrupts(env);
      if ((rc = pthread_mutex_lock(&mkcl_interrupt_thread_lock)))
	mkcl_lose("mk_mt_thread_join failed on pthread_mutex_lock.");

      if ( thread->thread.interrupt_count )
	{
	  base_thread = thread->thread.interrupted_threads[0].tid;
	}
      else
	{
	  base_thread = thread->thread.thread;
	}
      
      if ((rc = pthread_mutex_unlock(&mkcl_interrupt_thread_lock)))
	mkcl_lose("mk_mt_thread_join failed on pthread_mutex_unlock.");
      mkcl_set_interrupt_status(env, old_intr);
    }
  else
    {
      base_thread = thread->thread.thread;
    }

  if ( thread->thread.result_value == @':detached' )
    { @(return thread->thread.result_value); }
  else if (thread == mkcl_core.initial_thread)
    {
      /* Boehm's GC dumps core if you try to "join" the initial thread.
	 So we try to bypass it. */
      if ((rc = _true_pthread_join(base_thread, &status)))
	{
	  mkcl_lose("mk_mt_thread_join failed on _true_pthread_join.");
	}
      /* This is too harsh! JCB */
    }
  else
    {
      if ((rc = pthread_join(base_thread, &status)))
	{
	  mkcl_lose("mk_mt_thread_join failed on pthread_join."); /* This is too harsh! JCB */
	}
    }

  result_value = thread->thread.result_value;
  if (result_value == MKCL_OBJNULL)
    result_value = @':invalid-value';

  {
    volatile bool locked = false;

    MKCL_UNWIND_PROTECT_BEGIN(env) {
      mkcl_interrupt_status old_intr;

      mkcl_get_interrupt_status(env, old_intr);
      mkcl_disable_interrupts(env);
      MKCL_THREAD_LIST_LOCK(); locked = true;
      mkcl_set_interrupt_status(env, old_intr);

      mkcl_core.threads = mkcl_remove_eq(env, thread, mkcl_core.threads);
    } MKCL_UNWIND_PROTECT_EXIT {
      if (locked) MKCL_THREAD_LIST_UNLOCK();
    } MKCL_UNWIND_PROTECT_END;
  }

  @(return result_value);
}

#elif defined(MKCL_WINDOWS)

mkcl_object
mk_mt_thread_join(MKCL, mkcl_object thread)
{
  pthread_t handle;
  mkcl_object result_value;
  DWORD exitCode;

  if (mk_mt_thread_active_p(env, thread) != mk_cl_Cnil )
    {
      mkcl_interrupt_status old_intr;
      
      mkcl_get_interrupt_status(env, old_intr);
      mkcl_disable_interrupts(env);
      switch (WaitForSingleObject(mkcl_interrupt_thread_lock, INFINITE))
	{
	case WAIT_OBJECT_0: break;
	case WAIT_TIMEOUT:
	case WAIT_ABANDONED:
	case WAIT_FAILED:
	default:
	  mkcl_FEwin32_error(env, "thread-join failed to acquire lock.", 0);
	}

      if ( thread->thread.interrupt_count )
	{
	  handle = thread->thread.interrupted_threads[0].tid;
	}
      else
	handle = thread->thread.thread;

      if (!ReleaseMutex(mkcl_interrupt_thread_lock))
	mkcl_FEwin32_error(env, "thread-join failed to release interrupt lock.", 0);
      mkcl_set_interrupt_status(env, old_intr);
    }
  else
    handle = thread->thread.thread;

  if ( thread->thread.result_value == @':detached' )
    { @(return thread->thread.result_value); }
  else
    {
      switch (WaitForSingleObject(handle, INFINITE))
	{
	case WAIT_OBJECT_0: break; /* The normal case. */
	case WAIT_TIMEOUT:
	case WAIT_FAILED:
	default:
	  mkcl_FEwin32_error(env, "mk_mt_thread_join failed on WaitForSingleObject", 0);
	  break;
	}

      if (!GetExitCodeThread(handle, &exitCode))
	mkcl_FEwin32_error(env, "mk_mt_thread_join failed on GetExitCodeThread", 0);

      if (exitCode == MKCL_THREAD_NORMAL_EXIT)
	result_value = thread->thread.result_value;
      else
	result_value = MKCL_MAKE_FIXNUM(-1);
    }

  {
    volatile bool locked = false;
    
    MKCL_UNWIND_PROTECT_BEGIN(env) {
      mkcl_interrupt_status old_intr;
    
      mkcl_get_interrupt_status(env, old_intr);
      mkcl_disable_interrupts(env);
      MKCL_THREAD_LIST_LOCK(); locked = true;
      mkcl_set_interrupt_status(env, old_intr);
      mkcl_core.threads = mkcl_remove_eq(env, thread, mkcl_core.threads);
    } MKCL_UNWIND_PROTECT_EXIT {
      if (locked) MKCL_THREAD_LIST_UNLOCK();
    } MKCL_UNWIND_PROTECT_END;
  }

  CloseHandle(handle);

  @(return result_value);
}

#endif /* defined(MKCL_WINDOWS) */

mkcl_object
mk_mt_thread_yield(MKCL)
{
#ifdef HAVE_SCHED_YIELD
  sched_yield();
#elif defined(MKCL_WINDOWS)
  SwitchToThread();
#else
  sleep(0); /* Use sleep(0) to yield to a >= priority thread */
#endif
  @(return);
}

#ifdef MKCL_WINDOWS
mkcl_object
mk_mt_thread_enable(MKCL, mkcl_object thread)
{
  mkcl_object output;
  HANDLE code;

  if (mk_mt_thread_active_p(env, thread) != mk_cl_Cnil)
    mkcl_FEerror(env, "Cannot enable an already running thread ~A.", 1, thread);

  {
    unsigned threadId;
    code = (HANDLE) _beginthreadex(NULL, 0, thread_entry_point, thread, 0, &threadId);

    if (code == 0)
      mkcl_FEwin32_error(env, "mt::thread-enable failed on _beginthreadex for thread ~S", 1, thread);
  }
  output = (thread->thread.thread = code) ? thread : mk_cl_Cnil;

  @(return output);
}

#elif defined(linux) /* MKCL_WINDOWS */

mkcl_object
mk_mt_thread_enable(MKCL, mkcl_object thread)
{
  mkcl_object output;
  int code;

  if (mk_mt_thread_active_p(env, thread) != mk_cl_Cnil)
    mkcl_FEerror(env, "Cannot enable an already running thread ~A.", 1, thread);

  /* Is there a race condition on thread.thread here? JCB */
  code = pthread_create(&thread->thread.thread, NULL, thread_entry_point, thread);
  thread->thread.base_thread = thread->thread.thread;

  switch (code)
    {
    case 0: output = thread; break;
    case EAGAIN:
    case EINVAL:
    case EPERM: /* These should raise a condition instead, with more info. JCB */
      mkcl_FEerror(env, "Thread creation failed for thread ~S", 1, thread);
      output = mk_cl_Cnil; break;
    default: /* maybe we should lose on this one. (its undocumented!) JCB */
      mkcl_FEerror(env, "Thread creation failed for thread ~S", 1, thread);
      output = mk_cl_Cnil; break;
    }
  @(return output);
}
#endif /* linux */


mkcl_object
mk_mt_exit_thread(MKCL, mkcl_object result_value)
{
  /* We simply undo the whole of the frame stack. This brings up
     back to the thread entry point, going through all possible
     UNWIND-PROTECT.
  */

  mkcl_disable_interrupts(env);
  if (env->own_thread->thread.result_value == MKCL_OBJNULL)
    env->own_thread->thread.result_value = result_value;

  mkcl_unwind(env, env->frs_org);

  /* If we get here it means that the thread failed to do a global unwind
     and is most probably damaged pretty much beyond repair!
     So, all we can really do at this point is to use
     very drastic means to bring this to an end. JCB
   */

  int i = env->own_thread->thread.interrupt_count;

  for (; i; --i)
    {
#ifdef linux
      if (pthread_cancel(env->own_thread->thread.interrupted_threads[i-1].tid))
	mkcl_lose("mk_mt_exit_thread failed on pthread_cancel.");
#elif defined(MKCL_WINDOWS)
      if (!TerminateThread(env->own_thread->thread.interrupted_threads[i-1].tid, 0))
	mkcl_lose("mk_mt_exit_thread failed on TerminateThread.");
      CloseHandle(env->own_thread->thread.interrupted_threads[i-1].tid);
#endif
    }

  /* If we get here it means that the FRS stack was empty. */
  mkcl_thread_exit(MKCL_THREAD_KILLED);
}

mkcl_object mk_mt_terminate_thread(MKCL)
{
  mk_mt_exit_thread(env, @':killed');
}

mkcl_object
mk_mt_all_threads(MKCL)
{
  /* No race condition here because this list is never destructively
   * modified. When we add or remove threads, we create new lists. */
  @(return mk_cl_copy_list(env, mkcl_core.threads));
}

mkcl_object
mk_mt_thread_name(MKCL, mkcl_object thread)
{
  mkcl_assert_type_thread(env, thread);
  @(return thread->thread.name);
}

mkcl_object
mk_mt_thread_active_p(MKCL, mkcl_object thread)
{
  mkcl_assert_type_thread(env, thread);
  @(return (thread->thread.active ? mk_cl_Ct : mk_cl_Cnil));
}

mkcl_object
mk_mt_thread_whostate(MKCL, mkcl_object thread)
{
  mkcl_assert_type_thread(env, thread);
  @(return (mkcl_core.null_string));
}

mkcl_object
mk_mt_thread_run_function(MKCL, mkcl_narg narg, mkcl_object name, mkcl_object function, ...)
{
  mkcl_object thread;
  mkcl_va_list args;
  mkcl_va_start(env, args, function, narg, 2);
  if (narg < 2)
    mkcl_FEwrong_num_arguments(env, @'mt::thread-run-function');
  if (MKCL_CONSP(name)) {
    /* This is useful to pass :initial-bindings to make-thread */
    thread = mk_cl_apply(env, 2, @'mt::make-thread', name);
  } else {
    thread = mk_mt_make_thread(env, 2, @':name', name);
  }
  mk_cl_apply(env, 4, @'mt::thread-preset', thread, function, mk_cl_grab_rest_args(env, args, FALSE));
  return mk_mt_thread_enable(env, thread);
}

/*----------------------------------------------------------------------
 * LOCKS or MUTEX
 */

@(defun mt::make-lock (&key name ((:recursive recursive) mk_cl_Ct))
  mkcl_object output;
@
  output = mkcl_alloc_object(env, mkcl_t_lock);
#ifdef MKCL_WINDOWS
  output->lock.name = name;
  output->lock.mutex = CreateMutex(NULL, FALSE, NULL);
  output->lock.holder = mk_cl_Cnil;
  output->lock.counter = 0;
  output->lock.recursive = (recursive != mk_cl_Cnil);
#else
  {
    pthread_mutexattr_t attr;

    pthread_mutexattr_init(&attr);
    output->lock.name = name;
    output->lock.holder = mk_cl_Cnil;
    output->lock.counter = 0;
    if (recursive == mk_cl_Cnil) {
      if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK))
	mkcl_lose("mk_mt_make_lock failed on pthread_mutexattr_settype.");
      output->lock.recursive = 0;
    } else {
      if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE))
	mkcl_lose("mk_mt_make_lock failed on pthread_mutexattr_settype.");
      output->lock.recursive = 1;
    }
    if (pthread_mutex_init(&output->lock.mutex, &attr))
      mkcl_lose("mk_mt_make_lock failed on pthread_mutex_init.");

    pthread_mutexattr_destroy(&attr);
  }
#endif
  mk_si_set_finalizer(env, output, mk_cl_Ct);
  @(return output);
@)

mkcl_object
mk_mt_recursive_lock_p(MKCL, mkcl_object lock)
{
  if (mkcl_type_of(lock) != mkcl_t_lock)
    mkcl_FEwrong_type_argument(env, @'mt::lock', lock);
  @(return (lock->lock.recursive ? mk_cl_Ct : mk_cl_Cnil));
}

mkcl_object
mk_mt_lock_name(MKCL, mkcl_object lock)
{
  if (mkcl_type_of(lock) != mkcl_t_lock)
    mkcl_FEwrong_type_argument(env, @'mt::lock', lock);
  @(return lock->lock.name);
}

mkcl_object
mk_mt_lock_holder(MKCL, mkcl_object lock)
{
  if (mkcl_type_of(lock) != mkcl_t_lock)
    mkcl_FEwrong_type_argument(env, @'mt::lock', lock);
  @(return lock->lock.holder);
}

mkcl_object
mk_mt_giveup_lock(MKCL, mkcl_object lock)
{
  mkcl_object own_thread = mk_mt_current_thread(env);
  if (mkcl_type_of(lock) != mkcl_t_lock)
    mkcl_FEwrong_type_argument(env, @'mt::lock', lock);
  if (lock->lock.holder != own_thread) {
    mkcl_FEerror(env, "Attempt to give up a lock ~S that is not owned by ~S.", 2,
	    lock, own_thread);
  }
  if (--lock->lock.counter == 0) {
    lock->lock.holder = mk_cl_Cnil;
  }
#ifdef MKCL_WINDOWS
  if (ReleaseMutex(lock->lock.mutex) == 0)
    mkcl_FEwin32_error(env, "Unable to release Win32 Mutex", 0);
#else
  int rc = pthread_mutex_unlock(&lock->lock.mutex);

  switch (rc)
    {
    case EPERM: mkcl_FEerror(env, "Not lock owner", 0); break;
    case EINVAL: mkcl_FEerror(env, "Invalid lock", 0); break;
    default: mkcl_lose("mk_mt_giveup_lock failed on pthread_mutex_unlock"); break;
    case 0: break;
    }
#endif
  @(return mk_cl_Ct);
}

@(defun mt::get-lock (lock &optional (timeout mk_cl_Ct))
@
  if (mkcl_type_of(lock) != mkcl_t_lock)
    mkcl_FEwrong_type_argument(env, @'mt::lock', lock);

#ifdef MKCL_WINDOWS
  {
    mkcl_object output = mk_cl_Cnil;
    DWORD delai = INFINITE;
    DWORD val;

    /* In Windows, all locks are recursive. We simulate the other case. */
    /* We complain if recursive=0 and this is an attempt to lock recursively. */
    if (!lock->lock.recursive && (lock->lock.holder == env->own_thread)) {
      mkcl_FEerror(env, "A recursive attempt was made to hold lock ~S", 1, lock);
    }

    if (mkcl_Null(timeout) || (MKCL_MAKE_FIXNUM(0) == timeout))
      delai = 0;
    else if (timeout == mk_cl_Ct)
      delai = INFINITE;
    else
      {
	/* INV: mkcl_minusp() makes sure `timeout' is real */
	if (mkcl_minusp(env, timeout))
	  mk_cl_error(env, 9, @'simple-type-error', @':format-control',
		   mkcl_make_simple_base_string(env, "Not a positive real ~S"),
		   @':format-arguments', mk_cl_list(env, 1, timeout),
		   @':expected-type', @'real', @':datum', timeout);

	{ 
	  double r = mkcl_to_double(env, timeout);

	  delai = floor(r * 1000.0); /* in milliseconds. */
	}
      }

    MKCL_NO_INTR(env, val = WaitForSingleObject(lock->lock.mutex, delai));
    switch (val)
      {
      case WAIT_OBJECT_0:
	lock->lock.holder = env->own_thread;
	lock->lock.counter++;
	output = mk_cl_Ct;
	break;
      case WAIT_TIMEOUT:
	output = mk_cl_Cnil;
	break;
      case WAIT_ABANDONED:
	mkcl_FEwin32_error(env, "Mutex was abandoned in a locked state", 0);
	break;
      case WAIT_FAILED:
      default:
	mkcl_FEwin32_error(env, "Unable to lock Win32 Mutex", 0);
	break;
      }
    @(return output);
  }
#else /* MKCL_WINDOWS */
  int rc;

  if (mkcl_Null(timeout) || (MKCL_MAKE_FIXNUM(0) == timeout))
    {
      MKCL_NO_INTR(env, rc = pthread_mutex_trylock(&lock->lock.mutex));

      if (rc == EBUSY)
	{ @(return mk_cl_Cnil); }
    }
  else if (timeout == mk_cl_Ct)
    {
      MKCL_NO_INTR(env, rc = pthread_mutex_lock(&lock->lock.mutex));
    }
  else
    {
      struct timespec ts;
      int error;

      /* INV: mkcl_minusp() makes sure `timeout' is real */
      if (mkcl_minusp(env, timeout))
	mk_cl_error(env, 9, @'simple-type-error', @':format-control',
		 mkcl_make_simple_base_string(env, "Not a positive real ~S"),
		 @':format-arguments', mk_cl_list(env, 1, timeout),
		 @':expected-type', @'real', @':datum', timeout);

      MKCL_NO_INTR(env, error = clock_gettime(CLOCK_REALTIME, &ts));
      if (error)
	mkcl_lose("mk_mt_get_lock failed on clock_gettime");

      /* Add `timeout' delta */
      { 
	double r = mkcl_to_double(env, timeout);
	ts.tv_sec += floor(r);
	ts.tv_nsec += ((r - floor(r)) * 1e9);
      }
      if (ts.tv_nsec >= 1000000000) {
	ts.tv_nsec -= 1000000000;
	ts.tv_sec++;
      }

      MKCL_NO_INTR(env, rc = pthread_mutex_timedlock(&lock->lock.mutex, &ts));

      if (rc == ETIMEDOUT)
	{ @(return mk_cl_Cnil); }
    }

  switch (rc)
    {
    case EINVAL: mkcl_FEerror(env, "Invalid lock", 0); break;
    case EAGAIN: mkcl_FEerror(env, "Out of available locks", 0); break;
    case EDEADLK: mkcl_FEerror(env, "Would deadlock", 0); break;
    default: mkcl_lose("mk_mt_get_lock failed on pthread_mutex_lock."); break;
    case 0:
      lock->lock.holder = env->own_thread;
      lock->lock.counter++;
      break;
    }
  @(return mk_cl_Ct);
#endif /* MKCL_WINDOWS */
@)


/*----------------------------------------------------------------------
 * RWLOCKS
 */

mkcl_object mk_mt_make_rwlock(MKCL)
{
  mkcl_object output = mkcl_alloc_object(env, mkcl_t_rwlock);
  
#ifdef MKCL_WINDOWS
# if defined(_MSC_VER) && (defined(WinVista) || defined(Win7))
  InitializeSRWLock(&output->rwlock.rwlock);
# else
  mkcl_FEerror(env, "RWLocks are supported under Windows Vista and successors.", 0);
# endif
#else
  int rc = pthread_rwlock_init(&output->rwlock.rwlock, NULL);

  switch (rc)
    {
    case EAGAIN:
    case ENOMEM:
    case EPERM:
    case EBUSY:
    case EINVAL:
      mkcl_lose("mk_mt_make_rwlock failed on pthread_rwlock_init");
      break;
    case 0: break;
    }
  mk_si_set_finalizer(env, output, mk_cl_Ct);
#endif
  @(return output);
}

@(defun mt::giveup-rwlock (rwlock &optional (read_or_write @':read'))
@
  {
    if (mkcl_type_of(rwlock) != mkcl_t_rwlock)
      mkcl_FEwrong_type_argument(env, @'mt::rwlock', rwlock);

#ifdef MKCL_WINDOWS
# if defined(_MSC_VER) && (defined(WinVista) || defined(Win7))
    if (MKCL_EQ(read_or_write, @':read'))
      ReleaseSRWLockShared(&rwlock->rwlock.rwlock);
    else if (MKCL_EQ(read_or_write, @':write'))
      ReleaseSRWLockExclusive(&rwlock->rwlock.rwlock);
    else
      mkcl_FEerror(env, "Invalid second argument to mt::giveup-rwlock.", 0);
# else
    mkcl_FEerror(env, "RWLocks are supported under Windows Vista and successors.", 0);
# endif
#else
    int rc = pthread_rwlock_unlock(&rwlock->rwlock.rwlock);

    switch (rc)
      {
      case EINVAL: mkcl_FEerror(env, "Invalid lock", 0); break;
      case EPERM: mkcl_FEerror(env, "Not lock owner", 0); break;
      default: mkcl_lose("mk_mt_giveup_rwlock failed on pthread_rwlock_unlock."); break;
      case 0: break;
      }
#endif
    @(return mk_cl_Ct);
  }
@)


@(defun mt::get-read-rwlock (rwlock &optional (timeout mk_cl_Ct))
@
  if (mkcl_type_of(rwlock) != mkcl_t_rwlock)
    mkcl_FEwrong_type_argument(env, @'mt::rwlock', rwlock);

#ifdef MKCL_WINDOWS
# if defined(_MSC_VER) && (defined(WinVista) || defined(Win7))
  AcquireSRWLockShared(&rwlock->rwlock.rwlock);
# else
  mkcl_FEerror(env, "RWLocks are supported under Windows Vista and successors.", 0);
# endif
#else
  int rc;

  if (mkcl_Null(timeout) || (MKCL_MAKE_FIXNUM(0) == timeout))
    {
      rc = pthread_rwlock_tryrdlock(&rwlock->rwlock.rwlock);

      if (rc == EBUSY)
	{ @(return mk_cl_Cnil); }
    }
  else if (timeout == mk_cl_Ct)
    {
      rc = pthread_rwlock_rdlock(&rwlock->rwlock.rwlock);
    }
  else
    {
      struct timespec   ts;

      /* INV: mkcl_minusp() makes sure `timeout' is real */
      if (mkcl_minusp(env, timeout))
	mk_cl_error(env, 9, @'simple-type-error', @':format-control',
		 mkcl_make_simple_base_string(env, "Not a positive real ~S"),
		 @':format-arguments', mk_cl_list(env, 1, timeout),
		 @':expected-type', @'real', @':datum', timeout);

      if (clock_gettime(CLOCK_REALTIME, &ts))
	mkcl_lose("mk_mt_get_read_rwlock failed on clock_gettime");

      /* Add `timeout' delta */
      { 
	double r = mkcl_to_double(env, timeout);
	ts.tv_sec += floor(r);
	ts.tv_nsec += ((r - floor(r)) * 1e9);
      }
      if (ts.tv_nsec >= 1000000000) {
	ts.tv_nsec -= 1000000000;
	ts.tv_sec++;
      }
      rc = pthread_rwlock_timedrdlock(&rwlock->rwlock.rwlock, &ts);

      if (rc == ETIMEDOUT)
	{ @(return mk_cl_Cnil); }
    }

  switch (rc)
    {
    case EINVAL: mkcl_FEerror(env, "Invalid lock", 0); break;
    case EAGAIN: mkcl_FEerror(env, "Out of available locks", 0); break;
    case EDEADLK: mkcl_FEerror(env, "Would deadlock", 0); break;
    default: mkcl_lose("mk_mt_get_read_rwlock failed on pthread_rwlock_rdlock."); break;
    case 0: break;
    }
#endif
  @(return mk_cl_Ct);
@)

@(defun mt::get-write-rwlock (rwlock &optional (timeout mk_cl_Ct))
@
  if (mkcl_type_of(rwlock) != mkcl_t_rwlock)
    mkcl_FEwrong_type_argument(env, @'mt::rwlock', rwlock);

#ifdef MKCL_WINDOWS
# if defined(_MSC_VER) && (defined(WinVista) || defined(Win7))
  AcquireSRWLockExclusive(&rwlock->rwlock.rwlock);
# else
  mkcl_FEerror(env, "RWLocks are supported under Windows Vista and successors.", 0);
# endif
#else
  int rc;

  if (mkcl_Null(timeout) || (MKCL_MAKE_FIXNUM(0) == timeout))
    {
      rc = pthread_rwlock_trywrlock(&rwlock->rwlock.rwlock);

      if (rc == EBUSY)
	{ @(return mk_cl_Cnil); }
    }
  else if (timeout == mk_cl_Ct)
    {
      rc = pthread_rwlock_wrlock(&rwlock->rwlock.rwlock);
    }
  else
    {
      struct timespec ts;

      /* INV: mkcl_minusp() makes sure `timeout' is real */
      if (mkcl_minusp(env, timeout))
	mk_cl_error(env, 9, @'simple-type-error', @':format-control',
		 mkcl_make_simple_base_string(env, "Not a positive real ~S"),
		 @':format-arguments', mk_cl_list(env, 1, timeout),
		 @':expected-type', @'real', @':datum', timeout);

      if (clock_gettime(CLOCK_REALTIME, &ts))
	mkcl_lose("mk_mt_get_write_rwlock failed on clock_gettime");

      /* Add `timeout' delta */
      { 
	double r = mkcl_to_double(env, timeout);
	ts.tv_sec += floor(r);
	ts.tv_nsec += ((r - floor(r)) * 1e9);
      }
      if (ts.tv_nsec >= 1000000000) {
	ts.tv_nsec -= 1000000000;
	ts.tv_sec++;
      }

      rc = pthread_rwlock_timedwrlock(&rwlock->rwlock.rwlock, &ts);

      if (rc == ETIMEDOUT)
	{ @(return mk_cl_Cnil); }
    }

  switch (rc)
    {
    case EINVAL: mkcl_FEerror(env, "Invalid lock", 0); break;
    case EAGAIN: mkcl_FEerror(env, "Out of available locks", 0); break;
    case EDEADLK: mkcl_FEerror(env, "Would deadlock", 0); break;
    default: mkcl_lose("mk_mt_get_write_rwlock failed on pthread_rwlock_wrlock."); break;
    case 0: break;
    }
#endif
  @(return mk_cl_Ct);
@)


/*----------------------------------------------------------------------
 * SEMAPHORES
 */

@(defun mt::make-semaphore (&optional count)
@
  mkcl_object output = mkcl_alloc_object(env, mkcl_t_semaphore);
  unsigned int value = 0;

  if (!mkcl_Null(count))
    {
      if (!MKCL_FIXNUMP(count))
	count = mkcl_type_error(env, @'mt::make-semaphore', "count",
			       count, @'integer');
      value = mkcl_fixnum_to_word(count);
    }

#ifdef MKCL_WINDOWS
  {
    HANDLE sem = CreateSemaphore(NULL, value, LONG_MAX, NULL);
    if (sem == NULL)
      mkcl_FEwin32_error(env,
			 "mt::make-semaphore failed on CreateSemaphore.", 0);

    output->semaphore.sem = sem;
    output->semaphore.name = mk_cl_Cnil;
    output->semaphore.count = value;
    output->semaphore.max_count = LONG_MAX;
  }
#else
  int rc = sem_init(&output->semaphore.sem, FALSE, value);
  
  switch (rc)
    {
    case EINVAL:
    default:
      mkcl_C_lose("mt::make-semaphore failed on sem_init.");
      break;
    case 0: break;
    }
  output->semaphore.name = mk_cl_Cnil;
  output->semaphore.count = value;
  output->semaphore.max_count = SEM_VALUE_MAX;
#endif

  mk_si_set_finalizer(env, output, mk_cl_Ct);

  @(return output);
@)

mkcl_object mk_mt_semaphore_count(MKCL, mkcl_object sem)
{
  if (mkcl_type_of(sem) != mkcl_t_semaphore)
    mkcl_FEwrong_type_argument(env, @'mt::semaphore', sem);
  
  {
#ifdef MKCL_WINDOWS
    long count;
  
    /*  wait(timeout 0) followed by release if it was signaled */
    switch (WaitForSingleObject(sem->semaphore.sem, 0))
      {
      case WAIT_OBJECT_0:
	if (!ReleaseSemaphore(sem->semaphore.sem, 1, &count))
	  mkcl_FEwin32_error(env, "mk_mt_semaphore_count failed on ReleaseSemaphore.", 0);
	count++;
	break;
      case WAIT_TIMEOUT: count = 0; break;
      case WAIT_FAILED:
      default:
	mkcl_FEwin32_error(env, "mk_mt_semaphore_count failed on WaitForSingleObject.", 0);
      }
#else
    int rc, count;

    rc = sem_getvalue(&sem->semaphore.sem, &count);
    
    switch (rc)
      {
      case EINVAL: mkcl_FEerror(env, "Invalid semaphore.", 0); break;
      default: mkcl_C_lose("mk_mt_semaphore_count failed on sem_getvalue."); break;
      case 0: break;
      }
#endif
    sem->semaphore.count = count;
    @(return mkcl_make_integer(env, count));
  }
}

@(defun mt::semaphore-signal (sem &optional count)
@
  int c_count = 0;

  if (mkcl_type_of(sem) != mkcl_t_semaphore)
    mkcl_FEwrong_type_argument(env, @'mt::semaphore', sem);
  
  if (mkcl_Null(count))
    c_count = 1;
  else if (!MKCL_FIXNUMP(count) || ((c_count = mkcl_fixnum_to_word(count)) < 0))
    mk_cl_error(env, 9, @'simple-type-error', @':format-control',
	     mkcl_make_simple_base_string(env, "Not a positive integer ~S"),
	     @':format-arguments', mk_cl_list(env, 1, count),
	     @':expected-type', @'integer', @':datum', count);
  
#ifdef MKCL_WINDOWS
  {
    long prev_count;
    if (!ReleaseSemaphore(sem->semaphore.sem, c_count, &prev_count))
      mkcl_FEwin32_error(env, "mk_mt_semaphore_count failed on ReleaseSemaphore.", 0);
    sem->semaphore.count = prev_count;
  }
#else
  int i;

  for (i = 0; i < c_count; i++)
    {
      int rc = sem_post(&sem->semaphore.sem);

      switch (rc)
	{
	case EINVAL: mkcl_FEerror(env, "Invalid semaphore.", 0); break;
	default: mkcl_C_lose("mk_mt_semaphore_signal failed on sem_post."); break;
	case 0: break;
	}
    }
#endif
  sem->semaphore.count += c_count;
  @(return mk_cl_Ct);
@)

@(defun mt::semaphore-wait (sem &optional (timeout mk_cl_Ct))
@
  if (mkcl_type_of(sem) != mkcl_t_semaphore)
    mkcl_FEwrong_type_argument(env, @'mt::semaphore', sem);
  
#ifdef MKCL_WINDOWS
  {
    DWORD milliSeconds;

    if (mkcl_Null(timeout))
      milliSeconds = 0;
    else if (timeout == mk_cl_Ct)
      milliSeconds = INFINITE;
    else
      {
	/* INV: mkcl_minusp() makes sure `timeout' is real */
	if (mkcl_minusp(env, timeout))
	  mk_cl_error(env, 9, @'simple-type-error', @':format-control',
		      mkcl_make_simple_base_string(env, "Not a positive real ~S"),
		      @':format-arguments', mk_cl_list(env, 1, timeout),
		      @':expected-type', @'real', @':datum', timeout);
	
	double r = mkcl_to_double(env, timeout);
	milliSeconds = r * 1000;
      }

    switch (WaitForSingleObject(sem->semaphore.sem, milliSeconds))
      {
      case WAIT_OBJECT_0:
	break;
      case WAIT_TIMEOUT: 
	@(return mk_cl_Cnil);
	break;
      case WAIT_FAILED:
      default:
	mkcl_FEwin32_error(env, "mk_mt_semaphore_wait failed on WaitForSingleObject.", 0);
      }
  }
#else  
  int rc;

  if (mkcl_Null(timeout) || (MKCL_MAKE_FIXNUM(0) == timeout))
    {
      do
	rc = sem_trywait(&sem->semaphore.sem);
      while (rc && errno == EINTR);

      if (rc && errno == EAGAIN)
	{ @(return mk_cl_Cnil); }
    }
  else if (timeout == mk_cl_Ct)
    {
      do
	rc = sem_wait(&sem->semaphore.sem);
      while (rc && errno == EINTR);
    }
  else
    {
      struct timespec   ts;

      /* INV: mkcl_minusp() makes sure `timeout' is real */
      if (mkcl_minusp(env, timeout))
	mk_cl_error(env, 9, @'simple-type-error', @':format-control',
		    mkcl_make_simple_base_string(env, "Not a positive real ~S"),
		    @':format-arguments', mk_cl_list(env, 1, timeout),
		    @':expected-type', @'real', @':datum', timeout);
      
      if (clock_gettime(CLOCK_REALTIME, &ts))
	mkcl_lose("mk_mt_semaphore_wait failed on clock_gettime");

      /* Add `timeout' delta */
      { 
	double r = mkcl_to_double(env, timeout);
	double r_sec = floor(r);

	ts.tv_sec += r_sec;
	ts.tv_nsec += ((r - r_sec) * 1e9);
      }
      if (ts.tv_nsec >= 1000000000) {
	ts.tv_nsec -= 1000000000;
	ts.tv_sec++;
      }
      do
	rc = sem_timedwait(&sem->semaphore.sem, &ts);
      while (rc && (errno == EINTR));

      if (rc && errno == ETIMEDOUT)
	{ @(return mk_cl_Cnil); }
    }

  if (rc)
    switch (errno)
      {
      case EINVAL: mkcl_FEerror(env, "Invalid semaphore.", 0); break;
      default: mkcl_C_lose("mk_mt_semaphore_wait failed on sem_wait."); break;
      }
#endif

  @(return mk_cl_Ct);
@)



/*----------------------------------------------------------------------
 * CONDITION VARIABLES
 */

mkcl_object
mk_mt_make_condition_variable(MKCL)
{
#ifdef MKCL_WINDOWS
  mkcl_FEerror(env, "Condition variables are not supported under Windows.", 0);
  @(return mk_cl_Cnil);
#else
  pthread_condattr_t attr;
  mkcl_object output;
  
  pthread_condattr_init(&attr);
  output = mkcl_alloc_object(env, mkcl_t_condition_variable);
  pthread_cond_init(&output->condition_variable.cv, &attr);
  pthread_condattr_destroy(&attr);
  mk_si_set_finalizer(env, output, mk_cl_Ct);
  @(return output);
#endif
}

@(defun mt::condition-wait (cv lock &optional (timeout mk_cl_Ct))
@
{
#ifdef MKCL_WINDOWS
  mkcl_FEerror(env, "Condition variables are not supported under Windows.", 0);
#else
  int rc;
  struct timespec   ts;

  if (mkcl_type_of(cv) != mkcl_t_condition_variable)
    mkcl_FEwrong_type_argument(env, @'mt::condition-variable', cv);
  if (mkcl_type_of(lock) != mkcl_t_lock)
    mkcl_FEwrong_type_argument(env, @'mt::lock', lock);

  if (mkcl_Null(timeout))
    timeout = MKCL_MAKE_FIXNUM(0);

  if (timeout == mk_cl_Ct)
    {
      rc = pthread_cond_wait(&cv->condition_variable.cv, &lock->lock.mutex);
      if (rc == 0)
	lock->lock.holder = mk_mt_current_thread(env);
    }
  else
    {
      /* INV: mkcl_minusp() makes sure `timeout' is real */
      if (mkcl_minusp(env, timeout))
	mk_cl_error(env, 9, @'simple-type-error', @':format-control',
		 mkcl_make_simple_base_string(env, "Not a non-negative number ~S"),
		 @':format-arguments', mk_cl_list(env, 1, timeout),
		 @':expected-type', @'real', @':datum', timeout);

      if (clock_gettime(CLOCK_REALTIME, &ts))
	mkcl_lose("mk_mt_condition_wait failed on clock_gettime");

      /* Add `timeout' delta */
      {
	double r = mkcl_to_double(env, timeout);

	ts.tv_sec += floor(r);
	ts.tv_nsec += ((r - floor(r)) * 1e9);
      }
      if (ts.tv_nsec >= 1000000000) {
	ts.tv_nsec -= 1000000000;
	ts.tv_sec++;
      }
      
      rc = pthread_cond_timedwait(&cv->condition_variable.cv, &lock->lock.mutex, &ts);

      if (rc == 0)
	lock->lock.holder = mk_mt_current_thread(env);
      else if (rc == ETIMEDOUT)
	{ @(return mk_cl_Cnil); }
    }
  
  switch (rc)
    {
    case EINVAL: mkcl_FEerror(env, "Invalid condition variable or lock", 0); break;
    case EPERM: mkcl_FEerror(env, "Not lock owner", 0); break;
    default: mkcl_lose("mk_mt_condition_wait failed on pthread_cond_wait"); break;
    case 0: break;
    }
#endif
  @(return mk_cl_Ct);
}
@)

mkcl_object
mk_mt_condition_signal(MKCL, mkcl_object cv)
{
#ifdef MKCL_WINDOWS
  mkcl_FEerror(env, "Condition variables are not supported under Windows.", 0);
#else
  int rc;

  if (mkcl_type_of(cv) != mkcl_t_condition_variable)
    mkcl_FEwrong_type_argument(env, @'mt::condition-variable', cv);
  rc = pthread_cond_signal(&cv->condition_variable.cv);

  switch (rc)
    {
    case EINVAL: mkcl_FEerror(env, "Invalid condition variable", 0); break;
    default: mkcl_lose("mk_mt_condition_signal failed on pthread_cond_signal"); break;
    case 0: break;
    }
#endif
  @(return mk_cl_Ct);
}

mkcl_object
mk_mt_condition_broadcast(MKCL, mkcl_object cv)
{
#ifdef MKCL_WINDOWS
  mkcl_FEerror(env, "Condition variables are not supported under Windows.", 0);
#else
  int rc;

  if (mkcl_type_of(cv) != mkcl_t_condition_variable)
    mkcl_FEwrong_type_argument(env, @'mt::condition-variable', cv);
  rc = pthread_cond_broadcast(&cv->condition_variable.cv);

  switch (rc)
    {
    case EINVAL: mkcl_FEerror(env, "Invalid condition variable", 0); break;
    default: mkcl_lose("mk_mt_condition_broadcast failed on pthread_cond_broadcast"); break;
    case 0: break;
    }
#endif
  @(return mk_cl_Ct);
}

/*----------------------------------------------------------------------
 * INITIALIZATION
 */

void
mkcl_init_early_threads(MKCL)
{
  mkcl_object thread;
  pthread_t main_thread;

#ifdef MKCL_WINDOWS
  mkcl_core.global_lock = CreateMutex(NULL, FALSE, NULL);
  if ( mkcl_core.global_lock == NULL )
    mkcl_FEwin32_error(env, "mkcl_init_thread failed to create global lock.", 0);

  mkcl_core.package_list_lock = CreateMutex(NULL, FALSE, NULL);
  if ( mkcl_core.package_list_lock == NULL )
    mkcl_FEwin32_error(env, "mkcl_init_thread failed to create package list lock.", 0);

  mkcl_core.thread_list_lock = CreateMutex(NULL, FALSE, NULL);
  if ( mkcl_core.thread_list_lock == NULL )
    mkcl_FEwin32_error(env, "mkcl_init_thread failed to create thread list lock.", 0);

  mkcl_core.special_index_lock = CreateMutex(NULL, FALSE, NULL);
  if ( mkcl_core.special_index_lock == NULL )
    mkcl_FEwin32_error(env, "mkcl_init_thread failed to create special's index lock.", 0);

  mkcl_imported_thread_pool_lock = CreateMutex(NULL, FALSE, NULL);
  if ( mkcl_imported_thread_pool_lock == NULL )
    mkcl_FEwin32_error(env, "mkcl_init_thread failed to create imported thread pool lock.", 0);
#else
  {
    pthread_mutexattr_t attr;

    if (pthread_mutexattr_init(&attr))
      mkcl_lose("mkcl_init_early_threads failed on pthread_mutexattr_init.");
    if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK))
      mkcl_lose("mkcl_init_early_threads failed on pthread_mutexattr_settype.");
      pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);

    if (pthread_mutex_init(&mkcl_core.global_lock, &attr))
      mkcl_lose("mkcl_init_early_threads failed on pthread_mutex_init.");
    if (pthread_mutex_init(&mkcl_core.package_list_lock, &attr))
      mkcl_lose("mkcl_init_early_threads failed on pthread_mutex_init.");
    if (pthread_mutex_init(&mkcl_core.thread_list_lock, &attr))
      mkcl_lose("mkcl_init_early_threads failed on pthread_mutex_init.");
    if (pthread_mutex_init(&mkcl_core.special_index_lock, &attr))
      mkcl_lose("mkcl_init_early_threads failed on pthread_mutex_init.");
    if (pthread_mutex_init(&mkcl_imported_thread_pool_lock, &attr))
      mkcl_lose("mkcl_init_early_threads failed on pthread_mutex_init.");

    pthread_mutexattr_destroy(&attr);
  }
#endif
  mkcl_core.threads = MKCL_OBJNULL;
  mkcl_core.top_special_index = 0;
  mkcl_core.imported_thread_pool = mk_cl_Cnil;

#ifdef linux
  pthread_sigmask(SIG_SETMASK, NULL, &mkcl_standard_sigmask);

  if (sem_init(&mkcl_interrupted_thread_suspended, 0, 0))
    mkcl_lose("mkcl_init_early_threads failed on sem_init.");
  if (sem_init(&mkcl_interrupted_thread_resumed, 0, 0))
    mkcl_lose("mkcl_init_early_threads failed on sem_init.");
  if (sem_init(&mkcl_run_interrupt_function, 0, 0))
    mkcl_lose("mkcl_init_early_threads failed on sem_init.");
  if (sem_init(&mkcl_imported_thread_pool_empty, 0, 0))
    mkcl_lose("mkcl_init_early_threads failed on sem_init.");
  if (sem_init(&mkcl_imported_thread_pool_full, 0, 0))
    mkcl_lose("mkcl_init_early_threads failed on sem_init.");

  {
    pthread_mutexattr_t attr;
    if (pthread_mutexattr_init(&attr))
      mkcl_lose("mkcl_init_early_threads failed on pthread_mutexattr_init.");
    if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK))
      mkcl_lose("mkcl_init_early_threads failed on pthread_mutexattr_settype.");
    if (pthread_mutex_init(&mkcl_interrupt_thread_lock, &attr))
      mkcl_lose("mkcl_init_early_threads failed on pthread_mutex_init.");
    pthread_mutexattr_destroy(&attr);
  }
#elif defined(MKCL_WINDOWS)
  mkcl_imported_thread_pool_empty = CreateSemaphore(NULL, 0, 1, NULL);
  if (mkcl_imported_thread_pool_empty == NULL)
    mkcl_FEwin32_error(env,
		       "mkcl_init_thread failed to create"
		       " imported thread pool empty semaphore.", 0);

  mkcl_imported_thread_pool_full = CreateSemaphore(NULL, 0, 1, NULL);
  if (mkcl_imported_thread_pool_empty == NULL)
    mkcl_FEwin32_error(env, 
		       "mkcl_init_thread failed to create"
		       " imported thread pool full semaphore.", 0);

  mkcl_interrupt_thread_lock = CreateMutex(NULL, FALSE, NULL);
  if ( mkcl_interrupt_thread_lock == NULL )
    mkcl_FEwin32_error(env,
		       "mkcl_init_thread failed to create"
		       " interrupt thread lock.", 0);
#endif /* linux */

  /* We have to set the environment before any allocation takes place,
   * so that the interrupt handling code works. */
#if !defined(WITH___THREAD)
# if defined(MKCL_WINDOWS)
  cl_env_key = TlsAlloc();
# else
  pthread_key_create(&cl_env_key, NULL);
# endif
#endif

  mkcl_set_thread_env(env);

#ifdef MKCL_WINDOWS
  DuplicateHandle(GetCurrentProcess(),
 		  GetCurrentThread(),
 		  GetCurrentProcess(),
 		  &main_thread,
 		  0,
 		  FALSE,
 		  DUPLICATE_SAME_ACCESS);
#else
  main_thread = pthread_self();
#endif
  thread = mkcl_alloc_object(env, mkcl_t_thread);
  /* Why is it that the initial (a.k.a. main) thread does not have
     the same bindings as any of the following threads? JCB */
  thread->thread.active = 1;
  thread->thread.name = mkcl_make_simple_base_string(env, "Initial");
  thread->thread.function = mk_cl_Cnil;
  thread->thread.args = mk_cl_Cnil;
  thread->thread.result_value = MKCL_OBJNULL;
  thread->thread.thread = main_thread;
  thread->thread.base_thread = main_thread;
  thread->thread.tid = mkcl_gettid();
  thread->thread.env = env;
  thread->thread.sleeping = false;
  thread->thread.interrupt = mk_cl_Cnil;
  thread->thread.initial_bindings = mk_cl_Cnil;
  thread->thread.sigmask_frs_marker = NULL;
#ifdef MKCL_WINDOWS
  thread->thread.saved_sigmask = 0;
#else
  sigemptyset(&thread->thread.saved_sigmask);
#endif
  thread->thread.resume_handler_ran = false;
  thread->thread.interrupt_count = 0;

  env->own_thread = thread;

  mkcl_core.initial_thread = thread;
  mkcl_core.threads = mkcl_list1(env, thread);
}

void
mkcl_init_late_threads(MKCL)
{
  mkcl_create_imported_thread_pool_filler_thread(env);
}


void mkcl_thread_exit(long status_code)
{
#if defined(linux)
  pthread_exit((void *) status_code);
#elif defined(MKCL_WINDOWS)
  _endthreadex(status_code);
#else
# error "Missing implementation for mkcl_thread_exit()."
#endif
  mkcl_lose("Thread termination failed.");
}

/* This here is to work around the wrapping Boehm's GC puts around pthread_join
   and thus get direct access to the real pthread version.
 */
#ifdef linux
#undef pthread_join
static int _true_pthread_join(pthread_t thread, void ** retval)
{
  return pthread_join(thread, retval);
}
#endif /* linux */

