/* -*- mode: c -*- */
/*
    time.c -- Time routines.
*/
/*
    Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
    Copyright (c) 1990, Giuseppe Attardi.
    Copyright (c) 2001, Juan Jose Garcia Ripoll.
    Copyright (c) 2010-2011, Jean-Claude Beaudoin.

    MKCL is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file '../../Copyright' for full details.
*/

#include <mkcl/mkcl.h>
#include <mkcl/mkcl-math.h>
#include <time.h>
#include <errno.h>
#ifdef HAVE_TIMES
# include <sys/times.h>
#endif
#ifdef HAVE_GETRUSAGE
# include <sys/time.h>
# include <sys/resource.h>
#endif
#ifdef HAVE_GETTIMEOFDAY
# include <sys/time.h>
#endif
#ifndef _MSC_VER
# include <unistd.h>
#endif
#include <mkcl/internal.h>

#include <stdlib.h>

#ifdef darwin
#undef HAVE_NANOSLEEP
#endif

#if defined(MKCL_WINDOWS) && !defined(__MINGW64__)
struct timespec {
  time_t tv_sec;
  long tv_nsec;
};
#endif

static struct timespec beginning;

static void
get_real_time(MKCL, struct timespec *ts)
{
#if linux
  if (clock_gettime(CLOCK_REALTIME, ts) == -1) {
    mkcl_FElibc_error(env, "get_real_time() failed on clock_gettime()", 0);
  }
#elif defined(MKCL_WINDOWS)
  /* This version will not wrap-around before year 30827 and has 100 nanoseconds resolution! */
  FILETIME system_time;
  ULARGE_INTEGER uli_system_time;

  GetSystemTimeAsFileTime(&system_time);
  uli_system_time.LowPart = system_time.dwLowDateTime;
  uli_system_time.HighPart = system_time.dwHighDateTime;
  {
    lldiv_t sec_and_100nsec = lldiv(uli_system_time.QuadPart, 1000 * 1000 * 10);

    ts->tv_sec = sec_and_100nsec.quot;
    ts->tv_nsec = sec_and_100nsec.rem * 100;
  }
#elif defined(HAVE_GETTIMEOFDAY)
  struct timezone tz;
  gettimeofday(ts, &tz);
  ts->tv_nsec = ts->tv_nsec * 1000;
#else
  time_t t = time(0);
  ts->tv_sec = t;
  ts->tv_usec = 0;
#endif
}

static void
get_run_time(MKCL, struct timespec *ts)
{
#if linux
  if (clock_gettime(CLOCK_THREAD_CPUTIME_ID, ts) == -1) {
    mkcl_FElibc_error(env, "get_run_time() failed on clock_gettime()", 0);
  }
#elif defined(MKCL_WINDOWS)
  FILETIME creation_time;
  FILETIME exit_time;
  FILETIME kernel_time;
  FILETIME user_time;
  ULARGE_INTEGER uli_kernel_time;
  ULARGE_INTEGER uli_user_time;
  if (!GetThreadTimes(GetCurrentThread(),
		      &creation_time,
		      &exit_time,
		      &kernel_time,
		      &user_time))
    mkcl_FEwin32_error(MKCL_ENV(), "GetThreadTimes() failed", 0);
  uli_kernel_time.HighPart = kernel_time.dwHighDateTime;
  uli_kernel_time.LowPart  = kernel_time.dwLowDateTime;
  uli_user_time.HighPart = user_time.dwHighDateTime;
  uli_user_time.LowPart  = user_time.dwLowDateTime;
  uli_kernel_time.QuadPart += uli_user_time.QuadPart;
  {
    lldiv_t sec_and_100nsec = lldiv(uli_kernel_time.QuadPart, 1000 * 1000 * 10);

    ts->tv_sec = sec_and_100nsec.quot;
    ts->tv_nsec = sec_and_100nsec.rem * 100;
  }
#elif defined(HAVE_GETRUSAGE)
  struct rusage r;
  getrusage(RUSAGE_SELF, &r);
  ts->tv_sec = r.ru_utime.tv_sec;
  ts->tv_nsec = r.ru_utime.tv_usec * 1000;
#elif defined(HAVE_TIMES)
  struct tms buf;
  times(&buf);
  ts->tv_sec = buf.tms_utime / CLK_TCK;
  ts->tv_nsec = (buf.tms_utime % CLK_TCK) * (1000 * 1000 * 1000);
#else
  get_real_time(ts);
#endif
}

mkcl_object
mk_cl_sleep(MKCL, mkcl_object z)
{
  bool fe_inexact_on = FE_INEXACT & fegetexcept();
#ifdef HAVE_NANOSLEEP
  struct timespec tm;
#endif
  /* INV: mkcl_minusp() makes sure `z' is real */
  if (mkcl_minusp(env, z))
    mk_cl_error(env, 9, @'simple-type-error', @':format-control',
	     mkcl_make_simple_base_string(env, "Not a non-negative number ~S"),
	     @':format-arguments', mk_cl_list(env, 1, z),
	     @':expected-type', @'real', @':datum', z);

  if (fe_inexact_on)
    fedisableexcept(FE_INEXACT);

#ifdef HAVE_NANOSLEEP
  {
    double r = mkcl_to_double(env, z);
    tm.tv_sec = (time_t)floor(r);
    tm.tv_nsec = (long)((r - floor(r)) * 1e9);

    if (fe_inexact_on)
      {
	feclearexcept(FE_INEXACT);
	feenableexcept(FE_INEXACT);
      }

    {
      int rc;
      struct timespec rem = tm;
      do
	{
	  struct timespec req = rem;
	  rc = nanosleep(&req, &rem);
	}
      while (rc && errno == EINTR);
      if (rc)
	mkcl_FElibc_error(env, "mk_cl_sleep() failed on nanosleep().", 0);
    }
  }
#elif defined(MKCL_WINDOWS)
  {
    double r = mkcl_to_double(env, z) * 1000;
    long duration = (long) r;

    if (fe_inexact_on)
      {
	feclearexcept(FE_INEXACT);
	feenableexcept(FE_INEXACT);
      }

    env->own_thread->thread.sleeping = true;
    Sleep(duration); /* This one sleeps hard and cannot be interrupted. JCB */
    env->own_thread->thread.sleeping = false;
  }
#else
  z = mkcl_round1(z);
  if (fe_inexact_on)
    {
      feclearexcept(FE_INEXACT);
      feenableexcept(FE_INEXACT);
    }
  if (MKCL_FIXNUMP(z))
    sleep(mkcl_fixnum_to_word(z)); /* FIXME! sleep() takes an "unsigned int", beware of truncation here. JCB */
#endif
  @(return mk_cl_Cnil);
}

static mkcl_object
timespec_to_time(MKCL, time_t sec, long nsec)
{
#if MKCL_WORD_BITS < 64
  if (sec < (MKCL_MOST_POSITIVE_FIXNUM / (1000 * 1000)))
    {
      mkcl_word x = (sec * (1000 * 1000)) + (nsec / 1000);
      return MKCL_MAKE_FIXNUM(x);
    }
  else
    {
      mkcl_object time = mkcl_make_integer(env, sec);

      time = mkcl_times(env, time , MKCL_MAKE_FIXNUM(1000 * 1000));
      time = mkcl_plus(env, time, MKCL_MAKE_FIXNUM(nsec / 1000));

      return time;
    }
#else
  if ( sec < (MKCL_MOST_POSITIVE_FIXNUM / (1000 * 1000 * 1000)))
    {
      mkcl_word x = (sec * (1000 * 1000 * 1000)) + nsec;
      return MKCL_MAKE_FIXNUM((mkcl_word)x);
    }
  else
    {
      mkcl_object time = mkcl_make_integer(env, sec);

      time = mkcl_times(env, time , MKCL_MAKE_FIXNUM(1000 * 1000 * 1000));
      time = mkcl_plus(env, time, MKCL_MAKE_FIXNUM(nsec));

      return time;
    }    
#endif
}

mkcl_object
mk_cl_get_internal_run_time(MKCL)
{
  struct timespec ts;
  get_run_time(env, &ts);
  @(return timespec_to_time(env, ts.tv_sec, ts.tv_nsec));
}

mkcl_object
mk_cl_get_internal_real_time(MKCL)
{
  struct timespec ts;
  get_real_time(env, &ts);
  @(return timespec_to_time(env, ts.tv_sec - beginning.tv_sec, ts.tv_nsec - beginning.tv_nsec));
}

mkcl_object
mk_cl_get_universal_time(MKCL)
{ /* On 32 bits POSIX machines this code will fail around January 18, 2038
     because of the value returned by time() will have wrapped-around. JCB */
  mkcl_object utc = mkcl_make_integer(env, time(0));
  @(return mkcl_plus(env, utc, mkcl_core.Jan1st1970UT));
}

void
mkcl_init_unixtime(MKCL)
{
  get_real_time(env, &beginning);
  
#if MKCL_WORD_BITS < 64
  MKCL_SET(@'internal-time-units-per-second', MKCL_MAKE_FIXNUM(1000 * 1000)); /* microseconds */
#else
  MKCL_SET(@'internal-time-units-per-second', MKCL_MAKE_FIXNUM(1000 * 1000 * 1000)); /* nanoseconds */
#endif
  
  /* This is the number of seconds between 00:00:00 January 1st, 1900 GMT and 00:00:00 January 1st, 1970 GMT. */
  mkcl_core.Jan1st1970UT = mkcl_times(env, MKCL_MAKE_FIXNUM(24 * 60 * 60), MKCL_MAKE_FIXNUM(17 + 365 * 70));
}
