/* -*- mode: c -*- */
/*
    unixfsys.c  -- Unix file system interface.
*/
/*
    Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
    Copyright (c) 1990, Giuseppe Attardi.
    Copyright (c) 2001, Juan Jose Garcia Ripoll.
    Copyright (c) 2010, Jean-Claude Beaudoin.

    MKCL is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file '../../Copyright' for full details.
*/

#include <mkcl/mkcl.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#ifdef unix
# include <unistd.h>
# include <pwd.h>
#elif defined(_MSC_VER)
# include <io.h>
# define F_OK 0
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <mkcl/mkcl-inl.h>
#include <mkcl/internal.h>
#include <dirent.h>

#ifdef MKCL_WINDOWS
# include <windows.h>
# include <direct.h>
# include <ctype.h>
#endif

#include <fcntl.h>
#include <errno.h>

static mkcl_object string_pushx_cstr(MKCL, mkcl_object s1, const char * s2);
static mkcl_object string_pushx_string(MKCL, mkcl_object s1, mkcl_object s2);
static int string_last(MKCL, mkcl_object str);
static mkcl_object mkcl_init_path(MKCL);
static mkcl_object reset_path(MKCL, mkcl_object path);
static mkcl_object copy_path(MKCL, mkcl_object old);
static mkcl_object back_path(MKCL, mkcl_object path);

static int is_slash(int c) { return MKCL_IS_DIR_SEPARATOR(c); }


static int
safe_chdir(MKCL, const char *path)
{
  int output;

  MKCL_NO_INTR(env, output = chdir(path));
  return output;
}

static int
safe_stat(MKCL, const char *path, struct stat *sb)
{
  int output;

  MKCL_NO_INTR(env, output = stat(path, sb));
  return output;
}

#ifdef HAVE_LSTAT
static int
safe_lstat(MKCL, const char *path, struct stat *sb)
{
  int output;

  MKCL_NO_INTR(env, output = lstat(path, sb));
  return output;
}
#endif

#ifdef unix
static char * safe_realpath(MKCL, mkcl_object orig_pathname, const char * path)
{
  char * output;

  MKCL_NO_INTR(env, output = realpath(path, NULL));
  if (output == NULL)
    {
      /* errno could be: EACCES, EINVAL, EIO, ELOOP, ENAMETOOLONG, ENOENT, ENOTDIR. */
      mk_cl_error(env, 3, @'file-error', @':pathname', orig_pathname);
      return NULL;
    }
  return output;    
}

#elif defined(MKCL_WINDOWS)

static char * safe_realpath(MKCL, mkcl_object orig_pathname, const char * path)
{
  DWORD size = MAX_PATH + 1;
  char * output;

  MKCL_NO_INTR(env, output = malloc(size * sizeof(char)));
  if (output == NULL)
    {
      mkcl_FEerror(env, "Out of memory", 0);
      return NULL;
    }

  MKCL_NO_INTR(env, size = GetFullPathName(path, size, output, NULL));
  if (size > MAX_PATH)
    { /* Not big enough */
      MKCL_NO_INTR(env, (output = realloc(output, size * sizeof(char))));
      if (output == NULL)
	{
	  mkcl_FEerror(env, "Out of memory", 0);
	  return NULL;
	}
      MKCL_NO_INTR(env, size = GetFullPathName(path, size, output, NULL));
    }
  if (size == 0)
    { /* Something went wrong. */
      mkcl_FEwin32_file_error(env, orig_pathname,
			      "GetFullPathname() failed in safe_realpath().", 0);
      return NULL;
    }

  return output;    
}
#endif


/*
 * string_to_pathanme, to be used when s is a real pathname
 */
mkcl_object
mkcl_cstring_to_pathname(MKCL, char *s)
{
  mkcl_object string = mkcl_make_simple_base_string(env, s);
  return mk_cl_parse_namestring(env, 1, string);
}

#ifdef MKCL_WINDOWS
/*
 * Finds current directory by using _getdcwd() with an adjustable
 * string which grows until it can host the whole path.
 */
static mkcl_object
current_dir_on_drive(MKCL, int drive)
{
  mkcl_object output;
  const char *ok;
#ifdef _MSC_VER
  unsigned char *c;
#endif
  mkcl_index size = 128;

  do {
    output = mkcl_alloc_adjustable_base_string(env, size);
    MKCL_NO_INTR(env, ok = _getdcwd(drive, (char*)output->base_string.self, size));
    size += 256;
  } while (ok == NULL && errno == ERANGE);
  if (ok == NULL)
    mkcl_FElibc_error(env, "current_dir_on_drive() failed on _getdcwd()", 0);
  size = strlen((char*)output->base_string.self);
  if ((size + 1 /* / */ + 1 /* 0 */) >= output->base_string.dim) {
    /* Too large to host the trailing '/' */
    mkcl_object other = mkcl_alloc_adjustable_base_string(env, size+2);
    strcpy((char*)other->base_string.self, (char*)output->base_string.self);
    output = other;
  }
#ifdef _MSC_VER
  for (c = output->base_string.self; *c; c++)
    if (*c == '\\')
      *c = '/';
#endif
  if (output->base_string.self[size-1] != '/') {
    output->base_string.self[size++] = '/';
    output->base_string.self[size] = 0;
  }
  output->base_string.fillp = size;
  return output;
}
#endif /* defined(MKCL_WINDOWS) */

/*
 * Finds current directory by using getcwd() with an adjustable
 * string which grows until it can host the whole path.
 */
static mkcl_object
current_dir(MKCL)
{
  mkcl_object output;
  const char *ok;
#ifdef _MSC_VER
  unsigned char *c;
#endif
  mkcl_index size = 128;

  do {
    output = mkcl_alloc_adjustable_base_string(env, size);
    MKCL_NO_INTR(env, ok = getcwd((char*)output->base_string.self, size));
    size += 256;
  } while (ok == NULL && errno == ERANGE);
  if (ok == NULL)
    mkcl_FElibc_error(env, "current_dir() failed on getcwd()", 0);
  size = strlen((char*)output->base_string.self);
  if ((size + 1 /* / */ + 1 /* 0 */) >= output->base_string.dim) {
    /* Too large to host the trailing '/' */
    mkcl_object other = mkcl_alloc_adjustable_base_string(env, size+2);
    strcpy((char*)other->base_string.self, (char*)output->base_string.self);
    output = other;
  }
#ifdef _MSC_VER
  for (c = output->base_string.self; *c; c++)
    if (*c == '\\')
      *c = '/';
#endif
  if (output->base_string.self[size-1] != '/') {
    output->base_string.self[size++] = '/';
    output->base_string.self[size] = 0;
  }
  output->base_string.fillp = size;
  return output;
}

/*
 * Using a certain path, guess the type of the object it points to.
 */

static mkcl_object
file_kind(MKCL, char *filename, bool follow_links)
{
  mkcl_object output;
#ifdef MKCL_WINDOWS
  DWORD dw;
  MKCL_NO_INTR(env, dw = GetFileAttributes( filename ));
  if (dw == INVALID_FILE_ATTRIBUTES)
    output = mk_cl_Cnil;
  else if ( dw & FILE_ATTRIBUTE_DIRECTORY )
    output = @':directory';
  else
    output = @':file';
#else /* !defined(MKCL_WINDOWS) */
  struct stat buf;
  if (
# ifdef HAVE_LSTAT
      (follow_links ? safe_stat : safe_lstat)(env, filename, &buf) < 0
# else
      safe_stat(filename, &buf) < 0
# endif
      )
    /* Here errno could be EACCES, EBADF, EFAULT, ELOOP,
       ENAMETOOLONG, ENOENT, EOVERFLOW. */
    /* ANSI requires us to raise a file-error on most of these. JCB */
    output = mk_cl_Cnil;
  else if (S_ISDIR(buf.st_mode))
    output = @':directory';
  else if (S_ISREG(buf.st_mode))
    output = @':file';
# ifdef HAVE_LSTAT
  else if (S_ISLNK(buf.st_mode))
    output = @':link';
# endif
  else
    output = @':special';
#endif /* !defined(MKCL_WINDOWS) */
  return output;
}

mkcl_object
mk_si_file_kind(MKCL, mkcl_object filename, mkcl_object follow_links) {
  filename = mk_si_coerce_to_filename(env, filename);
  @(return file_kind(env, (char*)filename->base_string.self,
		     !mkcl_Null(follow_links)));
}



/*
 * Search the actual name of the directory of a pathname,
 * going through links if they exist. Default is
 * current directory
 */
mkcl_object
mk_cl_truename(MKCL, mkcl_object orig_pathname)
{
  mkcl_object true_path = mkcl_init_path(env);
  mkcl_object pathname = orig_pathname;
  mkcl_object filename = mk_si_coerce_to_filename(env, pathname);
  char * real_path = safe_realpath(env,
				   orig_pathname,
				   (char *) filename->base_string.self);

  string_pushx_cstr(env, true_path, real_path);
  if (real_path)
    free(real_path); /* Yes this bypasses the GC since it was malloced by realpath(). */

  if (@':directory' == file_kind(env, (char*)true_path->base_string.self, FALSE))
    if (!MKCL_IS_DIR_SEPARATOR(string_last(env, true_path)))
      {
	char str[] = { MKCL_DIR_SEPARATOR, 0 };
	string_pushx_cstr(env, true_path, str);
      }
  return mk_cl_pathname(env, true_path);
}

int
mkcl_backup_open(MKCL, const char *filename, int option, int mode)
{
  int fd;
  int error;
  char *backupfilename = mkcl_alloc(env, strlen(filename) + 5);

  if (backupfilename == NULL) {
    mkcl_FEerror(env, "Cannot allocate memory for backup filename", 0);
  }

  strcat(strcpy(backupfilename, filename), ".BAK");
#ifdef MKCL_WINDOWS
  /* MS-Windows rename doesn't remove an existing file */
  MKCL_NO_INTR(env, error = (_access(backupfilename, F_OK) == 0
			     && _unlink(backupfilename)));
  if (error) {
    mkcl_object f = mkcl_make_base_string_copy(env, backupfilename);
    mkcl_FElibc_file_error(env, f, "Cannot remove the file ~S", 1, f);
  }
#endif
  MKCL_NO_INTR(env, error = rename(filename, backupfilename));
  if (error) {
    mkcl_object f = mkcl_make_base_string_copy(env, (char *) filename);
    mkcl_FElibc_file_error(env, f, "Cannot rename the file ~S to ~S.",
			   2, f, mkcl_make_base_string_copy(env, backupfilename));
  }
  MKCL_NO_INTR(env, fd = open(filename, option, mode));
  /* error value of fd must be handled by caller. */
  /* mkcl_dealloc(env, backupfilename); */
  return fd;
}

mkcl_object
mkcl_file_len(MKCL, int f)
{
  struct stat filestatus;
  int not_ok;

  MKCL_NO_INTR(env, not_ok = fstat(f, &filestatus));
  if (not_ok)
    mkcl_FElibc_error(env, "fstat() failed in mkcl_file_len().", 0);

  return mkcl_make_integer(env, filestatus.st_size);
}


mkcl_object mk_cl_rename_file(MKCL, mkcl_object old_filespec, mkcl_object new_name)
{
  /* 1) Get the old filename, and complain if it has wild components,
   *    or if it does not exist. Notice that the filename to be renamed
   *    is not the truename, because we might be renaming a symbolic link.
   */
  mkcl_object old_filename = mk_si_coerce_to_filename(env, old_filespec);
  mkcl_object old_truename = mk_cl_truename(env, old_filespec);
  mkcl_object new_filename;
  mkcl_interrupt_status old_intr;
  
  /* 2) Create the new file name. */
  new_name = mkcl_merge_pathnames(env, new_name, old_filespec, @':newest');
  new_filename = mk_si_coerce_to_filename(env, new_name);
  
  mkcl_get_interrupt_status(env, old_intr);
  mkcl_disable_interrupts(env);
  {
#ifdef MKCL_WINDOWS
    int error_mode = SetErrorMode(0);

    if (MoveFile((char*)old_filename->base_string.self,
		 (char*)new_filename->base_string.self)) {
      SetErrorMode(error_mode);
      goto SUCCESS;
    }
    switch (GetLastError())
      {
      case ERROR_ALREADY_EXISTS:
      case ERROR_FILE_EXISTS:
	break;
      default:
	goto FAILURE_CLOBBER;
      };
    if (MoveFileEx((char*)old_filename->base_string.self,
		   (char*)new_filename->base_string.self,
		   MOVEFILE_REPLACE_EXISTING)) {
      SetErrorMode(error_mode);
      goto SUCCESS;
    }
    /* hack for win95/novell */
    chmod((char*)old_filename->base_string.self, 0777);
    chmod((char*)new_filename->base_string.self, 0777);
    SetFileAttributesA((char*)new_filename->base_string.self,
		       FILE_ATTRIBUTE_NORMAL);
    SetFileAttributesA((char*)new_filename->base_string.self,
		       FILE_ATTRIBUTE_TEMPORARY);
    if (MoveFile((char*)old_filename->base_string.self,
		 (char*)new_filename->base_string.self)) {
      SetErrorMode(error_mode);
      goto SUCCESS;
    }
    /* fallback on old behavior */
    (void)DeleteFileA((char*)new_filename->base_string.self);
    if (MoveFile((char*)old_filename->base_string.self,
		 (char*)new_filename->base_string.self)) {
      SetErrorMode(error_mode);
      goto SUCCESS;
    }
    /* fall through */
#else /* def MKCL_WINDOWS */
    if (rename((char*)old_filename->base_string.self,
	       (char*)new_filename->base_string.self) == 0) {
      goto SUCCESS;
    }
    /* fall through */
#endif /* def MKCL_WINDOWS */
  }
 FAILURE_CLOBBER:
  mkcl_set_interrupt_status(env, old_intr);
#ifdef MKCL_WINDOWS
  mkcl_FEwin32_file_error(env, old_filespec, "Cannot rename the file ~S to ~S.", 2, old_filespec, new_name);
#else
  mkcl_FElibc_file_error(env, old_filespec, "Cannot rename the file ~S to ~S.", 2, old_filespec, new_name);
#endif
  
 SUCCESS:
  mkcl_set_interrupt_status(env, old_intr);
  @(return new_name old_truename mk_cl_truename(env, new_name));
}

mkcl_object
mk_cl_delete_file(MKCL, mkcl_object file)
{
  mkcl_object filename = mk_si_coerce_to_filename(env, file);
  int ok;

  MKCL_NO_INTR(env, ok = unlink((char*)filename->base_string.self));

  if (ok < 0)
    mkcl_FElibc_file_error(env, file, "Cannot delete file.", 0);
  @(return mk_cl_Ct);
}

mkcl_object
mk_cl_probe_file(MKCL, mkcl_object file)
{
  /* INV: Both SI:FILE-KIND and TRUENAME complain if "file" has wildcards */
  @(return (mk_si_file_kind(env, file, mk_cl_Ct) != mk_cl_Cnil 
	    ? mk_cl_truename(env, file) : mk_cl_Cnil))
}

mkcl_object
mk_cl_file_write_date(MKCL, mkcl_object file)
{
  mkcl_object time, filename = mk_si_coerce_to_filename(env, file);
  struct stat filestatus;
  if (safe_stat(env, (char*)filename->base_string.self, &filestatus) < 0)
    mk_cl_error(env, 3, @'file-error', @':pathname', file);
  else
    time = MKCL_UTC_time_to_universal_time(env, filestatus.st_mtime);
  @(return time);
}

mkcl_object
mk_cl_file_author(MKCL, mkcl_object file)
{
#ifdef unix
  mkcl_object output = mk_cl_Cnil;
  mkcl_object filename = mk_si_coerce_to_filename(env, file);
  struct stat filestatus;

  if (safe_stat(env, (char*)filename->base_string.self, &filestatus) < 0)
    mkcl_FElibc_file_error(env, file, "mk_cl_file_author: Cannot get file status.", 0);

  {
    struct passwd *pwent;
    MKCL_NO_INTR(env, (errno = 0, pwent = getpwuid(filestatus.st_uid)));
    if (pwent == NULL)
      mkcl_FElibc_file_error(env, file,
			     "mk_cl_file_author: Cannot get information for UID (~D).",
			     1, mkcl_make_unsigned_integer(env, filestatus.st_uid));
    else
      output = mkcl_make_base_string_copy(env, pwent->pw_name);
  }

  /* Beware that this is the file owner not its author. */
  @(return output);
#else /* def unix */
  /* If we want to duplicate the unix behavior on MS-Windows
     we could call GetFileSecurity() then do GetSecurityDescriptorOnwer()
     and finally LookupAccountSid to get the name of the file owner.
   */
  @(return mk_cl_Cnil);
#endif /* def unix */
}

mkcl_object
mkcl_homedir_pathname(MKCL, mkcl_object user)
{
  mkcl_index i;
  mkcl_object namestring;
  char *h;
#ifdef MKCL_WINDOWS
  char *d;
#endif

  if (!mkcl_Null(user)) {
    char *p;
    /* This ensures that our string has the right length
       and it is terminated with a '\0' */
    user = mk_si_copy_to_simple_base_string(env, user);
    p = (char*)user->base_string.self;
    i = user->base_string.fillp;
    if (i > 0 && *p == '~') {
      p++;
      i--;
    }
    if (i == 0)
      return mkcl_homedir_pathname(env, mk_cl_Cnil);
    else
      {
#ifdef unix
	struct passwd *pwent = NULL;
	pwent = getpwnam(p); /* getpwnam_r could do better here. JCB */
	if (pwent == NULL)
	  mkcl_FEerror(env, "Unknown user ~S.", 1, user);
	namestring = mkcl_make_base_string_copy(env, pwent->pw_dir);
#else /* def unix */
	namestring = mkcl_core.null_string;
#endif /* def unix */
      }
  } else if ((h = getenv("HOME"))) {
    namestring = mkcl_make_base_string_copy(env, h);
#ifdef MKCL_WINDOWS
  } else if ((h = getenv("HOMEPATH")) && (d = getenv("HOMEDRIVE"))) {
    namestring =
      mk_si_concatenate_base_strings(env, 2,
				     mkcl_make_simple_base_string(env, d),
				     mkcl_make_simple_base_string(env, h));
#endif /* def MKCL_WINDOWS */
  } else {
    namestring = mkcl_make_simple_base_string(env, "/");
  }
  if (namestring->base_string.self[0] == '~') {
    mkcl_FEerror(env, "Not a valid home pathname ~S", 1, namestring);
  }
  i = namestring->base_string.fillp;
  if (!MKCL_IS_DIR_SEPARATOR(namestring->base_string.self[i-1]))
    namestring = mk_si_concatenate_base_strings(env, 2, namestring, MKCL_CODE_CHAR(MKCL_DIR_SEPARATOR));
  return mk_cl_parse_namestring(env, 3, namestring, mk_cl_Cnil, mk_cl_Cnil);
}

@(defun user_homedir_pathname (&optional host)
@
  /* Ignore optional host argument. */
  @(return mkcl_homedir_pathname(env, mk_cl_Cnil));
@)

#ifdef unix
#define not_eq_ch(c1, c2) ((c1) != (c2))
#elif defined(MKCL_WINDOWS)
static inline bool not_eq_ch(char c1, char c2)
{
  return(toupper(c1) != toupper(c2));
}
#endif /* def MKCL_WINDOWS */
/*
 * Take two C strings and check if the first one matches
 * against the pattern given by the second one. The pattern
 * is that of a Unix shell except for brackets and curly
 * braces.
 */ /* Which is too generous for the rules of ANSI CL! JCB */
static bool
string_match(const char *s, const char *p)
{
  const char *next;
  while (*s) {
    switch (*p) {
    case '*':
      /* Match any group of characters */
      next = p+1;
      if (*next != '?') {
	if (*next == '\\')
	  next++;
	while (*s && not_eq_ch(*s,*next)) s++;
      }
      if (string_match(s,next))
	return TRUE;
      /* starts back from the '*' */
      if (!*s)
	return FALSE;
      s++;
      break;
    case '?':
      /* Match any character */
      s++, p++;
      break;
    case '\\':
      /* Interpret a pattern character literally.
         Trailing slash is interpreted as a slash. */
      if (p[1]) p++;
      if (not_eq_ch(*s,*p))
	return FALSE;
      s++, p++;
      break;
    default:
      if (not_eq_ch(*s,*p))
	return FALSE;
      s++, p++;
      break;
    }
  }
  while (*p == '*')
    p++;
  return (*p == 0);
}

/*
 * list_directory() lists the files and directories which are contained
 * in directory designated by 'path'. If ONLY_DIR is
 * true, the list is made of only the directories -- 
 * a propert which is checked by following the symlinks.
 */

#ifdef MKCL_WINDOWS

static mkcl_object
list_directory(MKCL, mkcl_object path, const char *mask, bool only_dir)
{
  mkcl_object out = mk_cl_Cnil;
  char *text;

  WIN32_FIND_DATA fd;
  HANDLE hFind = NULL;
  BOOL found = FALSE;
  mkcl_object full_name = mkcl_init_path(env);
  mkcl_object path_wild = mkcl_init_path(env);

  string_pushx_string(env, path_wild, path);
  string_pushx_cstr(env, path_wild, "*");

  MKCL_NO_INTR(env, hFind = FindFirstFile((char *) path_wild->base_string.self, &fd));
  if (hFind == INVALID_HANDLE_VALUE) {
    return mk_cl_Cnil;
  }

  do
    {
      text = fd.cFileName;
      reset_path(env, full_name);
      string_pushx_string(env, full_name, path);
      string_pushx_cstr(env, full_name, text);

      if (text[0] == '.' &&
	  (text[1] == '\0' ||
	   (text[1] == '.' && text[2] == '\0')))
	goto NEXT_ENTRY;
	/* continue; */
      if (mask && !string_match(text, mask))
	goto NEXT_ENTRY;
	/* continue; */
      if (only_dir 
	  && (file_kind(env, (char *) full_name->base_string.self, TRUE) 
	      != @':directory'))
	goto NEXT_ENTRY;
	/* continue; */
      out = mkcl_cons(env, copy_path(env, full_name), out);
    NEXT_ENTRY:
      MKCL_NO_INTR(env, (found = FindNextFile(hFind, &fd)));
    }
  while (found);

  if (GetLastError() != ERROR_NO_MORE_FILES)
    mkcl_FEwin32_file_error(env, path, "list_directory failed on FindNextFile", 0);

  if (!FindClose(hFind))
    mkcl_FEwin32_file_error(env, path, "list_directory failed on FindClose", 0);

  return mk_cl_nreverse(env, out);
}

#elif unix

static mkcl_object
list_directory(MKCL, mkcl_object path, const char *mask, bool only_dir)
{
  mkcl_object out = mk_cl_Cnil;
  char *text;
  DIR *dir;
  struct dirent *entry;
  mkcl_object full_name = mkcl_init_path(env);

  if (path->base_string.fillp == 0)
    /* 'path' is empty we assume we want the current directory. */
    { MKCL_NO_INTR(env, dir = opendir(".")); }
  else
    { MKCL_NO_INTR(env, dir = opendir((char *) path->base_string.self)); }
  if (dir == NULL) {
    /* opendir failed, errno should say why.
       One of EACCES, EBADF, EMFILE, ENFILE, ENOENT, ENOMEM, ENOTDIR.
     */
    return mk_cl_Cnil;
  }

  MKCL_NO_INTR(env, (errno = 0, entry = readdir(dir)));
  while (entry) {
    text = entry->d_name;
    reset_path(env, full_name);
    string_pushx_string(env, full_name, path);
    string_pushx_cstr(env, full_name, entry->d_name);

    if (text[0] == '.' &&
	(text[1] == '\0' ||
	 (text[1] == '.' && text[2] == '\0')))
      goto NEXT_ENTRY;
      /* continue; */
    if (mask && !string_match(text, mask))
      goto NEXT_ENTRY;
      /* continue; */
    if (only_dir 
	&& (file_kind(env, (char *) full_name->base_string.self, TRUE)
	    != @':directory'))
      goto NEXT_ENTRY;
      /* continue; */
    out = mkcl_cons(env, copy_path(env, full_name), out);
  NEXT_ENTRY:
    MKCL_NO_INTR(env, (errno = 0, entry = readdir(dir)));
  }
  if (errno != 0)
    {
      /* Something went wrong with readdir */
      mkcl_FElibc_file_error(env, path, "list_directory failed on readdir", 0);
    }
  if (closedir(dir))
    mkcl_FElibc_file_error(env, path, "list_directory failed on closedir", 0);

  return mk_cl_nreverse(env, out);
}

#endif


static bool is_dot_pathname_p(MKCL, mkcl_object pathname)
{
  mkcl_object name = pathname->pathname.name;
  mkcl_object type = pathname->pathname.type;
  
  if (mkcl_Null(type)
      && !mkcl_Null(name)
      && name->base_string.fillp == 1
      && name->base_string.self[0] == '.')
    return true;

  return false;
}

static bool is_dot_dot_pathname_p(MKCL, mkcl_object pathname)
{
  mkcl_object name = pathname->pathname.name;
  mkcl_object type = pathname->pathname.type;

  if (!mkcl_Null(name)
      && name->base_string.fillp == 1
      && name->base_string.self[0] == '.'
      && !mkcl_Null(type)
      && type->base_string.fillp == 0)
    return true;

  return false;
}

static mkcl_object dir_real_pathname(MKCL, mkcl_object path)
{
  char * cur_dir = "./";
  mkcl_object true_path = mkcl_init_path(env);
  char * real_path = safe_realpath(env, path,
				   ((path->base_string.fillp != 0)
				    ? (char *) path->base_string.self
				    : cur_dir));

  string_pushx_cstr(env, true_path, real_path);
  if (real_path)
    free(real_path); /* Yes this bypasses the GC since it was malloced by realpath(). */

  if (!MKCL_IS_DIR_SEPARATOR(string_last(env, true_path)))
    {
      char str[] = { MKCL_DIR_SEPARATOR, 0 };
      string_pushx_cstr(env, true_path, str);
    }
  return mk_cl_pathname(env, true_path);
}

/*
 * dir_files() lists all files which are contained in the current directory and
 * which match the masks in PATHNAME. This routine is essentially a wrapper for
 * list_directory(), which transforms the list of strings into a list
 * of pathnames. BASEDIR is the real name of the current directory and it is
 * used to build these pathnames.
 */
static mkcl_object
dir_files(MKCL, bool follow_symlinks, mkcl_object path, mkcl_object pathname)
{
  mkcl_object all_files, output = mk_cl_Cnil;
  mkcl_object mask;
  mkcl_object name = pathname->pathname.name;
  mkcl_object type = pathname->pathname.type;
  mkcl_object basedir = dir_real_pathname(env, path);

  if (name == mk_cl_Cnil && type == mk_cl_Cnil)
    {
      mkcl_object it = copy_path(env, path);
      mkcl_object kind = file_kind(env, (char *) it->base_string.self, FALSE);

      back_path(env, it);
      if (!follow_symlinks && @':link' == kind)
	{
	  mkcl_object raw_pathname = mk_cl_pathname(env, it);
	  mkcl_object back = back_path(env, it);

	  mkcl_object true_pathname 
	    = mk_cl_make_pathname(env, 4,
			       @':name', raw_pathname->pathname.name,
			       @':defaults', mk_cl_truename(env, back));

	  mkcl_object true_namestring
	    = mkcl_namestring_base(env, true_pathname, TRUE);

	  if (@':directory'
	      == file_kind(env, (char *) true_namestring->base_string.self, TRUE))
	    {
	      reset_path(env, it);
	      string_pushx_string(env, it, true_namestring);
	      string_pushx_cstr(env, it, "/");
	      return mk_cl_list(env, 1, mk_cl_pathname(env, it));
	    }
	  else
	    return mk_cl_list(env, 1, true_pathname);
	}
      else if (!mkcl_Null(kind))
	return mk_cl_list(env, 1, basedir);
      else
	return mk_cl_Cnil;
    }
  if (is_dot_pathname_p(env, pathname))
    {
      return mk_cl_list(env, 1, basedir);
    }
  if (is_dot_dot_pathname_p(env, pathname))
    {
      string_pushx_cstr(env, path, "/..");
      return mk_cl_list(env, 1, mk_cl_truename(env, path));
    }

  mask = mkcl_make_pathname(env, mk_cl_Cnil, mk_cl_Cnil, mk_cl_Cnil,
			   name, type, pathname->pathname.version);
  all_files = list_directory(env, path, NULL, FALSE);
  mkcl_loop_for_in(env, all_files) {
    mkcl_object new = MKCL_CAR(all_files);
    char *text = (char*)new->base_string.self;
#ifdef unix
    if (mkcl_stringp(env, new) && mkcl_member_char(env, ':', new)) {
      /* File names are allowed to have ':', but MKCL
       * interprets colons as separators for device names
       * By prepending the name with a ':', we set the device
       * to NIL and parse the file name properly */
      /* This hack is not possible on MS-Windows! It is unix only at best. JCB */
      new = mk_si_concatenate_base_strings(env, 2,
					  mkcl_make_simple_base_string(env, ":"),
					  new);
    }
#endif
    new = mk_cl_pathname(env, new);
    if (mkcl_Null(mk_cl_pathname_match_p(env, new, mask)))
      continue;

    new->pathname.host = basedir->pathname.host;
    new->pathname.device = basedir->pathname.device;
    new->pathname.directory = basedir->pathname.directory;
#ifdef HAVE_LSTAT
    /* Resolve symbolic links */
    if (follow_symlinks && file_kind(env, text, FALSE) == @':link'){
      new = mk_cl_truename(env, new);
    } 
#endif
    {
      mkcl_object namestring = mkcl_namestring_base(env, new, TRUE);
      
      if (@':directory'
	  == file_kind(env, (char *) namestring->base_string.self, TRUE)) {
	mkcl_object new_path = mkcl_init_path(env);
	string_pushx_string(env, new_path, namestring);
	string_pushx_cstr(env, new_path, "/");
	new = mk_cl_pathname(env, new_path);
      }
    }
    output = MKCL_CONS(env, new, output);
  } mkcl_end_loop_for_in;
  return output;
}

/*
 * dir_recursive() performs the dirty job of DIRECTORY. The routine moves
 * through the filesystem looking for files and directories which match
 * the masks in the arguments PATHNAME and DIRECTORY, collecting them in a
 * list.
 */

static mkcl_object
dir_recursive(MKCL, bool follow_symlinks,
	      mkcl_object path, mkcl_object pathspec, mkcl_object dir_spec)
{
  mkcl_object item, next_dir, output = mk_cl_Cnil;

  /* There are several possibilities here:
   *
   * 1) The list of subdirectories DIR_SPEC is empty, and only PATHSPEC
   * remains to be inspected. If there is no file name or type, then
   * we simply output the truename of the current directory. Otherwise
   * we have to find a file which corresponds to the description.
   */
  if (dir_spec == mk_cl_Cnil) {
    return dir_files(env, follow_symlinks, path, pathspec);
  }
  /*
   * 2) We have not yet exhausted the DIR_SPEC component of the
   * pathspec. We have to enter some subdirectory, determined by
   * MKCL_CAR(DIR_SPEC) and scan it.
   */
  item = MKCL_CAR(dir_spec);

  if (mkcl_type_of(item) == mkcl_t_base_string || item == @':wild') {
    /*
     * 2.1) If MKCL_CAR(DIR_SPEC) is a string or :WILD, we have to
     * enter & scan all subdirectories in our curent directory.
     */
    next_dir = list_directory(env,
			      path,
			      (item == @':wild') 
			      ? "*" : (const char *)item->base_string.self,
			      TRUE);
    mkcl_loop_for_in(env, next_dir) {
      mkcl_object path = MKCL_CAR(next_dir);
      string_pushx_cstr(env, path, "/");
      item = dir_recursive(env, follow_symlinks, path, pathspec, MKCL_CDR(dir_spec));
      output = mkcl_nconc(env, item, output);
    } mkcl_end_loop_for_in;
  } else if (item == @':up') {
    /*
     * 2.4) If MKCL_CAR(DIR_SPEC) is :UP, we have to scan the directory
     * which contains this one.
     */

    string_pushx_cstr(env, path, "../");
    output = dir_recursive(env, follow_symlinks, path, pathspec, MKCL_CDR(dir_spec));
  } else if (item == @':back') {
    back_path(env, path);
    output = dir_recursive(env, follow_symlinks, path, pathspec, MKCL_CDR(dir_spec));
  } else if (item == @':wild-inferiors') {
    /*
     * 2.5) If MKCL_CAR(DIR_SPEC) is :WILD-INFERIORS, we have to do
     * scan all subdirectories from _all_ levels, looking for a
     * tree that matches the remaining part of DIR_SPEC.
     */
    next_dir = list_directory(env, path, "*", TRUE);
    mkcl_loop_for_in(env, next_dir) {
      mkcl_object path = MKCL_CAR(next_dir);
      string_pushx_cstr(env, path, "/");
      item = dir_recursive(env, follow_symlinks, path, pathspec, dir_spec);
      output = mkcl_nconc(env, item, output);
    } mkcl_end_loop_for_in;
    output = mkcl_nconc(env, output, dir_recursive(env, follow_symlinks,
						   path, pathspec, MKCL_CDR(dir_spec)));
  } else {
    mkcl_lose("dir_recursive: invalid directory component");
  }
  return output;
}

static mkcl_object mkcl_init_path(MKCL)
{
  return mkcl_alloc_adjustable_base_string(env, mkcl_core.path_max);
}

static mkcl_object reset_path(MKCL, mkcl_object path)
{
  path->base_string.fillp = 0;
  path->base_string.self[0] = 0;
  return path;
}

static mkcl_object back_path(MKCL, mkcl_object path)
{
  if (path->base_string.fillp == 1
      || (path->base_string.fillp == 2 && path->base_string.self[1] == '/')
      )
    return path;  /* we are at root already. */
  else
    {
      size_t i = path->base_string.fillp - 1;

      if (path->base_string.self[i] == '/')
	while (i > 0 && path->base_string.self[i] == '/') i--;
      else
	while (i > 0 && path->base_string.self[i] != '/') i--;

      if (i == 0 && path->base_string.self[0] != '/')
	path->base_string.self[i = 1] = '/';

      path->base_string.self[i+1] = 0;
      path->base_string.fillp = i+1;

      return path;
    }
}

static mkcl_object copy_path(MKCL, mkcl_object old)
{
  mkcl_object new = mkcl_init_path(env);
  return string_pushx_string(env, new, old);
}

@(defun directory (mask &key (follow_symlinks mk_cl_Cnil) &allow_other_keys)
  mkcl_object output;
@
  {
    mkcl_object path_spec = mkcl_coerce_to_file_pathname(env, mask);
    mkcl_object dir_spec = path_spec->pathname.directory;
    mkcl_object path = mkcl_init_path(env);

#ifdef MKCL_WINDOWS
    mkcl_object host = path_spec->pathname.host;
    mkcl_object device = path_spec->pathname.device;

    if (!mkcl_Null(host)
	&& !mkcl_string_E(env, host, mkcl_core.localhost_string))
      {
	mkcl_object share = device;

	string_pushx_cstr(env, path, "//");
	string_pushx_string(env, path, host);
	string_pushx_cstr(env, path, "/");
	if (!mkcl_Null(share))
	  {
	    string_pushx_string(env, path, share);
	    string_pushx_cstr(env, path, "/");
	  }
      }
    else if (!mkcl_Null(device))
      {
	string_pushx_string(env, path, device);
	string_pushx_cstr(env, path, ":");
      }
#endif
    if (MKCL_CAR(dir_spec) == @':absolute')
      string_pushx_cstr(env, path, "/");
    else if (MKCL_CAR(dir_spec) == @':relative')
      string_pushx_cstr(env, path, "./");
    output = dir_recursive(env, !mkcl_Null(follow_symlinks),
			   path, path_spec, MKCL_CDR(dir_spec));
  }
  @(return output);
@)

@(defun ext::getcwd (&key
		     (change_default_pathname_defaults mk_cl_Cnil)
		     (all_drives mk_cl_Cnil)
		     )
  mkcl_object output;
@
  output = mk_cl_parse_namestring(env, 3,
				  current_dir(env), mk_cl_Cnil, mk_cl_Cnil);
  if (!mkcl_Null(change_default_pathname_defaults)) {
    MKCL_SETQ(env, @'*default-pathname-defaults*', output);
  }

#ifdef MKCL_WINDOWS
  {
    mkcl_object all_cwd = mk_cl_Cnil;
    ULONG uDriveMask = _getdrives();
    int drive = 1;

    while (uDriveMask)
      {
	if (uDriveMask & 1)
	  {
	    if (!mkcl_Null(all_drives) || drive > 2) /* Skip floppies. */
	      {
		mkcl_object cwd = current_dir_on_drive(env, drive);
	      
		cwd = mk_cl_parse_namestring(env, 3,
					     cwd, mk_cl_Cnil, mk_cl_Cnil);
		all_cwd = MKCL_CONS(env, cwd, all_cwd);
	      }
	  }
	drive++;
	uDriveMask >>= 1;
      }


    @(return output mk_cl_nreverse(env, all_cwd));
  }
#else /* def MKCL_WINDOWS */
  @(return output);
#endif /* def MKCL_WINDOWS */
@)

static void warn_early(MKCL, char * msg)
{
  fprintf(stderr, "\nMKCL warning: %s\n", msg);
  fflush(stderr);
}

mkcl_object
mk_si_get_library_pathname(MKCL)
{
  mkcl_object s = mkcl_core.library_pathname;

  if (mkcl_Null(s))
    {
      const char *v = getenv("MKCL_LIBDIR");

      if (v)
	{
	  int len = strlen(v);

	  if ( len && v[len - 1] != '/' )
	    {  /* ensure there is a / at the end. */
	      if ( (len + 1) >= mkcl_core.path_max )
		mkcl_lose("Value of MKCL_LIBDIR is too long to be"
			  " a valid MKCL's library directory");
	    
	      s = mkcl_alloc_simple_base_string(env, mkcl_core.path_max);
	      len = sprintf((char *) s->base_string.self, "%s/", v);
	      s->base_string.dim = s->base_string.fillp = len;
	    }
	  else
	    s = mkcl_make_base_string_copy(env, v);
	}
      else
	{
	  mkcl_object true_pathname;
	  /* mkcl_object lib_device  */
	  /*   = mk_cl_pathname_device(env, 1, mkcl_core.self_truename); */
	  mkcl_object lib_dir_root =
	    mk_cl_directory_namestring(env, mkcl_core.self_truename);
	  mkcl_object lib_dir_tip
	    = mkcl_make_simple_base_string(env, "../lib/mkcl-" MKCL_VERSION "/");

	  lib_dir_root = mk_cl_make_pathname(env, 8,
					     @':name', mk_cl_Cnil,
					     @':type', mk_cl_Cnil,
					     @':version', mk_cl_Cnil,
					     @':defaults', mkcl_core.self_truename);


	  s = mk_cl_merge_pathnames(env, 2, lib_dir_tip, lib_dir_root);

	  true_pathname = mk_cl_probe_file(env, s);

	  if (mkcl_Null(true_pathname))
	    {
	      s = mkcl_make_simple_base_string(env, MKCL_LIBDIR "/");
	    }
	  else
	    /* Produce a string */
	    s = mkcl_namestring(env, s, FALSE);
	}

      mkcl_core.library_pathname = s;
    }
  @(return s);
}

@(defun ext::chdir (directory &optional (change_d_p_d mk_cl_Cnil))
  mkcl_object previous = mk_si_getcwd(env, 0);
  mkcl_object namestring;
@
  /* This will fail if the new directory does not exist */
  directory = mk_cl_truename(env, directory);
  if (directory->pathname.name != mk_cl_Cnil ||
      directory->pathname.type != mk_cl_Cnil)
    mkcl_FEerror(env, "~A is not a directory pathname.", 1, directory);
  namestring = mkcl_namestring_base(env, directory, TRUE);
  if (namestring == mk_cl_Cnil) {
    mkcl_FEerror(env, "Pathname ~A does not have a physical namestring", 1, directory);
  }
  if (safe_chdir(env, (char*)namestring->base_string.self) <0)
    mkcl_FElibc_file_error(env, namestring, 
			   "Can't change the current directory to ~A", 1, namestring);
  if (change_d_p_d != mk_cl_Cnil)
    MKCL_SETQ(env, @'*default-pathname-defaults*', directory);
  @(return previous);
@)

mkcl_object
mk_si_mkdir(MKCL, mkcl_object directory, mkcl_object mode)
{
  int ok;
  mkcl_object filename = mk_si_coerce_to_filename(env, directory);
#ifdef unix
  mkcl_index modeint = mkcl_fixnum_in_range(env, @'si::mkdir',"mode",mode,0,0777);
#endif

  /* Make sure we are NULL terminated. */
  if (filename->base_string.hasfillp)
    filename->base_string.self[filename->base_string.fillp] = '\0';

#ifdef MKCL_WINDOWS
  MKCL_NO_INTR(env, ok = mkdir((char*)filename->base_string.self));
#else
  MKCL_NO_INTR(env, ok = mkdir((char*)filename->base_string.self, modeint));
#endif

  if (ok < 0)
    mkcl_FElibc_file_error(env, filename, "Could not create directory.", 0);
  @(return filename);
}

mkcl_object
mk_si_mkstemp(MKCL, mkcl_object template)
{
  mkcl_object output;
  mkcl_index l;

#ifdef MKCL_WINDOWS
  mkcl_object phys, dir, file;
  char strTempDir[MAX_PATH];
  char strTempFileName[MAX_PATH];
  char *s;
  UINT ok = 0;

  phys = mk_cl_translate_logical_pathname(env, 1, template);
  dir = mk_cl_make_pathname(env, 8,
			    @':type', mk_cl_Cnil,
			    @':name', mk_cl_Cnil,
			    @':version', mk_cl_Cnil,
			    @':defaults', phys);
  dir = mk_si_coerce_to_filename(env, dir);
  file = mk_cl_file_namestring(env, phys);
	
  l = dir->base_string.fillp;
  if ( l <= (MAX_PATH - 14) )
    {
      memcpy(strTempDir, dir->base_string.self, l);
      strTempDir[l] = 0;
      for (s = strTempDir; *s; s++)
	if (*s == '/')
	  *s = '\\';
      
      MKCL_NO_INTR(env, ok = GetTempFileName(strTempDir,
					     (char*)file->base_string.self,
					     0,
					     strTempFileName));
    }
  if (!ok) {
    output = mk_cl_Cnil;
  } else {
    output = mkcl_make_base_string_copy(env, strTempFileName);
  }
#else /* def MKCL_WINDOWS */
  int fd;
  int ok;

  template = mk_si_coerce_to_filename(env, template);
  l = template->base_string.fillp;
  output = mkcl_alloc_simple_base_string(env, l + 6);
  memcpy(output->base_string.self, template->base_string.self, l);
  memcpy(output->base_string.self + l, "XXXXXX", 6);
  output->base_string.self[l + 6] = '\0';

  MKCL_NO_INTR(env, fd = mkstemp((char*)output->base_string.self));

  if (fd < 0) {
    output = mk_cl_Cnil;
  } else {
    mkcl_safe_close(env, fd, mk_cl_Cnil);
  }
#endif /* def MKCL_WINDOWS */
  @(return (mkcl_Null(output) ? output : mk_cl_truename(env, output)));
}

mkcl_object
mk_si_rmdir(MKCL, mkcl_object directory)
{
  int code;
  directory = mk_si_coerce_to_filename(env, directory);

  MKCL_NO_INTR(env, code = rmdir((char*)directory->base_string.self));

  if (code != 0)
    mkcl_FElibc_file_error(env, directory, "Can't remove directory.", 0);
  @(return mk_cl_Cnil);
}

mkcl_object
mk_si_copy_file(MKCL, mkcl_object orig, mkcl_object dest)
{
  mkcl_interrupt_status old_intr;
  FILE *in, *out;
  int ok = 0;

  orig = mk_si_coerce_to_filename(env, orig);
  dest = mk_si_coerce_to_filename(env, dest);

  mkcl_get_interrupt_status(env, old_intr);
  mkcl_disable_interrupts(env);
  in = fopen((char*)orig->base_string.self, "r");
  if (in) {
    out = fopen((char*)dest->base_string.self, "w");
    if (out) {
      const size_t buffer_size = 64 * 1024;
      unsigned char *buffer = mkcl_alloc_atomic_unprotected(buffer_size);
      size_t out_size;
      /* size_t size; */
      do {
	size_t in_size = buffer_size;
	unsigned char * in_base = buffer;
	unsigned char * out_base = buffer;

	do {
	  size_t size = fread(in_base, 1, in_size, in);
	  in_base += size;
	  in_size -= size;
	} while ( in_size && !ferror(in) && !feof(in) );

	out_size = buffer_size - in_size;
	do {
	  size_t size = fwrite(out_base, 1, out_size, out);
	  out_base += size;
	  out_size -= size;
	} while ( out_size && !ferror(out) );
      } while ( !ferror(in) && !feof(in) && !out_size && !ferror(out) );
      
      mkcl_set_interrupt_status(env, old_intr);

      if (ferror(in) || out_size || ferror(out))
	mkcl_FEerror(env, "In si::copy-file, error while copying from ~S to ~S.", 2, orig, dest);
      else
	ok = 1;
      mkcl_safe_fclose(env, out, dest);
    } else {
      mkcl_set_interrupt_status(env, old_intr);
      mkcl_FEerror(env, "In si::copy-file, failed to open for reading file ~S.", 1, orig);
    }
    mkcl_safe_fclose(env, in, orig);
  }
  else {
    mkcl_set_interrupt_status(env, old_intr);
    mkcl_FEerror(env, "In si::copy-file, failed to open for writing file ~S.", 1, dest);
  }

  mkcl_set_interrupt_status(env, old_intr);
  @(return (ok ? mk_cl_Ct : mk_cl_Cnil));
}



static mkcl_object string_pushx_string(MKCL, mkcl_object s1, mkcl_object s2)
{
  const mkcl_index len = s2->base_string.fillp;
  mkcl_index i = 0;

  for (; i < len; i++)
    mkcl_string_push_extend(env, s1, mkcl_char(env, s2, i));
  return s1;
}

static mkcl_object string_pushx_cstr(MKCL, mkcl_object s1, const char * s2)
{
  size_t i = 0;
  mkcl_base_char c = s2[0];
  
  for (c = s2[i]; c; c = s2[++i])
    mkcl_string_push_extend(env, s1, c);
  return s1;
}

static int string_last(MKCL, mkcl_object str)
{
  mkcl_index i = str->base_string.fillp;

  if (i)
    return str->base_string.self[i - 1];
  else
    return 0;
}
