/* -*- mode: c -*- */
/*
    unixsys.s  -- Unix shell interface.
*/
/*
    Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
    Copyright (c) 1990, Giuseppe Attardi.
    Copyright (c) 2001, Juan Jose Garcia Ripoll.
    Copyright (c) 2010, Jean-Claude Beaudoin.

    MKCL is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file '../../Copyright' for full details.
*/

#include <mkcl/mkcl.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <mkcl/internal.h>

#if defined(MKCL_WINDOWS)
# include <windows.h>
#endif

#ifdef HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
# ifdef linux
#  include <sys/syscall.h>
# endif
#endif

#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

void mkcl_safe_close(MKCL, int fd, mkcl_object name)
{
#if linux
  int status = 0;
  mkcl_interrupt_status old_intr;

  mkcl_get_interrupt_status(env, old_intr);
  mkcl_disable_interrupts(env);
  while ((status = close(fd)) < 0)
    if (errno != EINTR) break;
  mkcl_set_interrupt_status(env, old_intr);
  if (status)
    mkcl_FElibc_file_error(env, name, "mkcl_safe_close failed.", 0);
#elif defined(MKCL_WINDOWS)
  if (_close(fd))
    mkcl_FEwin32_file_error(env, name, "mkcl_safe_close failed.", 0);
#else
#error "Don't know how to implement mkcl_safe_close()!"
#endif
}

void mkcl_safe_fclose(MKCL, FILE * s, mkcl_object name)
{
#if linux
  int status = 0;
  mkcl_interrupt_status old_intr;

  mkcl_get_interrupt_status(env, old_intr);
  mkcl_disable_interrupts(env);
  while ((status = fclose(s)) == EOF)
    if ((errno != EINTR) && (errno != EAGAIN) && (errno != EWOULDBLOCK)) break;
  mkcl_set_interrupt_status(env, old_intr);

  if (status)
    mkcl_FElibc_file_error(env, name, "mkcl_safe_fclose() failed.", 0);
#elif defined(MKCL_WINDOWS)
  if (fclose(s) == EOF)
    mkcl_FEwin32_file_error(env, name, "mkcl_safe_fclose() failed.", 0);
#else
#error "Don't know how to implement mkcl_safe_fclose()!"
#endif
}

#ifdef MKCL_WINDOWS

static mkcl_object read_command_output(MKCL, HANDLE child_stdout_read)
{
  mkcl_object output = mk_cl_Cnil;
  mkcl_object out_tail = mk_cl_Cnil;

  while (TRUE)
    {
      CHAR buffer[16 * 1024];
      DWORD read_count = 0;
      mkcl_object paragraph;

      if (!ReadFile(child_stdout_read, buffer,
		    sizeof(buffer) - sizeof(buffer[0]),
 		    &read_count, NULL)
	  || (read_count == 0))
	{
	  if (GetLastError() == ERROR_BROKEN_PIPE)
	    break; /* EOF on pipe - normal exit. */
	  else
	    mkcl_FEwin32_error(env, "read_command_output failed on ReadFile", 0);
	}

      buffer[read_count/sizeof(buffer[0])] = 0; /* mark end of string. */
      paragraph = mkcl_make_base_string_copy(env, buffer);

      /* accumulate output */
      if (mkcl_Null(out_tail))
	output = out_tail = MKCL_CONS(env, paragraph, mk_cl_Cnil);
      else
	{
	  mkcl_object this = MKCL_CONS(env, paragraph, mk_cl_Cnil);

	  MKCL_RPLACD(out_tail, this);
	  out_tail = this;
	}
    }
  return output;
}

@(defun si::run-command (command &key directory)
@
  char * new_directory = NULL;
  mkcl_object command_bs = mk_si_copy_to_simple_base_string(env, command);
  mkcl_object output;
  STARTUPINFO StartUp;
  PROCESS_INFORMATION ProcInfo;
  DWORD ExitCode;
  BOOL success;

  HANDLE child_stdout_read, child_stdout_read_tmp, child_stdout_write;
  HANDLE child_stdin, child_stderr;
  SECURITY_ATTRIBUTES sec_attr;
  HANDLE current = GetCurrentProcess(); /* pseudo handle */

  sec_attr.nLength = sizeof(SECURITY_ATTRIBUTES);
  sec_attr.lpSecurityDescriptor = NULL;
  sec_attr.bInheritHandle = TRUE;

  /* create stdout */
  if (!CreatePipe(&child_stdout_read_tmp, &child_stdout_write, &sec_attr, 0))
    mkcl_FEwin32_error(env, "si::run-command failed on CreatePipe", 0);

  /* create stderr, handle has bInheritHandle to TRUE. */
  if (!DuplicateHandle(current, child_stdout_write, current, &child_stderr,
		       0, TRUE, DUPLICATE_SAME_ACCESS))
    mkcl_FEwin32_error(env, "si::run-command failed on DuplicateHandle", 0);

  /* create stdin */
  child_stdin = CreateFile("NUL", GENERIC_READ,
			   FILE_SHARE_READ | FILE_SHARE_WRITE,
			   &sec_attr, OPEN_EXISTING, 0, NULL);
  if (child_stdin == INVALID_HANDLE_VALUE)
    mkcl_FEwin32_error(env, "si::run-command failed to open NUL device", 0);

  /* create stdout, handle has bInheritHandle to FALSE. */
  if (!DuplicateHandle(current, child_stdout_read_tmp, current, &child_stdout_read,
		       0, FALSE, DUPLICATE_SAME_ACCESS))
    mkcl_FEwin32_error(env, "si::run-command failed on DuplicateHandle", 0);
  if (!CloseHandle(child_stdout_read_tmp))
    mkcl_FEwin32_error(env, "si::run-command failed on CloseHandle", 0);


  ZeroMemory(&StartUp, sizeof(STARTUPINFO));
  StartUp.cb = sizeof(STARTUPINFO);
  StartUp.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
  StartUp.wShowWindow = SW_HIDE;
  StartUp.hStdOutput = child_stdout_write;
  StartUp.hStdInput = child_stdin;
  StartUp.hStdError = child_stderr;

  /* Microsoft's documentation says that the maximum size for
     the "command" string is 32768 characters, all included.
     And the first component of it must be less than MAX_PATH.
     FIXME!
  */
  if (!mkcl_Null(directory))
    {
      mkcl_object directory_bs
	= mk_si_copy_to_simple_base_string(env, directory);

      new_directory = (char *) directory_bs->base_string.self;
    }

  success = CreateProcess(NULL,
			  (char *) command_bs->base_string.self,
			  NULL,
			  NULL,
			  TRUE, /* bInheritHandles */
			  CREATE_NEW_CONSOLE, /* dwCreationFlags */
			  NULL,
			  new_directory,
			  &StartUp,
			  &ProcInfo);
  if (success == 0)
    mkcl_FEwin32_error(env,
		       "unable to create subprocess to execute command: ~A", 
		       1, command);

  if (!CloseHandle(ProcInfo.hThread))
    mkcl_FEwin32_error(env, "si::run-command failed on CloseHandle", 0);

  if (!CloseHandle(child_stdin))
    mkcl_FEwin32_error(env, "si::run-command failed on CloseHandle", 0);
  if (!CloseHandle(child_stdout_write))
    mkcl_FEwin32_error(env, "si::run-command failed on CloseHandle", 0);
  if (!CloseHandle(child_stderr))
    mkcl_FEwin32_error(env, "si::run-command failed on CloseHandle", 0);

  /* Recover child output. */
  output = read_command_output(env, child_stdout_read);

  switch (WaitForSingleObject(ProcInfo.hProcess, INFINITE))
    {
    case WAIT_OBJECT_0: break;
    default:
      mkcl_FEwin32_error(env, "si::run-command failed on WaitForSingleObject", 0);
    }

  if (!GetExitCodeProcess(ProcInfo.hProcess, &ExitCode))
    mkcl_FEwin32_error(env, "si::run-command failed on GetExitCodeProcess", 0);

  if (!CloseHandle(child_stdout_read))
    mkcl_FEwin32_error(env, "si::run-command failed on CloseHandle", 0);

  if (!CloseHandle(ProcInfo.hProcess))
    mkcl_FEwin32_error(env, "si::run-command failed on CloseHandle", 0);

#ifdef WIN64
  { @(return MKCL_MAKE_FIXNUM(ExitCode) output); }
#else
  { @(return mkcl_make_unsigned_integer(env, ExitCode) output); }
#endif
@)

@(defun si::system (cmd_string &key directory)
@
  @(return mk_si_run_command(env, 3, cmd_string, @':directory', directory));
@)

#elif defined(unix)  /* MKCL_WINDOWS */

static mkcl_object read_command_output(MKCL, int parent_read)
{
  mkcl_object output = mk_cl_Cnil;
  mkcl_object out_tail = mk_cl_Cnil;
  bool at_eof = false;

  while (!at_eof)
    {
      char buffer[16 * 1024];
      int read_count;
      mkcl_object paragraph;

      read_count = read(parent_read, buffer, sizeof(buffer) - sizeof(buffer[0]));
      switch (read_count)
	{
	case -1:
	  if (errno != EINTR)
	    mkcl_FElibc_error(env, "read_command_output failed on read().", 0);
	  break;
	case 0: at_eof = true; break;
	default:
	  buffer[read_count/sizeof(buffer[0])] = 0; /* mark end of string. */
	  paragraph = mkcl_make_base_string_copy(env, buffer);
	  
	  /* accumulate output */
	  if (mkcl_Null(out_tail))
	    output = out_tail = MKCL_CONS(env, paragraph, mk_cl_Cnil);
	  else
	    {
	      mkcl_object this = MKCL_CONS(env, paragraph, mk_cl_Cnil);
	      
	      MKCL_RPLACD(out_tail, this);
	      out_tail = this;
	    }
	  break;
	}
    }
  return output;
}

@(defun si::system (cmd_string &key directory)
@
  /* On unix we ignore 'directory' since it plays
     no role in the executable search sequence.
     Everything depends on the PATH environment variable.
   */
  mkcl_object cmd = mk_si_copy_to_simple_base_string(env, cmd_string);

  if (cmd->base_string.fillp >= mkcl_core.arg_max)
    mkcl_FEerror(env, "Too long command line: ~S.", 1, cmd);

  /* This call to 'system' has the side-effect, for its duration, of
     blocking SIGCHLD and ignoring both SIGINT and SIGQUIT! JCB */
  /* It also shares, between parent and child, all file descriptors
     (among others: stdin, stdout, stderr). JCB */
  {
    int code = system((const char *)(cmd->base_string.self));

    if (WIFEXITED(code))
      { @(return MKCL_MAKE_FIXNUM(WEXITSTATUS(code))); }
    else if (WIFSIGNALED(code))
      { @(return MKCL_MAKE_FIXNUM((-1 & 0xff) + WTERMSIG(code))); }
    else
      { @(return MKCL_MAKE_FIXNUM(-1)); }
  }
@)

@(defun si::run-command (cmd_string &key directory)
@
  /* On unix we ignore 'directory' since it plays
     no role in the executable search sequence.
     Everything depends on the PATH environment variable.
   */
  int parent_read;
  int child_stdin, child_stdout, child_stderr;
  int fd[2];
  int status;
  pid_t pid;
  mkcl_object output = mk_cl_Cnil;
  mkcl_object cmd = mk_si_copy_to_simple_base_string(env, cmd_string);

  if (cmd->base_string.fillp >= mkcl_core.arg_max)
    mkcl_FEerror(env, "Too long command line: ~S.", 1, cmd);

  /* prepare redirection of stdout (and stderr). */
  if (pipe(fd) < 0)
    mkcl_FElibc_error(env, "ext:run-command failed on pipe [output].", 0);
  parent_read = fd[0];
  child_stdout = fd[1];
  child_stderr = child_stdout;

  /* This code paragraph comes from the book
     "Advanced Programming in the UNIX Environment"
     by Richard Stevens, page 223.

     Redirection of stdin, stdout, stderr was added later.
  */
  if ((pid = fork()) < 0) { /* error */
    mkcl_safe_close(env, child_stdout, mk_cl_Cnil);
    mkcl_safe_close(env, parent_read, mk_cl_Cnil);
    status = -1;
  } else if (pid == 0) { /* child */
    /* redirect stdin to /dev/null. */
    mkcl_safe_close(env, 0, mk_cl_Cnil);
    if ((child_stdin = open("/dev/null", O_RDONLY)) < 0)
      mkcl_FElibc_error(env, "ext:run-command failed on open. [child_stdin].", 0);
    /* redirect stdout. */
    mkcl_safe_close(env, 1, mk_cl_Cnil);
    if (dup(child_stdout) == -1)
      mkcl_FElibc_error(env, "ext:run-command child failed on dup(child_stdout).", 0);;
    /* redirect stderr. */
    mkcl_safe_close(env, 2, mk_cl_Cnil);
    if (dup(child_stderr) == -1)
      mkcl_FElibc_error(env, "ext:run-command child failed on dup(child_stderr).", 0);

    mkcl_safe_close(env, child_stdout, mk_cl_Cnil);
    mkcl_safe_close(env, parent_read, mk_cl_Cnil);
    
    execl("/bin/sh", "sh", "-c", cmd->base_string.self, (char *) 0);
    _exit(127); /* execl failed! Let's report it like system() does. */
  } else { /* parent */
    mkcl_safe_close(env, child_stdout, mk_cl_Cnil);

    /* Here we must read what the child wrote to its stdout. */
    output = read_command_output(env, parent_read);

    mkcl_safe_close(env, parent_read, mk_cl_Cnil);

    while (waitpid(pid, &status, 0) < 0)
      if (errno != EINTR) {
	status = -1;
	break;
      }

  } /* end of quote from Stevens. */

  if (status == -1)
    { @(return MKCL_MAKE_FIXNUM(status) output); }
  else if (WIFEXITED(status))
    { @(return MKCL_MAKE_FIXNUM(WEXITSTATUS(status)) output); }
  else if (WIFSIGNALED(status))
    { @(return MKCL_MAKE_FIXNUM((-1 & 0xff) + WTERMSIG(status)) output); }
  else
    { @(return MKCL_MAKE_FIXNUM(-1) output); }
@)

#endif /* MKCL_WINDOWS */



mkcl_object
mk_si_getpid(MKCL)
{
  @(return MKCL_MAKE_FIXNUM(getpid()));
}

pid_t mkcl_gettid(void)
{
#ifdef linux
  return syscall(SYS_gettid);
#elif defined(MKCL_WINDOWS)
  return GetCurrentThreadId();
#else
  return getpid();
#endif
}

mkcl_object
mk_si_gettid(MKCL)
{
  @(return MKCL_MAKE_FIXNUM(mkcl_gettid()));
}

mkcl_object
mk_si_getuid(MKCL)
{
#if defined(MKCL_WINDOWS)
  /* GetUserName() followed by LookupAccountName(). JCB */
  @(return MKCL_MAKE_FIXNUM(0));
#else
  @(return mkcl_make_integer(env, getuid()));
#endif
}

mkcl_object
mk_si_make_pipe(MKCL)
{
  mkcl_object output;
  int fds[2], ret;
#if defined(MKCL_WINDOWS)
  ret = _pipe(fds, 4096, _O_BINARY);
#else
  ret = pipe(fds);
#endif
  if (ret < 0) {
    mkcl_FElibc_error(env, "Unable to create pipe", 0);
    output = mk_cl_Cnil;
  } else {
    mkcl_object fake_in_name
      = mkcl_make_simple_base_string(env, "PIPE-READ-ENDPOINT");
    mkcl_object in = mkcl_make_file_stream_from_fd(env, fake_in_name,
						   fds[0], mkcl_smm_input, 8,
						   MKCL_STREAM_DEFAULT_FORMAT,
						   mk_cl_Cnil);
    mkcl_object fake_out_name
      = mkcl_make_simple_base_string(env, "PIPE-WRITE-ENDPOINT");
    mkcl_object out = mkcl_make_file_stream_from_fd(env, fake_out_name,
						    fds[1], mkcl_smm_output, 8,
						    MKCL_STREAM_DEFAULT_FORMAT,
						    mk_cl_Cnil);
    output = mk_cl_make_two_way_stream(env, in, out);
  }
  @(return output);
}

@(defun ext::run-program (command argv &key (input @':stream') (output @':stream')
	  		  (error @'t') (wait @'t'))
	int parent_write = 0, parent_read = 0;
	int child_pid;
	mkcl_object stream_write;
	mkcl_object stream_read;
	mkcl_object exit_status = mk_cl_Cnil;
@
  command = mk_si_copy_to_simple_base_string(env, command);
  argv = mk_cl_mapcar(env, 2, @'si::copy-to-simple-base-string', argv);
#if defined(MKCL_WINDOWS)
  {
    BOOL ok;
    STARTUPINFO st_info;
    PROCESS_INFORMATION pr_info;
    HANDLE child_stdout, child_stdin, child_stderr;
    HANDLE current = GetCurrentProcess(); /* pseudo handle */
    SECURITY_ATTRIBUTES attr;

    /* Enclose each argument, as well as the file name
       in double quotes, to avoid problems when these
       arguments or file names have spaces */
    command =
      mk_cl_format(env, 4, mk_cl_Cnil,
		   mkcl_make_simple_base_string(env, "~S~{ ~S~}"),
		   command, argv);
    command = mk_si_copy_to_simple_base_string(env, command);
    command = mkcl_null_terminated_base_string(env, command);

    attr.nLength = sizeof(SECURITY_ATTRIBUTES);
    attr.lpSecurityDescriptor = NULL;
    attr.bInheritHandle = TRUE;
    if (input == @':stream') {
      /* Creates a pipe that we can read from what the child
	 writes to it. We duplicate one end of the pipe
	 so that the child does not inherit it. */
      HANDLE tmp;
      ok = CreatePipe(&child_stdin, &tmp, &attr, 0);
      if (ok) {
	ok = DuplicateHandle(current, tmp, current,
			     &tmp, 0, FALSE,
			     DUPLICATE_CLOSE_SOURCE |
			     DUPLICATE_SAME_ACCESS);
	if (ok) {
	  parent_write = _open_osfhandle((intptr_t) tmp, _O_WRONLY /*| _O_TEXT*/);
	  if (parent_write < 0)
	    mkcl_FElibc_error(env, "ext:run-program failed on _open_osfhandle() for parent_write", 0);
	}
	else
	  mkcl_FEwin32_error(env, "ext:run-program failed on DuplicateHandle() for parent_write", 0);
      }
      else
	mkcl_FEwin32_error(env, "ext:run-program failed on CreatePipe() for child_stdin", 0);
    } else if (input == @'t') {
      /* The child inherits a duplicate of our input
	 handle. Creating a duplicate avoids problems when
	 the child closes it */
      mkcl_object input_stream = mkcl_symbol_value(env, @'*standard-input*');
      int stream_handle = mkcl_stream_to_handle(env, input_stream, 0);
      if (stream_handle >= 0)
	{
	  ok = DuplicateHandle(current,
			       (HANDLE) _get_osfhandle(stream_handle),
			       current,
			       &child_stdin,
			       0,
			       TRUE,
			       DUPLICATE_SAME_ACCESS);
	  if (!ok) mkcl_FEwin32_error(env, "ext:run-program failed on DuplicateHandle() for child_stdin", 0);
	}
      else
	child_stdin = NULL;
    } else {
      child_stdin = CreateFile("NUL", GENERIC_READ,
			       FILE_SHARE_READ | FILE_SHARE_WRITE,
			       &attr, OPEN_EXISTING, 0, NULL);
      if (child_stdin == INVALID_HANDLE_VALUE)
	mkcl_FEwin32_error(env, "ext:run-program failed to open NUL device for child_stdin", 0);
    }
    if (output == @':stream') {
      /* Creates a pipe that we can write to and the
	 child reads from. We duplicate one end of the
	 pipe so that the child does not inherit it. */
      HANDLE tmp;
      ok = CreatePipe(&tmp, &child_stdout, &attr, 0);
      if (ok) {
	ok = DuplicateHandle(current, tmp, current,
			     &tmp, 0, FALSE,
			     DUPLICATE_CLOSE_SOURCE |
			     DUPLICATE_SAME_ACCESS);
	if (ok) {
	  parent_read = _open_osfhandle((intptr_t) tmp, _O_RDONLY /*| _O_TEXT*/);
	  if (parent_read < 0)
	    mkcl_FElibc_error(env, "ext:run-program failed on _open_osfhandle() for parent_read", 0);
	}
	else
	  mkcl_FEwin32_error(env, "ext:run-program failed on DuplicateHandle() for parent_read", 0);
      }
      else
	mkcl_FEwin32_error(env, "ext:run-program failed on CreatePipe() for child_stdout", 0);
    } else if (output == @'t') {
      /* The child inherits a duplicate of our output
	 handle. Creating a duplicate avoids problems when
	 the child closes it */
      mkcl_object output_stream = mkcl_symbol_value(env, @'*standard-output*');
      int stream_handle = mkcl_stream_to_handle(env, output_stream, 1);
      if (stream_handle >= 0)
	{
	  ok = DuplicateHandle(current,
			       (HANDLE) _get_osfhandle(stream_handle),
			       current, &child_stdout, 0, TRUE,
			       DUPLICATE_SAME_ACCESS);
	  if (!ok) mkcl_FEwin32_error(env, "ext:run-program failed on DuplicateHandle() for child_stdout", 0);
	}
      else
	child_stdout = NULL;
    } else {
      child_stdout = CreateFile("NUL", GENERIC_WRITE,
				FILE_SHARE_READ | FILE_SHARE_WRITE,
				&attr, OPEN_EXISTING, 0, NULL);
      if (child_stdout == INVALID_HANDLE_VALUE)
	mkcl_FEwin32_error(env, "ext:run-program failed to open NUL device for child_stdout", 0);
    }
    if (error == @':output') {
      /* The child inherits a duplicate of its own output handle.*/
      ok = DuplicateHandle(current, child_stdout, current,
			   &child_stderr, 0, TRUE,
			   DUPLICATE_SAME_ACCESS);
      if (!ok) mkcl_FEwin32_error(env, "ext:run-program failed on DuplicateHandle() for child_stderr", 0);
    } else if (error == @'t') {
      /* The child inherits a duplicate of our output
	 handle. Creating a duplicate avoids problems when
	 the child closes it */
      mkcl_object error_stream = mkcl_symbol_value(env, @'*error-output*');
      int stream_handle = mkcl_stream_to_handle(env, error_stream, 1);
      if (stream_handle >= 0)
	{
	  ok = DuplicateHandle(current,
			       (HANDLE) _get_osfhandle(stream_handle),
			       current, &child_stderr, 0, TRUE,
			       DUPLICATE_SAME_ACCESS);
	  if (!ok) mkcl_FEwin32_error(env, "ext:run-program failed on DuplicateHandle() for child_stderr", 0);
	}
      else
	child_stderr = NULL;
    } else {
      child_stderr = CreateFile("NUL", GENERIC_WRITE,
				FILE_SHARE_READ | FILE_SHARE_WRITE,
				&attr, OPEN_EXISTING, 0, NULL);
      if (child_stderr == INVALID_HANDLE_VALUE)
	mkcl_FEwin32_error(env, "ext:run-program failed to open NUL device for child_stderr", 0);
    }

    ZeroMemory(&st_info, sizeof(STARTUPINFO));
    st_info.cb = sizeof(STARTUPINFO);
    st_info.lpTitle = NULL; /* No window title, just exec name */
    st_info.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW; /* Specify std{in,out,err} */
    st_info.wShowWindow = SW_HIDE;
    st_info.hStdInput = child_stdin;
    st_info.hStdOutput = child_stdout;
    st_info.hStdError = child_stderr;
    ZeroMemory(&pr_info, sizeof(PROCESS_INFORMATION));
    ok = CreateProcess(NULL, (char *) command->base_string.self,
		       NULL, NULL, /* lpProcess/ThreadAttributes */
		       TRUE, /* Inherit handles (for files) */
		       /*CREATE_NEW_CONSOLE |*/
		       0 /*(input == mk_cl_Ct || output == mk_cl_Ct || error == mk_cl_Ct ? 0 : CREATE_NO_WINDOW)*/,
		       NULL, /* Inherit environment */
		       NULL, /* Current directory */
		       &st_info, /* Startup info */
		       &pr_info); /* Process info */

    /* Child handles must be closed in the parent process */
    /* otherwise the created pipes are never closed       */
    if (child_stdin) CloseHandle(child_stdin);
    if (child_stdout) CloseHandle(child_stdout);
    if (child_stderr) CloseHandle(child_stderr);
    if (ok) {
      DWORD exitcode;
      CloseHandle(pr_info.hThread);
      child_pid = pr_info.dwProcessId;
      if (wait != mk_cl_Cnil) {
	switch (WaitForSingleObject(pr_info.hProcess, INFINITE))
	  {
	  case WAIT_OBJECT_0: break;
	  default:
	    mkcl_FEwin32_error(env, "ext:run-program failed on WaitForSingleObject() for subprocess", 0);
	  }

	if (GetExitCodeProcess(pr_info.hProcess, &exitcode) &&
	    STILL_ACTIVE != exitcode) {
	  exit_status = MKCL_MAKE_FIXNUM(exitcode);
	}
      }
      CloseHandle(pr_info.hProcess);
    } else {
#if 0
      mkcl_FEwin32_error(env, "ext:run-program failed on CreateProcess() for command ~S", 1, command);
#endif
      child_pid = -1;
    }
  }
  if (child_pid < 0) {
    if (parent_write) 
      if (_close(parent_write))
	mkcl_FElibc_error(env, "ext:run-program failed on _close(parent_write).", 0);
    if (parent_read) 
      if (_close(parent_read))
	mkcl_FElibc_error(env, "ext:run-program failed on _close(parent_read).", 0);
    parent_write = 0;
    parent_read = 0;
    mkcl_FEerror(env, "ext:run-program could not spawn subprocess to run ~S.", 1, command);
  }
#else /* def MKCL_WINDOWS */
  {
    int child_stdin, child_stdout, child_stderr;
    argv = MKCL_CONS(env, command, mkcl_nconc(env, argv, mkcl_list1(env, mk_cl_Cnil)));
    argv = mk_cl_funcall(env, 3, @'coerce', argv, @'vector');
    if (input == @':stream') {
      int fd[2];
      if (pipe(fd) < 0)
	mkcl_FElibc_error(env, "ext:run-program failed on pipe [input].", 0);
      parent_write = fd[1];
      child_stdin = fd[0];
    } else {
      child_stdin = -1;
      if (input == @'t') {
	mkcl_object input_stream = mkcl_symbol_value(env, @'*standard-input*');
	child_stdin = mkcl_stream_to_handle(env, input_stream, 0);
      }
      if (child_stdin >= 0) {
	if ((child_stdin = dup(child_stdin)) < 0)
	  mkcl_FElibc_error(env, "ext:run-program failed on dup(child_stdin).", 0);
      } else {
	if ((child_stdin = open("/dev/null", O_RDONLY)) < 0)
	  mkcl_FElibc_error(env, "ext:run-program failed on open. [child_stdin].", 0);
      }
    }
    if (output == @':stream') {
      int fd[2];
      if (pipe(fd) < 0)
	mkcl_FElibc_error(env, "ext:run-program failed on pipe [output].", 0);
      parent_read = fd[0];
      child_stdout = fd[1];
    } else {
      child_stdout = -1;
      if (output == @'t') {
	mkcl_object output_stream = mkcl_symbol_value(env, @'*standard-output*');
	child_stdout = mkcl_stream_to_handle(env, output_stream, 1);
      }
      if (child_stdout >= 0) {
	if ((child_stdout = dup(child_stdout)) < 0)
	  mkcl_FElibc_error(env, "ext:run-program failed on dup(child_stdout).", 0);
      } else {
	if ((child_stdout = open("/dev/null", O_WRONLY)) < 0)
	  mkcl_FElibc_error(env, "ext:run-program failed on open. [child_stdout].", 0);
      }
    }
    if (error == @':output') {
      child_stderr = child_stdout;
    } else if (error == @'t') {
      mkcl_object error_stream = mkcl_symbol_value(env, @'*error-output*');
      child_stderr = mkcl_stream_to_handle(env, error_stream, 1);
    } else {
      child_stderr = -1;
    }
    if (child_stderr < 0) {
      if ((child_stderr = open("/dev/null", O_WRONLY)) < 0)
	mkcl_FElibc_error(env, "ext:run-program failed on open. [child_stderr].", 0);
    } else {
      if ((child_stderr = dup(child_stderr)) < 0)
	mkcl_FElibc_error(env, "ext:run-program failed on dup(child_stderr).", 0);
    }
    child_pid = fork();
    if (child_pid < 0)
      { /* error */
#if 0
	fprintf(stderr, "\nfork failed for command = %s.\n", command->base_string.self);
	fflush(stderr);
#endif
      }
    else if (child_pid == 0) {
      /* Child */
      int j;
      char **argv_ptr = (char **)argv->vector.self.t;

      if (parent_read) mkcl_safe_close(env, parent_read, mk_cl_Cnil);
      if (parent_write) mkcl_safe_close(env, parent_write, mk_cl_Cnil);

      mkcl_safe_close(env, 0, mk_cl_Cnil);
      if (dup(child_stdin) == -1)
	mkcl_FElibc_error(env, "ext:run-program child failed on dup(child_stdin).", 0);
      mkcl_safe_close(env, child_stdin, mk_cl_Cnil);
      mkcl_safe_close(env, 1, mk_cl_Cnil);
      if (dup(child_stdout) == -1)
	mkcl_FElibc_error(env, "ext:run-program child failed on dup(child_stdout).", 0);;
      mkcl_safe_close(env, child_stdout, mk_cl_Cnil);
      mkcl_safe_close(env, 2, mk_cl_Cnil);
      if (dup(child_stderr) == -1)
	mkcl_FElibc_error(env, "ext:run-program child failed on dup(child_stderr).", 0);
      mkcl_safe_close(env, child_stderr, mk_cl_Cnil);

      for (j = 0; j < argv->vector.fillp; j++) {
	mkcl_object arg = argv->vector.self.t[j];
	if (arg == mk_cl_Cnil) {
	  argv_ptr[j] = NULL;
	} else {
	  argv_ptr[j] = (char *) arg->base_string.self;
	}
      }
      execvp((char*)command->base_string.self, argv_ptr);
      /* at this point exec has failed */
      perror("exec");
      abort();
    } else {
      /* Parent */
      mkcl_safe_close(env, child_stdin, mk_cl_Cnil);
      mkcl_safe_close(env, child_stdout, mk_cl_Cnil);
      mkcl_safe_close(env, child_stderr, mk_cl_Cnil);
      if (child_pid > 0 && wait != mk_cl_Cnil) {
	int status;
	while (waitpid(child_pid, &status, 0) < 0)
	  if (errno != EINTR) {
	    status = -1;
	    break;
	  }
	if (status == -1)
	  mkcl_lose("ext:run-program failed on waitpid().");
	else
	  exit_status = MKCL_MAKE_FIXNUM(WEXITSTATUS(status));
      }
    }
    if (child_pid < 0) { /* error */
      int fork_errno = errno;
      
      mkcl_safe_close(env, child_stdin, mk_cl_Cnil);
      mkcl_safe_close(env, child_stdout, mk_cl_Cnil);
      mkcl_safe_close(env, child_stderr, mk_cl_Cnil);
      
      if (parent_write) mkcl_safe_close(env, parent_write, mk_cl_Cnil);
      if (parent_read) mkcl_safe_close(env, parent_read, mk_cl_Cnil);
      parent_write = 0;
      parent_read = 0;
      errno = fork_errno;
      mkcl_FElibc_error(env, "Could not spawn subprocess to run ~S.", 1, command);
    }
  }
#endif  /* def MKCL_WINDOWS */
  if (parent_write > 0) {
    stream_write = mkcl_make_file_stream_from_fd(env, command, parent_write,
						 mkcl_smm_output, 8,
						 MKCL_STREAM_DEFAULT_FORMAT, mk_cl_Ct);
  } else {
    parent_write = 0;
    stream_write = mkcl_core.null_stream;
  }
  if (parent_read > 0) {
    stream_read = mkcl_make_file_stream_from_fd(env, command, parent_read,
						mkcl_smm_input, 8,
						MKCL_STREAM_DEFAULT_FORMAT, mk_cl_Ct);
  } else {
    parent_read = 0;
    stream_read = mkcl_core.null_stream;
  }
  @(return ((parent_read || parent_write)
	    ? mk_cl_make_two_way_stream(env, stream_read, stream_write)
	    : mk_cl_Cnil)
    exit_status);
@)

