;;;;  -*- Mode: Lisp; Syntax: Common-Lisp; Package: C -*-
;;;;
;;;;  Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
;;;;  Copyright (c) 1990, Giuseppe Attardi.
;;;;  Copyright (c) 2001, Juan Jose Garcia Ripoll.
;;;;  Copyright (c) 2010-2011, Jean-Claude Beaudoin.
;;;;
;;;;    This program is free software; you can redistribute it and/or
;;;;    modify it under the terms of the GNU Lesser General Public
;;;;    License as published by the Free Software Foundation; either
;;;;    version 3 of the License, or (at your option) any later version.
;;;;
;;;;    See file '../../Copyright' for full details.

;;;; CMPMAIN  Compiler main program.

(in-package "COMPILER")

(defvar *trace-cc* nil)

(defun safe-system (string)
  (cmpnote "Doing: ~A" string)
  (when *trace-cc*
    (format t "~&Doing: ~A~%" string)
    (finish-output))
  (let ((work-dir #+windows (namestring (make-pathname :name nil
						       :type nil
						       :version nil
						       :defaults (si:self-truename)))
		  #-windows nil))
    (multiple-value-bind (result output)
      (si:run-command string :directory work-dir)
      (unless (zerop result)
	(cerror "Continues anyway."
		"(si:run-command ~S) returned non-zero value ~D~%~
                 Command output:~%~{~A~}~%"
		string result output))
      (when (and *trace-cc* output)
	(format t "~{~A~}~%" output)
	(finish-output)
	)
      (values result output))))

(defun compile-file-pathname (name &key (output-file T) (type nil type-supplied-p)
                              verbose print c-file h-file data-file
                              (fasl-p t) current-directory
			      external-format)
  (declare (ignore verbose print c-file h-file data-file external-format))
  (let* ((format '())
         (extension '()))
    (unless type-supplied-p
      (setf type (if fasl-p :fasl :object)))
    (case type
      ((:shared-library :dll) (setf format +shared-library-format+))
      ((:static-library :library :lib) (setf format +static-library-format+))
      (:data (setf extension "data"))
      (:sdata (setf extension "sdat"))
      (:c (setf extension "c"))
      (:h (setf extension "h"))
      (:object (setf extension +object-file-extension+))
      (:program (setf format +executable-file-format+))
      #+msvc
      (:import-library (setf extension "implib"))
      ((:fasl :fas) (setf extension "fas"))
      (:fasb (setq extension "fasb")))
    (merge-pathnames
     (cond ((not (or (eq output-file T)
		     (eq output-file nil)))
	    output-file)
	   (format
	    (merge-pathnames (format nil format (pathname-name name)) name))
	   (t
	    (make-pathname :type extension :defaults name)))
     (or current-directory #P""))))
  
#+msvc
(defun delete-msvc-generated-files (output-pathname)
  (loop for i in '("lib" "exp" "ilk" "pdb")
        do (let ((the-pathname (merge-pathnames (make-pathname :type i) output-pathname)))
	     (when (probe-file the-pathname)
	       (cmp-delete-file the-pathname)))))

(defun cmp-delete-file (file)
  (cond ((null *delete-files*))
	(*debug-compiler*
	 (cmpprogress "~%Postponing deletion of ~A" file)
	 (push file *files-to-be-deleted*))
	(t
	 (and (probe-file file)
	      (delete-file file)))))

(push #'(lambda () (mapc #'delete-file *files-to-be-deleted*))
      si::*exit-hooks*)


(defun linker-cc (o-pathname &rest options)
  (safe-system
   (format nil
	   *ld-format*
	   *ld*
	   (si::coerce-to-filename o-pathname)
	   options
	   *ld-flags*)))

(defun shared-cc (o-pathname &rest options)
  (safe-system
   (format nil
	   *ld-format*
	   *ld*
	   (si::coerce-to-filename o-pathname)
	   options
	   *ld-shared-flags*)))

(defun bundle-cc (o-pathname init-name &rest options)
  (declare (ignorable init-name))
  (safe-system
   (format nil
	   *ld-format*
	   *ld*
	   (si::coerce-to-filename o-pathname)
	   options
	   #-msvc *ld-bundle-flags*
	   #+msvc (concatenate 'string *ld-bundle-flags*
			       " /EXPORT:" init-name
			       " /LIBPATH:" (mkcl-library-directory)
			       " /IMPLIB:"
			       (si::coerce-to-filename
				(compile-file-pathname
				 o-pathname :type :import-library))))))

(defconstant +lisp-program-header+ "
#include <mkcl/mkcl.h>

#ifdef __cplusplus
#define MKCL_CPP_TAG \"C\"
#else
#define MKCL_CPP_TAG
#endif

~{	extern MKCL_CPP_TAG void ~A(MKCL, mkcl_object, mkcl_object);~%~}

")


(defconstant +lisp-program-init-export+ "
#ifdef __cplusplus
extern \"C\"
#else
extern
#endif
__declspec(dllexport) void ~A(MKCL, mkcl_object cblock, mkcl_object fasl_filename);
~%")

;;
;; This format string contains the structure of the code that initializes
;; a program, a library, a module, etc. Basically, it processes a codeblock
;; just like in a normal compiled file, but then adds all the codeblocks of
;; its corresponding modules.
;;
;; IMPORTANT: Notice how the modules are linked to the parent forming a
;; circular chain. This disables the garbage collection of the library until
;; _ALL_ functions in all modules are unlinked.
;;
(defconstant +lisp-program-init+ "
#ifdef __cplusplus
extern \"C\"
#endif
void ~A(MKCL, mkcl_object cblock, mkcl_object fasl_filename)
{
	static mkcl_object Cblock;
        if (!MKCL_FIXNUMP(cblock)) {
		Cblock = cblock;
		cblock->cblock.data_text = compiler_data_text;
		cblock->cblock.data_text_size = compiler_data_text_size;
		cblock->cblock.data_size = VM;
		return;
	}
	~A
{
	mkcl_object current, next = Cblock;
~:[~{	current = mkcl_read_VV(env, MKCL_OBJNULL, ~A, fasl_filename);
        current->cblock.next = next; next = current; ~%~}
	Cblock->cblock.next = current;
~;~{	~A(env, Cblock);~%~}~]
}
	~A
}~%")

(defconstant +lisp-program-main+ "
int
main(int argc, char **argv)
{
  mkcl_env env;
  ~A

  env = mkcl_boot(argc, argv);

  mkcl_read_VV(env, MKCL_OBJNULL, ~A, mk_cl_Cnil);
  MKCL_CATCH_ALL_BEGIN(env) {
  ~A
  } MKCL_CATCH_ALL_END;
  env->own_thread->thread.active = 0;
  mkcl_thread_exit(MKCL_THREAD_KILLED);
  return(-1); /* This line is normally never reached. */
}~%")

#+windows
(defconstant +lisp-program-winmain+ "
#include <windows.h>
int
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
  char **argv;
  int argc;
  mkcl_env env;
  ~A

  mkcl_get_commandline_args(&argc, &argv);
  env = mkcl_boot(argc, argv);

  mkcl_read_VV(env, MKCL_OBJNULL, ~A, mk_cl_Cnil);
  MKCL_CATCH_ALL_BEGIN(env) {
  ~A
  } MKCL_CATCH_ALL_END;
  env->own_thread->thread.active = 0;
  mkcl_thread_exit(MKCL_THREAD_KILLED);
  return(-1);
}~%")

(defun init-function-name (s &key (kind :object))
  (flet ((translate-char (c)
	   (cond ((and (char>= c #\a) (char<= c #\z))
		  (char-upcase c))
		 ((and (char>= c #\A) (char<= c #\Z))
		  c)
		 ((or (eq c #\-) (eq c #\_))
		  #\_)
		 ((eq c #\*)
		  #\x)
		 ((eq c #\?)
		  #\a)
		 ((digit-char-p c)
		  c)
		 (t
		  #\p)))
	 (disambiguation (c)
	   (declare (ignore c))
	   (case kind
	     (:object "")
             (:program "exe_")
	     ((:fasl :fas) "fas_")
	     ((:library :shared-library :dll :static-library :lib) "lib_")
	     (otherwise (error "Not a valid argument to INIT-FUNCTION-NAME: kind = ~S"
			       kind)))))
    (setq s (map 'string #'translate-char (string s)))
    (concatenate 'string
		 "mkcl_init_"
		 (disambiguation kind)
		 (map 'string #'translate-char (string s)))))

(defun guess-kind (pathname)
  "Given a file name, guess whether it is an object file, a library, a program
or a loadable module."
  (let ((record (assoc (pathname-type pathname)
		       '(("o" :object) ("obj" :object) ("c" :c)
			 ("lib" :static-library)
			 ("a" :static-library)
			 ("dll" :shared-library)
			 ("so" :shared-library)
			 ("fas" :fasl))
		       :test #'string-equal)))
    (if record
	(second record)
	(progn
	  (warn "File ~s is of no known file type. Assuming it is an object file."
		pathname)
	  :object))))

(defun guess-ld-flags (pathname &key (kind (guess-kind pathname)))
  "Given a file name, return the compiler command line argument to link this file in."
  (case kind
    ((:object :c)
     (si::coerce-to-filename pathname))
    ((:fasl :fas)
     nil)
    ((:static-library :lib)
     (si::coerce-to-filename pathname))
    ((:shared-library :dll)
     (si::coerce-to-filename pathname))
    ((:program)
     nil)
    (otherwise
     (error "C::BUILDER cannot accept files of kind ~s" kind))))

(defun system-ld-flag (library)
  "Given a symbol, try to find a library that matches it, either by looking in the
filesystem or in the database of ASDF modules."
  (let ((asdf (find-package "ASDF"))
        system)
    (labels ((asdfsym (x) (find-symbol (string x) asdf))
             (asdfcall (fun &rest rest) (apply (asdfsym fun) rest))
             (system-output (system type)
               (let ((build (make-instance (asdfsym :build-op) :type type)))
                 (first (asdfcall :output-files build system))))
             (existing-system-output (system type)
               (let ((o (system-output system type)))
                 (and o (setf o (probe-file o)) (namestring o))))
             (find-archive (system)
                 (or (existing-system-output system :library)
                     (existing-system-output system :shared-library)))
             (fallback () (format nil #-msvc "-l~S" #+msvc "~A.lib" (string-downcase library))))
      (or (and asdf
               (setf system (asdfcall :find-system library nil))
               (find-archive system))
        (fallback)))))

(defun builder (target output-name &key lisp-files ld-flags
		(init-name nil)
		(prologue-code "")
		(epilogue-code (when (eq target :program) '(SI::TOP-LEVEL)))
		#+windows (system :console)
		&aux
		(cwd (si:getcwd))
		(*suppress-compiler-messages* (or *suppress-compiler-messages*
						  (not *compile-verbose*))))

  (when *suppress-compiler-notes*
    (setf *suppress-compiler-messages*
	  `(or ,*suppress-compiler-messages* compiler-note)))
  (when *suppress-compiler-warnings*
    (setf *suppress-compiler-messages*
	  `(or ,*suppress-compiler-messages* compiler-warning)))

  ;;
  ;; The epilogue-code can be either a string made of C code, or a
  ;; lisp form.  In the latter case we add some additional C code to
  ;; clean up, and the lisp form is stored in a text representation,
  ;; to avoid using the compiler.
  ;;
  (cond ((null epilogue-code)
	 (setf epilogue-code ""))
	((stringp epilogue-code)
	 )
	(t
	 (with-standard-io-syntax
	   (setq epilogue-code
		 (with-output-to-string (stream)
		   (princ "{ char *lisp_code = " stream)
		   (wt-filtered-data (write-to-string epilogue-code) stream)
		   (princ ";
  mkcl_object output = mk_cl_Cnil;
  mk_si_select_package(env, mkcl_make_simple_base_string(env, \"CL-USER\"));
  output = mk_si_safe_eval(env, 3, mkcl_fast_read_from_cstring(env, lisp_code), mk_cl_Cnil, MKCL_OBJNULL);
"
			  stream)
		   (when (eq target :program)
		     (princ "
  env->own_thread->thread.result_value = output;
  env->own_thread->thread.active = 0;
  mkcl_thread_exit(MKCL_THREAD_NORMAL_EXIT);
" 
			    stream))
		   (princ #\} stream)
		   )))))
  ;;
  ;; When a module is built out of several object files, we have to
  ;; create an additional object file that initializes those ones.
  ;; This routine is responsible for creating this file.
  ;;
  ;; To avoid name clashes, this object file will have a temporary
  ;; file name (tmp-name).
  ;;
  (when (or (symbolp output-name) (stringp output-name))
    (setf output-name (compile-file-pathname output-name :type target)))
  (setq output-name (merge-pathnames output-name cwd)) ;; make output-name absolute path

  (let* ((tmp-name (si::mkstemp (make-pathname :name "MKCLINIT" :type nil :defaults output-name)))
	 (c-name (si::coerce-to-filename
		  (compile-file-pathname tmp-name :type :c)))
	 (o-name (si::coerce-to-filename
		  (compile-file-pathname tmp-name :type :object)))
	 submodules
	 c-file)
    (dolist (item (reverse lisp-files))
      (etypecase item
        (symbol
         (push (system-ld-flag item) ld-flags)
         (push (init-function-name item :kind :lib) submodules))
        ((or string pathname)
	 (let* ((path (merge-pathnames item cwd)) ;; make item absolute path
		(kind (guess-kind path)))
	   (unless (member kind '(:shared-library :dll :static-library :lib
				  :object :c))
	     (error "C::BUILDER does not accept a file ~s of kind ~s" item kind))
	   (let ((init-fn (guess-init-name path))
		 (flags (guess-ld-flags path)))
	     (when flags (push flags ld-flags))
	     (push init-fn submodules))))))
    (setq c-file (open c-name :direction :output :external-format :default))
    (format c-file +lisp-program-header+ submodules)
    (format c-file "~@
                    #define compiler_data_text NULL~@
                    #define compiler_data_text_size 0~@
                    static mkcl_object * VV = NULL;~@
                    #define VM 0~%"
		   c-file)
    ;; (when (or (symbolp output-name) (stringp output-name))
    ;;   (setf output-name (compile-file-pathname output-name :type target)))
    ;; (setq output-name (merge-pathnames output-name cwd)) ;; make output-name absolute path
    (unless init-name
      (setf init-name (compute-init-name output-name :kind target)))
    (ecase target
      (:program
       (format c-file +lisp-program-init+ init-name "" nil
	       submodules "")
       (format c-file #+windows (ecase system (:console +lisp-program-main+)
					   (:windows +lisp-program-winmain+))
	              #-windows +lisp-program-main+
		      prologue-code init-name epilogue-code)
       (close c-file)
       (compiler-cc c-name o-name)
       #+mingw32
       (ecase system
	      (:console (push "-mconsole" ld-flags))
	      (:windows (push "-mwindows" ld-flags)))
       (apply #'linker-cc output-name (namestring o-name) ld-flags))
      ((:library :static-library :lib)
       (let ((output-filename (si::coerce-to-filename output-name)))
	 (format c-file +lisp-program-init+ init-name prologue-code
		 nil submodules epilogue-code)
	 (close c-file)
	 (compiler-cc c-name o-name)
	 (when (probe-file output-filename) (delete-file output-filename))
	 #-msvc
	 (progn
	   (safe-system (format nil "~S cr ~S ~S ~{~S ~}"
				*ar* output-filename o-name ld-flags))
	   (safe-system (format nil "~S ~S" *ranlib* output-filename)))
	 #+msvc
	 (unwind-protect
	     (progn
	       (with-open-file (f "static_lib.tmp" :direction :output :if-does-not-exist :create :if-exists :supersede)
			       (format f "/DEBUGTYPE:CV /OUT:~A ~A ~{~&\"~A\"~}"
				       output-filename o-name ld-flags))
	       (safe-system "link -lib @static_lib.tmp"))
	   (when (probe-file "static_lib.tmp")
	     (cmp-delete-file "static_lib.tmp")))
	 )
       )
      ((:shared-library :dll)
       #+windows (format c-file +lisp-program-init-export+ init-name)
       (format c-file +lisp-program-init+ init-name prologue-code
	       nil submodules epilogue-code)
       (close c-file)
       (compiler-cc c-name o-name)
       (apply #'shared-cc output-name o-name ld-flags))
      ((:fasl :fasb)
       #+windows (format c-file +lisp-program-init-export+ init-name)
       (format c-file +lisp-program-init+ init-name prologue-code nil
	       submodules epilogue-code)
       (close c-file)
       (compiler-cc c-name o-name)
       (apply #'bundle-cc output-name init-name o-name ld-flags)))
    (cmp-delete-file tmp-name)
    (cmp-delete-file c-name)
    (cmp-delete-file o-name)
    output-name))

(defun build-fasl (&rest args)
  (handler-case 
   (apply #'builder :fasl args)
   (condition (condition)
      (format t "~&build-fasl failed: ~A~%" condition))))

(defun build-bundle (&rest args)
  (handler-case 
   (apply #'builder :fasb args)
   (condition (condition)
      (format t "~&build-bundle failed: ~A~%" condition))))

(defun build-program (&rest args)
  (handler-case 
   (apply #'builder :program args)
   (condition (condition)
      (format t "~&build-program failed: ~A~%" condition))))

(defun build-static-library (&rest args)
  (handler-case 
   (apply #'builder :static-library args)
   (condition (condition)
      (format t "~&build-static-library failed: ~A~%" condition))))

(defun build-shared-library (&rest args)
  (handler-case 
   (apply #'builder :shared-library args)
   (condition (condition)
      (format t "~&build-shared-library failed: ~A~%" condition))))

(defun read-safely (input eof)
  (handler-case
   (si::read-object-or-ignore input eof)
   (condition (condition)
     (format t "~&;;; Read Error: in file ~S, near position ~S,~@
                  ;;; ~A~%"
	     (pathname input)
	     (file-position input) condition)
     (handle-warning/error condition)
     nil)))


(defvar *c-file* nil) ;; default value of the :c-file argument to compile-file
(defvar *h-file* nil) ;; default value of the :h-file argument to compile-file
(defvar *data-file* nil) ;; default value of the :data-file argument to compile-file
  

(defun compile-file (input-pathname 
		     &rest args
		     &key
		     ((:verbose *compile-verbose*) *compile-verbose*)
		     ((:print *compile-print*) *compile-print*)
		     (c-file *c-file*)
		     (h-file *h-file*)
		     (data-file *data-file*)
		     (fasl-p t)
		     (external-format :default)
		     output-file
		     &aux
		     (*standard-output* *standard-output*)
		     (*error-output* *error-output*)
		     (*compiler-in-use* *compiler-in-use*)
		     (*package* *package*)
		     (*print-pretty* nil)
		     (*compile-file-pathname* nil)
		     (*compile-file-truename* nil)
		     (*suppress-compiler-messages*
		      (or *suppress-compiler-messages* (not *compile-verbose*)))
		     (cwd (si:getcwd))
		     input-file
		     init-name)
  (declare (notinline compiler-cc))
  "Compiles the file specified by INPUT-PATHNAME and generates a compiled file.
The produced compiled file is in FASL format if :FASL-P is true (the default),
otherwise its format is the native system binary object file format. 
If the filetype is not specified in INPUT-PATHNAME, then \".lsp\" is used
as the default file type for the source file.
The :OUTPUT-FILE, :C-FILE, :H-FILE, and :DATA-FILE keyword parameters allow you to
control the intermediate files generated by the MKCL compiler. If the file was
compiled successfully, returns the pathname of the compiled file.
Keyword :external-format is currently ignored."

  (declare (ignore external-format)) ;; :output-file is processed through "args".

  (when *suppress-compiler-notes*
    (setf *suppress-compiler-messages*
	  `(or ,*suppress-compiler-messages* compiler-note)))
  (when *suppress-compiler-warnings*
    (setf *suppress-compiler-messages*
	  `(or ,*suppress-compiler-messages* compiler-warning)))

  ;; the compiler needs absolute pathnames.
  (setq input-pathname (merge-pathnames input-pathname cwd))
  (when output-file
    (setq output-file (merge-pathnames output-file cwd)))
  (when (and c-file (not (eq c-file T)))
    (setq c-file (merge-pathnames c-file cwd)))
  (when (and h-file (not (eq h-file T)))
    (setq h-file (merge-pathnames h-file cwd)))
  (when (and data-file (not (eq data-file T)))
    (setq data-file (merge-pathnames data-file cwd)))
  
  (setq *compile-file-pathname* (pathname (merge-pathnames input-pathname)))
  (unless (probe-file *compile-file-pathname*)
    (if (pathname-type input-pathname)
	(error 'file-error :pathname input-pathname)
      (dolist (ext '("lsp" "LSP" "lisp" "LISP")
		   (error 'file-error :pathname input-pathname))
	(setq *compile-file-pathname* (make-pathname :type ext :defaults input-pathname))
	(when (probe-file *compile-file-pathname*)
	  (return)))))
  (setq input-file (truename *compile-file-pathname*)
        *compile-file-truename* input-file)

  (cmpprogress "~&;;; Compiling ~a." (namestring input-pathname))

  (let* ((eof '(NIL))
	 (*compiler-in-use* *compiler-in-use*)
	 (*load-time-values* nil) ;; Load time values are compiled
         (output-file (apply #'compile-file-pathname input-file :current-directory cwd args))
         (c-pathname (apply #'compile-file-pathname output-file :output-file c-file
                            :current-directory cwd :type :c args))
         (h-pathname (apply #'compile-file-pathname output-file :output-file h-file
                            :current-directory cwd :type :h args))
         (data-pathname (apply #'compile-file-pathname output-file
                               :current-directory cwd :output-file data-file :type :data args))
	 (compiler-conditions nil)
         (to-delete (nconc (unless c-file (list c-pathname))
                           (unless h-file (list h-pathname))
                           (unless data-file (list data-pathname)))))

    (with-compiler-env (compiler-conditions)

     (data-init)

     (with-open-file 
      (*compiler-input* *compile-file-pathname*)
      (do* ((ext:*source-location* (cons *compile-file-pathname* 0))
	    (form (read-safely *compiler-input* eof)
		  (read-safely *compiler-input* eof))
	    (*compile-file-end-position* (file-position *compiler-input*)
					 (file-position *compiler-input*))
	    )
	   ((eq form eof))
	   (when form
	     (setf (cdr ext:*source-location*) *compile-file-end-position*)
	     (t1expr form))))

     (progn
       (setf init-name (compute-init-name output-file :kind (if fasl-p :fasl :object)))
       (compiler-pass2 c-pathname h-pathname data-pathname init-name
		       :input-designator (namestring input-pathname))
       (data-dump data-pathname)
       )

     (if fasl-p
	 (let ((o-path (compile-file-pathname output-file
					      :current-directory cwd :type :object)))
	   (push o-path to-delete)
	   (compiler-cc c-pathname o-path)
	   (bundle-cc (si::coerce-to-filename output-file)
		      init-name
		      (si::coerce-to-filename o-path)))
       (compiler-cc c-pathname output-file))

     (if (setf output-file (probe-file output-file))
	 (cmpprogress "~&;;; Finished compiling ~a.~%" (namestring input-pathname))
       (cmperr "The C compiler failed to compile the intermediate file."))

     (mapc #'cmp-delete-file to-delete)

     ) ;; with-compiler-env

    (compiler-output-values output-file compiler-conditions)
    )
  )

(defun compiler-output-values (main-value conditions)
  (loop for i in conditions
     with style-warning-p = nil
     with failure-p = nil
     do (cond ((typep i 'style-warning)
	       (setf style-warning-p t))
	      ((typep i 'warning)
	       (setf failure-p t))
	      ((typep i 'error)
	       (setf failure-p t))
	      ((eq i :aborted)
	       (setf failure-p t))
	      )
     finally (return (values (and (not failure-p) main-value) (or style-warning-p failure-p) failure-p))))

(defvar *gazonk-counter* 0)

(defun compile (name &optional (def nil supplied-p)
                      &aux form data-pathname
                      (*suppress-compiler-messages* (or *suppress-compiler-messages*
							(not *compile-verbose*)))
                      (*compiler-in-use* *compiler-in-use*)
                      (*standard-output* *standard-output*)
                      (*error-output* *error-output*)
                      (*package* *package*)
                      (*compile-print* nil)
		      (*print-pretty* nil)
		      (*compiler-constants* t))
  "Args: (name &optional definition)

If DEFINITION is NIL, NAME must be the name of a not-yet-compiled function.
In this case, COMPILE compiles the function, installs the compiled function as
the global function definition of NAME, and returns NAME.  If DEFINITION is
non-NIL, it must be a lambda expression and NAME must be a symbol.  COMPILE
compiles the lambda expression, installs the compiled function as the function
definition of NAME, and returns NAME.  There is only one exception for this:
If NAME is NIL, then the compiled function is not installed but is simply
returned as the value of COMPILE.  In any case, COMPILE creates temporary
files, whose filenames begin with \"gazonk\", which are automatically deleted
after compilation."
  (unless (symbolp name) (error "~s is not a symbol." name))

  (when *suppress-compiler-notes*
    (setf *suppress-compiler-messages*
	  `(or ,*suppress-compiler-messages* compiler-note)))
  (when *suppress-compiler-warnings*
    (setf *suppress-compiler-messages*
	  `(or ,*suppress-compiler-messages* compiler-warning)))

  (cond ((and supplied-p def)
	 (when (functionp def)
	   (unless (function-lambda-expression def)
	     (return-from compile (values def nil nil)))
	   (setf def (function-lambda-expression def)))
         (setq form (if name
                        `(setf (symbol-function ',name) #',def)
                        `(set 'GAZONK #',def))))
	((not (fboundp name))
	 (error "Symbol ~s is unbound." name))
	((null (setq form (function-lambda-expression (setq def (symbol-function name)))))
	 (warn "We have lost the original function definition for ~s. ~
                Compilation to C failed" name)
	 (return-from compile (values def t nil)))
	(t
	 (setq form `(setf (symbol-function ',name) #',form))))

  (let ((template (format nil "TMP:MKCL~3,'0x" (incf *gazonk-counter*))))
    (unless (setq data-pathname (si::mkstemp template))
      (error "Unable to create temporay file~%~
	~AXXXXXX
Make sure you have enough free space in disk, check permissions or set~%~
the environment variable TMPDIR to a different value." template)
      (return-from compile (values nil t t))))

  (let*((*load-time-values* 'values) ;; Only the value is kept
	(c-pathname (compile-file-pathname data-pathname :type :c))
	(h-pathname (compile-file-pathname data-pathname :type :h))
	(o-pathname (compile-file-pathname data-pathname :type :object))
	(so-pathname (compile-file-pathname data-pathname))
	(init-name (compute-init-name so-pathname :kind :fasl))
	(compiler-conditions nil))

    (with-compiler-env (compiler-conditions)
      (data-init)
      (t1expr form)
      (let (#+(or mingw32 msvc cygwin)(*self-destructing-fasl* t))
	(compiler-pass2 c-pathname h-pathname data-pathname init-name :input-designator (format nil "~A" def)))
      (setf *compiler-constants* (data-dump data-pathname))

      (compiler-cc c-pathname o-pathname)
      (bundle-cc (si::coerce-to-filename so-pathname)
		 init-name
		 (si::coerce-to-filename o-pathname))

      (cmp-delete-file c-pathname)
      (cmp-delete-file h-pathname)
      (cmp-delete-file o-pathname)
      (cmp-delete-file data-pathname)

      (cond ((probe-file so-pathname)
	     (load so-pathname :verbose nil)
	     #-(or mingw32 msvc cygwin)
	     (cmp-delete-file so-pathname)
	     #+msvc
	     (delete-msvc-generated-files so-pathname)
	     (setf name (or name (symbol-value 'GAZONK)))
	     ;; By unsetting GAZONK we avoid spurious references to the
	     ;; loaded code.
	     (set 'GAZONK nil)
	     (values name nil nil))
	    (t
	     (cmperr "The C compiler failed to compile the intermediate code for ~s." name)))
      ) ; with-compiler-env


    (when (probe-file c-pathname) (cmp-delete-file c-pathname))
    (when (probe-file h-pathname) (cmp-delete-file h-pathname))
    (when (probe-file so-pathname) (cmp-delete-file so-pathname))
    (when (probe-file data-pathname) (cmp-delete-file data-pathname))

    #+msvc
    (delete-msvc-generated-files so-pathname)
    (compiler-output-values name compiler-conditions)))

;;; This thing here that pretends to implement #'disassemble has next
;;; to no chance of producing any useful result. Use at your own risk! JCB
(defun disassemble (thing &key (h-file nil) (data-file nil)
			  &aux def disassembled-form
			  (cwd (si:getcwd))
			  (*compiler-in-use* *compiler-in-use*)
			  (*print-pretty* nil))
  "Compiles the form specified by THING and prints the intermediate C language
code for that form.  But does not install the result of compilation.  If THING
is NIL, then the previously DISASSEMBLEd form is re-DISASSEMBLEd.  If THING is
a symbol that names a function not yet compiled, the function definition is
disassembled.  If THING is a lambda expression, it is disassembled as a
function definition.  Otherwise, THING itself is disassembled as a top-level
form.  H-FILE and DATA-FILE specify intermediate files to build a fasl file
from the C language code.  NIL means \"do not create the file\"."
  
  (when h-file (setq h-file (merge-pathnames h-file cwd)))
  (when data-file (setq data-file (merge-pathnames data-file cwd)))
  (when (si::valid-function-name-p thing)
    (setq thing (fdefinition thing)))
  (cond ((null thing))
	((functionp thing)
	 (unless (si::bc-disassemble thing)
	   (warn "Cannot disassemble the binary function ~S because I do not have its source code." thing)
	   (return-from disassemble nil)))
	((atom thing)
	 (error 'simple-type-error
		:datum thing
		:expected-type '(OR FUNCTION (SATISFIES SI:VALID-FUNCTION-NAME-P))
		:format-control "DISASSEMBLE cannot accept ~A"
		:format-arguments (list thing)))
	((eq (car thing) 'LAMBDA)
	 (setq disassembled-form `(defun gazonk ,@(cdr thing))))
	((eq (car thing) 'EXT:LAMBDA-BLOCK)
	 (setq disassembled-form `(defun ,@(rest thing))))
	(t
	 (error 'simple-type-error
		:datum thing
		:expected-type '(OR FUNCTION (SATISFIES SI:VALID-FUNCTION-NAME-P))
		:format-control "DISASSEMBLE cannot accept ~A"
		:format-arguments (list thing))))

  (let* ((null-stream (make-broadcast-stream))
         (*compiler-output1* null-stream)
         (*compiler-output2* (if h-file
				 (open h-file :direction :output :external-format :default)
				 null-stream))
         (t3local-fun (symbol-function 'T3LOCAL-FUN))
	 (compiler-conditions nil))
    (with-compiler-env (compiler-conditions)
      (unwind-protect
	   (progn
	     (setf (symbol-function 'T3LOCAL-FUN)
		   #'(lambda (&rest args)
		       (let ((*compiler-output1* *standard-output*))
			 (apply t3local-fun args))))
	     (data-init)
	     (t1expr disassembled-form)
	     (ctop-write (compute-init-name "foo" :kind :fasl)
			 (if h-file h-file "")
			 (if data-file data-file ""))
	     (data-dump data-file))
	(setf (symbol-function 'T3LOCAL-FUN) t3local-fun)
	(when h-file (close *compiler-output2*)))))
  nil)

(defun compiler-pass2 (c-pathname h-pathname data-pathname init-name
		       &key input-designator)
  (with-open-file (*compiler-output1* c-pathname :direction :output)
    (wt-comment-nl "Compiler: ~A ~A" (lisp-implementation-type) (lisp-implementation-version))
    #-mkcl-min
    (multiple-value-bind (second minute hour day month year)
        (get-decoded-time)
      (wt-comment-nl "Date: ~D/~D/~D ~2,'0D:~2,'0D:~2,'0D (yyyy/mm/dd)" year month day hour minute second)
      (wt-comment-nl "Machine: ~A ~A ~A" (software-type) (software-version) (machine-type)))
    (wt-comment-nl "Source: ~A" input-designator)
    (with-open-file (*compiler-output2* h-pathname :direction :output)
      (wt-nl1 "#include " *cmpinclude*)
      (catch *cmperr-tag* (ctop-write init-name
				      h-pathname
				      data-pathname))
      (terpri *compiler-output1*)
      (terpri *compiler-output2*))))

(defun mkcl-include-directory ()
  "Finds the directory in which the header files were installed."

  (let* ((bin-dir (make-pathname :name nil :type nil :version nil
				 :defaults (si:self-truename)))
	 (inc-dir (merge-pathnames "../include/" bin-dir))
	 )
    (cond ((probe-file (merge-pathnames "mkcl/mkcl.h" inc-dir))
	   inc-dir
	   )
	  ((and *mkcl-include-directory*
		(probe-file (merge-pathnames "mkcl/mkcl.h" *mkcl-include-directory*)))
	   *mkcl-include-directory*)
	  ((probe-file "SYS:mkcl;mkcl.h")
	   (setf *mkcl-include-directory* 
		 (namestring (translate-logical-pathname "SYS:"))))
	  ((error "Unable to find include directory")))))

(defun mkcl-library-directory ()
  "Finds the directory in which the MKCL core library was installed."
  (let* ((bin-dir (make-pathname :name nil :type nil :version nil
				 :defaults (si:self-truename)))
	 (lib-dir (merge-pathnames "../lib/" bin-dir))
	 )
    (cond ((probe-file (merge-pathnames (compile-file-pathname
					 "mkcl" :type
					 :shared-library)
					lib-dir))
	   lib-dir
	   )
	  ((and *mkcl-library-directory*
		(probe-file (merge-pathnames (compile-file-pathname
					      "mkcl" :type
					      :shared-library)
					     *mkcl-library-directory*)))
	   *mkcl-library-directory*)
	  ((probe-file "SYS:BUILD-STAMP")
	   (setf *mkcl-library-directory*
		 (namestring (translate-logical-pathname "SYS:"))))
	  ((error "Unable to find library directory")))))

(defun compiler-cc (c-pathname o-pathname)
  (safe-system
   (format nil
	   *cc-format*
	   *cc* *cc-flags* (>= (cmp-env-optimization 'speed) 2) *cc-optimize*
	   (namestring (mkcl-include-directory))
	   (si::coerce-to-filename c-pathname)
	   (si::coerce-to-filename o-pathname))
   ))


(si::package-lock "CL" nil)

(provide 'cmp)


;;; During bootstrap these need to be redefined at some point
;;; and thus cannot be load-time evaluated.
#+mkcl-bootstrap
(proclaim '(notinline clos::find-slot-definition
		      cl::ensure-generic-function
		      clos::method-p
		      cl::make-method
		      cl::add-method
		      cl::find-method
		      cl::remove-method
		      clos::register-method-as-spec-user
		      clos::compute-effective-method
		      clos::setf-find-class
		      cl::warn
		      si::universal-error-handler))

;;; Since CMP is an autoloaded optional module these cannot be load-time evaluated
(proclaim '(notinline cl::proclaim 
		      cl::compile-file-pathname
		      cl::compile-file
		      cl::compile
		      cl::disassemble))

